/*
 * Copyright 1988 Silicon Graphics, Inc.  All rights reserved.
 *
 * Driver for Silicon Graphics IP6 lance ethernet controller
 * This version transmits directly to/from mbufs, saving the
 * overhead of copying to/from a fixed buffer.
 */
#ident "$Revision: 1.50 $"

#ifdef IP6	/* compile for IP6 only */

#include "tcp-param.h"
#include "sys/types.h"
#include "sys/param.h"
#include "sys/systm.h"
#include "sys/sysmacros.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/errno.h"
#include "sys/immu.h"
#include "sys/invent.h"
#include "sys/kopt.h"
#include "sys/mbuf.h"
#include "sys/sbd.h"
#include "sys/socket.h"
#include "sys/edt.h"
#include "bstring.h"
#include "lance.h"

extern struct ifnet loif;
extern time_t	lbolt;

static void ec_print(char*);
static int ec_start(int);


/* Be noisy about some arbitrary things:
 *   0  = report conditions which a user should be
 *	  concerned about.
 *   1  = report conditions which are bad but may
 *	  occur in normal operations.
 *   2  = report error conditions which may happen
 *	  in normal opperations and are recoverable.
 *   3  = report information to help developers debug.
 *
 *   -- when the ifconfig debug option is selected
 *   -- and noise level is at least 2, conditions
 *   -- which are not errors but which are interesting
 *   -- to a developer will be displayed.
 */


#ifdef OS_METER
/* keep count of errors that don't print a msg. on the console */
u_int	ec_istat_missed;
u_int	ec_istat_frame;
u_int	ec_istat_crc;
u_int	ec_istat_oflo;
#endif /* OS_METER */

/* first byte of SGI's assigned ethernet addresses */
#define SGI_ETHER_ADDR 0x08

#define DOG		(2*IFNET_SLOWHZ)  /* watchdog duration in seconds */

/* XXX move this to some header file, for all drivers */
#define EMSG_DELAY	(HZ*60*1)	/* suppress messages this long */

/* defines for the receive and transmit descriptors */
#define LOG_RMD         6		/* log2 # receive memory descriptors */
#define LOG_TMD         7		/* log2 # transmit memory descriptors */
#define NUM_RMD         (1<<LOG_RMD)    /* # of receive buffers */
#define NUM_TMD         (1<<LOG_TMD)    /* # of transmit buffers */

/*
 * Memory for first transmission chunk must be contiguous.  Reserve extra
 * space at the front for the padded header.
 */
#define	MIN_TBUF	(RAW_HDRPAD(EHDR_LEN) + MIN_TPKT_C)
#define	SPACE_FOR_EHDR	(EHDR_LEN + RAW_HDRPAD(EHDR_LEN))
#define MAX_CHUNKS	4
#define MIN_CHUNK_SIZE	8

/*
 * On the IP6, the 7990 lance registers are mapped into memory along
 * with the memory map registers.  These registers are all defined
 * by the I/O channel (IOC) multiplexer specification.
 */

/* 7990 hardware registers */
#define RDP_ADDR (volatile u_short *)PHYS_TO_K1(0x1f950000) /* reg data port */
#define RAP_ADDR (volatile u_short *)PHYS_TO_K1(0x1f950100) /* reg addr port */

/* 7990 hardware lines IOC mapped to memory */
#define ETHERRST_ADDR  (volatile u_char *)PHYS_TO_K1(0x1f960004) /* reset on */
#define ETHERRDY_ADDR  (volatile u_char *)PHYS_TO_K1(0x1f960000) /* reset off */

/* start of IOC host <--> 7990 memory mapping registers (256 in all) */
#define EMAP	((volatile u_short *)PHYS_TO_K1(0x1f920802))

/*
 * Macros for mapping addresses so that the 7990 can exchange data with
 * the host.  NB: map is addressed as alternating shorts.
 */
#define EMAPI(x)		(2 * (x))
#define EADDR(index,addr)	(((index) << 12) | ((u_int)(addr) & 0xfff))
#define	EADDR_INVERSE(haddr,laddr)	(((u_int)(laddr) >> 12) + ((u_int)(haddr) << 4))

/*
 * Organize the ether map registers.  2 entries for init block and rec/xmit
 * descriptors.  Receiving only requires one descriptor/buffer since the
 * buffers are always page aligned.  All of the rest are used for xmit
 * entries (which can cross page boundaries).
 */
#define EMAPLEN		256			/* # of entries in the emap */
#define IBMAP_INDEX	0			/* init blk and descrptors*/
#define RBUF_INDEX	(IBMAP_INDEX+2)		/* receive mbufs */
#define TBUF_INDEX	(RBUF_INDEX+NUM_RMD)	/* xmit map entry index */
#define TMAP_CNT	(EMAPLEN-TBUF_INDEX)	/* number of xmit map entries */

/*
 * Receive buffer structure and initializer.
 */
struct rcvbuf {
	struct etherbufhead	ebh;
	char			data[MAX_RDAT];
};

#define INITRCVBUF(ei, rmd, rmdnum, m) { \
	struct rcvbuf *rb = mtod(m, struct rcvbuf *); \
	struct ether_header *ehdr = &rb->ebh.ebh_ether; \
	int index = RBUF_INDEX + (rmdnum); \
\
	IF_INITHEADER(rb, &(ei)->ei_if, sizeof rb->ebh); \
	EMAP[EMAPI(index)] = btotp(kvtophys(ehdr)); \
	(rmd)->rmd_ladr = LO_ADDR(EADDR(index, ehdr)); \
	(rmd)->rmd_hadr = HI_ADDR(EADDR(index, ehdr)); \
	(rmd)->rmd_own = 1; \
}

/*
 * Data structures associated the the lance chip which must be kept
 * in local memory.
 */
static struct {
	struct ib	m_ib;
	union {
	    struct rmd	rmd[NUM_RMD];
	    double	align;
	} m_r;
	union {
	    struct tmd	tmd[NUM_TMD];
	    double	align;
	} m_t;
} mem;

static volatile struct rmd *rmdvec;
static volatile struct tmd *tmdvec;

/*
 * Driver-specific 7990 CSR values.
 */
#define CSR1		LO_ADDR(EADDR(IBMAP_INDEX, &mem.m_ib))
#define CSR2		HI_ADDR(EADDR(IBMAP_INDEX, &mem.m_ib))

/*
 * Space to save the addresses of the mbufs that are being transmitted
 * and received so that they may be subsequently freed.
 */
static struct mbuf *r_mbuf_addr[NUM_RMD];

static struct mbuf *t_mbuf_head;
static struct mbuf *t_mbuf_tail;

/*
 * Each interface is represented by a network interface structure, ec_info,
 * which the routing code uses to locate the interface.
 */
struct ec_info {
	struct lanceif	ei_lif;		/* lance network interface */

	char		*ei_terr;	/* recent transmit error msg */
	time_t		ei_terr_lbolt;	/* when to forget it */

	u_int		ei_ract;	/* next active receive buffer */
	u_int		ei_tact;	/* 1st active transmit buffer */
	u_int		ei_tnum;	/* # of active transmit buffers */
	u_int		ei_tfree;	/* next free transmit buffer */
	u_int		ei_tmapfree;	/* next free xmit map entry */
	u_int		ei_tmaplast;	/* index of the last freed xmit map */

	u_int		ei_numtdr;	/* # of TDR samples */
	u_long		ei_tdr;		/* sum of all TDR samples */
	u_int		ei_rxwatch;	/* timer for watching receives */
	u_int		ei_rxwatch_back;    /* backoff */
	u_int		ei_frzcount;	/* count suspected chip freezes */
	u_int		ei_pac_of;	/* error information breakdown */
	u_int		ei_rbuf_err;
	u_int		ei_pac_uf;
	u_int		ei_tbuf_err;
	u_int		ei_late_coll;
} ec_info;

#define	ei_eif		ei_lif.lif_eif		/* common ethernet stuff */
#define	ei_ib		ei_lif.lif_ib		/* initialization block ptr */
#define	ei_laf		ei_lif.lif_laf		/* logical address filter */
#define	ei_ac		ei_eif.eif_arpcom	/* common arp stuff */
#define ei_if           ei_ac.ac_if		/* network-visible interface */
#define ei_curaddr      ei_ac.ac_enaddr		/* current ethernet address */
#define	ei_addr		ei_eif.eif_addr		/* board's official address */
#define	ei_rawif	ei_eif.eif_rawif	/* raw domain interface */
#define	ei_lostintrs	ei_eif.eif_lostintrs	/* count lost interrupts */
#define	ei_resets	ei_eif.eif_resets	/* count board resets */
#define	ei_sick		ei_eif.eif_sick		/* 2 => reset on next bark */

/* constants used in the chip freeze detection algorithm */
#define WORRY_TICKS	(2*HZ)
#define	RX_WATCH_TICKS	(32*HZ)		/* > RIP broadcast timeout */


/*
 * Reset the lance chip by toggling reset.  Interrupts must be off.
 */
#ifdef lint
#define LINTREF(n) lintref = n;
static int lintref;
#else
#define LINTREF(n)
#endif
#define	ECRESET() { \
	u_char reset_t = *ETHERRST_ADDR;    /* ethernet chip reset on */ \
	DELAYBUS(5);							\
	reset_t = *ETHERRDY_ADDR;	/* ethernet chip reset off */	\
	LINTREF(reset_t);}

/*
 * Initialize the chip.  Interrupts must already be off.
 */
/* ARGSUSED */
static int				/* return 0 or an errno */
ec_init(eif, flags)
	struct etherif *eif;
	int flags;
{
	int error;

	ec_info.ei_rxwatch_back = 0;
	error = ec_start(flags);
	if (!error)
		ec_info.ei_if.if_timer = DOG;	/* turn on watchdog */
	return error;
}

static int				/* return 0 or an errno */
ec_start(int flags)
{
	int i;
	u_int epage, epage2, rindex, tindex;
	struct mbuf *m;

	volatile struct rmd *rmd = rmdvec;
	volatile struct tmd *tmd = tmdvec;
	volatile u_short *rdp = RDP_ADDR;

	/* reset lance */
	ec_info.ei_resets++;
	ECRESET();

	ec_info.ei_sick = 0;		/* stop watchdog resets */
	ec_info.ei_terr_lbolt = 0;	/* force error messages */

	ec_info.ei_ract = 0;		/* start with the 1st buffers */
	ec_info.ei_tact = 0;
	ec_info.ei_tnum = 0;
	ec_info.ei_tfree = 0;
	ec_info.ei_tmapfree = TBUF_INDEX;
	ec_info.ei_tmaplast = TBUF_INDEX;
	ec_info.ei_rxwatch = lbolt + (RX_WATCH_TICKS
				      << ++ec_info.ei_rxwatch_back);

	/*
	 * Map the location of mem.m_ib and the buffer descriptors
	 * for the lance chip.  The ib might cross a page boundary so
	 * map in the following page.
	 */
	epage = btotp(K0_TO_PHYS(&mem.m_ib));
	EMAP[EMAPI(IBMAP_INDEX)] = epage;
	EMAP[EMAPI(IBMAP_INDEX+1)] = epage + 1;

	epage2 = btotp(K1_TO_PHYS(rmd));
	rindex = (epage == epage2) ? IBMAP_INDEX : IBMAP_INDEX + 1;

	epage2 = btotp(K1_TO_PHYS(tmd));
	tindex = (epage == epage2) ? IBMAP_INDEX : IBMAP_INDEX + 1;

	/*
	 * Setup initialization block.
	 */
	lance_init(&ec_info.ei_lif, flags,
		   EADDR(rindex, rmd), EADDR(tindex, tmd),
		   LOG_RMD, LOG_TMD);

	/*
	 * Setup receive descriptors and IOC ethernet address map.
	 */
	bzero((char*)rmd, NUM_RMD * sizeof *rmd);
	for (i = 0; i < NUM_RMD; i++, rmd++) {
		m = r_mbuf_addr[i];
		if (m == 0) {
			m = m_vget(M_DONTWAIT, sizeof(struct rcvbuf),MT_DATA);
			if (m == 0) {
				ether_stop(&ec_info.ei_eif,
					   ec_info.ei_if.if_flags);
				return ENOBUFS;
			}
			r_mbuf_addr[i] = m;
		}

		rmd->rmd_bcnt = -MAX_RPKT;
		INITRCVBUF(&ec_info, rmd, i, m);
	}

	/*
	 * Initialize transmit descriptors.
	 */
	bzero((char*)tmd, NUM_TMD * sizeof *tmd);

	/*
	 * Throw away the pending data if we're resetting the interface.
	 */
	while ((m = t_mbuf_head) != 0) {
		t_mbuf_head = m->m_act;
		m_freem(m);
	}
	t_mbuf_tail = 0;

	for (i = 0; i < NUM_TMD; i++, tmd++) {
		tmd->tmd_stp_enp = STP|ENP;
		tmd->tmd_bcnt = 0xf000;	/* spec says high nibble must be 1s */
	}

	/*
	 * start the chip.
	 */
	*RAP_ADDR = 1;
	*rdp = CSR1;

	*RAP_ADDR = 2;
	*rdp = CSR2;

	*RAP_ADDR = 3;
	*rdp = CSR3;

	*RAP_ADDR = 0;
	*rdp = CSR0_INIT;
	while (!(*rdp & CSR0_IDON))
		/* wait */;

	*rdp = CSR0_IDON;
	*rdp = CSR0_INEA|CSR0_STRT;

	return 0;
}

/*
 * Reset the interface. Must be done at splimp level.
 */
static void
ec_reset(eif)
	struct etherif *eif;
{
	struct mbuf *m;
	int i;

	ec_info.ei_rxwatch_back = 0;
	eiftoifp(eif)->if_timer = 0;	/* turn off watchdog */
	/* free the receive mbufs */
	for (i = 0; i < NUM_RMD; i++) {
		if (0 != (m = r_mbuf_addr[i])) {
			r_mbuf_addr[i] = 0;
			m_free(m);
		}
	}
	/* free the transmit mbufs */
	while ((m = t_mbuf_head) != 0) {
		t_mbuf_head = m->m_act;
		m_freem(m);
	}
	t_mbuf_tail = 0;

	ECRESET();
}

/*
 * Periodically poll the controller for input packets in case
 * an interrupt gets lost.
 */
/* ARGSUSED */
static int
ec_watchdog(unit)
	int unit;
{
	int s;

	s = splimp();
	if (if_ecintr())
		ec_info.ei_lostintrs++;
	if (ec_info.ei_sick > 1) {
		dri_printf("ec0: died and restarted\n");
		(void) ec_init(&ec_info.ei_eif, ec_info.ei_if.if_flags);
		ec_info.ei_if.if_oerrors++;
	} else if (ec_info.ei_tnum != 0) {	/* still have output? */
		ec_info.ei_sick++;		/* yes, we may be sick */
	}

	ec_info.ei_if.if_timer = DOG;
	splx(s);
}

/*
 * Worry more often about the AMD 7990 hardware problem which causes
 * the chip to freeze.  This function is called from places which cause
 * bus errors which can make the chip freeze.
 */
ec_worry()
{
	int s = splimp();
	int worry_time = lbolt + WORRY_TICKS;	/* check for freeze */

	if (worry_time < ec_info.ei_rxwatch)
		ec_info.ei_rxwatch = worry_time;
	splx(s);
}

/*
 * Ethernet interface interrupt.  Come here at splimp().
 */
int
if_ecintr()
{
	int i;
	volatile struct rmd *rmd;
	volatile struct tmd *tmd;
	int found = 0;
	volatile u_short *rdp = RDP_ADDR;

	/* ignore early interrupts */
	if (ec_info.ei_curaddr[0] == 0 && ec_info.ei_curaddr[1] == 0
	    && ec_info.ei_curaddr[2] == 0 && ec_info.ei_curaddr[3] == 0
	    && ec_info.ei_curaddr[4] == 0 && ec_info.ei_curaddr[5] == 0) {
		dri_printf("ec0: early interrupt\n");
		return 1;
	}

	if (iff_dead(ec_info.ei_if.if_flags)) {	/* if down don't process */
		ether_stop(&ec_info.ei_eif, ec_info.ei_if.if_flags);
		dri_printf("ec0: interrupt during down\n");
		return 1;
	}

	for (;;) {
		*RAP_ADDR = 0;
		i = *rdp;

		if (!(i & CSR0_INTR))		/* intr pending? */
			break;

		found = 1;
		ec_info.ei_if.if_timer = DOG;	/* yes, postpone watchdog */

		if (i & (CSR0_RINT|CSR0_TINT)) {	/* note data int. */
			*rdp = CSR0_INEA|CSR0_RINT|CSR0_TINT;
		} else if (i & CSR0_MERR) {
			dri_printf("ec0: memory timeout\n");
			(void) ec_start(ec_info.ei_if.if_flags);
			ec_info.ei_if.if_ierrors++;
		} else if (i & CSR0_BABL) {
			dri_printf("ec0: babbling\n");
			(void) ec_start(ec_info.ei_if.if_flags);
			ec_info.ei_if.if_oerrors++;
		} else if (i & CSR0_MISS) {	/* count dropped packets */
			*rdp = CSR0_INEA|CSR0_MISS;
			ec_info.ei_if.if_idrops++;
			METER(ec_istat_missed++);
		} else if (i & CSR0_IDON) {	/* forget about init. int. */
			*rdp = CSR0_INEA|CSR0_IDON;
		} else {
			dri_printf("ec0: unknown interrupt\n");
			(void) ec_start(ec_info.ei_if.if_flags);
			ec_info.ei_if.if_oerrors++;
		}
	}
	i = ec_info.ei_ract;		/* process received packets */
	rmd = &rmdvec[i];
	while (!rmd->rmd_own) {
		ec_info.ei_rxwatch = lbolt + RX_WATCH_TICKS;
		ec_info.ei_rxwatch_back = 0;
		ec_info.ei_if.if_ipackets++;

		if (rmd->rmd_stp_enp != (STP|ENP)) {
			/* part of a chain of packets */
			if (rmd->rmd_stp_enp == STP) {
				/* count error on first packet */
				ec_info.ei_if.if_ierrors++;
				ec_info.ei_pac_of++;
				ec_print("ec0: rcv packet overflow\n");
			}
			rmd->rmd_stp_enp = STP|ENP;
			rmd->rmd_own = 1;
		} else {
			struct mbuf *n, *m;
			struct mbuf **mp;
			struct ether_header *eh;
			int len;
			int snoopflags = 0;

			if (rmd->rmd_err) {	/* map error state */
				if (rmd->rmd_buff) {
					ec_info.ei_if.if_ierrors++;
					ec_info.ei_rbuf_err++;
					ec_print("ec0: rcv buffer error\n");
					(void)ec_start(ec_info.ei_if.if_flags);
					break;
				}

				snoopflags = SN_ERROR;
				if (rmd->rmd_fram) {
					snoopflags |= SNERR_FRAME;
					METER(ec_istat_frame++);
				}
				if (rmd->rmd_crc)  {
					snoopflags |= SNERR_CHECKSUM;
					METER(ec_istat_crc++);
				}
				if (rmd->rmd_oflo)  {
					snoopflags |= SNERR_OVERFLOW;
					METER(ec_istat_oflo++);
				}
			}

			/*
			 * Allocate a new mbuf for the lance to receive into.
			 * Initialize an ifheader at the front of it.
			 */
			n = m_vget(M_DONTWAIT, sizeof(struct rcvbuf), MT_DATA);
			if (n == 0) {
				if (ec_info.ei_if.if_flags & IFF_DEBUG)
					dri_printf("ec0: out of mbufs\n");
				break;
			}
			mp = &r_mbuf_addr[i];
			m = *mp;
			ASSERT(m);
			*mp = n;

			INITRCVBUF(&ec_info, rmd, i, n);

			/*
			 * Set eh and len and flush any packet cache lines
			 * invalidated by this receive.
			 */
			eh = &mtod(m, struct rcvbuf *)->ebh.ebh_ether;
			len = rmd->rmd_mcnt;
			vflush((caddr_t) eh, len);

			/*
			 * Make sure the packet length is valid.
			 */
			if (len < MIN_RPKT) {
				ec_info.ei_terr_lbolt = 0;
				dri_printf("ec0: packet too small (%d)\n", len);
				snoopflags |= SN_ERROR|SNERR_TOOSMALL;
				if (len < EHDR_LEN + CRC_LEN)
					len = EHDR_LEN + CRC_LEN;
			} else if (len > MAX_RPKT) {
				ec_info.ei_terr_lbolt = 0;
				dri_printf("ec0: packet too big (%d)\n", len);
				snoopflags |= SN_ERROR|SNERR_TOOBIG;
				len = MAX_RPKT;
			}
			ec_info.ei_if.if_ibytes += len;

			/*
			 * Trim off the tail & CRC from the incoming packet.
			 */
			len -= CRC_LEN;
			m_adj(m, len - MAX_RPKT);

			/*
			 * Pass m to the ethernet decapsulator.
			 */
			ether_input(&ec_info.ei_eif, snoopflags, m);
		}

		i = (i+1) % NUM_RMD;
		ec_info.ei_ract = i;
		rmd = &rmdvec[i];
	}

	/*
	 * Process completed transmissions.
	 */
	while (ec_info.ei_tnum) {
		struct mbuf *m;

		tmd = &tmdvec[ec_info.ei_tact];
		if (tmd->tmd_own)
			break;
		ec_info.ei_sick = 0;	/* stop watchdog resets */
		if (tmd->tmd_buff) {
			if (!tmd->tmd_lcol && !tmd->tmd_rtry) {
				ec_info.ei_if.if_oerrors++;
				ec_info.ei_tbuf_err++;
				ec_print("ec0: xmit buffer error\n");
				(void) ec_start(ec_info.ei_if.if_flags);
				break;
			}
		}

		if (tmd->tmd_err) {
			char *terr = 0;

			ec_info.ei_if.if_oerrors++;
			if (tmd->tmd_uflo) {
				ec_info.ei_if.if_oerrors++;
				ec_info.ei_pac_uf++;
				ec_print("ec0: xmit underflow\n");
				(void)ec_start(ec_info.ei_if.if_flags);
				break;
			}
			if (tmd->tmd_rtry) {
				ec_info.ei_if.if_collisions += 16;
				ec_info.ei_tdr += tmd->tmd_tdr;
				ec_info.ei_numtdr++;
			}
			if (tmd->tmd_lcol) {
				terr = "ec0: late collision\n";
				ec_info.ei_late_coll++;
			} else if (tmd->tmd_lcar) {
				/*
				 * Added some code so that we print out this message
				 * separately from the rest of the 'terr' messages.
				 * It's easier to resolve here, and we supposedly don't
				 * fall into the loop below that does dri_printf(terr).
				 */
				terr = "ec0: no carrier: check Ethernet cable\n";
				cmn_err(CE_ALERT, "%s", terr);
				ec_info.ei_terr_lbolt = lbolt+EMSG_DELAY;
				ec_info.ei_terr = terr;
			}
			if (terr != 0
			    && (terr != ec_info.ei_terr
				|| ec_info.ei_terr_lbolt <= lbolt)) {
				dri_printf(terr);
				ec_info.ei_terr_lbolt = lbolt+EMSG_DELAY;
				ec_info.ei_terr = terr;
			}
		} else {
			ec_info.ei_if.if_collisions += tmd->tmd_try;
		}

		if ((tmd->tmd_stp_enp & ENP) && (m = t_mbuf_head) != 0) {
			if (0 == (t_mbuf_head = m->m_act))
				t_mbuf_tail = 0;
			m_freem(m);
		}

		ec_info.ei_tmaplast = EADDR_INVERSE(tmd->tmd_hadr,
						    tmd->tmd_ladr);
		ec_info.ei_tact = (ec_info.ei_tact+1) % NUM_TMD;
		ec_info.ei_tnum--;
	}

	/*
	 * If we have no output in progress, and if we have had no recent
	 * interrupts, then the chip might have been frozen by a bus error.
	 *
	 * This message happens after lboot probes for VME devices and
	 * causes bus errors, which cause the chip to freeze.
	 */
	if (ec_info.ei_tnum == 0
	    && ec_info.ei_rxwatch < lbolt
	    && (iff_alive(ec_info.ei_if.if_flags))) {
		ec_info.ei_frzcount++;
		if (ec_info.ei_if.if_flags & IFF_DEBUG)
			dri_printf("ec0: suspect chip freeze\n");
		(void)ec_start(ec_info.ei_if.if_flags);
	}

	return found;
}

/*
 * Queue the mbuf chain to be transmitted via the lance.
 */
/* ARGSUSED */
static int				/* 0 or errno */
ec_transmit(struct etherif *eif,	/* on this interface */
	struct etheraddr *edst,		/* with these addresses */
	struct etheraddr *esrc,
	u_short type,			/* of this type */
	struct mbuf *m)			/* send this chain */
{
	struct mbuf *n;
	int i, totlen;
	struct ether_header *eh;
	caddr_t cp;
	int len, mapfree;
	u_int eaddr;
	volatile struct tmd *tmd, *first_tmd;
	struct ec_info *ei = &ec_info;

	/*
	 * Make sure there are enough EMAP entries and enough tmds for the
	 * packet.  The number of EMAP entries needed is computed by doubling
	 * the number of mbufs (to cover page boundary crossing) and adding
	 * one (in case wrap the table when needing two maps for mbuf)
	 * and adding another one as sentinel.
	 */
	mapfree = ei->ei_tmapfree;
	i = ei->ei_tmaplast - mapfree;
	if (i <= 0)
		i += TMAP_CNT;
	if (MAX_CHUNKS*2+1+1 > i) {
		if (ei->ei_if.if_flags & IFF_DEBUG)
			dri_printf("ec0: dropped transmit packet--no map\n");
		m_freem(m);
		return ENOBUFS;
	}

	/* Make room for ethernet header.
	 *	In one common case, we will probably use m_pullup() to
	 *	copy 100 bytes from a cluster into a little mbuf, with
	 *	the MAC address.  In the other common case, all of
	 *	the data will be in a single little mbuf, and it will have
	 *	room for the MAC address.
	 *
	 */
	if (M_HASCL(m) || m->m_off <= MMINOFF + SPACE_FOR_EHDR) {
		n = m_get(M_DONTWAIT, MT_DATA);
		if (n == 0) {
			if (ei->ei_if.if_flags & IFF_DEBUG)
				dri_printf("ec0: transmit ehdr m_get failed\n");
			m_freem(m);
			return ENOBUFS;
		}
		n->m_len = SPACE_FOR_EHDR;
		n->m_next = m;
		m = n;
	} else {
		m->m_off -= SPACE_FOR_EHDR;
		m->m_len += SPACE_FOR_EHDR;
	}

	/* Count the number of buffers in the packet.
	 */
	i = 0;
	totlen = -SPACE_FOR_EHDR;
	len = ETHERMTU+1;
	for (n = m; n; n = n->m_next) {
		i++;
		totlen += n->m_len;
		if (len > n->m_len)
			len = n->m_len;
	}

	/*
	 * Force the first transmit chunk to be contiguous.
	 * Also ensure that there are not too many chunks and that no chunk
	 * except the last is too small.  These additional reasons
	 * for byte copying prevent transmitter underruns.
	 * Use m_pullup() when possible to avoid copying 1500 byte packets.
	 */
	if (m->m_next != 0) {
		if (i > MAX_CHUNKS || len < MIN_CHUNK_SIZE) {
			len = totlen+SPACE_FOR_EHDR;
			n = m_vget(M_DONTWAIT, len, MT_DATA);
			if (n == 0) {
				m_freem(m);
				if (ei->ei_if.if_flags & IFF_DEBUG)
					printf("ec0: xmit combine failed\n");
				return ENOBUFS;
			}
			(void)m_datacopy(m, 0,len, mtod(n, caddr_t));
			m_freem(m);
			m = n;
		} else if (m->m_len < MIN_TBUF) {
			m = m_pullup(m, MIN(totlen+SPACE_FOR_EHDR, MIN_TBUF));
			if (m == 0) {
				if (ei->ei_if.if_flags & IFF_DEBUG)
					printf("ec0: xmit mpullup failed\n");
				return ENOBUFS;
			}
		}
	}

	cp = mtod(m, caddr_t);
	eh = RAW_HDR(cp, struct ether_header);
	*(struct etheraddr *)eh->ether_dhost = *edst;
	*(struct etheraddr *)eh->ether_shost = *esrc;
	eh->ether_type = type;

	/* skip over the raw header padding */
	m->m_off += RAW_HDRPAD(EHDR_LEN);
	m->m_len -= RAW_HDRPAD(EHDR_LEN);
	cp += RAW_HDRPAD(EHDR_LEN);

	/*
	 * Check whether snoopers want to copy this packet.
	 */
	if (RAWIF_SNOOPING(&ei->ei_rawif)
	    && snoop_match(&ei->ei_rawif, cp, m->m_len)) {
		ether_selfsnoop(&ei->ei_eif, (struct ether_header *)cp,
				m, sizeof *eh, totlen);
	}

	/* save mbuf chain to release it after the DMA is finished */
	m->m_act = 0;
	if (t_mbuf_tail == 0)
		t_mbuf_head = m;
	else {
		t_mbuf_tail->m_act = m;
	}
	t_mbuf_tail = m;


	/* create the chain of pointers in the Transmit Descriptor ring
	 *
	 * Lie about the length of the 1st buffer to keep the ether happy.
	 *   The first buffer must be at least the minimum packet size.
	 *   The first buffer of a chain must be the minimum LANCE size.
	 */
	if (totlen <= MIN_TPKT_U - EHDR_LEN) {
		/* first & only buffer */
		len = MIN_TPKT_U;
	} else if (totlen <= MIN_TPKT_C - EHDR_LEN) {
		/* this is the first & only buffer */
		len = totlen + EHDR_LEN;
	} else {
		len = m->m_len;
	}
	tmd = first_tmd = &tmdvec[ei->ei_tfree];
	for (;;) {
		/* set up mapping */
		eaddr = EADDR(mapfree, cp);
		if (IS_KSEG2(cp)) {
			EMAP[EMAPI(mapfree)] = kvtokptbl(cp)->pte.pg_pfn;
		} else {
			EMAP[EMAPI(mapfree)] = btotp(K0_TO_PHYS(cp));
		}
		mapfree++;
		if (btoct(cp) != btoct(cp + len-1)) {
			if (IS_KSEG2(cp)) {
			   EMAP[EMAPI(mapfree)]=kvtokptbl(cp+len-1)->pte.pg_pfn;
			} else {
			    EMAP[EMAPI(mapfree)] = btotp(K0_TO_PHYS(cp+len-1));
			}
			mapfree++;
		}

		if (mapfree >= EMAPLEN-1)	/* do not use last map entry */
			mapfree = TBUF_INDEX;	/* to make the loop easier */

		tmd->tmd_ladr = LO_ADDR(eaddr);
		tmd->tmd_hadr = HI_ADDR(eaddr);
		tmd->tmd_stp_enp = 0;
		tmd->tmd_bcnt = -len;

		ei->ei_tnum++;
		ei->ei_tfree = (ei->ei_tfree+1) % NUM_TMD;
		do {
			m = m->m_next;
			if (m == 0)
				goto done;
			len = m->m_len;
		} while (len == 0);

		cp = mtod(m, caddr_t);
		if (++tmd >= &tmdvec[NUM_TMD])
			tmd = &tmdvec[0];
		tmd->tmd_own = 1;		/* set OWN on all but 1st */
	}
done:

	tmd->tmd_stp_enp = ENP;
	first_tmd->tmd_stp_enp |= STP;
	first_tmd->tmd_own = 1;			/* let the LANCE have it */
	*RAP_ADDR = 0;
	*RDP_ADDR = CSR0_INEA|CSR0_TDMD;	/* poke the board */

	ei->ei_tmapfree = mapfree;
	ei->ei_if.if_obytes += totlen+EHDR_LEN;
	return 0;
}

/*
 * attach a single IP6 onboard-ethernet connection.
 */
static struct etherifops ecops =
    { ec_init, ec_reset, ec_watchdog, ec_transmit, lance_ioctl };

/* ARGSUSED */
if_ecedtinit(struct edt *edtp)
{
	int s;
	struct etheraddr ea;

	/*
	 * Of course every IP6 has an ethernet connection,
	 * so no probe is necessary.
	 */

	if (ec_info.ei_resets) {
		dri_printf("ec0: ec_info structure non-zero\n");
		return;
	}

	prom_eaddr(ea.ea_vec);
	if (showconfig) {
		dri_printf("ec0: hardware ethernet address %s\n",
			   ether_sprintf(ea.ea_vec));
	}

	s = splimp();
	if (ea.ea_vec[0] != SGI_ETHER_ADDR) {
		dri_printf("ec0: machine has bad ethernet address: %s\n",
			   ether_sprintf(ea.ea_vec));
		splx(s);
		return;
	}

	ether_attach(&ec_info.ei_eif, "ec", 0, (caddr_t) &ec_info, &ecops, &ea,
		     INV_ETHER_EC, 0);

	ec_info.ei_ib = &mem.m_ib;
	rmdvec = (struct rmd *)K0_TO_K1(mem.m_r.rmd);
	tmdvec = (struct tmd *)K0_TO_K1(mem.m_t.tmd);

	ECRESET();			/* put chip into clean state */
	splx(s);
}

/*
 * Do not complain about DMA errors unless debugging
 */
static void
ec_print(char *msg)
{
	if (ec_info.ei_if.if_flags & IFF_DEBUG
	    && (msg != ec_info.ei_terr
		|| ec_info.ei_terr_lbolt <= lbolt)) {
		dri_printf(msg);
		ec_info.ei_terr_lbolt = lbolt+EMSG_DELAY;
		ec_info.ei_terr = msg;
	}
}


#ifdef NODEF
/*
 * for debugging
 */
static void
tdesc_print(struct tmd *td)
{
	dri_printf("\n\ntmd_ladr = 0x%x  ", td->tmd_ladr);
	dri_printf("tmd_own = 0x%x  ", td->tmd_own);
	dri_printf("tmd_err = 0x%x  ", td->tmd_err);
	dri_printf("tmd_try = 0x%x\n", td->tmd_try);
	dri_printf("tmd_def = 0x%x  ", td->tmd_def);
	dri_printf("tmd_stp_enp = 0x%x  ", td->tmd_stp_enp);
	dri_printf("tmd_hadr = 0x%x  ", td->tmd_hadr);
	dri_printf("tmd_bcnt = 0x%x\n", td->tmd_bcnt);
	dri_printf("tmd_buff = 0x%x  ", td->tmd_buff);
	dri_printf("tmd_uflo = 0x%x  ", td->tmd_uflo);
	dri_printf("tmd_lcol = 0x%x  ", td->tmd_lcol);
	dri_printf("tmd_lcar = 0x%x\n", td->tmd_lcar);
	dri_printf("tmd_rtry = 0x%x  ", td->tmd_rtry);
	dri_printf("tmd_tdr = 0x%x\n", td->tmd_tdr);
}

static void
rdesc_print(struct rmd *rd)
{
	dri_printf("\n\nrmd_ladr = 0x%x  ", rd->rmd_ladr);
	dri_printf("rmd_own = 0x%x  ", rd->rmd_own);
	dri_printf("rmd_err = 0x%x  ", rd->rmd_err);
	dri_printf("rmd_fram = 0x%x\n", rd->rmd_fram);
	dri_printf("rmd_oflo = 0x%x  ", rd->rmd_oflo);
	dri_printf("rmd_crc = 0x%x  ", rd->rmd_crc);
	dri_printf("rmd_buff = 0x%x  ", rd->rmd_buff);
	dri_printf("rmd_stp_enp = 0x%x  ", rd->rmd_stp_enp);
	dri_printf("rmd_hadr = 0x%x  ", rd->rmd_hadr);
	dri_printf("rmd_bcnt = 0x%x\n", rd->rmd_bcnt);
	dri_printf("rmd_mcnt = 0x%x\n", rd->rmd_mcnt);
}
#endif /* DEBUG */

#endif	/* IP6 */
