/*
 * Copyright 1988 Silicon Graphics, Inc.  All rights reserved.
 *
 * Driver for Silicon Graphics Integral Clover2 Ethernet Controller.
 */
#ident "$Revision: 1.45 $"

/*
 * Compile for IP5/IP17 only.
 */
#if defined(IP5) || defined(IP17)

#include "tcp-param.h"
#include "sys/types.h"
#include "sys/cpu.h"
#include "sys/param.h"
#include "sys/systm.h"
#include "sys/sysmacros.h"
#include "sys/cmn_err.h"
#include "sys/debug.h"
#include "sys/edt.h"
#include "sys/pio.h"
#include "sys/errno.h"
#include "sys/immu.h"
#include "sys/pda.h"
#include "sys/invent.h"
#include "sys/kopt.h"
#include "sys/mbuf.h"
#include "sys/sbd.h"
#include "sys/socket.h"
#include "bstring.h"
#include "lance.h"


extern struct ifnet	loif;
extern time_t	lbolt;

/* 
 * counter in microseconds that delays lance interrupts to batch packets
 */
extern int lance_holdoff; 

#define MAX_LANCE_HOLDOFF    15300	 /* maximum lance interrupt delay */
#define DEFAULT_LANCE_HOLDOFF  500       /* default used if no holdoff chosen */


/* Be noisy about some arbitrary things:
 *   0  = report conditions which a user should be
 *	  concerned about.
 *   1  = report conditions which are bad but may
 *	  occur in normal operations.
 *   2  = report error conditions which may happen
 *	  in normal opperations and are recoverable.
 *   3  = report information to help developers debug.
 *
 *   -- when the ifconfig debug option is selected
 *   -- and noise level is at least 2, conditions
 *   -- which are not errors but which are interesting
 *   -- to a developer will be displayed.
 */

#define ET3_DEBUG 0
#if ET3_DEBUG >= 2
# define DBG_PRINT(a)		printf a
#else
# define DBG_PRINT(a)
#endif

/* keep count of errors that don't print a msg. on the console */
u_int	et3_istat_frame;
u_int	et3_istat_crc;
u_int	et3_istat_oflo;
u_int	et3_istat_missed;

/* first byte of SGI's assigned ethernet addresses */
#define SGI_ETHER_ADDR 0x08

#define DOG	(2*IFNET_SLOWHZ)	/* watchdog duration in seconds */
#define	RX_WATCH_TICKS	(31*HZ)

/* XXX move this to some header file, for all drivers */
#define EMSG_DELAY	(HZ*60*1)	/* suppress messages this long */

#define LOG_RMD		6		/* log2 # rcv descriptors */
#define LOG_TMD		7		/* log2 # xmit descriptors */
#define NUM_RMD		(1<<LOG_RMD)	/* # of receive buffers */
#define NUM_TMD		(1<<LOG_TMD)	/* # of transmit buffers */

#define MAXBD	2	/* 2 IO3s can be in one system */

/*
 * Memory for first transmission chunk must be contiguous.  Reserve extra
 * space at the front for the padded header.
 */
#define	MIN_TBUF	(RAW_HDRPAD(EHDR_LEN) + MIN_TPKT_C)
#define	SPACE_FOR_EHDR	(EHDR_LEN + RAW_HDRPAD(EHDR_LEN))

#define IDPROM_LEN	32
/*
 * Shape of I/O space of a single board.
 *
 *	starts at 0xb7f28000 -- F-slot
 *		  0xb7b28000 -- E-slot
 */
typedef volatile struct {
	u_int	io_mrrd;		/* 4: Mapper Ram Read Data */
	u_short io_pad0[0xbffe];	/* 0x40000 */
	u_char  mode0;
	u_char  mode1;
	u_char  mode2;
	u_char  mode3;
	u_int	io_pad[0x7FFF];
	u_int	io_mc1vector;		/* 4: Intvec-0 is ENET's stat */
	u_short io_pad1[0x1fffe];	/* 0x3fffc */
	u_int	io_mrw;			/* Mapper Ram write */
	u_short io_pad2[0x7ffe];	/* 0xfffc */
	u_int	io_mrr;			/* Mapper Ram Read */
	u_short io_pad3[0x7fff];	/* 0xfffe */
	u_short	io_rdp;			/* 0x00002 7990 register data port */
	u_short	io_pad4[3];
	u_short	io_rap;			/* 0x0000a 7990 register address port */
	u_short io_pad5[0x17ffa];	/* 0x2fff6 */
	/* to make struct word aligned */
	struct {			/* 0x30002-0x30102 enet address/stat */
		u_short ad_pad1;
		u_char	ad_byte;	/* enet addr */
		u_char	ad_stat;	/* enet stat */
		u_int	ad_pad2;
	} io_ad[IDPROM_LEN];		/* 1st 6 are enet addr */
	u_short io_pad6[0x3f80];	/* 0x7efe */
	u_char io_lihc;			/* Lance Interrupt Holdoff Counter */
	u_char	io_pad7[0x7fef];
	u_char	io_bid;			/* Board ID: IO[23]_REV[12]_ID */
} *ET3IO;
#define io_stat io_ad[0].ad_stat	/* status of enet */

/*
 * Receive buffer structure and initializer.
 */
struct rcvbuf {
	struct etherbufhead	ebh;
	char			data[MAX_RDAT]; /* 1504 */
};


/*
 * Organize the ether map registers.  2 entries for init block and rec/xmit
 * descriptors.  Receiving only requires one descriptor/buffer since the
 * buffers are always page aligned.  All of the rest are used for xmit
 * entries (which can cross page boundaries).
 */
#define EMAPBASE	0x1e00			/* 0X1E00 - 0X1EFF */
#define EMAPLEN		256			/* # of entries in the emap */
#define RBUF_INDEX	0			/* receive mbufs */
#define TBUF_INDEX	(RBUF_INDEX+NUM_RMD)	/* xmit map entry index */
#define EMAP_TINDEX(x)	((EMAPBASE + TBUF_INDEX + x) << 16)
#define EMAP_RINDEX(x)	((EMAPBASE + RBUF_INDEX + x) << 16)
#define EADDR(index,addr)	\
			(((index)<<IO_PNUMSHFT) | ((u_int)(addr)&IO_POFFMASK))

#define INITRCVBUF(ei, et3rmd, rmdnum, m) {		\
	struct ether_header *ehdr;			\
	struct rcvbuf *rb = mtod(m, struct rcvbuf *);	\
							\
	IF_INITHEADER(rb, &(ei)->ei_if, sizeof rb->ebh); \
	ehdr = (struct ether_header *)(kvtophys(&rb->ebh.ebh_ether)); \
	ei->ei_io->io_mrw = EMAP_RINDEX(rmdnum) + io_btoct(ehdr); \
	ehdr = (struct ether_header *)(EADDR(rmdnum, ehdr)); \
	(et3rmd)->et3rmd_ladr = LO_ADDR(ehdr);		\
	set3rmd_hadr(et3rmd, HI_ADDR(ehdr));		\
}

/*
 * IO3 lance control memory has 8 byte gap between every
 * half words.
 */
#define LANCE_BYTEOFF	8
#define LANCE_HWOFF	(LANCE_BYTEOFF/sizeof(u_short))
typedef volatile struct {
	u_short reg;
	u_short pad[LANCE_HWOFF - 1];
} io3_short;

/* 7990 logical address filter is a bit-set of address hash values */
struct et3lafilter {
	io3_short	laf_vec[4];
};
#define	et3laf_vec(x)	laf_vec[(x)].reg

struct et3ib {
	io3_short	ib_mode;
	io3_short		ib_padr[3];	/* shuffled ethernet address */
	struct et3lafilter	ib_laf;		/* logical address filter */
	io3_short	ib_lrdra;	/* lsb receive ring address */
	union {
		io3_short reg;
		u_short rdata;
		/* NOT USED since et3 must be accessed as 16 bit quatities */
		struct {
			u_short ib_rlen:3;	/* receive ring address */
			u_short ib_rpad:5;
			u_short ib_hrdra:8;	/* msb receive ring address */
		} b;
	} U1;
	io3_short	ib_ltdra;	/* lsb transmit ring address */
	union {
		io3_short reg;
		u_short tdata;
		/* NOT USED since et3 must be accessed as 16 bit quatities */
		struct {
			u_short	ib_tlen:3;	/* transmit ring length */
			u_short ib_tpad:5;
			u_short	ib_htdra:8;	/* msb transmit ring address */
		} b;
	} U2;
};
#define IB_RLEN		0xe000
#define IB_HRDRA	0x00ff
#define IB_TLEN		0xe000
#define IB_HTDRA	0x00ff
#define et3ib_mode	ib_mode.reg

#define	et3ib_lrdra	ib_lrdra.reg
#define set3ib_rlen(x,y) ((x)->U1.rdata = ((x)->U1.rdata & ~IB_RLEN) | (y << 13))
#define set3ib_hrdra(x,y) ((x)->U1.rdata = ((x)->U1.rdata & ~IB_HRDRA) | y)
#define	et3ib_ltdra	ib_ltdra.reg
#define set3ib_tlen(x,y) ((x)->U2.tdata = ((x)->U2.tdata & ~IB_TLEN) | (y << 13))
#define set3ib_htdra(x,y) ((x)->U2.tdata = ((x)->U2.tdata & ~IB_HTDRA) | y)

/* receive descriptors, 8 byte aligned */
struct et3rmd {
	io3_short		rmd_ladr;	/* lo buffer address */
	union {
		io3_short	reg;
		u_short		stat;
		/* NOT USED since et3 must be accessed as 16 bit quatities */
		struct rmdstat {
			u_short rmd_own:1;	/* 0x80 0=owned by cpu,1=7990 */
			u_short rmd_err:1;	/* 0x40 error */
			u_short rmd_fram:1;	/* 0x20 framing error */
			u_short rmd_oflo:1;	/* 0x10 silo overflow */
			u_short rmd_crc:1;	/* 0x08 CRC error */
			u_short rmd_buff:1;	/* 0x04 buffer chaining bug */
			u_short rmd_stp_enp:2;	/* 0x03 start/end of packet */
			u_short rmd_hadr:8;	/* high byte of address */
		} b;
	} U1;
	union {
		io3_short	reg;
		short		rmd_bcnt;	/* negative buffer size */
	} U2;
	io3_short	rmd_mcnt;	/* message byte count */
};
#define RMD_OWN		(0x80 << 8)
#define RMD_ERR		(0x40 << 8)
#define RMD_FRAM	(0x20 << 8)
#define RMD_OFLO	(0x10 << 8)
#define RMD_CRC		(0x08 << 8)
#define RMD_BUFF	(0x04 << 8)
#define RMD_STP_ENP	(0x03 << 8)
#define	et3rmd_ladr	rmd_ladr.reg
#define et3rmd_own(x)	(((x)->U1.stat & RMD_OWN) == RMD_OWN)
#define set3rmd_own(x)	((x)->U1.stat |= RMD_OWN)
#define et3rmd_err(x)	(((x)->U1.stat & RMD_ERR) == RMD_ERR)
#define set3rmd_err(x)	((x)->U1.stat |= RMD_ERR)
#define et3rmd_fram(x)	(((x)->U1.stat & RMD_FRAM) == RMD_FRAM)
#define set3rmd_fram(x)	((x)->U1.stat |= RMD_FRAME)
#define et3rmd_oflo(x)	(((x)->U1.stat & RMD_OFLO) == RMD_OFLO)
#define set3rmd_oflo(x)	((x)->U1.stat |= RMD_OFLO)
#define et3rmd_crc(x)	(((x)->U1.stat & RMD_CRC) == RMD_CRC)
#define set3rmd_crc(x)	((x)->U1.stat |= RMD_CRC)
#define et3rmd_buff(x)	(((x)->U1.stat & RMD_BUFF) == RMD_BUFF)
#define set3rmd_buff(x)	((x)->U1.stat |= RMD_BUFF)
#define et3rmd_stp_enp(x)	(((x)->U1.stat & RMD_STP_ENP) >> 8)
#define set3rmd_stp_enp(x,y)	((x)->U1.stat = ((x)->U1.stat & ~RMD_STP_ENP) | ((y) << 8))
#define et3rmd_hadr(x)	((x)->U1.stat & 0xFF)
#define set3rmd_hadr(x,y)	((x)->U1.stat = ((x)->U1.stat & ~0xFF) | (y))
#define	et3rmd_bcnt	U2.rmd_bcnt
#define	et3rmd_mcnt	rmd_mcnt.reg

/* transmit descriptors, 8 byte aligned */
struct et3tmd {
	io3_short	tmd_ladr;	/* lo buffer address */
	union {
		io3_short reg;
		u_short	stat;
		/* NOT USED since et3 must be accessed as 16 bit quatities */
		struct {
			u_short tmd_own:1;	/*0x80 0=owned by cpu, 1=7990 */
			u_short tmd_err:1;	/*0x40 error */
			u_short tmd_res:1;	/*0x20 (reserved) */
			u_short tmd_try:2;	/*0x18 retries required */
			u_short tmd_def:1;	/*0x04 chip defered to others */
			u_short tmd_stp_enp:2;	/*0x03 start/end of packet */
			u_short tmd_hadr:8;	/*high byte of address */
		} b;
	} U1;
	union {
		io3_short reg;
		short		tmd_bcnt;	/*negative buffer size */
	} U2;
	union {
		io3_short reg;
		u_short	stat;
		/* NOT USED since et3 must be accessed as 16 bit quatities */
		struct {
			u_short tmd_buff:1;	/*0x8000 buffer chaining bug */
			u_short tmd_uflo:1;	/*0x4000 silo underflow */
			u_short tmd_res2:1;	/*0x2000 */
			u_short tmd_lcol:1;	/*0x1000 late collision */
			u_short tmd_lcar:1;	/*0x0800 loss of carrier */
			u_short tmd_rtry:1;	/*0x0400 retried but failed */
			u_short tmd_tdr:10;	/*0x03ff time reflectometry */
		} b;
	} U3;
};
#define TMD_OWN		(0x80 << 8)
#define TMD_ERR		(0x40 << 8)
#define TMD_RES		(0x20 << 8)
#define TMD_TRY		(0x18 << 8)
#define TMD_DEF		(0x04 << 8)
#define TMD_STP_ENP	(0x03 << 8)
#define TMD_BUFF	0x8000
#define TMD_UFLO	0x4000
#define TMD_RES2	0x2000
#define TMD_LCOL	0x1000
#define TMD_LCAR	0x0800
#define TMD_RTRY	0x0400
#define TMD_TDR		0x03ff
#define et3tmd_ladr	tmd_ladr.reg
#define et3tmd_own(x)	(((x)->U1.stat & TMD_OWN) == TMD_OWN)
#define set3tmd_own(x)	((x)->U1.stat |= TMD_OWN)
#define et3tmd_err(x)	(((x)->U1.stat & TMD_ERR) == TMD_ERR)
#define set3tmd_err(x)	((x)->U1.stat |= TMD_ERR)
#define et3tmd_res(x)	(((x)->U1.stat & TMD_RES) == TMD_RES)
#define set3tmd_res(x)	((x)->U1.stat |= TMD_RES)
#define et3tmd_try(x)	(((x)->U1.stat & TMD_TRY) >> (8+3))
#define et3tmd_def(x)	(((x)->U1.stat & TMD_DEF) == TMD_DEF)
#define set3tmd_def(x)	((x)->U1.stat |= TMD_DEF)
#define et3tmd_stp_enp(x)	(((x)->U1.stat & TMD_STP_ENP) >> 8)
#define set3tmd_stp_enp(x,y)	((x)->U1.stat = ((x)->U1.stat & ~TMD_STP_ENP) | ((y) << 8))
#define soret3tmd_stp_enp(x,y)	((x)->U1.stat = (x)->U1.stat | ((y) << 8))
#define et3tmd_hadr(x)	((x)->U1.stat & 0xFF)
#define set3tmd_hadr(x,y)	((x)->U1.stat = ((x)->U1.stat & ~0xFF) | (y))

#define et3tmd_buff(x)	(((x)->U3.stat & TMD_BUFF) == TMD_BUFF)
#define et3tmd_uflo(x)	(((x)->U3.stat & TMD_UFLO) == TMD_UFLO)
#define et3tmd_res2(x)	(((x)->U3.stat & TMD_RES2) == TMD_RES2)
#define et3tmd_lcol(x)	(((x)->U3.stat & TMD_LCOL) == TMD_LCOL)
#define et3tmd_lcar(x)	(((x)->U3.stat & TMD_LCAR) == TMD_LCAR)
#define et3tmd_rtry(x)	(((x)->U3.stat & TMD_RTRY) == TMD_RTRY)
#define et3tmd_tdr(x)	((x)->U3.stat & TMD_TDR)

#define et3tmd_bcnt	U2.tmd_bcnt

typedef volatile struct {
	struct et3ib	et3ib;
	struct et3rmd	et3rmd[NUM_RMD];
	struct et3tmd	et3tmd[NUM_TMD];
} *ET3MEM;

/*
 * Each interface is represented by a network interface structure, et3_info,
 * which the routing code uses to locate the interface.
 */
struct et3_info {
	struct lanceif	ei_lif;		/* common ethernet structures */

	char		*ei_terr;	/* recent transmit error msg */
	time_t		ei_terr_lbolt;	/* when to forget it */

	u_int		ei_ract;	/* next active receive buffer */

	u_int		ei_tact;	/* 1st active transmit buffer */
	u_int		ei_tnum;	/* # of active transmit buffers */
	u_int		ei_tfree;	/* next free transmit buffer */

	u_int		ei_numtdr;	/* # of TDR samples */
	u_long		ei_tdr;		/* sum of all TDR samples */

	u_int		ei_rxwatch;	/* timer for watching receives */
	u_int		ei_frzcount;	/* count suspected chip freezes */
	u_int		ei_pac_of;	/* error information breakdown */
	u_int		ei_rbuf_err;
	u_int		ei_pac_uf;
	u_int		ei_tbuf_err;
	u_int		ei_late_coll;

	ET3IO		ei_io;		/* I/O addresses */
	ET3MEM		ei_mem;		/* start of memory */

	/*
	 * Space to save the addresses of the mbufs that are being transmitted
	 * and received so that they may be subsequently freed.
	 */
	struct mbuf *r_mbuf_addr[NUM_RMD];
	struct mbuf *t_mbuf_head;
	struct mbuf *t_mbuf_tail;

	/*
	 * Copy of ib
	 */
	struct ib	ei_localib;

} et3_info[MAXBD];

#define	ei_eif		ei_lif.lif_eif
#define	ei_ib		ei_lif.lif_ib
#define	ei_laf		ei_lif.lif_laf
#define	ei_ac		ei_eif.eif_arpcom
#define ei_if           ei_ac.ac_if		/* network-visible interface */
#define ei_curaddr	ei_ac.ac_enaddr		/* current ethernet address */
#define	ei_addr		ei_eif.eif_addr
#define	ei_rawif	ei_eif.eif_rawif
#define	ei_lostintrs	ei_eif.eif_lostintrs
#define	ei_resets	ei_eif.eif_resets
#define	ei_sick		ei_eif.eif_sick

#define	ET3_INFO(eif)	((struct et3_info *)(eif)->eif_private)


#define ET3_RESET(io)		{ HWREG_USELOCK(); HWREG_LOCK(); \
				io->mode3 |= ET_NRESET; \
				DELAYBUS(5); \
				io->mode3 &= ~ET_NRESET; \
				DELAYBUS(5); \
				HWREG_UNLOCK(); }
/*
 * IO3 lance memory space
 */
typedef volatile struct {
	struct ib	ib;
	struct rmd	rmd[NUM_RMD];
	struct tmd	tmd[NUM_TMD];
} *LAN_ET3MEM;

/* start board data structures here in the local RAM */
#define BASE	((LAN_ET3MEM)0x800000)

/* 7990 CSR 1 & 2 values */
#define CSR1		LO_ADDR(&BASE->ib)
#define CSR2		HI_ADDR(&BASE->ib)


/*
 * local ib kept for public use.
 * Due to the holes in IO3 mem, it can not be used as is.
 */
static void
lance_cpin(volatile void *src, void *dst, uint len)
{
	void *enddst;

	enddst = (void*)((char*)dst + len);
	while (enddst != dst) {
		*(ushort*)dst = *(ushort*)src;
		dst = (void*)((ushort*)dst + 1);
		src = (volatile void*)((ushort*)src+LANCE_HWOFF);
	}
}

/*
 * lance mem is HW with gaps between each HW.
 */
static void
lance_clearmem(volatile u_short *mem, uint len)
{
	volatile u_short *endmem;

	endmem = (volatile u_short*)((char*)mem + len);
	while (mem != endmem) {
		*mem = 0;
		mem += LANCE_HWOFF;
	}
}

static int
et3_start(struct et3_info *ei, int flags, int canwait)
{
	int i;
	volatile struct et3ib *et3ib;
	volatile struct et3rmd *et3rmd;
	volatile struct et3tmd *et3tmd;
	ET3IO io;
	ET3MEM mem;

	DBG_PRINT(("et%d: start\n", ei->ei_if.if_unit));
	/* reset lance */
	ET3_RESET(ei->ei_io);		/* reset LAN chip */

	ei->ei_sick = 0;		/* stop watchdog resets */
	ei->ei_terr_lbolt = 0;		/* force error messages */

	ei->ei_ract = 0;		/* start with the 1st buffers */
	ei->ei_tact = 0;
	ei->ei_tnum = 0;
	ei->ei_tfree = 0;
	ei->ei_rxwatch = lbolt + RX_WATCH_TICKS;    /* set rx time */

	io = ei->ei_io;
	mem = ei->ei_mem;

	/*
	 * setup initialization block
	 */
	et3ib = &mem->et3ib;
	lance_clearmem((u_short *)et3ib, sizeof(*et3ib));
	if (flags & IFF_PROMISC)
		et3ib->et3ib_mode = IB_PROM;
	et3ib->ib_padr[0].reg = ei->ei_curaddr[1]<<8|ei->ei_curaddr[0];
	et3ib->ib_padr[1].reg = ei->ei_curaddr[3]<<8|ei->ei_curaddr[2];
	et3ib->ib_padr[2].reg = ei->ei_curaddr[5]<<8|ei->ei_curaddr[4];

	DBG_PRINT(("et%d: hardware ethernet address %s\n",
		   ei->ei_if.if_unit, ether_sprintf(ei->ei_curaddr)));

	if (flags & IFF_ALLMULTI) {
		et3ib->ib_laf.et3laf_vec(0) = 0xffff;
		et3ib->ib_laf.et3laf_vec(1) = 0xffff;
		et3ib->ib_laf.et3laf_vec(2) = 0xffff;
		et3ib->ib_laf.et3laf_vec(3) = 0xffff;
	} else {
		et3ib->ib_laf.et3laf_vec(0) = ei->ei_laf.laf_vec[0];
		et3ib->ib_laf.et3laf_vec(1) = ei->ei_laf.laf_vec[1];
		et3ib->ib_laf.et3laf_vec(2) = ei->ei_laf.laf_vec[2];
		et3ib->ib_laf.et3laf_vec(3) = ei->ei_laf.laf_vec[3];
	}

	set3ib_rlen(et3ib, LOG_RMD);
	et3ib->et3ib_lrdra = LO_ADDR(((u_int)&BASE->rmd[0]));
	set3ib_hrdra(et3ib, HI_ADDR(((u_int)&BASE->rmd[0])));
	set3ib_tlen(et3ib, LOG_TMD);
	et3ib->et3ib_ltdra = LO_ADDR(((u_int)&BASE->tmd[0]));
	set3ib_htdra(et3ib, HI_ADDR(((u_int)&BASE->tmd[0])));
	/*
	 * local ib kept for public use.
	 * Due to the holes in IO3 mem, it can not be used as is.
	 */
	lance_cpin(&ei->ei_mem->et3ib, &ei->ei_localib,
		   sizeof(struct ib));

	/*
	 * Setup receive descriptors and IO3 ethernet address map.
	 */
	et3rmd = &mem->et3rmd[0];
	lance_clearmem((u_short *)et3rmd, NUM_RMD*sizeof(*et3rmd));
	for (i = 0; i < NUM_RMD; i++, et3rmd++) {
		struct mbuf *m;

		if (ei->r_mbuf_addr[i] == 0) {
			m = m_vget(canwait, sizeof(struct rcvbuf), MT_DATA);
			if (m == 0) {
				DBG_PRINT(("et%d: no buf for recv(%d)\n",
					ei->ei_if.if_unit, i));
				ether_stop(&ei->ei_eif, ei->ei_if.if_flags);
				return ENOBUFS;
			}
			ei->r_mbuf_addr[i] = m;
		} else
			m = ei->r_mbuf_addr[i];

		INITRCVBUF(ei, et3rmd, i, m);
#ifdef IP17
		/*
		 * write back the cache line for the future
		 * buffer before any data might get clobbered.
		 */
		dki_dcache_wbinval(mtod(m,caddr_t), sizeof(struct rcvbuf));
#endif
		et3rmd->et3rmd_bcnt = -MAX_RPKT;
		set3rmd_own(et3rmd);
	}

	/*
	 * Initialize transmit descriptors.
	 */
	et3tmd = &mem->et3tmd[0];
	lance_clearmem((u_short *)et3tmd, NUM_TMD*sizeof(*et3tmd));
	for (i = 0; i < NUM_TMD; i++, et3tmd++) {
		set3tmd_stp_enp(et3tmd, STP|ENP);
		et3tmd->et3tmd_bcnt = 0xf000;	/* high nibble must be 1s */
	}

	/*
	 * Throw away all of the pending data if we are resetting
	 * the interface.
	 */
	while (ei->t_mbuf_head != 0) {
		struct mbuf *m;

		m = ei->t_mbuf_head;
		ei->t_mbuf_head = m->m_act;
		m_freem(m);
	}
	ei->t_mbuf_tail = 0;

	/*
	 * start the chip.
	 */
	io->io_rap = 1; DELAYBUS(0);
	io->io_rdp = CSR1; DELAYBUS(0);

	io->io_rap = 2; DELAYBUS(0);
	io->io_rdp = CSR2; DELAYBUS(0);

	io->io_rap = 3; DELAYBUS(0);
	io->io_rdp = CSR3; DELAYBUS(0);

	io->io_rap = 0; DELAYBUS(0);
	io->io_rdp = CSR0_INEA|CSR0_INIT|CSR0_STRT;

	return 0;
}


/*
 * Initialize a board.	Interrupts must already be off.
 */
static int				/* 0 or an errno */
et3_init(struct etherif *eif, int flags)
{
	int error;

	error = et3_start(ET3_INFO(eif), flags, M_WAIT);
	if (!error)
		eif->eif_arpcom.ac_if.if_timer = DOG;
	return error;
}


/*
 * Reset the board.
 */
static void
et3_reset(eif)
	struct etherif *eif;
{
	register struct et3_info *ei;
	register int i;

	ei = ET3_INFO(eif);
	ASSERT(ei->ei_io);
	ei->ei_if.if_timer = 0;		/* turn off watchdog */
	DBG_PRINT(("et%d: reset\n", ei->ei_if.if_unit));

	ET3_RESET(ei->ei_io);			/* reset LAN chip */

	/* free the receive mbufs */
	for (i = 0; i < NUM_RMD; i++) {
		if (ei->r_mbuf_addr[i]) {
			m_free(ei->r_mbuf_addr[i]);
			ei->r_mbuf_addr[i] = 0;
		}
	}
	/* free the transmit mbufs */
	while (ei->t_mbuf_head != 0) {
		struct mbuf *m;

		m = ei->t_mbuf_head;
		ei->t_mbuf_head = m->m_act;
		m_freem(m);
	}
	ei->t_mbuf_tail = 0;
}


/*
 * Ethernet interface interrupt.
 */

if_et3intr(int unit)
{
	struct et3_info *ei;
	ET3IO io;
	int found = 0;
	int s = splimp();		

	ASSERT(unit >= 0 && unit < MAXBD);
	ei = &et3_info[unit];
	io = ei->ei_io;
	if (io == 0) {				/* ignore early interrupts */
		printf("et%d: early interrupt\n", unit);
		splx(s);
		return 1;
	}
	IFNET_LOCKNOSPL(&ei->ei_if);
	found = et3_poll(ei, unit);
	IFNET_UNLOCKNOSPL(&ei->ei_if);
	splx(s);
	return(found);
}

et3_poll(struct et3_info *ei, int unit)
{
	ET3MEM mem;
	ET3IO io;
	volatile struct et3rmd *et3rmd;
	volatile struct et3tmd *et3tmd;
	int i;
	int found = 0;
	extern int enetdma_err(int unit);

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));
	mem = ei->ei_mem;
	io = ei->ei_io;

	if (iff_dead(ei->ei_if.if_flags)) {	/* if down don't process */
		ether_stop(&ei->ei_eif, ei->ei_if.if_flags);
		printf("et%d: interrupt during down\n", unit);
		return 1;
	}

	if (enetdma_err(unit)) {
		(void) et3_start(ei, ei->ei_if.if_flags, M_DONTWAIT);
		ei->ei_if.if_oerrors++;
		return(1);
	}

	io->io_rap = 0; DELAYBUS(0);

	while ((i = io->io_rdp) & CSR0_INTR) {
		found = 1;
		ei->ei_if.if_timer = DOG; /* yes, postpone watchdog */

		if (i & (CSR0_RINT|CSR0_TINT)) {	/* note data int. */
			io->io_rdp = CSR0_INEA|CSR0_RINT|CSR0_TINT;
		} else if (i & CSR0_MERR) {
			printf("et%d: memory timeout\n", unit);
			(void) et3_start(ei, ei->ei_if.if_flags, M_DONTWAIT);
			ei->ei_if.if_ierrors++;
			return(1);
		} else if (i & CSR0_BABL) {
			printf("et%d: babbling(%x)\n", unit, i);
			(void) et3_start(ei, ei->ei_if.if_flags, M_DONTWAIT);
			ei->ei_if.if_oerrors++;
			return(1);
		} else if (i & CSR0_MISS) {	/* count dropped packets */
			io->io_rdp = CSR0_INEA|CSR0_MISS;
			ei->ei_if.if_idrops++;
			et3_istat_missed++;
		} else if (i & CSR0_IDON) {	/* forget about init. int. */
			io->io_rdp = CSR0_INEA|CSR0_IDON;
		} else {
			printf("et%d: unknown interrupt\n", unit);
			(void) et3_start(ei, ei->ei_if.if_flags, M_DONTWAIT);
			ei->ei_if.if_oerrors++;
		}
	}


	i = ei->ei_ract;		/* process received packets */
	et3rmd = &mem->et3rmd[i];
	while (!et3rmd_own(et3rmd)) {
		ei->ei_rxwatch = lbolt + RX_WATCH_TICKS; /* set rx time */
		ei->ei_if.if_ipackets++;

		if (et3rmd_stp_enp(et3rmd) != (STP|ENP)) {
			/* part of chain of packets? */
			if (et3rmd_stp_enp(et3rmd) == STP) {
				/* count error on 1st packet */
				ei->ei_terr_lbolt = 0;
				printf("et%d: receive packet overflow\n",
					   ei->ei_if.if_unit);
				ei->ei_if.if_ierrors++;
				ei->ei_pac_of++;
			}
			set3rmd_stp_enp(et3rmd, STP|ENP);
			set3rmd_own(et3rmd);
		} else {
			struct mbuf *n, *m;
			struct mbuf **mp;
			register struct ether_header *eh;
			register int len;
			int snoopflags = 0;

			if (et3rmd_err(et3rmd)) {
				if (et3rmd_buff(et3rmd)) {
					ei->ei_terr_lbolt = 0;
					ei->ei_if.if_ierrors++;
					ei->ei_rbuf_err++;
					printf("et%d: receive buff error\n",
						   ei->ei_if.if_unit);
					(void)et3_start(ei,ei->ei_if.if_flags,
							M_DONTWAIT);
					break;
				}

				snoopflags = SN_ERROR;
				if (et3rmd_fram(et3rmd)) {
					snoopflags |= SNERR_FRAME;
					et3_istat_frame++;
				}
				if (et3rmd_crc(et3rmd))  {
					snoopflags |= SNERR_CHECKSUM;
					et3_istat_crc++;
				}
				if (et3rmd_oflo(et3rmd))  {
					snoopflags |= SNERR_OVERFLOW;
					et3_istat_oflo++;
				}
			}

			/*
			 * Allocate a new mbuf for the lance to receive into.
			 * Initialize an ifheader at the front of it.
			 */
			n = m_vget(M_DONTWAIT, sizeof(struct rcvbuf),MT_DATA);
			if (n == 0) {
				if (ei->ei_if.if_flags & IFF_DEBUG) {
					printf("et%d: out of mbufs\n",unit);
				}
				DBG_PRINT(("et%d: out of mbufs\n",unit));
				break;
			}
			mp = &ei->r_mbuf_addr[i];
			m = *mp;
			ASSERT(m);
			*mp = n;

			eh = &mtod(m, struct rcvbuf *)->ebh.ebh_ether;
			len = et3rmd->et3rmd_mcnt;
			INITRCVBUF(ei, et3rmd, i, n);
#ifdef IP17
			/*
			 * write back the cache line for the future
			 * buffer before any data might get clobbered.
			 */
			dki_dcache_wbinval(mtod(n,caddr_t), sizeof(struct rcvbuf));
#else
			/*
			 * flush any cache lines invalidated by this receive.
			 */
			vflush(eh, len);
#endif

			/*
			 * let the LANCE start playing with the new buffer
			 */
			set3rmd_own(et3rmd);

			/*
			 * Make sure the packet length is valid.
			 */
			len = et3rmd->et3rmd_mcnt;
			if (len < MIN_RPKT) {
				ei->ei_terr_lbolt = 0;
				printf("et%d: packet too small (%d)\n",
					   ei->ei_if.if_unit, len);
				snoopflags |= SNERR_TOOSMALL | SN_ERROR;
				if (len < EHDR_LEN + CRC_LEN)
					len = EHDR_LEN + CRC_LEN;
			} else if (len > MAX_RPKT) {
				ei->ei_terr_lbolt = 0;
				printf("et%d: packet too large (%d)\n",
					   ei->ei_if.if_unit, len);
				snoopflags |= SNERR_TOOBIG | SN_ERROR;
				len = MAX_RPKT;
			}
			ei->ei_if.if_ibytes += len;

			/*
			 * Trim off the tail & CRC from the incoming packet.
			 */
			len -= CRC_LEN;
			m_adj(m, len - MAX_RPKT);
			ether_input(&ei->ei_eif, snoopflags, m);
		}
		i = (i+1) % NUM_RMD;
		ei->ei_ract = i;
		et3rmd = &mem->et3rmd[i];
	}


	/*
	 * process completed transmissions
	 */
	while (ei->ei_tnum != 0) {
		et3tmd = &mem->et3tmd[ei->ei_tact];
		if (et3tmd_own(et3tmd))	/* only completed packets */
			break;

		ei->ei_sick = 0;	/* stop watchdog resets */
		if (et3tmd_buff(et3tmd)) {
			if (!et3tmd_lcol(et3tmd) && !et3tmd_rtry(et3tmd)) {
				ei->ei_terr_lbolt = 0;
				printf("et%d: transmit buff error\n", unit);
				(void)et3_start(ei, ei->ei_if.if_flags,
						M_DONTWAIT);
				ei->ei_if.if_oerrors++;
				ei->ei_tbuf_err++;
				break;
			}
		}

		if (et3tmd_err(et3tmd)) {
			register char *terr = NULL;

			ei->ei_if.if_oerrors++;
			if (et3tmd_uflo(et3tmd)) {
				ei->ei_terr_lbolt = 0;
				printf("et%d: transmit underflow\n", unit);
				(void)et3_start(ei, ei->ei_if.if_flags,
						M_DONTWAIT);
				ei->ei_pac_uf++;
				ei->ei_if.if_oerrors++;
				break;
			}
			if (et3tmd_rtry(et3tmd)) {
				ei->ei_if.if_collisions += 16;
				ei->ei_tdr += et3tmd_tdr(et3tmd);
				ei->ei_numtdr++;
			}
			if (et3tmd_lcol(et3tmd)) {
				terr = "et%d: late collision\n";
				ei->ei_late_coll++;
			} else if (et3tmd_lcar(et3tmd)) {
                /*
                 * Added some code so that we print out this message
                 * separately from the rest of the 'terr' messages.
                 * It's easier to resolve here, and we supposedly don't
                 * fall into the loop below that does printf(terr, unit).
                 */
				terr = "et%d: no carrier: check Ethernet cable\n";
				cmn_err(CE_ALERT, terr, unit);
				ei->ei_terr_lbolt = lbolt+EMSG_DELAY;
				ei->ei_terr = terr;
			}
			if (terr != NULL
			    && (terr != ei->ei_terr
				|| ei->ei_terr_lbolt <= lbolt)) {
				printf(terr, unit);
				ei->ei_terr_lbolt = lbolt+EMSG_DELAY;
				ei->ei_terr = terr;
			}
		} else {
			ei->ei_if.if_collisions += et3tmd_try(et3tmd);
		}

		if ((et3tmd_stp_enp(et3tmd) & ENP) && ei->t_mbuf_head != 0) {
			struct mbuf *m;

			m = ei->t_mbuf_head;
			ei->t_mbuf_head = ei->t_mbuf_head->m_act;
			if (m == ei->t_mbuf_tail)
				ei->t_mbuf_tail = 0;
			m_freem(m);
		}

		ei->ei_tact = (ei->ei_tact+1) % NUM_TMD;
		ei->ei_tnum--;		/* advance to next ring entry */
	}

	io->io_lihc = (u_char)((lance_holdoff + 30)/60);
	return found;
}

/*
 * Periodically poll the controller for input packets in case
 * an interrupt gets lost.
 */
static
et3_watchdog(register int unit)
{
	register struct et3_info *ei = &et3_info[unit];

	ASSERT(IFNET_ISLOCKED(&ei->ei_if));
	if (et3_poll(ei, unit))
		ei->ei_lostintrs++;
	if (ei->ei_sick > 1) {
		printf("et%d: died and restarted\n", unit);
		if ((et3_start(ei, ei->ei_if.if_flags, M_DONTWAIT)) == 0)
			ei->ei_if.if_timer = DOG;

		ei->ei_if.if_oerrors++;
	} else if (ei->ei_tnum != 0) {	/* still have output? */
		ei->ei_sick++;		/* yes, we may be sick */
		ei->ei_if.if_timer = DOG;
	}
}

/*
 * Set up DMA from from mbuf chain.
 */
static int					/* 0 or errno */
et3_transmit(struct etherif *eif,		/* on this interface */
	struct etheraddr *edst,			/* with these addresses */
	struct etheraddr *esrc,
	u_short type,				/* of this type */
	struct mbuf *m)				/* send this chain */
{
	register struct et3_info *ei;
	register ET3MEM mem;
	register volatile struct et3tmd *et3tmd, *first_et3tmd;
	register struct ether_header *eh;
	struct mbuf *n;
	int i, mlen;
	caddr_t cp;
	int len;
	u_int eaddr;

	ei = ET3_INFO(eif);

	/* make room for ethernet header */
	if (M_HASCL(m) || m->m_off <= MMINOFF + SPACE_FOR_EHDR) {
		n = m_get(M_DONTWAIT, MT_DATA);
		if (n == 0) {
			DBG_PRINT(("et%d: transmit dropped packet. no HDR\n",
					ei->ei_if.if_unit));
			m_freem(m);
			return ENOBUFS;
		}
		n->m_len = SPACE_FOR_EHDR;
		n->m_next = m;
		m = n;
	} else {
		m->m_off -= SPACE_FOR_EHDR;
		m->m_len += SPACE_FOR_EHDR;
	}
	/*
	 * Count the number of buffers to go out in the packet.  Conservative
	 * estimate (may be fewer if multiple mbufs span MIN_TPKT_C).
	 */
	i = 0;
	mlen = -SPACE_FOR_EHDR;
	for (n = m; n; n = n->m_next) {
		i++;
		mlen += n->m_len;
	}

	ASSERT(mlen <= MAX_TPKT - EHDR_LEN);

	cp = mtod(m, caddr_t);
	eh = RAW_HDR(cp, struct ether_header);
	*(struct etheraddr *)eh->ether_dhost = *edst;
	*(struct etheraddr *)eh->ether_shost = *esrc;
	eh->ether_type = type;

	/* force the first transmit chunk to be contiguous */
	if (m->m_len < MIN_TBUF && m->m_next != 0) {
		m = m_pullup(m, MIN(mlen+SPACE_FOR_EHDR, MIN_TBUF));
		if (m == 0) {
			DBG_PRINT(("et%d: xmit mpullup failed\n",
				ei->ei_if.if_unit));
			/* mpullup does the free */
			return ENOBUFS;
		}
		cp = mtod(m, caddr_t);
	}

	/*
	 * Make sure there are enough IOC emap entries and
	 * enough tmds to support this packet. The number
	 * of emaps needed is computed by doubling the number
	 * of mbufs (to cover page boundary crossing) and
	 * adding one (in case wrap the table when needing
	 * two maps for mbuf) and adding another one as sentinel.
	 */
	if (ei->ei_tnum + (i*2)+1+1 > NUM_TMD) {
		DBG_PRINT(("et%d: dropped packet no tmd\n",
			   ei->ei_if.if_unit));
		IF_DROP(&ei->ei_if.if_snd);
		m_freem(m);
		return ENOBUFS;
	}
	/* skip over the raw header padding */
	m->m_off += RAW_HDRPAD(EHDR_LEN);
	m->m_len -= RAW_HDRPAD(EHDR_LEN);
	cp += RAW_HDRPAD(EHDR_LEN);

	if ((m->m_len < MIN_TPKT_U) && (m->m_next == 0)) {
		m->m_len = MIN_TPKT_U;
	}

	/*
	 * Check whether snoopers want to copy this packet.
	 */
	if (RAWIF_SNOOPING(&ei->ei_rawif)
	    && snoop_match(&ei->ei_rawif, cp, m->m_len)) {
		ether_selfsnoop(&ei->ei_eif, eh, m, sizeof *eh, mlen);
	}

	/* Save mbuf chain to release it after DMA is finished */
	m->m_act = 0;

	if (ei->t_mbuf_tail == 0)
		ei->t_mbuf_tail = ei->t_mbuf_head = m;
	else {
		ei->t_mbuf_tail->m_act = m;
		ei->t_mbuf_tail = m;
	}

	/* prepare to enter tmd set-up loop */
	mem = ei->ei_mem;
	et3tmd = first_et3tmd = &mem->et3tmd[ei->ei_tfree];
	for (;;) {
		int	singlepage;

		/* check for split page */
		len = m->m_len;
		ASSERT(len > 0);
		if (io_btoct(cp) == io_btoct(cp + len - 1)) {
			singlepage = 1;
		} else {
			singlepage = 0;
			len = IO_NBPP - io_poff(cp);
			m->m_len -= len;
			m->m_off += len;
		}

#ifdef IP17
		/* ensure that the cache has been written back */
		dki_dcache_wb(cp,len);
#endif
		/* set up mapping */
		eaddr = kvtophys(cp);
		ei->ei_io->io_mrw = EMAP_TINDEX(ei->ei_tfree) + io_btoct(eaddr);

		/* set tmd */
		eaddr = EADDR(ei->ei_tfree+TBUF_INDEX, eaddr);
		et3tmd->et3tmd_ladr = LO_ADDR(eaddr);
		set3tmd_hadr(et3tmd, HI_ADDR(eaddr));
		set3tmd_stp_enp(et3tmd, 0);
		et3tmd->et3tmd_bcnt = -len;

		/* prepare for next mbuf */
		ei->ei_tnum++;
		ei->ei_tfree = (ei->ei_tfree+1) % NUM_TMD;
		if (singlepage) {
			do {
				m = m->m_next;
				if (m == 0)
					goto done;
			} while (m->m_len == 0);
		}
		cp = mtod(m, caddr_t);
		if (++et3tmd >= &mem->et3tmd[NUM_TMD])
			et3tmd = &mem->et3tmd[0];
		set3tmd_own(et3tmd); /* set OWN=LANCE on all but first */
	}
done:
	set3tmd_stp_enp(et3tmd, ENP);
	soret3tmd_stp_enp(first_et3tmd, STP);
	set3tmd_own(first_et3tmd);	/* let the LANCE have whole chain */

	ei->ei_io->io_rap = 0; DELAYBUS(0);
	ei->ei_io->io_rdp = CSR0_INEA|CSR0_TDMD;    /* poke the LANCE */

	ei->ei_if.if_obytes += mlen;
	return 0;
}



char	XXXetheraddr[6];
char	XXXetheraddr_initialized;

struct etherifops et3ops = {
	et3_init, et3_reset, et3_watchdog, et3_transmit, lance_ioctl
};


/*
 * check out ET hardware.
 */
static int
et3_diag(int unit, struct et3_info *ei)
{
#	define RAM_SIZE (2*1024)	/* total HW-RAM on a board */
	int i;
	u_short *p;

	p = (u_short *)ei->ei_mem;
	for (i = 0; i < RAM_SIZE/LANCE_HWOFF; i++, p += LANCE_HWOFF) {
		*p = 0x1357;
	}

	p = (u_short *)ei->ei_mem;
	for (i = 0; i < RAM_SIZE/LANCE_HWOFF; i++, p += LANCE_HWOFF) {
		if (*p != 0x1357) {
			printf("et%d: failed IO3 RAM test1\n", unit);
			return(1);
		}
	}

	p = (u_short *)ei->ei_mem;
	for (i = 0; i < RAM_SIZE/LANCE_HWOFF; i++, p += LANCE_HWOFF) {
		*p = (u_short)(~0x1357);
	}

	p = (u_short *)ei->ei_mem;
	for (i = 0; i < RAM_SIZE/LANCE_HWOFF; i++, p += LANCE_HWOFF) {
		if (*p != (u_short)(~0x1357)) {
			printf("et%d: failed IO3 RAM test2\n", unit);
			return(1);
		}
	}

	return(0);
}


/*
 * early init for ethernet
 */
if_et3edtinit(struct edt *edtp)
{
	uint unit;
	struct et3_info *ei;
	struct etheraddr ea;
	int i;
	int s;
	piomap_t *piomap[2];


	/* map I/O address space */

	for (i = 0; i < 2; i++) {
		piomap[i] = pio_mapalloc(edtp->e_bus_type, edtp->e_adap, 
					&edtp->e_space[i], PIOMAP_FIXED, "et3");
		edtp->e_space[i].ios_vaddr = 
			pio_mapaddr(piomap[i], edtp->e_space[i].ios_iopaddr);
	}
	unit = et3unit(&((ET3IO)edtp->e_base2)->mode0);
	if (MAXBD <= unit) {
		printf("et%u: bad IO3 EDT entry\n", unit);
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		return;
	}

	ei = &et3_info[unit];
	s = splimp();
	if (ei->ei_io) {
		printf("et%u: duplicate IO3 EDT entry\n", unit);
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		splx(s);
		return;
	}

	/*
	 * PROBE the IOC version number
	 * Board ID: 2=IO2, 3=IO3
	 */
	if (badaddr(&((ET3IO)edtp->e_base2)->io_bid,
		    sizeof(((ET3IO)edtp->e_base2)->io_bid))
	    || ((ET3IO)edtp->e_base2)->io_bid != IO3_REV1_ID) {
		if (showconfig)
			printf("et%u: IO3 missing\n", unit);
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		splx(s);
		return;
	}

	ei->ei_mem = (ET3MEM)edtp->e_base;
	ei->ei_io = (ET3IO)edtp->e_base2;

	ET3_RESET(ei->ei_io);			/* reset LANCE chip */

	if (et3_diag(unit, ei) != 0) {
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		splx(s);
		return;
	}

	/* get ethernet address from board, 1 byte at a time */
	for (i = 0; i < sizeof ea.ea_vec; i++)
		ea.ea_vec[i] = ei->ei_io->io_ad[ETHERADDRLEN-i-1].ad_byte;

	if (showconfig) {
		printf("et%u: hardware ethernet address %s\n",
			   unit, ether_sprintf(ea.ea_vec));
	}

	if (ea.ea_vec[0] != SGI_ETHER_ADDR) {
		printf("et%u: board has bad ethernet address: %s\n",
			   unit, ether_sprintf(ea.ea_vec));
		pio_mapfree(piomap[0]);
		pio_mapfree(piomap[1]);
		splx(s);
		return;
	}

	if (!XXXetheraddr_initialized) {
		bcopy(ea.ea_vec, XXXetheraddr, sizeof ea.ea_vec);
		XXXetheraddr_initialized = 1;
	}

	DBG_PRINT(("et%u: MBINIT DONE\n", unit));
	ether_attach(&ei->ei_eif, "et", unit, (caddr_t)ei, &et3ops, &ea,
		     INV_ETHER_ET, 3);

	ei->ei_ib = &ei->ei_localib;
	if (lance_holdoff > MAX_LANCE_HOLDOFF)
		lance_holdoff = MAX_LANCE_HOLDOFF;
        else if (lance_holdoff < 0)
                lance_holdoff = DEFAULT_LANCE_HOLDOFF;
	splx(s);
}

#endif /* IP5 || IP17 */
