/*
 * Copyright (c) 1994 Silicon Graphics, Inc.
 *
 * This collection of procedures implement the allocation/free'ing and
 * modification of network buffers called 'mbufs'. The major client of
 * mbufs is the BSD sockets based networking stack.
 *
 * mbufs come from two sources, system virtual memory or network controllers
 * which provide off-board RAM. They come in two major variety and are
 * typically classified as one of the following:
 *
 * 1. regular mbuf which contains up to 'MLEN' bytes of contiguous data.
 * 2. cluster mbuf which is a regular mbuf but contains a pointer to a
 *    buffer 'NBPP' bytes in length.
 * 3. cluster mbuf which is a regular mbuf but contains a pointer to an
 *    externally allocated buffer whose length is stored in the mbuf header
 *
 * All mbufs contain the address'es of procedures which are called when the
 * generic mbuf 'free' and 'dup' functions are called, mfree(), mfreem()
 * and mdup() respectively. Hence the code is rather general in regards to
 * the distinction between cluster mbufs in regards to whom allocated the
 * memory for the data buffer.
 *
 * The allocation and locking scheme in this implementation differ's from
 * previous IRIX releases. The current scheme has been designed to use a
 * set of fast locking procedures which are supported on the R4000 series
 * microprocessors to aid in reducing the locking overhead for MP machines.
 * In the older R3000 series MP special locking hardware was provided so
 * the new scheme provides a general set of entry points which either the
 * new R4000 locking primitives are used if available or the older R3000
 * locking procedures in the R3000 machines. In uniprocessor machines the
 * locking is compiled so that the traditional raising of the processor spl
 * is employed.
 *
 * The new storage allocation scheme has been designed to support a variety
 * of buffer sizes in order to provide a general framework to implement BOTH
 * BSD networking mbufs and System V Streams mblk's.
 *
 * The basic design is to use a set of linked lists, a copy of which is
 * maintained on each CPU. The vectors are stored in a private data structure,
 * the address os which is found in the one page PDA associated with each cpu.
 *
 * The machine specific locking primitives are called by the various
 * procedures when required to manipulate the free list's associated
 * with each buffer size on each processor.
 *
 * The objectives of the new scheme are:
 *
 * 1. Support fast allocation and deallocation of network buffers
 * 2. Provide a common set of locking primitives which can be optimized in
 *    the case of the R4000 or the old spin-locks in R3000 based machines.
 * 3. Support processor cache affinity to avoid excessive secondary cache
 *    thrashing
 * 4. Utilize and maintain a global pool of pages so that any processor
 *    can quickly obtain a page from the global pool to tile for the specific
 *    buffer size when any individual linked-list needs replenishing.
 * 5. Support a general framework to re-combine like sized nodes back into
 *    pages and return the page to the global page pool depending on the
 *    reclaim algorithm(s) deemed appropriate. The goal is to reclaim full
 *    pages used by individual processors and return them to the global page
 *    pool so that the number of pages used by any specific processor
 *    is bounded.
 * 6. Use the normal kernel VM page allocator to obtain kernel pages to tile
 *    to the appropriate size.
 * 7. Support page fliping for the FDDI and ATM network controllers for
 *    both the socket and stream head clients.
 */
#ident "$Revision: 4.67 $"

#include "tcp-param.h"

#include "sys/types.h"
#include "sys/param.h"
#include "sys/debug.h"

#include "sys/immu.h"
#include "sys/pfdat.h"
#include "sys/sysmacros.h"
#include "sys/tcpipstats.h"
#include "listvec.h"
#include "sys/mbuf.h"

#include "sys/proc.h"
#include "sys/user.h"
#include "sys/systm.h"
#include "sys/pda.h"
#include "sys/immu.h"
#include "sys/kmem.h"
#include "sys/sema.h"
#include "sys/cmn_err.h"
#include "sys/sysinfo.h"
#include "ether.h"
#include "bstring.h"

#ifdef EVEREST
extern int	io4ia_war;
#endif

/*
 * External Referenced Procedures
 */
extern int atomicAddInt(int *, int);
extern void *kvpalloc(u_int size, int flags, int color);
extern void kvpfree(void *vaddr, u_int size);
extern caddr_t pattach(caddr_t);

extern int splimp();

/*
 * Forward Referenced Procedures
 */
void mbuf_init_memsize(void);
void mbuf_init_masterlfvec(void);
void m_reclaim(void);

struct lf_elt *lf_addpage(struct lf_listvec *, void *);
struct lf_elt *lf_get(struct lf_listvec *, int);
void lf_free(struct lf_listvec *, struct lf_elt *);

/*
 * External Referenced Variables
 */
extern int nm_clusters;		/* Max num pages for mbufs */
extern int mbuf_max_clfree;	/* Max num free cluster mbufs */

/*
 * Global defined constants
 */
#define	NMBPCL  (CLBYTES/MSIZE)	/* number small mbufs per page */
#define MIN_CLUSTERS	32	/* Minimum number pages for all mbufs */

/*
 * Forward Referenced Procedures
 */
void mbuf_init_lfvec(struct knetvec *);
void mbuf_init_alloc_bufs(struct knetvec *);

struct lf_elt *m_getpage_tiled(int canwait, int nodesize, u_int vmflags);

struct mbuf *m_free(struct mbuf *m);
void m_freem(struct mbuf *m);

void m_freepage(int lfvec_index, caddr_t page);
void m_trim(struct mbuf *m, int len);

/*
 * type and structure definitions local to this module
 *
 * Cluster buffer page information.
 * Supports cluster mbuf reference counting and kernel-to-user
 * address space page flipping.
 *
 * NOTE: This node must be a power of 2 in length due to the constraints
 * of the storage allocator to evenly tile a page with nodes.
 */
struct cpageinfo {
	struct cpageinfo *next;	/* next node in list */
	caddr_t	virtaddr;	/* Kernel virtual address of buffer */
	caddr_t	usr_virtaddr;	/* Virt addr in user's address to map page */
#if (_MIPS_SZLONG == 32)
	u_short	nodesize;	/* cluster size in bytes per buffer in page */
	u_short	refct;		/* reference count for this cluster buffer */
#endif
#if (_MIPS_SZLONG == 64)	/* use int to pad struct size to 32 bytes */
	u_int	nodesize;	/* cluster size in bytes per buffer in page */
	u_int	refct;		/* reference count for this cluster buffer */
#endif
};

/*
 * Constants for various node size's for the vectors associated with each cpu.
 */
#define MBUF_CPI_INDEX		0
#define MBUF_CPI_OFFSET		0
#define	MBUF_CPI_THRESHOLD	(NBPP/sizeof(struct cpageinfo))
#define MBUF_CPI_MIN_PREALLOC	(NBPP/sizeof(struct cpageinfo))

#define MBUF_INDEX		1
#define MBUF_OFFSET		sizeof(struct lf_listvec)*MBUF_INDEX
#define	MBUF_THRESHOLD		((NBPP/MSIZE)/2)
#define MBUF_MIN_PREALLOC	((NBPP/MSIZE) * 2)

#define MBUF_MISC_INDEX		2
#define MBUF_MISC_OFFSET	sizeof(struct lf_listvec)*MBUF_MISC_INDEX
#define	MBUF_MISC_THRESHOLD	((NBPP/MSIZE)/2)
#define MBUF_MISC_MIN_PREALLOC	((NBPP/MSIZE) * 2)

#define MBUFCL_INDEX		3
#define MBUFCL_OFFSET		(sizeof(struct lf_listvec)*MBUFCL_INDEX)
#define	MBUFCL_THRESHOLD	6
#define MBUFCL_MIN_PREALLOC	((NBPP/MBUFCL_SIZE) * 2)

#define MBUFCL_PAGE_INDEX	4
#define MBUFCL_PAGE_OFFSET	(sizeof(struct lf_listvec)*MBUFCL_PAGE_INDEX)
#define	MBUFCL_PAGE_THRESHOLD	3
#define MBUFCL_PAGE_MIN_PREALLOC 2

#define MBUF_INDEXMAX	MBUFCL_PAGE_INDEX + 1

/*
 * Major macro to get the mbuf list vectors local to each cpu
 */
#define MBUF_LOCVEC(x) (((struct knetvec *)(private.knetvecptr))->lfvectors[x])

#ifdef DEBUG
/*
 * Assert that a function address in a cluster mbuf is 'valid'
 */
extern start(), etext();
#define GOODFUNC(func) { ASSERT((((__psunsigned_t)func)&3)==0);	\
	ASSERT((__psunsigned_t)func > (__psunsigned_t)start && \
		(__psunsigned_t)func < (__psunsigned_t)etext); }
#else
#define GOODFUNC(func)

#endif /* DEBUG */

/*
 * Global Vectors for mbuf lists used by all cpu's.
 * In addition we statically allocate a set for the master cpu since it is
 * too early in the startup sequence to call the kernel heap allocator.
 */
struct lf_listvec lf_mastercpu_localvec[MBUF_INDEXMAX]; /* master cpu only */

struct lf_listvec lf_globalvec[MBUF_INDEXMAX]; /* Used by all cpu's */

/*
 * Global variables which control releasing physical pages
 * which have been recycled from the various global mbuf list vectors.
 */
struct lf_elt *lf_reclaim_list;	/* Global list of reclaimed pages */
u_int  lf_nreclaimed;		/* Number pages on global reclaim list */
lock_t lf_reclaim_lock;		/* Global reclaim list spin-lock */

/*
 * Following global variables are the system limits for mbuf page useage.
 *  Notice no attempt is made to balance, compute or maintain accurate counts
 *  for each type of mbuf. We allow the system to dynamically assign them.
 *
 * NOTE: All of the global variables are protected by the reclaim spin lock.
 */
lock_t mbuf_cnt_lock;	/* lock protecting global mbuf counts and variables */
lock_t mbuf_refcnt_lock;/* Global lock for cluster mbuf ref cnt operations */

int mbuf_page_got;	/* Number of pages got from kernel allocator */
int mbuf_page_rel;	/* Number of pages returned to kernel */

int mbuf_page_cur;	/* Current number of pages allocated for all mbufs */
int mbuf_page_min;	/* Mininum number of pages for mbufs (read-only) */
int mbuf_page_max;	/* Maximum number of pages for mbufs (read-only) */

time_t page_keep_lbolt;		/* keep page if >= lbolt */

#define PAGE_KEEP_LBOLT	(HZ*4)	/* Keep new pages at least this num seconds */
#define PAGE_FREE_LBOLT	(HZ*4)	/* Release old pages after this num seconds */

#ifdef DEBUG

struct mbuf_procstats mbuf_procstats;

#define MBUF_CALLSTATS(field) atomicAddInt(((int *)&(mbuf_procstats.field)),1)
#define LV_STAT(x) x++
#define LV_STAT_ATOMIC(x) atomicAddInt((int *)&(x),1)

#else

#define MBUF_CALLSTATS(field)
#define LV_STAT(x)
#define LV_STAT_ATOMIC(x)

#endif /* DEBUG */

#ifdef MBUF_DEBUG
/*
 * mbuf statistics macro's. Although each set of statistics are maintained
 * on a per cpu basis we use and atomic update to protect from interrupts
 * since mbuf's are allocated and free'd by calls originating in interrupt
 * handlers.
 */
#define MT_CPI		14	/* cpi node debug */
#define MT_CLDATA	15	/* cluster debug used to be MT_DN_DRBUF */
#define MT_CLPAGE	16	/* cluster page buffer debug */
#endif

#define MBSTAT_INC(x, c)	\
	atomicAddInt(&(((struct kna *)(private.knaptr))->mbstat.x),c)

#define MBSTAT_DEC(x, c)	\
	atomicAddInt(&(((struct kna *)(private.knaptr))->mbstat.x),-c)

/*
 * flag and semaphore for user process'es waiting for mbuf memory
 */
sema_t mbuf_page_sema;	/* Semaphore for waiting for a returned page */

/*
 * Does our os.a support the new way of having pflip() and pattach() not
 * do a tlb sync? If yes, set this variable to be true. This is a backport
 * from kudzu into banyan and ficus. The changes are important for
 * any process that is sproced/threaded and does page flipping. sosend()
 * and soreceive() have the option, with the new os.a, of not having
 * os.a sync the TLBs after the pageflip. sosend() and soreceive() then
 * take on the responsibility of initiating the sync later, hopefully
 * after they have flipped several additional pages. This is only true,
 * of course, if these changes to bsd.a are accompanied by the
 * appropriately updated os.a, and that's what we check here.
 */
int	os_has_new_pflip = 0;

int
mflipinit()
{
	/* if pflip(1,1) returns 1, then we've got the new os.a support */

	os_has_new_pflip =
		(((size_t)pflip((caddr_t)(size_t)1, (caddr_t)(size_t)1)) == 1);
	return 0;
}


/*
 * This structure provides values for constants that may be used by
 * administrative applications.
 */
struct mbufconst _mbufconst = {
	/* m_msize */		MSIZE,
	/* m_mbufclsize */	MBUFCL_SIZE,
	/* m_mclbytes */	MCLBYTES,
	/* m_mminoff */		MMINOFF,
	/* m_mtail */		MTAIL,
	/* m_maxoff */		MMAXOFF,
	/* m_mlen */		MLEN
};

/*
 * Procedure description:
 *	Return the constant values used by the mbuf system.  Return the
 *	error number, if any.
 */
int
get_mbufconst(void *user_address, int insize, int *outsize)
{
	int error = 0;
	int size = min(insize, sizeof(_mbufconst));
	
	error = copyout(&_mbufconst, user_address, size);
	if (outsize) {
		error = copyout(&size, outsize, sizeof(int));
	}
	return error;
}

/*
 * Procedure Description:
 *   Initialize the mbuf data structures
 *
 *   This procedure is now a historical place holder since lot's of network
 *   device drivers still call it to initialize the mbuf data structures.
 *   This is a historical artifact since the device drivers are now NO longer
 *   responsible for calling the mbuf initialization procedure. The mbuf data
 *   structures are now initialized at cpu startup time.
 */
void
mbinit(void)
{
	return;
}

/*
 * Procedure Description:
 *   Allocate storage for a control block.  This function supports the
 *   replacement of using mbufs for control blocks.  It maintains the
 *   statistics counters though as if mbufs were being allocated for
 *   the user level network statistics program's assume them.
 *
 * Returns:
 *   Void * of Storage allocated for control block
 *
 * Parameters:
 *   canwait => M_WAIT => ok to sleep to get the data block
 *    NOTE: M_WAIT and M_DONTWAIT (the values for "canwait") have exactly
 *    the OPPOSITE sense of KM_SLEEP and KM_NOSLEEP (the values for the
 *    second parameter to kmem_alloc).
 *   size => size in bytes of required data area.
 *   type => Type of intended usage (for maintaining mbuf counts)
 */
void *
mcb_get(int canwait, int size, int type)
{
	void * result;

	if (result = kmem_zalloc(size, !canwait)) {
		MBSTAT_INC(m_mtypes[type], 1);
	}
	return result;
}

/*
 * Procedure Description:
 *   Allocate storage for a control block.  This function supports the
 *   replacement of using mbufs for control blocks.  It maintains the
 *   statistics counters though as if mbufs were being allocated for
 *   the user level network statistics program's assume them.
 *
 * Returns:
 *   Void
 *
 * Parameters:
 *    mcb => Address of storage allocated
 *    size => size in bytes of allocated data area
 *    type => Type of intended usage (for maintaining mbuf counts)
 */
void
mcb_free(void *mcb, int size, int type)
{
	kmem_free(mcb, size);
	MBSTAT_DEC(m_mtypes[type], 1);
	return;
}

/*
 * Procedure Description:
 *  Allocate the initial set of mbuf pages to the regular and cluster mbuf
 *  storage vectors for this cpu.
 *
 * Returns:
 *  Void
 */
void
mbuf_init_alloc_bufs(struct knetvec *knetvecptr)
{
	register struct lf_listvec *q;
	register char *p;
	register u_short i;

	/*
	 * Preallocate some space for cluster page info blocks
	 */
	q = knetvecptr->lfvectors[MBUF_CPI_INDEX];
	ASSERT(q != (struct lf_listvec *)0);

	i = 0;
	while (i < MBUF_CPI_MIN_PREALLOC) {

		if (p = (char *)m_getpage_tiled(M_DONTWAIT,
			sizeof(struct cpageinfo), VM_DIRECT|VM_STALE)) {
			/*
			 * ok got page full of mbuf's so put on free list
			 */
			((struct lf_elt *)(p + NBPP
				- sizeof(struct cpageinfo)))->next =
				  (struct lf_elt *)q->get;

			q->get = p;
			q->getcnt += q->e_per_page;
			i += q->e_per_page;
		} else { /* failed */
			break;
		}
	}

	/*
	 * Preallocate some space for regular mbufs used for network
	 * packet data and packet headers ONLY!
	 */
	q = knetvecptr->lfvectors[MBUF_INDEX];
	ASSERT(q != (struct lf_listvec *)0);

	i = 0;
	while (i < MBUF_MIN_PREALLOC) {

		if (p = (char *)m_getpage_tiled(M_DONTWAIT, MSIZE,
				VM_DIRECT|VM_STALE)) {
			/*
			 * ok got page full of mbuf's so put on free list
			 */
			((struct lf_elt *)(p + NBPP - MSIZE))->next =
				(struct lf_elt *)q->get;

			q->get = p;
			q->getcnt += q->e_per_page;
			i += q->e_per_page;

		} else { /* failed */
			break;
		}
	}

	/*
	 * Preallocate some space for miscellaneous mbuf uses, such as
	 * routing tables, protocol control blocks, sockets, etc.
	 */
	q = knetvecptr->lfvectors[MBUF_MISC_INDEX];
	ASSERT(q != (struct lf_listvec *)0);

	i = 0;
	while (i < MBUF_MISC_MIN_PREALLOC) {

		if (p = (char *)m_getpage_tiled(M_DONTWAIT, MSIZE,
				VM_DIRECT|VM_STALE)) {
			/*
			 * ok got page full of mbuf's so put on free list
			 */
			((struct lf_elt *)(p + NBPP - MSIZE))->next =
				(struct lf_elt *)q->get;

			q->get = p;
			q->getcnt += q->e_per_page;
			i += q->e_per_page;

		} else { /* failed */
			break;
		}
	}

	/*
	 * Preallocate some space for cluster mbufs
	 */
	q = knetvecptr->lfvectors[MBUFCL_INDEX];
	ASSERT(q != (struct lf_listvec *)0);

	i = 0;
	while (i < MBUFCL_MIN_PREALLOC) {

		if (p = (char *)m_getpage_tiled(M_DONTWAIT, MBUFCL_SIZE,
				VM_DIRECT|VM_STALE)) {
			/*
			 * page worth of cluster mbuf's so put on free list
			 */
			((struct lf_elt *)(p + NBPP - MBUFCL_SIZE))->next =
				(struct lf_elt *)q->get;

			q->get = p;
			q->getcnt += q->e_per_page;
			i += q->e_per_page;

		} else { /* failed */
			break;
		}
	}
	return;
}

/*
 * Procedure Description:
 *  Initialize the network buffer data structures for the master cpu.
 *  Called during the initialization of the master cpu AFTER the knetvec
 *  (kernel network vector area) has been allocated. For the master
 *  cpu the kna structure is global since the kernel heap allocator
 *  hasn't been initialized at that time of the call to this procedure.
 *
 * Returns:
 *   Zero => Success
 *   Non-Zero => Failure
 */
/* ARGSUSED */
int
mbuf_init_master(cpuid_t cpu)
{
	register u_short i;

	/* Initialize the mbuf list vectors local to the master cpu */
	mbuf_init_masterlfvec();

	/*
	 * NOTE: We can't allocate a set of tile pages for regular and
	 * cluster mbuf's for the master cpu at this point as the vm
	 * system has not been totally initialized. So we simply initialize
	 * the vector structures themselves.
	 */
	MBUF_LOCVEC(MBUF_CPI_INDEX) = &(lf_mastercpu_localvec[MBUF_CPI_INDEX]);
	MBUF_LOCVEC(MBUF_INDEX) = &(lf_mastercpu_localvec[MBUF_INDEX]);
	MBUF_LOCVEC(MBUF_MISC_INDEX)=&(lf_mastercpu_localvec[MBUF_MISC_INDEX]);

	MBUF_LOCVEC(MBUFCL_INDEX) = &(lf_mastercpu_localvec[MBUFCL_INDEX]);
	MBUF_LOCVEC(MBUFCL_PAGE_INDEX) =
		&(lf_mastercpu_localvec[MBUFCL_PAGE_INDEX]);
	/*
	 * Initialize:
	 * 1. global lock protecting the reclaim list
	 * 2. global mbuf lock protecting the global mbuf counts
	 * 3. global mbuf need a page semaphore
	 * 4. global page reclaim list head
	 * 5. global count of number of pages on reclaim list
	 * 6. global keep page lbolt time.
	 */
	spinlock_init(&lf_reclaim_lock, 0);
	spinlock_init(&mbuf_cnt_lock, 0);
	spinlock_init(&mbuf_refcnt_lock, 0);

#ifdef DEBUG
	bzero(&mbuf_procstats, sizeof(struct mbuf_procstats));
#endif
	initnsema(&mbuf_page_sema, 0, "mbufpg");

	lf_reclaim_list = (struct lf_elt *)0;
	lf_nreclaimed = 0;
	page_keep_lbolt = lbolt;

	/*
	 * Initialize global mbuf list vectors for all sizes used by all cpu's
	 */
	for (i=0; i< MBUF_INDEXMAX; i++) {
		bcopy (&lf_mastercpu_localvec[i], &lf_globalvec[i],
		       sizeof(struct lf_listvec));

		/* Initialize lock address since the copied one's invalid */
		spinlock_init(&(lf_globalvec[i].lock), 0);
	}

	/*
	 * The validation of the mbuf tuning parameters is done in the
	 * startup code for the master cpu when it is initialized. Hence
	 * a call to 'mbuf_init_memsize()' here isn't valid as physical
	 * memory isn't sized until after this procedure is called.
	 */
	return 0;
}

/*
 * Procedure Description:
 *  Initialize the global network buffer parameters specific to the
 *  memory size of the machine. This is called during the initialization
 *  of the master cpu but AFTER the physical memory size of the machine
 *  has been determined.
 *
 *  NOTE: This procedure requires that the global 'physmem' variable
  *  be set to a valid value upon entry.
 *
 * Returns:
 *   Void
 */
void
mbuf_init_memsize(void)
{
	/*
	 * Validate the mbuf tuning parameters potentially
	 * set by the system administrator.
	 */
	if (nm_clusters == 0) {	/* autoconfig to 1/8 of memory */		
		nm_clusters = physmem / 8;
	}

	if (mbuf_max_clfree >= nm_clusters)
		mbuf_max_clfree = nm_clusters - 1;

	/* add one extra page for "small" mbufs and save away */
	mbuf_page_min = mbuf_max_clfree + 1;

	if (mbuf_page_min < MIN_CLUSTERS)
		mbuf_page_min = MIN_CLUSTERS;

	mbuf_page_max = nm_clusters;

	return;
}

/*
 * Procedure Description:
 *  Allocate and initialize the mbuf list vectors local to this cpu.
 *  Called during the initialization of a slave cpu AFTER the knetvec
 *  (kernel network vector area) has been allocated and the wired tlb entry
 *  for the private pda area has been set for this cpu.
 *
 * Returns:
 *  Void
 */
void
mbuf_init_lfvec(struct knetvec *knetvecptr)
{
	/*
	 * Initialize the cluster page info local vectors for this cpu.
	*/
	knetvecptr->lfvectors[MBUF_CPI_INDEX]->threshold = MBUF_CPI_THRESHOLD;
	knetvecptr->lfvectors[MBUF_CPI_INDEX]->size = sizeof(struct cpageinfo);
	knetvecptr->lfvectors[MBUF_CPI_INDEX]->offset = MBUF_CPI_OFFSET;
	knetvecptr->lfvectors[MBUF_CPI_INDEX]->e_per_page =
		NBPP/sizeof(struct cpageinfo);
	spinlock_init(&(knetvecptr->lfvectors[MBUF_CPI_INDEX]->lock), 0);
#ifdef DEBUG
	knetvecptr->lfvectors[MBUF_CPI_INDEX]->node_mask = -sizeof(struct cpageinfo);
#endif

	/*
	 * Initialize regular data mbuf local vectors for this cpu.
	*/
	knetvecptr->lfvectors[MBUF_INDEX]->threshold = MBUF_THRESHOLD;
	knetvecptr->lfvectors[MBUF_INDEX]->size = MSIZE;
	knetvecptr->lfvectors[MBUF_INDEX]->offset = MBUF_OFFSET;
	knetvecptr->lfvectors[MBUF_INDEX]->e_per_page = NBPP/MSIZE;
	spinlock_init(&(knetvecptr->lfvectors[MBUF_INDEX]->lock), 0);
#ifdef DEBUG
	knetvecptr->lfvectors[MBUF_INDEX]->node_mask = -MSIZE;
#endif

	/*
	 * Initialize regular data mbuf local vectors for this cpu.
	*/
	knetvecptr->lfvectors[MBUF_MISC_INDEX]->threshold =MBUF_MISC_THRESHOLD;
	knetvecptr->lfvectors[MBUF_MISC_INDEX]->size = MSIZE;
	knetvecptr->lfvectors[MBUF_MISC_INDEX]->offset = MBUF_MISC_OFFSET;
	knetvecptr->lfvectors[MBUF_MISC_INDEX]->e_per_page = NBPP/MSIZE;
	spinlock_init(&(knetvecptr->lfvectors[MBUF_MISC_INDEX]->lock), 0);
#ifdef DEBUG
	knetvecptr->lfvectors[MBUF_MISC_INDEX]->node_mask = -MSIZE;
#endif

	/*
	 * Initialize cluster mbufs local vectors
	*/
	knetvecptr->lfvectors[MBUFCL_INDEX]->threshold = MBUFCL_THRESHOLD;
	knetvecptr->lfvectors[MBUFCL_INDEX]->size = MBUFCL_SIZE;
	knetvecptr->lfvectors[MBUFCL_INDEX]->offset = MBUFCL_OFFSET;
	knetvecptr->lfvectors[MBUFCL_INDEX]->e_per_page = NBPP/MBUFCL_SIZE;
	spinlock_init(&(knetvecptr->lfvectors[MBUFCL_INDEX]->lock), 0);
#ifdef DEBUG
	knetvecptr->lfvectors[MBUFCL_INDEX]->node_mask = -MBUFCL_SIZE;
#endif

	/*
	 * Initialize page size cluster mbufs local vectors
	*/
	knetvecptr->lfvectors[MBUFCL_PAGE_INDEX]->threshold =
		MBUFCL_PAGE_THRESHOLD;
	knetvecptr->lfvectors[MBUFCL_PAGE_INDEX]->size = NBPP;
	knetvecptr->lfvectors[MBUFCL_PAGE_INDEX]->offset = MBUFCL_PAGE_OFFSET;
	knetvecptr->lfvectors[MBUFCL_PAGE_INDEX]->e_per_page = 1;
	spinlock_init(&(knetvecptr->lfvectors[MBUFCL_PAGE_INDEX]->lock), 0);
#ifdef DEBUG
	knetvecptr->lfvectors[MBUFCL_PAGE_INDEX]->node_mask = -NBPP;
#endif
	return;
}

/*
 * Procedure Description:
 *  Allocate the mbuf list vectors local to this cpu.
 *  Called during the initialization of a slave cpu AFTER the kna
 *  (kernel network activity area) has been allocated and the wired tlb entry
 *  for the private pda area has been set for this cpu.
 *
 * Returns:
 *  Zero => Success
 *  Non-Zero => Failure
 */
int
mbuf_init_alloc_lfvec(struct knetvec *knetvecptr)
{
	register struct lf_listvec *q;
	int err = 0;
	u_short i;

	q = (struct lf_listvec *)kern_calloc(MBUF_INDEXMAX,
			sizeof(struct lf_listvec));
	if (q) {
		bzero(q, sizeof(struct lf_listvec)*MBUF_INDEXMAX);
		for (i=0; i < MBUF_INDEXMAX; i++) {
			knetvecptr->lfvectors[i] = q;
			q++;
		}
	} else {
		err = 1;
	}
	return err;
}

/*
 * Procedure Description:
 *  Allocate and initialize the mbuf list vectors local to this cpu.
 *  Called during the initialization of a slave cpu AFTER the kna
 *  (kernel network activity area) has been allocated and the wired tlb entry
 *  for the private pda area has been set for this cpu.
 *
 * Returns:
 *  Zero => Success
 *  Non-Zero => Failure
 */
void
mbuf_init_masterlfvec(void)
{
	/*
	 * Initialize regular mbufs local vectors for this cpu.
	*/
	bzero(&lf_mastercpu_localvec[MBUF_CPI_INDEX],
		sizeof(struct lf_listvec));
	lf_mastercpu_localvec[MBUF_CPI_INDEX].threshold = MBUF_CPI_THRESHOLD;
	lf_mastercpu_localvec[MBUF_CPI_INDEX].size = sizeof(struct cpageinfo);
	lf_mastercpu_localvec[MBUF_CPI_INDEX].offset = MBUF_CPI_OFFSET;
	lf_mastercpu_localvec[MBUF_CPI_INDEX].e_per_page =
		NBPP/sizeof(struct cpageinfo);
	spinlock_init(&(lf_mastercpu_localvec[MBUF_CPI_INDEX].lock), 0);
#ifdef DEBUG
	lf_mastercpu_localvec[MBUF_CPI_INDEX].node_mask = -sizeof(struct cpageinfo);
#endif

	/*
	 * Initialize regular mbufs local vectors for this cpu.
	*/
	bzero(&lf_mastercpu_localvec[MBUF_INDEX], sizeof(struct lf_listvec));
	lf_mastercpu_localvec[MBUF_INDEX].threshold = MBUF_THRESHOLD;
	lf_mastercpu_localvec[MBUF_INDEX].size = MSIZE;
	lf_mastercpu_localvec[MBUF_INDEX].offset = MBUF_OFFSET;
	lf_mastercpu_localvec[MBUF_INDEX].e_per_page = NBPP/MSIZE;
	spinlock_init(&(lf_mastercpu_localvec[MBUF_INDEX].lock), 0);
#ifdef DEBUG
	lf_mastercpu_localvec[MBUF_INDEX].node_mask = -MSIZE;
#endif

	/*
	 * Initialize regular mbufs local vectors for this cpu.
	*/
	bzero(&lf_mastercpu_localvec[MBUF_MISC_INDEX],
		sizeof(struct lf_listvec));
	lf_mastercpu_localvec[MBUF_MISC_INDEX].threshold = MBUF_MISC_THRESHOLD;
	lf_mastercpu_localvec[MBUF_MISC_INDEX].size = MSIZE;
	lf_mastercpu_localvec[MBUF_MISC_INDEX].offset = MBUF_MISC_OFFSET;
	lf_mastercpu_localvec[MBUF_MISC_INDEX].e_per_page = NBPP/MSIZE;
	spinlock_init(&(lf_mastercpu_localvec[MBUF_MISC_INDEX].lock), 0);
#ifdef DEBUG
	lf_mastercpu_localvec[MBUF_MISC_INDEX].node_mask = -MSIZE;
#endif

	/*
	 * Initialize cluster mbufs local vectors
	*/
	bzero(&lf_mastercpu_localvec[MBUFCL_INDEX], sizeof(struct lf_listvec));
	lf_mastercpu_localvec[MBUFCL_INDEX].threshold = MBUFCL_THRESHOLD;
	lf_mastercpu_localvec[MBUFCL_INDEX].size = MBUFCL_SIZE;
	lf_mastercpu_localvec[MBUFCL_INDEX].offset = MBUFCL_OFFSET;
	lf_mastercpu_localvec[MBUFCL_INDEX].e_per_page = NBPP/MBUFCL_SIZE;
	spinlock_init(&(lf_mastercpu_localvec[MBUFCL_INDEX].lock), 0);
#ifdef DEBUG
	lf_mastercpu_localvec[MBUFCL_INDEX].node_mask = -MBUFCL_SIZE;
#endif

	/*
	 * Initialize page size cluster mbufs local vectors
	*/
	bzero(&lf_mastercpu_localvec[MBUFCL_PAGE_INDEX],
		sizeof(struct lf_listvec));
	lf_mastercpu_localvec[MBUFCL_PAGE_INDEX].threshold =
		MBUFCL_PAGE_THRESHOLD;
	lf_mastercpu_localvec[MBUFCL_PAGE_INDEX].size = NBPP;
	lf_mastercpu_localvec[MBUFCL_PAGE_INDEX].offset = MBUFCL_PAGE_OFFSET;
	lf_mastercpu_localvec[MBUFCL_PAGE_INDEX].e_per_page = 1;
	spinlock_init(&(lf_mastercpu_localvec[MBUFCL_PAGE_INDEX].lock), 0);
#ifdef DEBUG
	lf_mastercpu_localvec[MBUFCL_PAGE_INDEX].node_mask = -NBPP;
#endif

	return;
}

/*
 * Procedure Description:
 *  Initialize the network buffer data structures for a slave cpu.
 *  Called during the initialization of the master cpu AFTER the kna
 *  (kernel network activity area) has been allocated and the wired tlb entry
 *  for the private pda area has been set for this cpu.
 *
 * Returns:
 *  Zero => Success
 *  Non-Zero => Failure
 */
int
mbuf_init_slave(struct knetvec *knetvecptr)
{
	int err;

	/* Allocate the storage for local list vectors for this cpu */
	if (err = mbuf_init_alloc_lfvec(knetvecptr)) {
		cmn_err(CE_PANIC,
			"mbuf_init_slave(): mbuf_init_alloc_lfvec failed\n");
		return err;
	}

	/* Initialize the local list vectors for this cpu */
	mbuf_init_lfvec(knetvecptr);

	/*
	 * Initially allocate a few pages for regular and cluster
	 * mbuf's for this particular cpu.
	 */
	mbuf_init_alloc_bufs(knetvecptr);
	return 0;
}

/*
 * Procedure Description:
 *   Adjust an mbuf by a specified number of bytes starting
 *   from either the head or tail of the mbuf chain.
 *
 * Parameters:
 *   mp => address of mbuf chain which to adjust
 *   if len >= 0 then remove 'len' bytes from the head
 *   if len < 0 then remove 'len' bytes from the tail
 *
 * Returns:
 *   void
 */
void
m_adj(struct mbuf *mp, int len)
{
	struct mbuf *m;

	if ((m = mp) == NULL)
		return;

	GOODMP(m);
	GOODMT(m->m_type);

	MBUF_CALLSTATS(m_adj_calls);

	if (len >= 0) { /* Trim from head */

		MBUF_CALLSTATS(m_adj_head);
		while (m != NULL && len > 0) {
			if (m->m_len <= len) {
				len -= m->m_len;
				m->m_len = 0;
				m = m->m_next;
			} else {
				m->m_len -= len;
				m->m_off += len;
				break;
			}
		}
	} else { /* Trim from tail */

		MBUF_CALLSTATS(m_adj_tail);

		len += m_length(mp);
		/*
		 * The correct length for the chain is "len".
		 * Find the mbuf with last data to retain, adjust its
		 * length and toss data from remaining mbufs on chain.
		 */
		for (m = mp; ; m = m->m_next) {

			len -= m->m_len;

			if (len <= 0) {
				/*
				 * Release entire mbuf chain from
				 * here on if anything is there to release.
				 */
				if (m->m_next != 0) {
					m_freem(m->m_next);
					m->m_next = 0;
				}
				if (len != 0) {
					/*
					 * Adjust length field of the last mbuf
					 */
					m_trim(m, -len);
				}
				return;
			}
		}
	}
}

/*
 * Procedure Description:
 *  Obtain a accurate snap-shot for the number of regular and cluster mbufs
 *  in use by the entire system and return in the mbstat structure passed in.
 *
 * Parameters:
 *  mbstatp => address of mbstat structure to return the current
 *   global mbuf allocation statistics for use by netstat command.
 *
 * Returns:
 *   The m_mbufs, m_clusters and m_clfree fields filled in the
 *   mbstat structure with the global system wide counts.
 *
 * NOTE: The counts returned are "approximate" for everthing but the number
 *       of pages currently allocated, that one is exact. The reason these
 *       counts are approximate as we do NOT maintain separate node counts
 *       for each free list on each cpu. In order to get counts which would
 *       be exact we'd have to walk and count the nodes on each free list.
 */
void
m_allocstats(struct mbstat *mbstatp)
{
	struct kna *knaptr;
	struct knetvec *knetvecptr;
	register struct lf_listvec *mbuf_lfvec;
	register int m_mbufs_alloc, m_mbufs_free, m_clfree;
	register u_short i, j;

	m_mbufs_alloc = m_mbufs_free = m_clfree = 0;

	for (i = 0; i < maxcpus; i++) {

	   if ((pdaindr[i].CpuId == -1) ||
	       !(knaptr = (struct kna *)pdaindr[i].pda->knaptr) ||
	       !(knetvecptr = (struct knetvec *)pdaindr[i].pda->knetvecptr)) {
			continue;
	   }
#ifdef LATER
	   /* add up number of allocated regular mbufs across all cpus */
	   for (j = 0; j < MT_MAX; j++) {
			/*
			 * We skip those slots which just keep counts because
			 * in these instances the kernel allocates heap
			 * memory rather than mbufs when the nodes are too
			 * large for a small mbuf. This covers the cases
			 * when calls to mcb_get() are made to get heap memory
			 * and to update the mbuf counts. Hence to accurate
			 * mbuf memory counts this special case has to be
			 * accounted.
			 * NOTE: In the 'struct tcpcb' MT_PCB case it is too
			 * big to fit into a mbuf so it's yet another special
			 * case where as all the other MT_PCB case fit.
			 */
			switch (j) {
			case MT_FREE:
			case MT_SOCKET:
			case MT_IFADDR:
				m_mbufs_notreally +=knaptr->mbstat.m_mtypes[j];
				break;
			default:
				m_mbufs_alloc += knaptr->mbstat.m_mtypes[j];
				break;
			}
		}
#else
	   /* add up number of allocated regular mbufs across all cpus */
	   for (j = MT_DATA; j < MT_MAX; j++) {
			m_mbufs_alloc += knaptr->mbstat.m_mtypes[j];
	    }
#endif /* LATER */

		/* add up number free regular mbufs on all cpus */
		mbuf_lfvec = knetvecptr->lfvectors[MBUF_INDEX];
		m_mbufs_free += mbuf_lfvec->getcnt;
		m_mbufs_free += mbuf_lfvec->putcnt;

		/* add up number free misc regular mbufs on all cpus */
		mbuf_lfvec = knetvecptr->lfvectors[MBUF_MISC_INDEX];
		m_mbufs_free += mbuf_lfvec->getcnt;
		m_mbufs_free += mbuf_lfvec->putcnt;

		/* add up number cluster mbufs on all cpus */
		mbuf_lfvec = knetvecptr->lfvectors[MBUFCL_INDEX];
		m_clfree += mbuf_lfvec->getcnt;
		m_clfree += mbuf_lfvec->putcnt;
	}

	/*
	 * Compute number of small mbufs allocated by system.
	 * NOTE: This count maybe a little low since we don't have completely
	 * accurate counts for the number of regular sized mbuf's that appear
	 * in the free lists associated with the global vectors for mbufs.
	 * No counts are maintained for the nodes are on the global bristle
	 * lists. This number can be derived from the total number of pages
	 * allocated MINUS those in use MINUS the number of free small mbufs
	 * MINUS the number of pages on the reclaim page list, MINUS the
	 * number of cluster mbufs, etc. Currently we don't compute that value.
	 */
	mbstatp->m_mbufs = m_mbufs_alloc + m_mbufs_free;

	/*
	 * clusters obtained from page pool => we return the
	 * number of currently allocated pages for mbufs.
	 */
	mbstatp->m_clusters = mbuf_page_cur;

	/*
	 * compute number of free clusters. This includes the 2K variety
	 * plus the number of pages on the reclaim free page list.
	 * NOTE: Since cluster mbuf's are NOT page-size we must convert
	 * this back to a page count for the returned mbstat counts.
	 */
	mbstatp->m_clfree = m_clfree/(NBPP/MBUFCL_SIZE) + lf_nreclaimed;

	/* return number of free regular mbufs in system */
	mbstatp->m_mtypes[MT_FREE] = m_mbufs_free;
	return;
}

/*
 * Procedure Description:
 *   Concatentate one chain of mbufs onto another
 *
 * Parameters:
 *   m => address of mbuf chain to append onto (head)
 *   n => address of mbuf chain to append
 *
 * Returns:
 *   Void
 */
void
m_cat(struct mbuf *m, struct mbuf *n)
{

	GOODMP(m);
	GOODMT(m->m_type);

	GOODMP(n);
	GOODMT(n->m_type);

	MBUF_CALLSTATS(m_cat_calls);

	while (m->m_next) {
		m = m->m_next;
	}

	while (n) {
		if (MBUF_IS_CLUSTER(m) ||
			(m->m_off + m->m_len + n->m_len > MMAXOFF)) {
			/*
			 * just join the two chains
			 */
			m->m_next = n;
			return;
		}
		/* copy data from one into the other */
		bcopy(mtod(n, caddr_t), mtod(m, caddr_t) + m->m_len,
			(u_int)n->m_len);
		m->m_len += n->m_len;
		n = m_free(n);
	}
}

/*
 * Procedure Description:
 *   Copy an mbuf chain and the associated data
 *
 * Parameters:
 *   m => address of mbuf chain to copy
 *   off => Offset in bytes in mbuf 'm' at which to start copying
 *   len => Length in bytes to copy
 *
 * Returns:
 *   Address of mbuf chain containing the copy
 */
struct mbuf *
m_copy(struct mbuf *m, int off, int len)
{
	struct mbuf *n, **np, *top;
	int mlen, s_refcnt;

#ifdef DEBUG
	inst_t *callerspc = (inst_t *)__return_address;
#endif
	ASSERT(off >= 0 && len >= 0);
	MBUF_CALLSTATS(m_copy_calls);

	if (len == 0)
		return (0);
	do {
		GOODMP(m);
		GOODMT(m->m_type);

		if (off < m->m_len)
			break;
		off -= m->m_len;
		m = m->m_next;
	} while (off >= 0);		/* skip 0-length mbufs */

	np = &top;
	top = 0;
	do {
		if (m == 0) {
			ASSERT(len == M_COPYALL);
			break;
		}
		if (m->m_len > 0) {
			n = m_get(M_DONTWAIT, m->m_type);

			if (n == (struct mbuf *)0) { /* no mbuf */
				m_freem(top);
				return (0);
			}

			*np = n;
			np = &n->m_next;

			mlen = m->m_len - off;
			if (len != M_COPYALL) {
				if (mlen > len)
					mlen = len;
				len -= mlen;
			}
			n->m_len = mlen;

			/*
			 * Don't know how long this lock is going to be
			 * held, so only acquire at splimp? --yohn
			 */
			s_refcnt = mutex_spinlock_spl(&mbuf_refcnt_lock,
							splimp);

			if (MBUF_IS_CLUSTER(m)) {

				GOODFUNC(m->m_dupfunc);
				n->m_off = (mtod(m,__psunsigned_t) -
						(__psunsigned_t)n) + off;
				n->m_flags = m->m_flags;
				n->m_u.m_us = m->m_u.m_us;
#ifdef DEBUG
				(*n->m_dupfunc)(n, callerspc);
#else
				(*n->m_dupfunc)(n);
#endif
			} else {
				ASSERT(m->m_off >= MMINOFF);
				bcopy(mtod(m, caddr_t)+off,
				      mtod(n, caddr_t), mlen);
			}

			mutex_spinunlock(&mbuf_refcnt_lock, s_refcnt);
			off = 0;
		}
		m = m->m_next;
	} while (len > 0);

	return (top);
}

/*
 * Procedure Description:
 *   Copy data from an mbuf chain starting "off" bytes from the beginning,
 *   continuing for "maxlen" bytes into the supplied buffer address 'cp'.
 *
 * Parameter Description:
 *   m => Address of mbuf chain
 *   off => Number of bytes to skip before starting data copying
 *   maxlen => Number of bytes to copy
 *   cp => Destination buffer address
 *
 * Returns:
 *   Number of bytes copied
 */
int
m_datacopy(struct mbuf *m, int off, int maxlen, caddr_t cp)
{
	int count, done;

	ASSERT(off >= 0 && maxlen >= 0);

	GOODMP(m);
	GOODMT(m->m_type);

	MBUF_CALLSTATS(m_datacopy_calls);

	done = 0;
	while (off > 0) {
		if (m == 0) {
			return done;
		}

		if (off < m->m_len)
			break;
		off -= m->m_len;
		m = m->m_next;
	}

	while (maxlen > 0) {
		if (m == 0) {
			return done;
		}
		count = MIN(m->m_len-off, maxlen);
		if (count != 0) {
			bcopy(mtod(m, char*) + off, cp, count);
			maxlen -= count;
			cp += count;
			done += count;
			off = 0;
		}
		m = m->m_next;
	}
	return done;
}

/*
 * Procedure Description:
 *   Free a mbuf allocated by the mbuf kernel supported allocation procedures.
 *   This procedure ONLY handles cluster mbuf deallocation. It deallocates
 *   only the buffer the mbuf header points to and NOT the mbuf header itself.
 *
 * Parameter Description:
 *   m => address of mbuf chain (cluster mbuf or we don;t get called)
 *   callerspc => Caller's PC address in the DEBUG case
 *
 * Returns:
 *   Void
 */
void
#ifdef DEBUG
_m_freefunction(struct mbuf *m, inst_t *callerspc)
#else
m_freefunction(struct mbuf *m)
#endif
{
	register struct lf_listvec *q_cluster, *q_cpi;
	register struct cpageinfo *cpi;
	int s_refcnt;
	u_short mdata;
#ifdef MBUF_DEBUG
	u_short newtype;
#endif

	MBUF_CALLSTATS(m_freefunction_calls);

	/* claim lock protecting ref cnt operation */
	s_refcnt = mutex_spinlock(&mbuf_refcnt_lock);

	cpi = (struct cpageinfo *)m->m_darg;

	ASSERT((cpi->refct > 0));

	if (cpi->refct == 1) { /* release if last reference */
#ifdef DEBUG
		cpi->refct--;
#endif
		/* release lock protecting ref cnt operation */
		mutex_spinunlock(&mbuf_refcnt_lock, s_refcnt);

		if (MBUF_IS_SHARED(m)) {

			/* turn off the cluster and shared flags */
			m->m_flags &= ~(M_CLUSTER | M_SHARED);

			MBUF_CALLSTATS(m_freefunction_shared);
			/*
			 * Release the mapping of the last virtual address
			 * associated with this cluster buffer.
			 */
			prelease(cpi->virtaddr);
		} else {

#ifdef DEBUG
			if ((m->m_nodesize != MBUFCL_SIZE) &&
			    (m->m_nodesize != NBPP)) {
			cmn_err(CE_PANIC,
		"m_freefunction: Bad node size; mbuf 0x%x, nodesize %d\n",
			  m, m->m_nodesize);
			}
#endif
			if (m->m_nodesize == MBUFCL_SIZE) {

				MBUF_CALLSTATS(m_freefunction_clusters);

				/* turn off the cluster flag */
				m->m_flags &= ~M_CLUSTER;

				/* cluster sized buffers arrive here */
				q_cluster = MBUF_LOCVEC(MBUFCL_INDEX);
				lf_free(q_cluster, (struct lf_elt *)m->m_farg);
#ifdef MBUF_DEBUG
				newtype = MT_CLDATA;
#else
				mdata = MBUFCL_SIZE/MSIZE;
#endif
			} else {	
				/* page sized buffers arrive here */
				MBUF_CALLSTATS(m_freefunction_pages);

				m_freepage(MBUFCL_PAGE_INDEX,
					   (caddr_t)m->m_farg);
#ifdef MBUF_DEBUG
				newtype = MT_CLPAGE;
#else
				mdata = NBPP/MSIZE;
#endif
			}
#ifdef MBUF_DEBUG
			MBSTAT_DEC(m_mtypes[newtype], 1);
#else
			MBSTAT_DEC(m_mtypes[m->m_type], mdata);
#endif
		}
#ifdef DEBUG
		m->m_freefunc = (void(*)()) (NULL+0xfeedbad);
		m->m_dupfunc = (void(*)()) (NULL+0xfeedbad);
		m->m_darg = (long)callerspc;
		m->m_farg = 0xfeeda1f;
		m->m_nodesize = 0;
#endif
		/*
		 * Release the cpi node
		 */
		q_cpi = MBUF_LOCVEC(MBUF_CPI_INDEX);
		lf_free(q_cpi, (struct lf_elt *)cpi);
#ifdef MBUF_DEBUG
		MBSTAT_DEC(m_mtypes[MT_CPI], 1);
#endif
	} else {
		ASSERT((cpi->refct > 1));
		cpi->refct--;

		/* release lock protecting ref cnt operation */
		mutex_spinunlock(&mbuf_refcnt_lock, s_refcnt);

		MBUF_CALLSTATS(m_freefunction_dec);
	}
	return;
}

/*
 * Procedure Description:
 *   Increment the reference count for a fixed-size cluster
 *   An internal procedure and is called only by the simple mbuf functions.
 *
 * Parameter Description:
 *   m => address of mbuf chain (m must be a cluster mbuf type)
 *
 * Returns:
 *   Void
 *
 * NOTE:
 * This procedure assumes that the 'mbuf_refcnt_lock' lock is held on entry.
 */
#ifdef DEBUG
/* ARGSUSED */
void
_m_dupfunction(struct mbuf *m, inst_t *callerspc)
#else
void
m_dupfunction(struct mbuf *m)
#endif
{
	register struct cpageinfo *cpi;

	MBUF_CALLSTATS(m_dupfunction_calls);

	cpi = (struct cpageinfo *)m->m_darg;
	cpi->refct++;
	return;
}

/*
 * Currently defined values are:
 * 0 => do it
 * 1 => fail in fault.c
 * 2 => fail immediately in the m_flip procedure
 */
int m_flip_off = 0;

/*
 * Procedure Description:
 *   Attempt to page flip (ie make Copy-On-Write) an mbuf's data buffer
 *   from the kernel's address space into the user's virtual address space.
 *   This procedure is called from the receive socket system call code,
 *   to flip/remap pages from the kernel's address space into the user's
 *   address space, thus avoiding a data copy from kernel buffers to the
 *   user's address space.
 *
 *   NOTE: The caller MUST has already determined that the lengths and 
 *   target address are suitable for this page oriented operation.
 *
 * Parameter Description:
 *   m => cluster mbuf containing the address of a data buffer to page flip
 *   ubase => Virtual address of user space page to flip
 *
 * Returns:
 *   NULL => Failure; !NULL => Pointer to unfreed old page
 */

caddr_t
m_flip(struct mbuf *m, caddr_t ubase, int do_sync)
{
	register struct lf_listvec *q_cpi;
	register struct cpageinfo *cpi;
	register caddr_t p;
	int s_refcnt;

	GOODMP(m);
	GOODMT(m->m_type);

	MBUF_CALLSTATS(m_flip_calls);

	if (m_flip_off > 1) {
		MBUF_CALLSTATS(m_flip_calls_optoff);
		return NULL;
	}

	/*
	 * The mbuf must satisfy the following:
	 * 1. Must be a cluster mbuf AND
	 * 2. Must not be a page which was obtain via pattach from m_shget AND
	 * 3. Must have a single user (ie reference count of one).
	 * Otherwise we return a failure indication.
	 */
	if (!(MBUF_IS_CLUSTER(m)) || MBUF_IS_SHARED(m)) {
		MBUF_CALLSTATS(m_flip_calls_cluster);
		return NULL;
	}
	/*
	 * fail if more than one user is using the cluster
	 */
	s_refcnt = mutex_spinlock(&mbuf_refcnt_lock);
	cpi = (struct cpageinfo *)m->m_darg;

	if (cpi->refct != 1) {
		mutex_spinunlock(&mbuf_refcnt_lock, s_refcnt);
		MBUF_CALLSTATS(m_flip_calls_refct);
		return NULL;
	}
	mutex_spinunlock(&mbuf_refcnt_lock, s_refcnt);

	/*
	 * pflip takes a kernel vaddr and user virtual address and returns
	 * the virtual address of the target OR zero on failure. On sucess
	 * the kernel page has become part of the user's address space and
	 * the page that may have been there before has been released.
	 *
	 * Upon return the kernel page that was under the cluster mbuf has
	 * been replaced and a new one returned to us.
	 *
	 * The m->m_darg is a pointer to the cluster page info node
	 * which contains the kernel virtual address of the buffer
	 * associated with this cluster.
	 */
	ASSERT(do_sync || os_has_new_pflip);
	if (do_sync)
		p = pflip(cpi->virtaddr, ubase);
	else {
		caddr_t addr = (caddr_t)(((char *)cpi->virtaddr) + 1);
		p = pflip(addr, ubase);
	}

	if (p == 0) { /* failed */
		MBUF_CALLSTATS(m_flip_calls_pflip);
		return NULL;
	}

	/*
	 * Now that the kernel page has been mapped into the user's address
	 * space, we now convert the cluster mbuf into a small mbuf BUT still
	 * retain the original mbuf length so that the upper level socket
	 * code can release and be happy.
	 */
	m->m_off = MMINOFF;
	m->m_flags &= ~M_CLUSTER;

	/*
	 * Call the procedure which decides if the new kernel page just
	 * returned to us gets added to one of our page free lists OR
	 * returned to the kernel vm allocator.
	 *
	 * Another alternative would be to attach to the cluster mbuf header
	 * a page which the paging code could give us to replace the cluster
	 * page we just re-mapped into the user's address space. That
	 * alternative isn't implemented here.
	 */
	if (do_sync)
		m_freepage(MBUFCL_PAGE_INDEX, p);

#ifdef MBUF_DEBUG
	MBSTAT_DEC(m_mtypes[MT_CLPAGE], 1);
#else
	MBSTAT_DEC(m_mtypes[m->m_type], NBPP/MSIZE);
#endif
	/* We now deallocate the cpi node since it's no longer needed */
	q_cpi = MBUF_LOCVEC(MBUF_CPI_INDEX);
	lf_free(q_cpi, (struct lf_elt *)cpi);

#ifdef MBUF_DEBUG
	MBSTAT_DEC(m_mtypes[MT_CPI], 1);
#endif
	MBUF_CALLSTATS(m_flip_calls_ok);

	/* success */
	return p;
}

/*
 * Procedure Description:
 *   Called from soreceive() to synchronize the tlb's after successful
 *   calls to m_flip() before returning from the read to the caller.
 *   The receive socket system call code, which has collected several
 *   flipped pages, passes us an array of all the pointers previously
 *   returned by m_flip() which now need to be freed.
 *
 * Parameter Description:
 *   pages => array of pointers to pages to be freed, or NULL
 *   count => number of addresses in pages, above.
 *
 * Returns:
 *   void
 */

void
m_flip_finished(caddr_t *pages, int count)
{
	ASSERT(os_has_new_pflip);
	
	pflip(NULL, NULL);	/* tell pflip to sync the tlbs now */

	if (pages == NULL)
		return;
	/*
	 * Call the procedure which decides if the new kernel page just
	 * returned to us gets added to one of our page free lists OR
	 * returned to the kernel vm allocator.
	 *
	 * Another alternative would be to attach to the cluster mbuf header
	 * a page which the paging code could give us to replace the cluster
	 * page we just re-mapped into the user's address space. That
	 * alternative isn't implemented here.
	 */

	while (count--)
		m_freepage(MBUFCL_PAGE_INDEX, pages[count]);
}

/*
 * Procedure Description:
 *   Free an mbuf, returning the next mbuf in the m_next chain.
 *
 * Parameters:
 *   m => Address of mbuf chain of which to free first item on list
 *
 * Returns:
 *   Address of remainder of the mbuf chain. Null if there is none.
 */
struct mbuf *
#ifdef DEBUG
_m_free(struct mbuf *m, inst_t *callerspc)
#else
m_free(struct mbuf *m)
#endif
{
	register struct lf_listvec *q_mbuf;
	register struct mbuf *n;
	register u_short index, m_type;

	GOODMP(m);
	GOODMT(m->m_type);

	MBUF_CALLSTATS(m_free_calls);

#ifdef EVEREST
	if (io4ia_war && m->m_len) {
		volatile char c;
		c = *(mtod(m, char*) + (m->m_len - 1));
	}
#endif

	/* save next mbuf in chain */
	n = m->m_next;
	m_type = m->m_type;

#ifdef DEBUG
	if (MBUF_IS_CLUSTER(m)) {

		GOODFUNC(m->m_freefunc);
		MBUF_CALLSTATS(m_free_cluster);
		(*m->m_freefunc)(m, callerspc);

	} else {
		MBUF_CALLSTATS(m_free_mbuf);
	}

	m->m_act = (struct mbuf*)callerspc;
	m->m_type = MT_FREE;
	m->m_next = 0;
#else
	if (MBUF_IS_CLUSTER(m)) {
		(*m->m_freefunc)(m);
	}
#endif

	/* Deallocate regular mbuf to correct free list */
	index = (m_type > MT_HEADER) ? MBUF_MISC_INDEX : MBUF_INDEX;
	q_mbuf = MBUF_LOCVEC(index);

	lf_free(q_mbuf, (struct lf_elt *)m);

	MBSTAT_DEC(m_mtypes[m_type], 1);
	return n;
}

#ifdef DEBUG
struct mbuf *
m_free(struct mbuf *m)
{
	struct mbuf *n;
	n = _m_free(m, (inst_t *)__return_address);
	return n;
}
#endif /* DEBUG */

/*
 * Procedure Description:
 *   Free an entire mbuf chain
 *
 * Parameters:
 *   m => Address of mbuf chain of which to free ENTIRE chain
 *
 * Returns:
 *   Void
 */
void
m_freem(struct mbuf *m)
{
#ifdef DEBUG
	inst_t *callerspc = (inst_t *)__return_address;
#endif
	MBUF_CALLSTATS(m_freem_calls);

	while (m) {
#ifdef DEBUG
		m = _m_free(m, callerspc);
#else
		m = m_free(m);
#endif
	}
}

/*
 * Procedure Description:
 *   Free a page buffer by either returning it to the kernel VM allocator
 *   OR putting it on either the designated list vector free list or the
 *   reclaim page list depending on the 'lfvec_index' parameter.
 *
 * Parameters:
 *   page => Address of page buffer to recycle, add to reclaim list OR
 *   return to kernel VM allocator
 *   lfvec_index => Destination to relaim page list if -1 OR valid lfvec_index
 *
 * Returns:
 *   Void
 */
void m_freepage(int lfvec_index, caddr_t page)
{
	register struct lf_listvec *q_cluster;
	register struct lf_elt *e;
	register int s, s_reclaim;

	MBUF_CALLSTATS(m_freepage_calls);

	s = mutex_spinlock(&mbuf_cnt_lock);

	if ((page_keep_lbolt < lbolt && mbuf_page_cur > mbuf_page_min)
		|| (mbuf_page_cur > mbuf_page_max)) {

		/*
		 * Return this page to system as we're:
		 * 1. Over the minimum page count AND
		 *    the minimum time interval has passed OR
		 * 2. We've exceed the maximum page count
		 */
		mbuf_page_cur--;
		mbuf_page_rel++;

		/* Update system global for # pages networking is using */
		MINFO.bsdnet -= NBPC;

		/* compute next time to allow page release */
		page_keep_lbolt = lbolt + PAGE_FREE_LBOLT;

		/* don't hold mbuf lock across kernel vm call */
		mutex_spinunlock(&mbuf_cnt_lock, s);

		MBUF_CALLSTATS(m_freepage_release);
		kvpfree(page, 1);

	} else {

		mutex_spinunlock(&mbuf_cnt_lock, s);

		if (lfvec_index < 0) { /* add to reclaim free page list */

			s_reclaim = mutex_spinlock(&lf_reclaim_lock);

			e = (struct lf_elt *)page;
			e->next = lf_reclaim_list;

			lf_reclaim_list = e;
			lf_nreclaimed++;

			mutex_spinunlock(&lf_reclaim_lock, s_reclaim);

			MBUF_CALLSTATS(m_freepage_reclaim);

		} else { /* add to designated list vector */

			q_cluster = MBUF_LOCVEC(lfvec_index);
			lf_free(q_cluster, (struct lf_elt *)page);

			MBUF_CALLSTATS(m_freepage_recycle);
		}
	}

	while (cvsema(&mbuf_page_sema)) ;
	return;
}

/*
 * Procedure Description:
 *   The main procedure to allocate a "little" mbuf.
 *   Called from both external procedures and from within this module.
 *
 * Parameter Description:
 *   canwait => flags meaning:
 *		M_WAIT => ok to wait if necessary
 *		M_DONTWAIT => Do NOT wait
 *   type  => Type of usage planned for this mbuf.
 *
 * Returns:
 *   Address of the allocated mbuf
 */
struct mbuf *
m_get(int canwait, int type)
{
	register struct lf_listvec *q_mbuf;
	register struct mbuf *m;
	register u_short index;

	GOODMT(type);
	MBUF_CALLSTATS(m_get_calls);

	index = (type > MT_HEADER) ? MBUF_MISC_INDEX : MBUF_INDEX;

	q_mbuf = MBUF_LOCVEC(index);
	ASSERT((q_mbuf != (struct lf_listvec *)0));

	m = (struct mbuf *)lf_get(q_mbuf, canwait);
	if (m == (struct mbuf *)0) { /* failed */

		m = (struct mbuf *)m_getpage_tiled(canwait, MSIZE,
			VM_DIRECT|VM_STALE);

		if (m) {
			/* add the page safely and return the first node */
			m = (struct mbuf *)lf_addpage(q_mbuf, (void *)m);
		} else {
			LV_STAT_ATOMIC(q_mbuf->get_nopage);

			/*
			 * Failed.  Try to get the protocol stacks to free
			 * up some memory and then try again.
			 */
			m_reclaim();

			m = (struct mbuf *)lf_get(q_mbuf, canwait);
			if (m == (struct mbuf *)0) {	/* failed */
				MBUF_CALLSTATS(m_get_calls_fail);
				return m;
			}
		}

	}
	MBSTAT_INC(m_mtypes[type], 1);
	/*
	 * The initialization of the following fields appear in order
	 * of structure declaration if that helps the code generator/optimizer.
	 */
	m->m_next = 0;
	m->m_off = MMINOFF;
	m->m_len = 0;
	m->m_flags = 0;
	m->m_type = type;
	m->m_act = 0;

	return m;
}

/*
 * Procedure Description:
 *   The main procedure to allocate a "little" mbuf and zero it's data buffer.
 *   Called from both external procedures and from within this module.
 *
 * Parameter Description:
 *   canwait => flags with M_DONTWAIT or M_WAIT
 *   type  => Type of usage planned for this mbuf.
 *
 * Returns:
 *   Address of the allocated mbuf
 */
struct mbuf *
m_getclr(int canwait, int type)
{
	register struct mbuf *m;

	MBUF_CALLSTATS(m_getclr_calls);

	if (m = m_get(canwait, type)) {
		bzero(mtod(m, caddr_t), MLEN);
	}
	return m;
}


/*
 * Procedure Description:
 *   Get a page from the kernel and tile it with equal sized nodes.
 *
 * Parameter Description:
 *   canwait => flags meaning:
 *		M_WAIT => ok to wait if necessary
 *		M_DONTWAIT => Do NOT wait
 *   nodesize  => Size of node in bytes
 *
 * Returns:
 *   Address of the allocated page already tiled.
 */
struct lf_elt *
m_getpage_tiled(int canwait, int nodesize, u_int vmflags)
{
	register struct lf_elt *page, *current;
	register char *next;
	register u_int flags, n;
	register int s;

retry:
	MBUF_CALLSTATS(m_getpage_calls);

	s = mutex_spinlock(&lf_reclaim_lock);

	if ((page = lf_reclaim_list)) {
		/*
		 * Get page from reclaim page free list
		 */
		lf_reclaim_list = lf_reclaim_list->next;
		lf_nreclaimed--;

		mutex_spinunlock(&lf_reclaim_lock, s);

		MBUF_CALLSTATS(m_getpage_calls_reclaim);
	} else {
		mutex_spinunlock(&lf_reclaim_lock, s);

		/*
		 * s = mutex_spinlock(&mbuf_cnt_lock);
		 *
		 * It is silly to acquire this lock, only to release it
		 * after checking mbuf_page_cur and mbuf_page_max -- these
		 * could change between the check and the following update.
		 */

		/*
		 * Check if we're under maximum number of pages allowed
		 */
		if (mbuf_page_cur < mbuf_page_max) { /* ok */

			/* don't hold lock over kernel proc which can sleep */
			/* mutex_spinunlock(&mbuf_cnt_lock, s); not needed */

			flags = vmflags;
			if (!canwait)
				flags |= VM_NOSLEEP;

			/* get a vm page from the kernel */
			page = (struct lf_elt *)kvpalloc(1, (int)flags, 0);

			if (page == (struct lf_elt *)0) { /* failed */
				MBUF_CALLSTATS(m_getpage_calls_failed);
				return page;
			}

			MBUF_CALLSTATS(m_getpage_calls_newpage);

			/* reclaim the mbuf count lock for book keeping */
			s = mutex_spinlock(&mbuf_cnt_lock);

			/* Update system global for # pages networking using */
			MINFO.bsdnet += NBPC;

			/* update number of pages we've ever got for mbufs */
			mbuf_page_got++;

			/* update current number of pages we have for mbufs */
			mbuf_page_cur++;

			/* compute next time to allow pages to be released */
			page_keep_lbolt = lbolt + PAGE_KEEP_LBOLT;

			mutex_spinunlock(&mbuf_cnt_lock, s);

		} else {
			/* mutex_spinunlock(&mbuf_cnt_lock, s); not needed */

			/*
			 * Exceeded maximum number of mbuf pages allowed
			 * Hence wait if allowed until one is freed.
			 */
			if (canwait) {
				MBUF_CALLSTATS(m_getpage_calls_maxwait);
				MBSTAT_INC(m_wait, 1);

				psema(&mbuf_page_sema, PZERO-1);
				goto retry;
			} else {
				MBUF_CALLSTATS(m_getpage_calls_maxnowait);
				MBSTAT_INC(m_drops, 1);
			}

			/* fail if can't wait and over limit */
			return ((struct lf_elt *)0);
		}
	}

	/*
	 * Now tile the page with nodes of the specified size
	 */
	n = (NBPP/nodesize) - 1;
	for (current = page, next = (char *)current; n; n--) {
		next += nodesize;
		current->next = (struct lf_elt *)next;
		current = (struct lf_elt *)next;
	}
	current->next = (struct lf_elt *)0;

	return page;
}

/*
 * Procedure Description:
 *   Get a variable-sized cluster up to a full page in length. In reality we
 *   turn the variable length requests into fixed length of one of three
 *   choices: MLEN, MBUFCL_SIZE or NBPP(entire page).
 *
 * Parameters:
 *   canwait => M_DONTWAIT or M_WAIT
 *   size => desired data buffer size in bytes
 *   type => Type of usage planned for the mbuf
 *
 * Returns:
 *   Address of a mbuf header or ZERO on failure
 */
struct mbuf *
m_vget(int canwait, int size, int type)
{
	register struct lf_listvec *q_cpi, *q_cluster;
	struct lf_listvec *q_mbuf;
	register struct lf_elt *p;
	struct cpageinfo *cpi;
	register struct mbuf *m;
	register u_short index, rsize, msize;
#ifdef MBUF_DEBUG
	u_short newtype;
#endif

	ASSERT(size >= 0 && size <= NBPP);

	MBUF_CALLSTATS(m_vget_calls);

	m = m_get(canwait, type);

	if (m == (struct mbuf *)0) {
		MBUF_CALLSTATS(m_vget_nombuf);
		return m;
	}

	m->m_len = size;
	if (size <= MLEN) {
		MBUF_CALLSTATS(m_vget_mbuf);
		return m;
	}

	q_mbuf = MBUF_LOCVEC(MBUF_INDEX);
	ASSERT(q_mbuf != (struct lf_listvec *)0);

	q_cpi = MBUF_LOCVEC(MBUF_CPI_INDEX);
	ASSERT(q_cpi != (struct lf_listvec *)0);

	if (!(cpi = (struct cpageinfo *)lf_get(q_cpi, canwait))) { /* fail */

		cpi = (struct cpageinfo *)m_getpage_tiled(canwait,
				sizeof(struct cpageinfo),VM_DIRECT|VM_STALE);

		if (cpi) { /* ok got a page of nodes */

			/* add the page safely and return the first node */
			cpi = (struct cpageinfo *)lf_addpage(q_cpi,
				(void *)cpi);
			cpi->next = (struct cpageinfo *)0;
		} else {

			LV_STAT_ATOMIC(q_cpi->get_nopage);

			lf_free(q_mbuf, (struct lf_elt *)m);

			/* account for free'ing the mbuf in the statistics */
			MBSTAT_DEC(m_mtypes[type], 1);

			MBUF_CALLSTATS(m_vget_nocpi);
			return ((struct mbuf *)0);
		}
	}

#ifdef MBUF_DEBUG
	MBSTAT_INC(m_mtypes[MT_CPI], 1);
#endif
	if (size <= MBUFCL_SIZE) {
		index = MBUFCL_INDEX;
		rsize = MBUFCL_SIZE;
#ifdef MBUF_DEBUG
		newtype = MT_CLDATA;
#else
		msize = MBUFCL_SIZE/MSIZE;
#endif
	} else {
		index = MBUFCL_PAGE_INDEX;
		rsize = NBPP;
#ifdef MBUF_DEBUG
		newtype = MT_CLPAGE;
#else
		msize = NBPP/MSIZE;
#endif
	}

	q_cluster = MBUF_LOCVEC(index);
	ASSERT(q_cluster != (struct lf_listvec *)0);

	if (!(p = lf_get(q_cluster, canwait))) { /* failed */
		/*
		 * Now try and allocate a new page
		 */
		if ((p = m_getpage_tiled(canwait, rsize, VM_STALE))) { /* ok */

			if (rsize < NBPP) {
				/*
				 * Node size less than an entire page so add
				 * the page of linked nodes safely to the get
				 * list and return the first node.
				 */
				p = lf_addpage(q_cluster, (void *)p);
			}
		} else {
			LV_STAT_ATOMIC(q_mbuf->get_nopage);

			/*
			 * Failed.  Try to get the protocol stacks to give up
			 * some memory.
			 */
			m_reclaim();
			if ((p = lf_get(q_cluster, canwait))) { /* ok */
				goto doit;
			}

			/* failed so cleanup by releasing headers */
			lf_free(q_mbuf, (struct lf_elt *)m);

			/* account for free'ing the mbuf in the statistics */
			MBSTAT_DEC(m_mtypes[type], 1);

			lf_free(q_cpi, (struct lf_elt *)cpi);
#ifdef MBUF_DEBUG
			MBSTAT_DEC(m_mtypes[MT_CPI], 1);
#endif
			MBUF_CALLSTATS(m_vget_nobuf);
			return ((struct mbuf *)0);
		}	
	}

doit:
	m->m_flags = M_CLUSTER;

	/* set address of buffer relative from the mbuf header */
	m->m_off = (__psunsigned_t)p - (__psunsigned_t)m;

#ifdef DEBUG
	m->m_freefunc = _m_freefunction;
	m->m_dupfunc = _m_dupfunction;
#else
	m->m_freefunc = m_freefunction;
	m->m_dupfunc = m_dupfunction;
#endif

#ifdef MBUF_DEBUG
	/* update number regular mbuf's used for the cluster buffer */
	MBSTAT_INC(m_mtypes[newtype], 1);
#else
	/* update number regular mbuf's used for the cluster buffer */
	MBSTAT_INC(m_mtypes[type], msize);
#endif
	m->m_farg = (long)p;
	m->m_darg = (long)cpi;

	/* save the actual size of buffer allocated to the cluster */
	m->m_nodesize = rsize;

	/* now fill in the cluster page info */
	cpi->virtaddr = (caddr_t)p;
	cpi->usr_virtaddr = 0;
	cpi->nodesize = rsize;
	cpi->refct = 1;

#ifdef DEBUG
	if (rsize == NBPP) {
		MBUF_CALLSTATS(m_vget_page);
	} else {
		MBUF_CALLSTATS(m_vget_cluster);
	}
#endif
	return m;
}

#ifdef DEBUG
/*
 * Following Copy-on-Write specific variable is used only when compiled in
 * -DDEBUG mode.
 *
 * Currently defined values are:
 * 0 => do it
 * 1 => fail immediately
 */
int m_shget_off = 0;
#endif

/*
 * Procedure Description:
 *
 *   Take a user virtual address and a length, and convert them into
 *   'shared cluster'.  Such clusters are user address space pages which
 *   have been marked Copy-on-Write if the kernel ever decides to write
 *   into the page. This procedure is called from the socket send system
 *   call code to avoid the copy in operation.
 *
 * Parameters:
 *   canwait => wait flags => M_DONTWAIT or M_WAIT
 *   type => Type of usage planned for mbuf
 *   vaddr => user virtual address
 *   size => length in bytes MUST be <= 1 page 
 *
 * Returns:
 *   Address of mbuf obtained; 0 => Failure
 */
struct mbuf *
m_shget(int canwait, int type, caddr_t vaddr, int size, int do_sync)
{
	register struct lf_listvec *q_cpi, *q_mbuf;
	register struct lf_elt *p;
	register struct cpageinfo *cpi;
	register struct mbuf *m;

#ifdef DEBUG
	MBUF_CALLSTATS(m_shget_calls);

	ASSERT(size >= 0 && size <= NBPP);

	if (m_shget_off > 0) { /* fail immediately */
		MBUF_CALLSTATS(m_shget_calls_optoff);
		return ((struct mbuf *)0);
	}
#endif
	m = m_get(canwait, type);
	if (m == (struct mbuf *)0) {
		MBUF_CALLSTATS(m_shget_calls_nombuf);
		return ((struct mbuf *)0);
	}

	q_mbuf = MBUF_LOCVEC(MBUF_INDEX);
	ASSERT(q_mbuf != (struct lf_listvec *)0);

	q_cpi = MBUF_LOCVEC(MBUF_CPI_INDEX);
	ASSERT(q_cpi != (struct lf_listvec *)0);

	if (!(cpi = (struct cpageinfo *)lf_get(q_cpi, canwait))) { /* failed */

		cpi = (struct cpageinfo *)m_getpage_tiled(canwait,
			sizeof(struct cpageinfo),VM_DIRECT|VM_STALE);

		if (cpi) { /* ok got a page of nodes */

			/* add the page safely and return the first one */
			cpi =(struct cpageinfo *)lf_addpage(q_cpi,(void *)cpi);
			cpi->next = (struct cpageinfo *)0;
		} else {

			LV_STAT_ATOMIC(q_mbuf->get_nopage);

			/* release mbuf header */
			lf_free(q_mbuf, (struct lf_elt *)m);

			/* account for free'ing the mbuf in the statistics */
			MBSTAT_DEC(m_mtypes[type], 1);

			MBUF_CALLSTATS(m_shget_calls_nocpi);
			return ((struct mbuf *)0);
		}
	}

#ifdef MBUF_DEBUG
	MBSTAT_INC(m_mtypes[MT_CPI], 1);
#endif
	/*
	 * Now map into the kernel's virtual memory the address supplied by
	 * the user from there virtual address space.
	 */
	ASSERT(do_sync || os_has_new_pflip);
	if (do_sync)
		p = (struct lf_elt *)pattach(vaddr);
	else {
		caddr_t addr = (caddr_t)(((char *)vaddr) + 1);
		p = (struct lf_elt *)pattach(addr);
	}

	if (p == 0) { /* failed */

		/* release header mbuf */
		lf_free(q_mbuf, (struct lf_elt *)m);

		/* account for free'ing the mbuf in the statistics */
		MBSTAT_DEC(m_mtypes[type], 1);

		/* release cluster page info node */
		lf_free(q_cpi, (struct lf_elt *)cpi);

#ifdef MBUF_DEBUG
		MBSTAT_DEC(m_mtypes[MT_CPI], 1);
#endif
		MBUF_CALLSTATS(m_shget_calls_pattach);
		return ((struct mbuf *)0);
	}

	m->m_off = (((__psunsigned_t)p - (__psunsigned_t)m) +
					((__psunsigned_t)vaddr % NBPP));
	m->m_len = m->m_nodesize = size;
	m->m_flags = (M_CLUSTER | M_SHARED);
	m->m_type = type;

#ifdef DEBUG
	m->m_freefunc = _m_freefunction;
	m->m_dupfunc = _m_dupfunction;
#else
	m->m_freefunc = m_freefunction;
	m->m_dupfunc = m_dupfunction;
#endif
	m->m_farg = (long)p;
	m->m_darg = (long)cpi;

	/* now fill in the cluster page info */
	cpi->virtaddr = (caddr_t)p;
	cpi->usr_virtaddr = vaddr;
	cpi->nodesize = size;
	cpi->refct = 1;

	MBUF_CALLSTATS(m_shget_calls_ok);
	return m;
}


/*
 * Procedure Description:
 *   Called from sosend() to synchronize the tlb's before unlocking
 *   the socket buffer, in order either to allow it to drain or to
 *   return to the caller. 
 *
 * Parameter Description:
 *   none
 *
 * Returns:
 *   void
 */

void
m_shget_finished()
{
	pflip(NULL, NULL);	/* tell pflip to sync the tlbs now */
}
	

/*
 * Procedure Description:
 *   Count the number of data bytes in an mbuf chain
 *
 * Parameters:
 *   m => Address of mbuf chain
 *
 * Returns:
 *   Number of bytes in mbuf chain
 */
int
m_length(struct mbuf *m)
{
	int len = 0;

	MBUF_CALLSTATS(m_length_calls);

	while (m != 0) {
		GOODMP(m);
		GOODMT(m->m_type);

		len += m->m_len;
		m = m->m_next;
	}
	return len;
}

/*
 * Procedure Description:
 *   Rearrange an mbuf chain so that len bytes are contiguous and in
 *   the data area of an mbuf. This is required so that the mtod
 *   macros will return a pointer to an item that is large enough.
 *
 *   If there is room, it will add up to MPULL_EXTRA bytes to the contiguous
 *   region in an attempt to avoid being called next time.
 *
 * Parameters:
 *   n => Address of mbuf chain
 *   len => Number of bytes which must be contiguous.
 *
 * Returns:
 *   On success => returns the resulting mbuf chain.
 *   On failure frees the ENTIRE mbuf chain 'n' and returns 0.
 */
struct mbuf *
m_pullup(struct mbuf *n, int len)
{
	struct mbuf *m;
	int count, space;

	GOODMP(n);
	GOODMT(n->m_type);
	ASSERT(len > 0);

	MBUF_CALLSTATS(m_pullup_calls);
	if (!(MBUF_IS_CLUSTER(n)) && n->m_len >= len) {
		/*
		 * No need to be called in the first place.
		 */
		return (n);
	}

	if (!(MBUF_IS_CLUSTER(n)) && n->m_off + len <= MMAXOFF && n->m_next) {
		m = n;
		n = n->m_next;
		len -= m->m_len;
	} else {
		if (len > MLEN)
			goto bad;

		m = m_get(M_DONTWAIT, n->m_type);
		if (m == 0)
			goto bad;

		m->m_len = 0;

		/* copy only the link-level flag information */
		m->m_flags = n->m_flags & M_COPYFLAGS;
	}
	space = MMAXOFF - m->m_off;

	for (;;) {
		count = MIN(MIN(space - m->m_len, len+MPULL_EXTRA), n->m_len);
		bcopy(mtod(n, caddr_t), mtod(m, caddr_t)+m->m_len, count);

		len -= count;
		m->m_len += count;
		n->m_len -= count;

		if (n->m_len != 0)
			n->m_off += count;
		else
			n = m_free(n);
		if (len <= 0)
			break;

		if (n == 0) {
			(void)m_free(m);
			goto bad;
		}
	}
	m->m_next = n;
	return (m);
bad:
	m_freem(n);
	return (0);
}


/*
 * Procedure Description:
 *
 *  Return the current statistics count for the designated mbuf type.
 *  NOTE: This is only used by the Security Audit Trail (SAT) to limit
 *  the maximum number of mbuf's in use by that subsystem before waiting
 *  on a semaphore to limit the sustem's usage.
 *
 * Parameters:
 *   m_type => One of the valid mbuf m_type values defined in mbuf header file
 *
 * Returns:
 *   integer count of the number of allocated mbuf's of the specified
 *   type across all cpu's in the system.
 */
int
m_typeget(u_short m_type)
{
	register struct kna *knaptr;
	register int i, mbcnt;

	if (m_type >= MT_MAX) { /* bogus type */
		return 0;
	}

	mbcnt = 0;
	for (i = 0; i < maxcpus; i++) {

		if ((pdaindr[i].CpuId == -1) ||
		   !(knaptr = (struct kna *)pdaindr[i].pda->knaptr)) {
			continue;
		}

		/* compute for all cpu's number of regular mbufs allocated */
		mbcnt += knaptr->mbstat.m_mtypes[m_type];
	}
	return mbcnt;
}

/*
 * Procedure Description:
 *
 *   Trim 'len' bytes from the end of a single mbuf and subtract
 *  'len' bytes from it's length.
 *
 * Parameters:
 *   n => Address of head of mbuf chain
 *   len => Number of bytes to trim from the chain
 *
 * Returns:
 *   Void
 */
void
m_trim(struct mbuf *m, int len)
{
	int newlen;

	MBUF_CALLSTATS(m_trim_calls);

	newlen = m->m_len - len;

	if (newlen < 0) {
		MBUF_CALLSTATS(m_trim_zerolen);
		newlen = 0;
	}
	MBUF_CALLSTATS(m_trim_newlen);
	m->m_len = newlen;
	return;
}

/*
 * Procedure Description:
 *   Add a page of nodes to the local list vector and return a node.
 *
 * Parameter Description:
 *   local_lfvecp => Address of the local listvec structure
 *
 * Returns:
 *   Address of the node.
 */
struct lf_elt *
lf_addpage(struct lf_listvec *local_lfvecp, void *page)
{
	register struct lf_elt *node;
	register int s_lfvec;

	MBUF_CALLSTATS(lf_addpage_calls);

	s_lfvec = mutex_spinlock(&local_lfvecp->lock);

	/*
	 * got a new page so add all but one to the get list
	 * unless somebody else has already added something in
	 * which case we append the new page of nodes.
	 */
	if (local_lfvecp->get == (char *)NULL) { /* still empty */

		local_lfvecp->get = (char *)(((struct lf_elt *)page)->next);
		local_lfvecp->getcnt = local_lfvecp->e_per_page - 1;
		MBUF_CALLSTATS(lf_addpage_empty);
	} else {

		node = (struct lf_elt *)((char *)page +
			NBPP - local_lfvecp->size);
		node->next = (struct lf_elt *)local_lfvecp->get;

		local_lfvecp->get = (char *)(((struct lf_elt *)page)->next);
		local_lfvecp->getcnt += local_lfvecp->e_per_page - 1;
		MBUF_CALLSTATS(lf_addpage_occupied);
	}
	node = (struct lf_elt *)page;
	LV_STAT(local_lfvecp->get_addpage);

	mutex_spinunlock(&local_lfvecp->lock, s_lfvec);
	return node;
}

/*
 * Procedure Description:
 * Get procedure that obtains a node from a list vector structure
 *
 * Parameters:
 *   local_lfvecp => Address of local list vector structure
 *   canwait => Integer canwait flag (not used yet)
 *
 * Returns:
 *   Address of node returned from list vector or NULL if failed.
 */
/* ARGSUSED */
struct lf_elt *
lf_get(local_lfvecp, canwait)
	register struct lf_listvec *local_lfvecp;
	register int canwait;
{
	register struct lf_listvec *global_lfvecp;
	register struct global_elt *e;
	register int s_lfvec;

	s_lfvec = mutex_spinlock_spl(&local_lfvecp->lock, splimp);

	LV_STAT(local_lfvecp->get_calls);

	if (local_lfvecp->get) { /* first try local get list */

		e = (struct global_elt *)local_lfvecp->get;
		local_lfvecp->get = (char *)e->next;
		local_lfvecp->getcnt--;

		/* Clear next field of returned node */
		e->next = 0;

		LV_STAT(local_lfvecp->get_lget);
	} else {

		if (local_lfvecp->put) { /* try local put list */
			/*
			 * Move local put list to local get list
			 * clear local put list and node count
			 */
			e = (struct global_elt *)local_lfvecp->put;
			local_lfvecp->get = (char *)e->next;
			e->next = 0;

			/* set local get count to put list count minus 1 */
			local_lfvecp->getcnt = local_lfvecp->putcnt - 1;

			/* clear local put list address and node count */
			local_lfvecp->put = NULL;
			local_lfvecp->putcnt = 0;

			LV_STAT(local_lfvecp->get_lput_to_lget);
		} else {
			/*
			 * Arrive here when BOTH local get and put lists
			 * are empty and we try and refresh them by looking
			 * at the global vector associated with this node size.
			 */	  
			global_lfvecp = (struct lf_listvec *)
				((char *)lf_globalvec + local_lfvecp->offset);

			mp_spinlock(&global_lfvecp->lock);

			e = (struct global_elt *)global_lfvecp->get;

			if (e) { /* get bristle from global get list */

				LV_STAT(local_lfvecp->get_gget_to_lget);

				/* get one bristle list and put on local get */
				global_lfvecp->get = (char *)e->remain;

				/*
				 * local list gets one list of bristles
				 * worth from the global list
				 */
				local_lfvecp->get = (char *)e->next;
				local_lfvecp->getcnt =
					local_lfvecp->threshold - 1;

				e->next = 0;
			} else {
				/*
				 * global get list of bristles is empty so
				 * check if global put list is non-null
				 */
				e = (struct global_elt *)global_lfvecp->put;
				if (e) {
					LV_STAT(local_lfvecp->get_gput_to_gget);
					global_lfvecp->get = (char *)e->remain;
					global_lfvecp->put = (char *)NULL;

					LV_STAT(local_lfvecp->get_gget_to_lget);
					local_lfvecp->get = (char *)e->next;
					local_lfvecp->getcnt =
						local_lfvecp->threshold - 1;

					e->next = (struct global_elt *)NULL;

				} else {
					LV_STAT(local_lfvecp->get_totalfails);
				}
			}
			mp_spinunlock(&global_lfvecp->lock);
		}
	}
	mutex_spinunlock(&local_lfvecp->lock, s_lfvec);

	return ((struct lf_elt *)e);
}

/*
 * Procedure Description:
 * Free procedure that returns a node to a lf list vector structure
 *
 * Parameters:
 *   local_lfvecp => Address of local list vector structure
 *   e => Address of node to return
 *
 * Returns:
 *   Void
 */
void
lf_free(local_lfvecp, e)
	register struct lf_listvec *local_lfvecp;
	register struct lf_elt *e;
{
	register struct lf_listvec *global_lfvecp;
	register int s_global, s_lfvec;
	register caddr_t page;

#ifdef DEBUG
	register __psunsigned_t invalid_addr;
 
	/*
	 * NOTE: Special case to avoid getting local listvec lock
	 * but in debug mode we want the call counts to be accurate.
	 */
	atomicAddInt(((int *)&(local_lfvecp->free_calls)), 1);

	if (!e) { /* fast exit case if null */
		atomicAddInt(((int *)&(local_lfvecp->free_nullnode)), 1);
		return;
	}
#else
	if (!e) { /* fast exit case if null */
		return;
	}
#endif
	global_lfvecp = (struct lf_listvec *)((char *)lf_globalvec
			+ local_lfvecp->offset);

	s_global = mutex_spinlock(&global_lfvecp->lock);

	if (global_lfvecp->e_per_page != 1) {

		/* compute base page address associated with node */
		page = (caddr_t)((__psunsigned_t)e & -NBPP);

		/*
		 * Node size not NBPP so check if reclaim page address set
		 * Set reclaim page address if it's null, then return.
		 */
		if (global_lfvecp->reclaim_page == (char *)NULL) {

			global_lfvecp->reclaim_page = page;
			global_lfvecp->nelts = 1;
#ifdef DEBUG
			atomicAddInt(((int *)&(local_lfvecp->free_greclaim)),
				1);
#endif
			mutex_spinunlock(&global_lfvecp->lock, s_global);
			return;
		}

		if (global_lfvecp->reclaim_page == page) {

			global_lfvecp->nelts++;
#ifdef DEBUG
			atomicAddInt(((int *)&(local_lfvecp->free_greclaim)),
				1);
#endif
			if (global_lfvecp->nelts >= global_lfvecp->e_per_page){
				/*
				 * Decide to either release or add this page
				 * to the reclaim page list.
				 */
				m_freepage(-1, page);

				global_lfvecp->nelts = 0;
				global_lfvecp->reclaim_page = (char *)NULL;

			}
			mutex_spinunlock(&global_lfvecp->lock, s_global);
			return;
		}
	}

	/* release global list lock */
	mutex_spinunlock(&global_lfvecp->lock, s_global);

#ifdef DEBUG
	/*
	 * NOTE: Since all node sizes are a power of two we check in
	 * debug mode that the appropriate bottom address bits are zero
	 * and fail if they are non-zero.
	 */
	invalid_addr = (__psunsigned_t)e & ~local_lfvecp->node_mask;

	if (invalid_addr) { /* bad address not proper power of 2 */
		cmn_err(CE_PANIC,
		 "lf_free(): Bad free addr 0x%x, node_mask 0x%x, lfvec 0x%x\n",
		  (__psunsigned_t)e, local_lfvecp->node_mask, local_lfvecp);
		return;
	}
#endif
	/* claim local list lock */
	s_lfvec = mutex_spinlock(&local_lfvecp->lock);

	/* 
	 * Thread returning node onto local put list. Bump local put list
	 * node count and check if threshold has been reached.
	 * When the threshold is reached move the entire bristle,
	 * list of nodes, to the global put list.
	 */
	e->next = (struct lf_elt *)local_lfvecp->put;

	local_lfvecp->put = (char *)e;
	local_lfvecp->putcnt++;

	if (local_lfvecp->putcnt < local_lfvecp->threshold) {

		LV_STAT(local_lfvecp->free_lput);
		/*
		 * zero remain field in case node ever becomes a bristle head
		 */
		((struct global_elt *)e)->remain = (struct global_elt *)NULL;

		mutex_spinunlock(&local_lfvecp->lock, s_lfvec);
		return;
	}

	LV_STAT(local_lfvecp->free_lput_to_gput);
	/*
	 * Append entire local put list onto the global
	 * put list of bristles, list of lists.
	 */
	local_lfvecp->put = (char *)NULL;
	local_lfvecp->putcnt = 0;

	/*
	 * get the global lock to update the global put list head
	 */
	nested_spinlock(&global_lfvecp->lock);

	/* save current put list bristle head in remain field */
	((struct global_elt *)e)->remain =
		(struct global_elt *)global_lfvecp->put;

	global_lfvecp->put = (char *)e;

	/* release global lock after putting bristle on put list */
	nested_spinunlock(&global_lfvecp->lock);

	/* now release local list lock */
	mutex_spinunlock(&local_lfvecp->lock, s_lfvec);
	return;
}

#ifdef EVEREST
/*
 * Bug 323277 - Challenge IO4 IA chip bug workaround.
 * Flush IA partial cache line by bringing data into processor cache.
 */
void
m_io4iawar(m)
	struct mbuf *m;
{
	volatile char c;

	for (; m; m = m->m_next)
		if (m->m_len > 0)
			c = *(mtod(m, char*) + (m->m_len - 1));
}
#endif

int mbuf_reclaim_needed = 0;

void
m_reclaim(void)
{
	if (mbuf_reclaim_needed) {
		/* Help is on the way */
		return;
	}
	atomicAddInt(&mbuf_reclaim_needed, 1);

	MBSTAT_INC(m_drain, 1);
}
