#ifndef __net_if__
#define __net_if__
/*
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)if.h	8.1 (Berkeley) 6/10/93
 */

/*
 * Structures defining a network interface, providing a packet
 * transport mechanism (ala level 0 of the PUP protocols).
 *
 * Each interface accepts output datagrams of a specified maximum
 * length, and provides higher level routines with input datagrams
 * received from its medium.
 *
 * Output occurs when the routine if_output is called, with three parameters:
 *	(*ifp->if_output)(ifp, m, dst)
 * Here m is the mbuf chain to be sent and dst is the destination address.
 * The output routine encapsulates the supplied datagram if necessary,
 * and then transmits it on its medium.
 *
 * On input, each interface unwraps the data received by it, and either
 * places it on the input queue of a internetwork datagram routine
 * and posts the associated software interrupt, or passes the datagram to a raw
 * packet input routine.
 *
 * Routines exist for locating interfaces by their addresses
 * or for locating a interface on a certain network, as well as more general
 * routing and gateway routines maintaining information used to locate
 * interfaces.  These routines live in the files if.c and route.c
 */

#include <sys/socket.h>
#include <sys/sema.h>
#include <sys/time.h>
#include <sys/mac_label.h>		/* Trusted IRIX */
#include <net/if_sec.h>
#include <net/if_arp.h>

#ifdef __STDC__
/*
 * Forward structure declarations for function prototypes [sic].
 */
struct	mbuf;
struct	socket;
struct	rtentry;
#endif

/*
 * Structure describing information about an interface
 * which may be of interest to management entities.
 * keep if_data separate to make C++ happy
 */
struct	if_data {

	/* generic interface information */
	u_char	ifi_type;		/* ethernet, tokenring, etc */
	u_char	ifi_addrlen;		/* media address length */
	u_char	ifi_hdrlen;		/* media header length */
	u_int	ifi_mtu;		/* maximum transmission unit */
	u_int	ifi_metric;		/* routing metric (external only) */
	u_int	ifi_baudrate;		/* linespeed */

	/* volatile statistics */
	u_int	ifi_ipackets;		/* packets received on interface */
	u_int	ifi_ierrors;		/* input errors on interface */
	u_int	ifi_opackets;		/* packets sent on interface */
	u_int	ifi_oerrors;		/* output errors on interface */
	u_int	ifi_collisions;		/* collisions on csma interfaces */
	u_int	ifi_ibytes;		/* total number of octets received */
	u_int	ifi_obytes;		/* total number of octets sent */
	u_int	ifi_imcasts;		/* packets received via multicast */
	u_int	ifi_omcasts;		/* packets sent via multicast */
	u_int	ifi_iqdrops;		/* dropped on input, this interface */
	u_int	ifi_odrops;		/* output pkts discarded w/o error */
	u_int	ifi_noproto;		/* destined for unsupported protocol */
	time_t	ifi_lastchange;		/* last updated */
};

struct ifnet {
	char	*if_name;		/* name, e.g. ``en'' or ``lo'' */
	struct	ifnet *if_next;		/* all struct ifnets are chained */
	caddr_t in_ifaddr;    /* (struct in_ifaddr *) primary interface addr */
	u_short	if_index;		/* numeric abbreviation for this if  */
	short	if_unit;		/* sub-unit for lower level driver */
	short	if_timer;		/* time until if_watchdog called */
	short	if_flags;		/* up/down, broadcast, etc. */
	struct if_data if_data;

	/* procedure handles */
	int	(*if_output)(struct ifnet *, struct mbuf *, struct sockaddr *);
					/* output routine */
	int	(*if_ioctl)(struct ifnet *, int, void *); /* ioctl routine */
	int	(*if_reset)(int, int);	/* bus reset routine */
	void	(*if_watchdog)(int);	/* timer routine */

	struct	ifqueue {
		struct	mbuf *ifq_head;
		struct	mbuf *ifq_tail;
		int	ifq_len;
		int	ifq_maxlen;
		u_int	ifq_drops;
		lock_t	ifq_lock;
	} if_snd;			/* output queue */

	/*
	 * Trusted IRIX: Any changes to the following three
	 * elements must be done at splnet or equivalent.
	 */
	struct	mbuf *if_lbl_cache;	/* label mapping cache */
	ifsec_t	*if_sec;		/* interface security information */
	uid_t 	if_uid;			/* default uid for dacless if */

	lock_t	if_lock;
};

#ifdef _KERNEL				/* all of these should go away */
#define	if_mtu		if_data.ifi_mtu
#define	if_type		if_data.ifi_type
#define	if_addrlen	if_data.ifi_addrlen
#define	if_hdrlen	if_data.ifi_hdrlen
#define	if_metric	if_data.ifi_metric
#define	if_baudrate	if_data.ifi_baudrate
#define	if_ipackets	if_data.ifi_ipackets
#define	if_ierrors	if_data.ifi_ierrors
#define	if_opackets	if_data.ifi_opackets
#define	if_oerrors	if_data.ifi_oerrors
#define	if_collisions	if_data.ifi_collisions
#define	if_ibytes	if_data.ifi_ibytes
#define	if_obytes	if_data.ifi_obytes
#define	if_imcasts	if_data.ifi_imcasts
#define	if_omcasts	if_data.ifi_omcasts
#define	if_iqdrops	if_data.ifi_iqdrops
#define if_odrops	if_data.ifi_odrops
#define	if_noproto	if_data.ifi_noproto
#define	if_lastchange	if_data.ifi_lastchange
#endif

#define	if_idrops	if_iqdrops	/* XXX compatibility */

#define	IFF_UP		0x1		/* interface is up */
#define	IFF_BROADCAST	0x2		/* broadcast address valid */
#define	IFF_DEBUG	0x4		/* turn on debugging */
#define	IFF_LOOPBACK	0x8		/* is a loopback net */

#define	IFF_POINTOPOINT	0x10		/* interface is point-to-point link */
#define	IFF_NOTRAILERS	0x20		/* avoid use of trailers */
#define	IFF_RUNNING	0x40		/* resources allocated */
#define	IFF_NOARP	0x80		/* no address resolution protocol */

#define	IFF_PROMISC	0x100		/* receive all packets */
#define	IFF_ALLMULTI	0x200		/* receive all multicast packets */
#define	IFF_FILTMULTI	0x400		/* need to filter multicast packets */
#define	IFF_INTELLIGENT	0x400		/* MIPS ABI - protocols on interface */
#define	IFF_MULTICAST	0x800		/* supports multicast */

#define IFF_CKSUM	0x1000		/* does checksumming */
#define IFF_ALLCAST	0x2000		/* does SRCROUTE broadcast */
#define	IFF_DRVRLOCK	0x4000		/* n/w driver does its own MP locking*/
#define IFF_PRIVATE     0x8000          /* MIPS ABI - do not advertise */

/*
 * The IFF_MULTICAST flag indicates that the network can support the
 * transmission and reception of higher-level (e.g., IP) multicast packets.
 * It is independent of hardware support for multicasting; for example,
 * point-to-point links or pure broadcast networks may well support
 * higher-level multicasts.
 *
 * The IFF_FILTMULTI flag indicates that the interface has imperfect
 * hardware multicast filtering. Additional filtering is required before
 * a multicast packet is accepted.
 *
 * The IFF_ALLCAST flag indicates that the interface supports
 * 802.5 source routing broadcast so that the broadcast msg can
 * cross bridges.
 */

/* flags set internally only: */
#define	IFF_CANTCHANGE	(IFF_BROADCAST | IFF_POINTOPOINT |   \
			    IFF_RUNNING | IFF_MULTICAST | IFF_FILTMULTI)

/*
 * Output queues (ifp->if_snd) and internetwork datagram level (pup level 1)
 * input routines have queues of messages stored on ifqueue structures
 * (defined above).  Entries are added to and deleted from these structures
 * by these macros, which should be called with ipl raised to splimp().
 */
#ifdef _MP_NETLOCKS
extern lock_t ifhead_lock;

#define IFHEAD_INITLOCK()	spinlock_init(&ifhead_lock, "ifheadlk")
#define IFHEAD_LOCK()		mp_spinlock(&ifhead_lock);
#define IFHEAD_UNLOCK()		mp_spinunlock(&ifhead_lock);

#define IFNET_INITLOCKS(ifp) { \
			spinlock_init(&(ifp)->if_lock, "if_lock"); \
			IFQ_INITLOCK(&(ifp)->if_snd); \
}
#define IFNET_ISLOCKED(ifp)	(spinlock_islocked(&(ifp)->if_lock))
#define IFQ_LOCK(q)		mp_spinlock(&(q)->ifq_lock)
#define IFQ_UNLOCK(q)		mp_spinunlock(&(q)->ifq_lock)
#define IFNET_UPPERLOCK(ifp,s)	{ 	\
	if (!(ifp->if_flags & (IFF_LOOPBACK | IFF_DRVRLOCK)))		\
		IFNET_LOCK(ifp, s); 	\
}
#define IFNET_UPPERUNLOCK(ifp,s) {	\
	if (!(ifp->if_flags & (IFF_LOOPBACK | IFF_DRVRLOCK)))		\
		IFNET_UNLOCK(ifp, s);	\
}
#define IFNET_LOCK(ifp,s)	s = mp_mutex_spinlock(&(ifp)->if_lock)
#define IFNET_UNLOCK(ifp,s)	mp_mutex_spinunlock(&(ifp)->if_lock, s)
#define IFNET_LOCKNOSPL(ifp) 	mp_spinlock(&(ifp)->if_lock)
#define IFNET_UNLOCKNOSPL(ifp) 	mp_spinunlock(&(ifp)->if_lock)
#define	IF_ENQUEUE_NOLOCK(ifq, m) { \
	(m)->m_act = 0; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_head = m; \
	else \
		(ifq)->ifq_tail->m_act = m; \
	(ifq)->ifq_tail = m; \
	(ifq)->ifq_len++; \
}
#define	IF_DEQUEUE_NOLOCK(ifq, m) { \
	(m) = (ifq)->ifq_head; \
	if (m) { \
		if (((ifq)->ifq_head = (m)->m_act) == 0) \
			(ifq)->ifq_tail = 0; \
		(m)->m_act = 0; \
		(ifq)->ifq_len--; \
	} \
}
#else
#define IFHEAD_INITLOCK()	
#define IFHEAD_LOCK()
#define IFHEAD_UNLOCK()	

#define IFNET_INITLOCKS(ifp) 
#define IFNET_ISLOCKED(ifp)	(1)
#define IFQ_LOCK(q)	
#define IFQ_UNLOCK(q)
#define IFNET_UPPERLOCK(ifp,s)
#define IFNET_UPPERUNLOCK(ifp,s)
#define IFNET_LOCK(ifp,s)	s = splimp()
#define IFNET_UNLOCK(ifp,s)	splx(s)
#define IFNET_LOCKNOSPL(ifp) 
#define IFNET_UNLOCKNOSPL(ifp)
#define	IF_ENQUEUE_NOLOCK(ifq, m) 	IF_ENQUEUE(ifq, m)
#define	IF_DEQUEUE_NOLOCK(ifq, m) 	IF_DEQUEUE(ifq, m)
#endif /* _MP_NETLOCKS */

#define IFQ_INITLOCK(ifq) spinlock_init(&(ifq)->ifq_lock, "ifq_lock") 
#define	IF_QFULL(ifq)		((ifq)->ifq_len >= (ifq)->ifq_maxlen)
#define	IF_DROP(ifq)		((ifq)->ifq_drops++)
#define	IF_ENQUEUE(ifq, m) { \
	IFQ_LOCK(ifq); \
	(m)->m_act = 0; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_head = m; \
	else \
		(ifq)->ifq_tail->m_act = m; \
	(ifq)->ifq_tail = m; \
	(ifq)->ifq_len++; \
	IFQ_UNLOCK(ifq); \
}
#define	IF_PREPEND(ifq, m) { \
	IFQ_LOCK(ifq); \
	(m)->m_act = (ifq)->ifq_head; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_tail = (m); \
	(ifq)->ifq_head = (m); \
	(ifq)->ifq_len++; \
	IFQ_UNLOCK(ifq); \
}
/*
 * Packets destined for level-1 protocol input routines have the following
 * structure prepended to the data.  IF_DEQUEUEHEADER extracts and returns
 * this structure's interface pointer, and skips over the space indicated
 * by it, when dequeueing the packet.  Otherwise M_SKIPIFHEADER should be
 * used to adjust for the ifheader.
 */
struct ifheader {
	struct ifnet	*ifh_ifp;	/* pointer to receiving interface */
	u_short		ifh_hdrlen;	/* byte size of this structure */
};

#define	M_ADJ(m, adj) { \
	(m)->m_off += (adj); \
	(m)->m_len -= (adj); \
	if ((m)->m_len == 0) \
		(m) = m_free(m); \
}
#define	M_SKIPIFHEADER(m) { \
	register int hdrlen; \
	hdrlen = mtod((m), struct ifheader *)->ifh_hdrlen; \
	M_ADJ(m, hdrlen); \
}
#define	IF_DEQUEUEHEADER(ifq, m, ifp) { \
	IFQ_LOCK(ifq); \
	(m) = (ifq)->ifq_head; \
	if (m) { \
		struct ifheader *ifh; \
		(ifq)->ifq_head = (m)->m_act; \
		if ((ifq)->ifq_head == 0) \
			(ifq)->ifq_tail = 0; \
		(m)->m_act = 0; \
		--(ifq)->ifq_len; \
		ifh = mtod((m), struct ifheader *); \
		(ifp) = ifh->ifh_ifp; \
		M_ADJ(m, ifh->ifh_hdrlen); \
	} \
	IFQ_UNLOCK(ifq); \
}

/*
 * Initialize a receive buffer's ifqueue header to point at ifp and to
 * have a data offset hdrlen bytes beyond buf.
 */
#define	IF_INITHEADER(buf, ifp, hdrlen) { \
	struct ifheader *ifh; \
	ifh = (struct ifheader *)(buf); \
	ifh->ifh_ifp = (ifp); \
	ifh->ifh_hdrlen = (hdrlen); \
}
#define	M_INITIFHEADER(m, ifp, hdrlen) { \
	struct ifheader *ifh; \
	ifh = mtod(m, struct ifheader *); \
	ifh->ifh_ifp = (ifp); \
	(m)->m_len = ifh->ifh_hdrlen = (hdrlen); \
}

#define	IF_DEQUEUE(ifq, m) { \
	IFQ_LOCK(ifq); \
	(m) = (ifq)->ifq_head; \
	if (m) { \
		if (((ifq)->ifq_head = (m)->m_act) == 0) \
			(ifq)->ifq_tail = 0; \
		(m)->m_act = 0; \
		(ifq)->ifq_len--; \
	} \
	IFQ_UNLOCK(ifq); \
}

#define	IFQ_MAXLEN	50
#define	IFNET_SLOWHZ	1		/* granularity is 1 second */

/*
 * The ifaddr structure contains information about one address
 * of an interface.  They are maintained by the different address families,
 * are allocated and attached when an address is set, and are stored in
 * a hash table. If all addresses for an interface need to be located
 * the hash table can be enumerated for entries matching the interface address
 */
struct ifaddr {
	struct	sockaddr *ifa_addr;	/* address of interface */
	struct	sockaddr *ifa_dstaddr;	/* other end of p-to-p link */
#define	ifa_broadaddr	ifa_dstaddr	/* broadcast address interface */

#ifdef _HAVE_SIN_LEN
	struct	sockaddr *ifa_netmask;	/* used to determine subnet */
#else
	struct	sockaddr_new *ifa_netmask;	/* used to determine subnet */
#endif
	struct	ifnet *ifa_ifp;		/* back-pointer to interface */
	struct	ifaddr *ifa_next;	/* next interface address in list */

	void	(*ifa_rtrequest)(int,struct rtentry *,struct sockaddr *);
					/* check for clean routes (+ or -)'d */

	u_short	ifa_flags;		/* mostly rt_flags for cloning */
	u_int	ifa_refcnt;		/* extra to malloc for link info */
	int	ifa_metric;		/* cost of going out this interface */

	/*
	 * since the ifaddr structure is embedded inside an in_ifaddr structure
	 * we store the starting address of the in_ifaddr for free operations
	 * since the starting address of the entire block is required.
	 * This save lots of weird computation when the code free's the ifaddr
	 * structure since it holds the ref count covering both structures.
	 * NOTE: It's a caddr_t to avoid header file circularity problems.
	 */
	caddr_t ifa_start_inifaddr;	/* address of in_ifaddr struct */
};

/*
 * 'struct ifaddr field ifa_flags values defined in route.h
 */
#define IFA_ROUTE	RTF_UP		/* route installed */

/* convert 'data' address given SIOCSIFADDR */
#define _IFADDR_SA(d) (((struct ifaddr*)(d))->ifa_addr)

struct ifstats {
	u_int	ifs_ipackets;		/* packets received on interface */
	u_int	ifs_opackets;		/* packets sent on interface */
	u_short	ifs_ierrors;		/* input errors on interface */
	u_short	ifs_oerrors;		/* output errors on interface */
	u_int	ifs_collisions;		/* collisions on csma interfaces */
};

/*
 * Message format for use in obtaining information about interfaces
 * from getkerninfo and the routing socket
 */
struct if_msghdr {
	u_short	ifm_msglen;	/* to skip over non-understood messages */
	u_char	ifm_version;	/* future binary compatability */
	u_char	ifm_type;	/* message type */
	int	ifm_addrs;	/* like rtm_addrs */
	int	ifm_flags;	/* value of if_flags */
	u_short	ifm_index;	/* index for associated ifp */
	struct	if_data ifm_data;/* statistics and other data about if */
};

/*
 * Message format for use in obtaining information about interface addresses
 * from getkerninfo and the routing socket
 */
struct ifa_msghdr {
	u_short	ifam_msglen;	/* to skip over non-understood messages */
	u_char	ifam_version;	/* future binary compatability */
	u_char	ifam_type;	/* message type */
	int	ifam_addrs;	/* like rtm_addrs */
	int	ifam_flags;	/* value of ifa_flags */
	u_short	ifam_index;	/* index for associated ifp */
	int	ifam_metric;	/* value of ifa_metric */
};

/*
 * Interface request structure used for socket
 * ioctl's.  All interface ioctl's must have parameter
 * definitions which begin with ifr_name.  The
 * remainder may be interface specific.
 */
struct	ifreq {
#define	IFNAMSIZ	16
	char	ifr_name[IFNAMSIZ];		/* if name, e.g. "en0" */
	union {
		struct	sockaddr ifru_addr;
		struct	sockaddr ifru_dstaddr;
		struct	sockaddr ifru_broadaddr;
		short	ifru_flags;
		int	ifru_metric;
		char	ifru_data[1];		/* MIPS ABI - unused by BSD */
		char	ifru_enaddr[6];		/* MIPS ABI */
		char	ifru_oname[IFNAMSIZ];	/* MIPS ABI */
		struct	ifstats ifru_stats;

		/* Trusted IRIX */
		struct {
			caddr_t	ifruv_base;
			int	ifruv_len;
		}	ifru_vec;
	} ifr_ifru;
#define	ifr_addr	ifr_ifru.ifru_addr	/* address */
#define	ifr_dstaddr	ifr_ifru.ifru_dstaddr	/* other end of p-to-p link */
#define	ifr_broadaddr	ifr_ifru.ifru_broadaddr	/* broadcast address */
#define	ifr_flags	ifr_ifru.ifru_flags	/* flags */
#define	ifr_metric	ifr_ifru.ifru_metric	/* metric */
#define	ifr_data	ifr_ifru.ifru_data	/* for use by interface */
#define ifr_enaddr      ifr_ifru.ifru_enaddr    /* ethernet address */
#define ifr_oname       ifr_ifru.ifru_oname     /* other if name */
#define	ifr_stats	ifr_ifru.ifru_stats	/* statistics */
/* Trusted IRIX */
#define ifr_base	ifr_ifru.ifru_vec.ifruv_base
#define ifr_len		ifr_ifru.ifru_vec.ifruv_len
};

struct ifaliasreq {
	char	ifra_name[IFNAMSIZ];		/* if name, e.g. "en0" */
	struct	sockaddr ifra_addr;
	struct	sockaddr ifra_broadaddr;
	struct	sockaddr ifra_mask;
#define ifra_dstaddr ifra_broadaddr
	int	cookie;			/* SIOCLIFADDR enumeration cookie */
};

/*
 * Structure used in SIOCGIFCONF request.
 * Used to retrieve interface configuration for the ystem.
 * Useful for programs which must know all networks accessible.
 */
struct	ifconf {
	int	ifc_len;		/* size of associated buffer */
	union {
		caddr_t	ifcu_buf;
		struct	ifreq *ifcu_req;
	} ifc_ifcu;
#define	ifc_buf	ifc_ifcu.ifcu_buf	/* buffer address */
#define	ifc_req	ifc_ifcu.ifcu_req	/* array of structures returned */
};

#ifdef _KERNEL

extern struct	ifqueue rawintrq;		/* raw packet input queue */
extern struct	ifnet *ifnet;

extern char	*ether_sprintf(u_char *);

extern void	if_attach(struct ifnet *);
extern void	if_down(struct ifnet *);
extern void	if_slowtimo(void);
extern int	ifconf(int, caddr_t);
extern void	ifinit(void);
extern int	ifioctl(struct socket *, int , caddr_t);
extern struct	ifnet *ifunit(char *);

extern struct	ifaddr *ifa_ifwithaddr(struct sockaddr *);
extern struct	ifaddr *ifa_ifwithaf(int);
extern struct	ifaddr *ifa_ifwithdstaddr(struct sockaddr *);
extern struct	ifaddr *ifa_ifwithnet(struct sockaddr *);
extern struct ifaddr *ifa_ifwithroute(int, struct sockaddr*,struct sockaddr *);
extern struct	ifaddr *ifaof_ifpforaddr(struct sockaddr *, struct ifnet *);
extern void	ifafree(struct ifaddr *);

extern int	loioctl(struct ifnet *, int, caddr_t);
extern int	looutput(struct ifnet *, struct mbuf *, struct sockaddr *);
extern void	loattach(void);

extern int      ifioctl(struct socket *, int , caddr_t);
extern int	if_dropall(struct ifqueue*);
extern __uint32_t in_cksum_sub(u_short *, int, __uint32_t);

#endif /* _KERNEL */
#endif	/* !__net_if__ */
