/*
 * Routines to handle "software level 2" (network) interrupts.
 *
 * Copyright 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

#ident "$Revision: 3.39 $"

#include "tcp-param.h"
#include "sys/types.h"
#include "sys/sbd.h"
#include "sys/immu.h"
#include "sys/pda.h"
#include "sys/param.h"
#ifdef _MP_NETLOCKS
#include "sys/proc.h"
#include "sys/par.h"
#endif
#include "sys/kthread.h"
#include "sys/debug.h"
#include "sys/sema.h"
#include "sys/errno.h"
#include "sys/systm.h"
#include "sys/siginfo.h"
#include "netisr.h"
#include "if.h"

extern int nextbit(u_int);		/* forward reference */
extern void vrelvm(void);

#ifdef _MP_NETLOCKS
extern struct ifqueue  ipintrq;		/* needed for "momentum" scheduling */
lock_t netisr_lock;
#endif

sv_t	netpoke;

/*
 * By default, the standard splnet handler, netintr, calls the protocol
 * input routines. The input routines run until all packets in their
 * queues have been removed.
 *
 * The following "netproc" routines create and use a process running at a
 * preemptable, nondegrading priority for packet processing.
 *
 * The kernel process is initialized by a user-level daemon that calls
 * netproc() via syssgi(). netproc_intr(), which just wakes the process up,
 * is called instead of setsoftnet() from hardware interrupt (schednetisr).
 * If the daemon is killed, the schednetisr() calls setsoftnet().
 *
 * To allow higher-priority real-time processes to preempt the process,
 * the protocol input routines return after handling 1 packet if the
 * netproc_mode flag is set.
 */

u_int	netisr;


/*
 * Network process routines.
 */
#ifdef DEBUG
static struct {
	u_long  loop;
	u_long  sleep;		/* # times no packets pending, slept */
	u_long  intr;		/* # of interrupts */
	u_long  wake;		/* # times process was woken */
	u_long  none;		/* # times process wasn't waiting */
} nproc_stats;
#endif

int		netproc_mode;   /* = 1, process 1 packet at a time,
				 * = 0, do as many as possible. */
#ifdef _MP_NETLOCKS

/*
 * Soft interrupts are not used yet on multi-processors
 */
void
netintr(void)
{
	ASSERT(0);
}

/*
 * Set to zero to disable. If non-zero, this should be an estimate of the ratio
 * between the cost of a context switch over the cost of processing a single
 * packet at the IP and transport layers. The number of running rtnetds is
 * targeted to this times the queue length. So the arithmetic for the second
 * argument to sv_threshold divides the queue length by the gain (rounding 
 * up), subtracted from the total number of rtnetds started (netproc_mode).
 * This is compared against the sv queue count, which is the number waiting.
 */
#define IPQ_GAIN 16

/*
 * This is called from network interface device drivers when packets come in.
 * Use momentum simulation to possibly wake up a new network process.
 */
void
schednetisr(int anisr)
{
	mp_spinlock(&netisr_lock);
	netisr |= SCHEDBIT(anisr);
	METER(nproc_stats.intr++);
	mp_spinunlock(&netisr_lock);
	if (anisr == NETISR_IP) {
		switch (sv_threshold(&netpoke, 
			      netproc_mode -
			      (ipintrq.ifq_len + IPQ_GAIN - 1)/IPQ_GAIN)) {
		    case -1:
			return;
		    case 0:
			METER(nproc_stats.none++);
			break;
		    default:
			METER(nproc_stats.wake++);
			break;
		}
	} else {
		if (sv_signal(&netpoke))
			METER(nproc_stats.wake++);
		else
			METER(nproc_stats.none++);
	}
}

/*
 * Multi-processor version of rtnetd
 */
void
netproc(void *pri)
{
	int	s, isr;
	int	bit;
	int	count;

	curthreadp->k_pri = (__psint_t)pri;
	curthreadp->k_rtpri = (__psint_t)pri;
	count = 0;
	for (;;) {
		s = mp_mutex_spinlock(&netisr_lock);
		isr = netisr;
		netisr = 0;
		if (isr == 0) {
			METER(nproc_stats.sleep++);
			sv_wait(&netpoke, PZERO-1, &netisr_lock, s);
			count = 0;
			continue;
		}
		mp_mutex_spinunlock(&netisr_lock, s);
		METER(nproc_stats.loop++);
		while (isr) {
			bit = nextbit(isr);
			ASSERT(netisr_tab[bit].func);
			while ((*netisr_tab[bit].func)()) {
				/* only run through scheduler now and then */
				if (count++ > 15) {
					preemptchk();
					count = 0;
				}
			}
			isr &= ~SCHEDBIT(bit);
		}
	}
	/* NOTREACHED */
}

#else  /* !_MP_NETLOCKS */
void
netproc_intr(void)
{
	if ((netisr & SCHEDBIT(NETISR_RUNNING)) == 0) {
		ASSERT(netisr != 0);
		netisr |= SCHEDBIT(NETISR_RUNNING);
		if (sv_signal(&netpoke))
			METER(nproc_stats.wake++);
		else
			METER(nproc_stats.none++);
	}
	METER(nproc_stats.intr++);
}

/* 	We want to cause a soft interrupt the first
 *	time a bit is set.  We want to avoid causing an interrupt otherwise
 *	so that we do not uselessly interrupt the CPU.
 */
void
schednetisr(int anisr)
{
	int isr_s = splimp();			
	unsigned int isr_n = netisr;	
	if ((isr_n & SCHEDBIT(anisr)) == 0) {	
		netisr |= SCHEDBIT(anisr);
		if (!isr_n)		
			if (netproc_mode)
				netproc_intr();
			else		
				setsoftnet();	
	}				
	splx(isr_s);		
}

/*
 * Original, non-preemptable splnet handler.
 * Keep running as long as there is more work to do.  Keep netisr!=0 as long
 *	as possible so that hardware interrupts on other CPUs will not
 *	schedule soft interrupts.  It is better to keep the soft interrupts
 *	on the CPU that is doing the protocol processing.  Reducing the
 *	number of soft interrupts does single processor systems no harm.
 */

void
netintr(void)
{
	int	bit;
	int	s = splimp();

	ASSERT(!netproc_mode);
	do {
		netisr |= SCHEDBIT(NETISR_RUNNING);
		while ((bit = nextbit(netisr & ~SCHEDBIT(NETISR_RUNNING))) >= 0) {
			netisr &= ~SCHEDBIT(bit);
			splx(s);

			s = splnet();
			ASSERT(netisr_tab[bit].func);
			(void)(*netisr_tab[bit].func)();
			splx(s);
			acksoftnet();
			s = splimp();
		}
		netisr &= ~SCHEDBIT(NETISR_RUNNING);
	} while (netisr);
	ASSERT(netisr == 0);
	splx(s);
	return;
}


/*
 * pre-emptable single processor network input handler
 */
int
netproc(void)
{
	int	s, isr;
	int	bit;

	s = splimp();

	if (netproc_mode) {		/* prevent multiple instances */
		(void) splx(s);
		return EBUSY;
	}

	sv_init(&netpoke, SV_DEFAULT, "netpoke");

	/* Release client memory; we never return to user */
	vrelvm();

	netproc_mode = 1;

	for (;;) {
		splx(s);
		s = splimp();		/* same spl as schednetisr() */

		isr = netisr & ~SCHEDBIT(NETISR_RUNNING);
		if (isr == 0) {
			netisr = 0;
			METER(nproc_stats.sleep++);
			if (mp_sv_wait_sig(&netpoke, PZERO+1, NULL, 0))
				break;
			isr = netisr & ~SCHEDBIT(NETISR_RUNNING);
			/* we should not wake up without work to do */
			ASSERT(isr != 0);
		}
		netisr = SCHEDBIT(NETISR_RUNNING);

		METER(nproc_stats.loop++);

		while (isr) {
			bit = nextbit(isr);
			splx(s);
			s = splnet();		/* do not hog the MP lock */
			ASSERT(netisr_tab[bit].func);
			while ((*netisr_tab[bit].func)()) {
				splx(s);
				preemptchk();
				s = splnet();
			}
			isr &= ~SCHEDBIT(bit);
		}
	}

	/* Clean up and restore original handler */
	netproc_mode = 0;
	sv_destroy(&netpoke);
	splx(s);

	netintr();			/* avoid races */
	curthreadp->k_error = EINTR;	/* not needed; may help debugging */
	exit(CLD_EXITED, 0);
	/* NOTREACHED */
	return 0;
}
#endif /* !_MP_NETLOCKS */

/*
 * Special case for IP and RAW.  This
 * will allow SGI protocols to have higher
 * priority that customer added profiles.  The
 * bit NETISR_RUNNING has been moved to give it
 * the lowest priority.
 *
 *	NETISR_IP	0
 *	NETISR_RAW	1
 *	NETISR_RUNNING	31
 *
 * This scheme assumes that there are 32 bits in
 * an unsigned integer.
 */

extern int ip_intr();
extern int rawintr();

netisr_tab_t netisr_tab[32] = {
	{ip_intr},
	{rawintr},
};

#define MIN_BIT		(NETISR_RAW + 1)
#define MAX_BIT		NETISR_RUNNING

int
register_isr(int (*func)())
{
	int	bit;
	int	s = splimp();

	for (bit = MIN_BIT; bit < MAX_BIT; bit++)
		if (!netisr_tab[bit].func)
			break;

	if (bit < MAX_BIT)
		netisr_tab[bit].func = func;
	else
		bit = -1;

	splx(s);
	return(bit);
}

void
unregister_isr(int bit)
{
	int	s = splimp();
	
	/*
	 * Don't allow anyone to remove the reserved
	 * functions.
	 */
	if (bit >= MIN_BIT && bit < MAX_BIT) {
		netisr_tab[bit].func = 0;
		netisr &= ~SCHEDBIT(bit);
	}

	splx(s);
	return;
}

#undef MIN_BIT
#undef MAX_BIT

/*
 * This code might need to be in assembler.
 */

static u_char
isrtable[] = { 0, 0, 1, 99, 2, 99, 99, 99, 3, 99, 99, 99, 99, 99, 99, 99, 4};

/*
 * return -1 if mask == 0
 */
int
nextbit(u_int mask)
{
	u_int bit;
	u_int n = 0;

	if (!mask)
		return(-1);
	bit = mask & -mask;
	if (bit & 0xfffffff0) {
		if (bit & 0xffff0000) {
			n += 16;
			bit >>= 16;
		}
		if (bit & 0xff00) {
			n += 8;
			bit >>= 8;
		}
		if (bit & 0xf0) {
			n += 4;
			bit >>= 4;
		}
	}
	n += isrtable[bit];
	return(n);
}

#ifdef _MP_NETLOCKS
/*
 * Multi-processor start-up of the network input daemons
 */
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sockd.h>
int max_netprocs;
void
netisr_init(void)
{
	int i;
	extern int numcpus;

	sv_init(&netpoke, SV_LIFO, "netpoke");
	spinlock_init(&netisr_lock, "netisrlk");
	if (max_netprocs == 0) {
		if (numcpus > 4)
		    max_netprocs = (numcpus  - 1)/4 + 2;
		else
		    max_netprocs = (numcpus + 1)/2;
	}
	for (i=0; i<max_netprocs; i++) {
		netproc_mode++;
		(void) kthread_create("rtnetd", NULL, 4096, netproc,
					(void *)(__psint_t)39, 0);
	}
}

static lock_t sockd_lock;
static struct sockd_entry *sockd_head, *sockd_tail, *sockd_free;
static zone_t *sockd_zone;
static void sockd_call(void (*)(), void *);
static void sockd(void *);
#define SOCKD_FREEMAX	15
int sockd_freecnt = SOCKD_FREEMAX;
int sockd_freemax = SOCKD_FREEMAX;

static int pffasttimo_ticks;	/* ticks till pffasttimo */
static int pfslowtimo_ticks;	/* ticks till pfslowtimo */
static int if_slowtimo_ticks;	/* ticks till if_slowtimo */
static int lastwait;		/* length of last wait */

static int
periodic_timeouts(void)
{
	int waitticks, t;	/* number of ticks to wait to run again */

        waitticks = HZ;
        if (pftimo_active) {
                pffasttimo_ticks += lastwait;
                if (pffasttimo_ticks >= HZ/5) {         /* 200 ms */
                        pffasttimo();
                        pffasttimo_ticks = 0;
                        t = HZ/5;
                } else  {
                        t = HZ/5 - pffasttimo_ticks;
                }
                waitticks = t;

                pfslowtimo_ticks += lastwait;
                if (pfslowtimo_ticks >= HZ/2) {         /* 500 ms */
                        pfslowtimo();
                        pfslowtimo_ticks = 0;
                        t = HZ/2;
                } else {
                        t = HZ/2 - pfslowtimo_ticks;
                }
                if (t < waitticks)
                        waitticks = t;
        }
        if (if_slowtimo_active) {
                if_slowtimo_ticks += lastwait;
                if (if_slowtimo_ticks >= IFNET_SLOWHZ * HZ) { /* 1 sec */
                        if_slowtimo();
                        if_slowtimo_ticks = 0;
                }
        }
	lastwait = waitticks;
	return (waitticks);
}

void
sockd_init(void)
{
	register int i;

        spinlock_init(&sockd_lock, "sockd_lock");
        sockd_zone = kmem_zone_init(sizeof(struct sockd_entry), "sockd entry");
	/* create back up free list */
	for (i=sockd_freemax; i; i--) {
        	struct sockd_entry *entry;

		entry = kmem_zone_alloc(sockd_zone, KM_NOSLEEP);
		entry->next = sockd_free;
		sockd_free = entry;
	}
        kthread_create("sockd", NULL, 4096, (void (*)(void *))sockd,
			(void *)(__psint_t)39, 0);
}

/*
 * sockd_timeout()
 */
int
sockd_timeout(void (*fun)(), void *arg, long tim)
{
        return (timeout(sockd_call, (void *)fun, tim, arg));
}

/*
 * sockd_call()
 */
static void
sockd_call(void (*fun)(), void *arg)
{
        struct sockd_entry *entry = kmem_zone_alloc(sockd_zone, KM_NOSLEEP);
	register int s = mutex_spinlock(&sockd_lock);

	if (!entry) {
		entry = sockd_free;
		if (!entry)
                	cmn_err(CE_PANIC,
				"sockd_call: could not allocate sockd entry");
		else {
			sockd_free = entry->next;
			sockd_freecnt--;
		}
	}

        entry->func = fun;
        entry->arg1 = arg;
        entry->arg2 = NULL;
        entry->next = 0;
        if (!sockd_tail) {
                ASSERT(!sockd_head);
                sockd_head = sockd_tail = entry;
        } else {
                sockd_tail->next = entry;
                sockd_tail = entry;
        }
	mutex_spinunlock(&sockd_lock, s);
}

/*
 * sockd()
 *
 * consolidates the handling of timeout events for networking;
 * many of the event handlers manipulate the same data structures
 * so they are done sequentially by sockd; periodic hardcoded
 * timeout events [pffasttimeo, pfslowtimo, and if_slowtimo] are
 * handled in a function which returns the number of ticks to the
 * next desired wakeup.  Others timeouts [ipfiltertimer and
 * nfs client side timeouts] are handled after; such events are
 * queued to be run by sockd when a timeout occurs by sockd_call.
 *
 * note: periodic timeouts will occur at least every 200ms,
 * a timeout queued to sockd may take up to that much longer
 * to happen.  If there are many events queued to sockd it
 * has the effect of lengthening periodic timeouts somewhat which
 * actually might be desired when the system is heavily loaded.
 */
static void
sockd(void *pri)
{
	int s, ticks;
        struct sockd_entry *entry;

	curprocp->p_pri = curprocp->p_rtpri = (__psint_t)pri;

	while (1) {
		ticks = periodic_timeouts();

		s = mutex_spinlock(&sockd_lock);
		while (entry = sockd_head) {
			sockd_head = entry->next;
			if (sockd_head == 0) {
				ASSERT(entry == sockd_tail);
				sockd_tail = 0;
			}
			mutex_spinunlock(&sockd_lock, s);

			entry->func(entry->arg1, entry->arg2);

			s = mutex_spinlock(&sockd_lock);
			if (sockd_freecnt < sockd_freemax) {
				entry->next = sockd_free;
				sockd_free = entry;
				sockd_freecnt++;
			} else {
				kmem_zone_free(sockd_zone, entry);
			}
		}
		mutex_spinunlock(&sockd_lock, s);

		ASSERT(ticks >= 1);
		curprocp->p_stat = SSLEEP;
		s = splhi();
		timeout(setrun, curprocp, ticks);
		swtch(RESCHED_Y);
		splx(s);
	}
}
#endif /* _MP_NETLOCKS */
