/*
 * Copyright (c) 1982, 1986, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)in.c	8.2 (Berkeley) 11/15/93
 */

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/sema.h"
#include "sys/hashing.h"
#include "sys/atomic_ops.h"
#include "sys/kmem.h"

#include "sys/errno.h"
#include "sys/mbuf.h"
#include "sys/socket.h"
#include "sys/socketvar.h"
#include "sys/kthread.h"
#include "sys/systm.h"
#include "sys/debug.h"

#include "net/soioctl.h"
#include "net/if.h"
#include "net/route.h"

#include "in_systm.h"
#include "in.h"
#include "in_var.h"
#include "ip_var.h"

#include "igmp_var.h"
#include "sys/capability.h"

#define	satosin(sa)	((struct sockaddr_in *)(sa))
#define	satosin_new(sa)	((struct sockaddr_in_new *)(sa))

/*
 * Forward reference procedures
 */
void in_ifscrub(struct ifnet *, struct in_ifaddr *);
int in_ifinit(struct ifnet *, struct in_ifaddr *, struct sockaddr_in *,
	int, int);
int inaddr_match(struct hashbucket *, caddr_t, caddr_t, caddr_t);

int in_broadcast_full_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2);
int in_broadcast_net_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2);
int in_broadcast_match(struct hashbucket *, caddr_t, caddr_t, caddr_t);
int in_bcast_find_match(struct hashbucket *, caddr_t, caddr_t, caddr_t);

int in_getalias_enummatch(struct hashbucket *, caddr_t, caddr_t, caddr_t);

/*
 * External Procedures
 */
extern int ip_accept_primary_broadcast(struct in_addr *);
extern int ip_check_broadcast(struct in_ifaddr *, struct in_addr *);

extern void rawif_fixlink(struct ifnet *);
extern void link_rtrequest(int, struct rtentry *, struct sockaddr *);

/*
 * Module defined global data structures
 *
 * hashtable_inaddr contains the head and write lock for each hash chain
 */
struct hashinfo hashinfo_inaddr;
struct hashtable hashtable_inaddr[HASHTABLE_SIZE];

struct hashinfo hashinfo_inaddr_bcast;
struct hashtable hashtable_inaddr_bcast[HASHTABLE_SIZE];

int	in_interfaces;		/* number of external internet interfaces */
int	in_ifaddr_count;	/* number configured address'es including lb */

struct	ifqueue	ipintrq;		/* ip packet input queue */

/*
 * External variables
 */
extern struct ifnet loif;	/* loop-back interface ifnet structure */
extern int subnetsarelocal;
extern time_t time;
extern zone_t *ipmember_zone;

/*
 * This is a hash lookup match procedure which compares the ifaddr
 * fields of the in_ifaddr structure passed in so that we can find
 * the hash table entry corresponding to this real address.
 *
 * int
 * ifafree_match(struct hashbucket *h, struct in_addr *key,
 *	struct ifnet *arg1, sockaddr_in *arg2)
 *
 * Parameters:
 *  key is a pointer to the struct in_addr, containing the IP address
 *   to lookup in the hash table.
 *  arg1 is the ifnet pointer
 *  arg2 is a NULL pointer
 *
 * NOTE: This procedure is called with ALL types of hash bucket entries which
 * are linked onto the hash chain, so we explictly check for the HTF_INET ones.
 */
/* ARGSUSED */
int
ifafree_match(struct hashbucket *h, caddr_t key, caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia;
	struct in_addr *addr;
	struct ifnet *ifp;
	int match;

	if ((h->flags & HTF_INET) == 0) { /* not unicast entry */
		return 0;
	}

	ia = (struct in_ifaddr *)h;
	ifp = (struct ifnet *)arg1;
	/*
	 * NOTE: 'addr' is the new IP address since the in_ifaddr structure
	 * has already been updated. Without this we can't be sure we found
	 * the correct entry since it was inserted using the old IP address
	 * to compute the hash table index.
	 */
	addr = (struct in_addr *)arg2;

	if ((ia->ia_ifp == ifp) &&
	    (satosin(&ia->ia_addr)->sin_addr.s_addr == addr->s_addr)) {
		match = 1;
	} else {
		match = 0;
	}
	return match;
}

/*
 * This is the generic in_ifaddr address hash refcnt function.
 * It returns 1 if after decrementing the refcnt the new value
 * is less than 1 or ZERO otherwise.
 *
 * NOTE: This function is called from the generic hash reference count
 * procedure holding the write lock on the hash chain this entry is linked.
 */
int
ifafree_refcnt(struct hashbucket *h)
{
	register struct in_ifaddr *ia = (struct in_ifaddr *)h;
	register int new_refval;

	/*
	 * Decrement the reference count and return status indicating
	 * whether it has returned to less than 1 and can be delinked
	 * and the storage reclaimed.
	 */
	new_refval = atomicAddUint(&(ia->ia_ifa.ifa_refcnt), -1);
	return ((new_refval < 1) ? 1 : 0);
}

#ifdef DEBUG
/* ARGSUSED */
int
ifafree_check1(struct in_ifaddr *ia, struct in_ifaddr *ia_start,
	struct ifaddr *ifa)
{
	return 1;
}

/* ARGSUSED */
int
ifafree_check2(struct in_ifaddr *ia, struct in_ifaddr *ia_start,
	       struct ifaddr *ifa)
{
	return 2;
}

/* ARGSUSED */
int
ifafree_check3(struct in_ifaddr *ia, struct in_ifaddr *ia_start,
	       struct ifaddr *ifa)
{
	return 3;
}
#endif /* DEBUG */

/*
 * Free procedure for an ifaddr structure
 * Assume on entry that the 'ifa' address is valid as is the ref count.
 * Original code was:
 *
 * define IFAFREE(ifa) { \
 *     ASSERT(ROUTE_ISLOCKED()); \
 *	        if ((ifa)->ifa_refcnt <= 0) \
 *	                ifafree(ifa); \
 *	        else \
 *	                (ifa)->ifa_refcnt--;}
 *							
 * void ifafree(register struct ifaddr *ifa) {
 *         ASSERT(ROUTE_ISLOCKED());
 *	          if (ifa == NULL)
 *	                    panic("ifafree");
 *	          if (ifa->ifa_refcnt == 0)
 *	                  free(ifa, M_IFADDR);
 *	          else
 *		          ifa->ifa_refcnt--;
 *	}
 */
void
ifafree(struct ifaddr *ifa)
{
	struct in_ifaddr *ia, *ia_start;

	ia_start = (struct in_ifaddr *)ifa->ifa_start_inifaddr;
	if (ia_start == NULL) {
		panic("ifafree: ifa_start_inifaddr NULL: ia 0x%x, ifa 0x%x\n",
			ia_start, ifa);
	}
	if (ia_start->hashbucket.flags & HTF_NOTINTABLE) { /* ok deallocate */
		ia = ia_start;
#ifdef DEBUG
		ifafree_check1(ia, ia_start, ifa);
#endif
	} else {
		/*
		 * The hash refcnt procedure under the write lock for the
		 * hash chain first calls 'ifafree_refcnt' to perform the
		 * atomic refcnt decrement. The 'hash_refcnt' procedure returns
		 * the address of the entry if the 'ifafree_refcnt' procedure
		 * returned 1 indicating we can deallocate the storage or
		 * zero if not.
		 */
		ia = (struct in_ifaddr *)hash_refcnt(&hashinfo_inaddr,
				ifafree_refcnt,
				&(ia_start->hashbucket));
#ifdef DEBUG
		if (ia) {
			ifafree_check2(ia, ia_start, ifa);
		}
#endif
	}
	if (ia) {
		mcb_free(ia, sizeof(struct in_ifaddr), MT_IFADDR);
	}
	return;
}

/*
 * Free procedure for an ifaddr structure which forces delinking it from the
 * hash table and noting that fact in the hash table entry.
 * If the reference count indicates we're the last then the storage is freed.
 * The only difference between this procedure and 'ifafree' is the forcable
 * delink operation.
 */
void
ifafree_delink(struct ifaddr *ifa)
{
	struct in_ifaddr *ia, *ia_start;

	ia_start = (struct in_ifaddr *)ifa->ifa_start_inifaddr;

	if (ia_start == NULL) {
		panic("ifafree: ifa_start_inifaddr NULL: ia 0x%x, ifa 0x%x\n",
			ia_start, ifa);
		return;
	}
	/*
	 * Check for the case where no lookup is required as we already
	 * delinked the hash entry via an ioctl for deleting an IP address.
	 *
	 * Otherwise we call the generic hash remove with refcnt function.
	 * which first gets the write lock for the hash bucket chain then
	 * calls the 'ifafree_refcnt' function here, to handle the reference
	 * count operation. The ifafree_refcnt() returns 1 to indicate the
	 * hash table entry should be deallocated and ZERO otherwise.
	 * The 'hash_remove_refcnt' returns a non-zero pointer to the hash
	 * table entry to free when the ifafree_refcnt() procedure returned
	 * 1 and NULL otherwise.
	 *
	 * NOTE: We always dequeue the 'struct in_ifaddr' from the IP address
	 * hash chain BUT we deallocate the storage ONLY when a NON_ZERO
	 * address is returned from 'hash_remove_refcnt' procedure.
	 *
	 * This case arises when the reference count has NOT returned to '1'
	 * so address record must stay since we can have routing table entries
	 * pointing to this address record. One case is an ioctl operation
	 * was requested tp delete the IP address or IP alias address but
	 * references still remain.
	 */
	if (ia_start->hashbucket.flags & HTF_NOTINTABLE) {
		/*
		 * NOTE: The HTF_NOTINTABLE is set by 'hash_remove_refcnt'
		 * in the case when we delink it from the hash table BUT
		 * the refcnt indicates the storage can't be freed.
		 */
		ia = ia_start;
#ifdef DEBUG
		ifafree_check1(ia, ia_start, ifa);
#endif
	} else {
		ia = (struct in_ifaddr *)hash_remove_refcnt(&hashinfo_inaddr,
				ifafree_match,
				ifafree_refcnt,
				(caddr_t)&(ia_start->ia_addr.sin_addr),
				(caddr_t)(ia_start->ia_ifp),
				(caddr_t)&(ia_start->ia_addr.sin_addr));
#ifdef DEBUG
		if (ia) {
			ifafree_check3(ia, ia_start, ifa);
		}
#endif
	}
	if (ia) { /* ok to free it */
		mcb_free(ia, sizeof(struct in_ifaddr), MT_IFADDR);
	}
	return;
}

/*
 * Initialization procedure for network interface utilities
 */
void
in_init(void)
{
	ipmember_zone= kmem_zone_init(sizeof(struct ip_membership), "ipmulti");

	hash_init(&hashinfo_inaddr, hashtable_inaddr,
		HASHTABLE_SIZE, sizeof(struct in_addr), HASH_MASK);

	hash_init(&hashinfo_inaddr_bcast, hashtable_inaddr_bcast,
		HASHTABLE_SIZE, sizeof(struct in_addr), HASH_MASK);
	return;
}

/*
 * This is a hash lookup match procedure which compares the ia_addr
 * fields of the in_ifaddr structure passed in so that we can
 * find the real inifaddr entry if one exists.
 *
 * int
 * in_adjust_match(struct hashbucket *h, struct in_addr *key,
 *	struct ifnet *arg1, sockaddr_in *arg2)
 *
 * Parameters:
 *  key is a pointer to the old struct in_addr, containing the IP address
 *   that was used to insert this entry.
 *  arg1 is the ifnet pointer
 *  arg2 is a pointer to the new IP address
 *
 * NOTE: This procedure is called with ALL types of hash bucket entries which
 * are linked onto the hash chain, so we explictly check for the HTF_INET ones.
 */
/* ARGSUSED */
int
in_adjust_match(struct hashbucket *h, caddr_t key, caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia;
	struct in_addr *addr;
	struct ifnet *ifp;
	int match;

	if ((h->flags & HTF_INET) == 0) { /* not unicast entry */
		return 0;
	}

	ia = (struct in_ifaddr *)h;
	ifp = (struct ifnet *)arg1;
	/*
	 * NOTE: 'addr' is the new IP address since the in_ifaddr structure
	 * has already been updated. Without this we can't be sure we found
	 * the correct entry since it was inserted using the old IP address
	 * to compute the hash table index.
	 */
	addr = (struct in_addr *)arg2;

	if ((ia->ia_ifp == ifp) &&
	    (satosin(&ia->ia_addr)->sin_addr.s_addr == addr->s_addr)) {
		match = 1;
	} else {
		match = 0;
	}
	return match;
}

/*
 * This procedure is used to "adjust/update the hash table entries for
 * either/both the in_addr and in_broadaddr address associated with the
 * 'struct in_ifaddr' IF the value of either address'es are different than
 * when they were originally inserted into the ip address hash table and
 * the ip broadcast hash table respectively.
 * The algorithm performs a hash lookup using the original value, removes
 * the old entry and inserts it back into the hash table using the new address.
 *
 * In the case that either 'ia_addr_old' or 'ia_broadaddr_old' is NULL then
 * the hash table entries indexed by that field are not updated.
 *
 * NOTE: This procedure assumes that the 'struct in_ifaddr' has been updated
 * to contain the new address'es.
 */
void
in_adjust_hashentries(struct ifnet *ifp,
	struct in_ifaddr *ia,
	struct sockaddr_in *ia_addr_old,
	struct sockaddr_in *ia_broadaddr_old,
	struct in_addr *in_netbroadcast_old)
{
	struct in_ifaddr *ia_old;
	struct in_bcast *in_bcast_old, *in_bcast_net_old;
	int fix_ia_addr = 0;

	if (ia_addr_old) { /* check if adj address required */

		if (ia->ia_addr.sin_addr.s_addr !=
		    ia_addr_old->sin_addr.s_addr) { /* different addr */

			/*
			 * Check if we must fix primary address in_bcast entry
			 * NOTE: ppp interfaces must be explicted checked
			 * since they don't have broadcast address entries
			 * in the bcast hash table.
			 */
			if ((ia->ia_addrflags & IADDR_PRIMARY) &&
			    ((ifp->if_flags & IFF_POINTOPOINT) == 0)) {
				fix_ia_addr = 1;
			}

			ia_old = (struct in_ifaddr *)hash_lookup(
					&hashinfo_inaddr,
					in_adjust_match,
					(caddr_t)&(ia_addr_old->sin_addr),
					(caddr_t)ifp,
					(caddr_t)&(ia->ia_addr.sin_addr));

			if (ia_old) {
				hash_remove(&hashinfo_inaddr,
					    &(ia_old->hashbucket));

				(void)hash_insert(&hashinfo_inaddr,
					&(ia_old->hashbucket),
					(caddr_t)&(ia->ia_addr.sin_addr));
			}
		}
	}

	/*
	 * Adjust full broadcast entry if request or primary address changed
	 * NOTE: ppp interfaces don't have broadcast address entries, so this
	 * is handled by the caller passing in 'ia_broadaddr_old' and
	 * 'in_netbroadcast_old' as NULL and the 'fix_ia_addr' variable not
	 * set in the case the primary address is being adjusted when 'ifp'
	 * corresponds to a ppp interface. This way the insertion of a
	 * broadcast address entry/entries is avoided!
	 */
	if (ia_broadaddr_old || fix_ia_addr) {
		/*
		 * Check if we have a different full broadcast address
		 */
		if (ia_broadaddr_old &&
		    (ia->ia_broadaddr.sin_addr.s_addr !=
		     ia_broadaddr_old->sin_addr.s_addr)) { /* yes different */

			in_bcast_old = (struct in_bcast *)hash_lookup(
					&hashinfo_inaddr_bcast,
					in_broadcast_full_match,
					(caddr_t)&(ia_broadaddr_old->sin_addr),
					(caddr_t)ifp,
					(caddr_t)0);

			if (in_bcast_old) {
				hash_remove(&hashinfo_inaddr_bcast,
					    &(in_bcast_old->hashbucket));
				/*
				 * update primary address in broadcast entry
				 * if flag set and refcnt one. If the refcnt
				 * is greater than 1 then we can't change
				 * the primary address out from under this one
				 */
				if ((fix_ia_addr) &&
				    (in_bcast_old->refcnt == 1)) {
					in_bcast_old->ia_addr = ia->ia_addr;
				}

				(void)hash_insert(&hashinfo_inaddr_bcast,
					&(in_bcast_old->hashbucket),
					(caddr_t)&(ia->ia_broadaddr.sin_addr));
			}
		}
	}

	/*
	 * Adjust network broadcast entry if request or primary address changed
	 */
	if (in_netbroadcast_old || fix_ia_addr) {
		/*
		 * Check if we have a different network broadcast address
		 */
		if (in_netbroadcast_old &&
		    (ia->ia_netbroadcast.s_addr !=
		     in_netbroadcast_old->s_addr)) { /* yes different */

			/* re-adjust network broadcast address entry */
			in_bcast_net_old = (struct in_bcast *)hash_lookup(
					&hashinfo_inaddr_bcast,
					in_broadcast_net_match,
					(caddr_t)in_netbroadcast_old,
					(caddr_t)ifp,
					(caddr_t)0);

			if (in_bcast_net_old) {
				hash_remove(&hashinfo_inaddr_bcast,
					    &(in_bcast_net_old->hashbucket));
				/*
				 * update primary address in broadcast entry
				 * if flag set and reference count is one
				 */
				if ((fix_ia_addr) &&
				    (in_bcast_net_old->refcnt == 1)) {
					in_bcast_net_old->ia_addr= ia->ia_addr;
				}

				(void)hash_insert(&hashinfo_inaddr_bcast,
					&(in_bcast_net_old->hashbucket),
				       (caddr_t)&(ia->ia_netbroadcast.s_addr));
			}
		}
	}
	return;
}

/*
 * Allocate and initialize an in_bcast structure
 */
struct in_bcast *
in_bcast_alloc(struct in_ifaddr *ia)
{
	register struct in_bcast *ia_bcast;

	/* allocate and initialize in_bcast structure */
	ia_bcast = (struct in_bcast *)mcb_get(M_WAIT,
		sizeof(struct in_bcast), MT_IFADDR);

	ia_bcast->hashbucket.flags = HTF_BROADCAST;
	ia_bcast->ia_net = ia->ia_net;
	ia_bcast->ia_netmask = ia->ia_netmask;
	ia_bcast->ia_subnet = ia->ia_subnet;
	ia_bcast->ia_subnetmask = ia->ia_subnetmask;
	ia_bcast->ia_netbroadcast = ia->ia_netbroadcast;
	ia_bcast->ia_dstaddr = ia->ia_dstaddr;
	ia_bcast->ia_addr = ia->ia_addr;
	ia_bcast->ifp = ia->ia_ifp;
	ia_bcast->refcnt = 1;

	return ia_bcast;
}

/*
 * This is a hash lookup match procedure which compares the important
 * fields of the broadcast hash table entry for a complete match.
 * If a match is found we atomically decrements the reference count in
 * the entry by 1 and returns a match indication of non-zero IF the
 * reference count is less-than-or-equal to 1.
 * This allows us to remove the broadcast address entries associated with
 * a network interface from the broadcast hash table.
 *
 * in_bcast_free_match(struct hashbucket *h, struct in_addr *key,
 *  struct ifnet *arg1, struct in_ifaddr *arg2)
 *
 * NOTE: This procedure is called with only HTF_BROADCAST hash bucket entries,
 */
/* ARGSUSED */
int
in_bcast_free_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_bcast *ia_bcast;
	struct ifnet *ifp;
	struct in_ifaddr *ia;
	int match;

	if ((h->flags & HTF_BROADCAST) == 0) { /* not broadcast entry */
		return 0;
	}

	/*
	 * The address supplied must match all of the following fields in the
	 * in_bcast entry: ia_broadaddr, ia_netbroadcast, ia_subnet and ia_net
	 *
	 * NOTE: There are TWO entries inserted into the broadcast hash table
	 * that correspond to the full broadcast address AND the network
	 * broadcast address, so that either can be found via the hash lookup.
	 *
	 * So we check that supplied lookup key matches that the one used when
	 * the entry was inserted before we check the rest of the fields to be
	 * sure that the correct one is matched.
	 */
	if (bcmp(h->key, key, sizeof(struct in_addr) != 0)) { /* no match */
		return 0;
	}

	ia_bcast = (struct in_bcast *)h;
	ifp = (struct ifnet *)arg1;
	ia = (struct in_ifaddr *)arg2;

	if ((ia_bcast->ifp == ifp) &&
	    (ia_bcast->ia_broadaddr.sin_addr.s_addr ==
		ia->ia_broadaddr.sin_addr.s_addr) &&
	    (ia_bcast->ia_netbroadcast.s_addr ==
		ia->ia_netbroadcast.s_addr) &&
	    (ia_bcast->ia_net == ia->ia_net) &&
	    (ia_bcast->ia_subnet == ia->ia_subnet)) {

		/* return match indication if entry can be removed */
		match = 1;
	} else {
		/* return NO match indication */
		match = 0;
	}
	return match;
}

/*
 * This is a hash lookup match procedure which compares the important
 * fields of the broadcast hash table entry for a complete match.
 * If a match is found we atomically decrements the reference count in
 * the entry by 1 and returns a match indication of non-zero IF the
 * reference count is less-than-or-equal to 1.
 * This allows us to remove the broadcast address entries associated with
 * a network interface from the broadcast hash table.
 *
 * in_bcast_free_refcnt(struct hashbucket *h, struct in_addr *key,
 *  struct ifnet *arg1, struct in_ifaddr *arg2)
 *
 * NOTE: This procedure is called with only HTF_BROADCAST hash bucket entries,
 */
/* ARGSUSED */
int
in_bcast_free_refcnt(struct hashbucket *h)
{
	struct in_bcast *in_bcastp;
	int new_refcnt;

	if ((h->flags & HTF_BROADCAST) == 0) { /* not broadcast entry */
		return 0;
	}

	in_bcastp = (struct in_bcast *)h;
	new_refcnt = atomicAddUint(&in_bcastp->refcnt, -1);

	/* return match indication if entry can be removed */
	return ((new_refcnt < 1) ? 1 : 0);
}

/*
 * Decrement in_bcast entries associated with this in_ifaddr main address.
 * The two broadcast address are the full and netbroadcast one.  We decrement
 * the reference count and free if the count is below the threshold. We return
 * the number of entries delinked and whose storage has been reclaimed.
 */
int
in_bcast_free(struct in_ifaddr *ia, struct ifnet *ifp)
{
	register struct in_bcast *in_bcastp_remove;
	register int num_removed = 0;

	/*
	 * lookup the full broadcast address and perform
	 * an atomic refcnt decrement operation on this entry
	 */
	in_bcastp_remove = (struct in_bcast *)hash_remove_refcnt(
				&hashinfo_inaddr_bcast,
				in_bcast_free_match,
				in_bcast_free_refcnt,
				(caddr_t)(&(ia->ia_broadaddr.sin_addr)),
				(caddr_t)ifp,
				(caddr_t)ia);
	if (in_bcastp_remove) {
		num_removed++;
		mcb_free(in_bcastp_remove, sizeof(struct in_bcast), MT_IFADDR);
	}

	if (ia->ia_broadaddr.sin_addr.s_addr == ia->ia_netbroadcast.s_addr) {
		return num_removed;
	}

	in_bcastp_remove = (struct in_bcast *)hash_remove_refcnt(
				&hashinfo_inaddr_bcast,
				in_bcast_free_match,
				in_bcast_free_refcnt,
				(caddr_t)(&(ia->ia_netbroadcast)),
				(caddr_t)ifp,
				(caddr_t)ia);
	if (in_bcastp_remove) {
		num_removed++;
		mcb_free(in_bcastp_remove, sizeof(struct in_bcast), MT_IFADDR);
	}
	return num_removed;
}

/*
 * This procedure is called to lookup the broadcast address entries
 * and to increment the reference counts if they already exist. In the
 * event that they do NOT exist, we allocate and insert into the broadcast
 * address table the required entries.
 * This procedure is called when IP alias address'es are created and the
 * corresponding 'struct in_ifaddr' parameter has been allocated and
 * initialized.
 *
 * NOTE: In the case say where lots of Class C IP alias'es are created
 * there will be only TWO broadcast address entries for this entire group
 * assuming all of couse the following fields all match: ifp, ia_broadaddr,
 * ia_netbroadcast, ia_net and ia_subnet.
 * In this case the broadcast address entry associated with the primary
 * address assigned to the network interface and the netbroadcast address'e
 * reference counts will get incremented.
 *
 * NOTE: This procedure ASSUMES on entry that the network interface 'ifp'
 * supports broadcast address'es. The caller must ensure this invariant.
 *
 * Returns
 * -------
 * NON_ZERO => New entries were allocated and inserted into the hash table
 * ZERO => Existing entries exist so the reference count was incremented
 */
int
in_bcast_refcnt(struct ifnet *ifp, struct in_ifaddr *ia)
{
	struct in_bcast *in_bcastp, *ia_bcast_net;
	int status = 0;

	/*
	 * Check if we already have a hash table entry containing the full
	 * broadcast address address. If found we increment the reference
	 * count and return it's address.
	 */
	in_bcastp = (struct in_bcast *)hash_lookup(
				&hashinfo_inaddr_bcast,
				in_bcast_find_match,
				(caddr_t)&(ia->ia_broadaddr.sin_addr),
				(caddr_t)ifp,
				(caddr_t)ia);
	if (in_bcastp) {
		/*
		 * Exact match found on ia_broadaddr so now check if
		 * we have a different ia_netbroadcast value.
		 * Is so we lookup the second one and just so that
		 * it's reference count gets incremented.
		 */
		if (ia->ia_broadaddr.sin_addr.s_addr !=
		    ia->ia_netbroadcast.s_addr) { /* different */

			ia_bcast_net = (struct in_bcast *)hash_lookup(
				&hashinfo_inaddr_bcast,
				in_bcast_find_match,
				(caddr_t)&(ia->ia_netbroadcast),
				(caddr_t)ifp,
				(caddr_t)ia);
		}
	} else {
		/*
		 * No exact match found so check if we have more than one
		 * broadcast address which must be allocated and inserted.
		 * The two are the full broadcast address and the
		 * network broadcast address
		 */
		in_bcastp = in_bcast_alloc(ia);

		(void)hash_insert(&hashinfo_inaddr_bcast,
				&(in_bcastp->hashbucket),
				(caddr_t)(&(ia->ia_broadaddr.sin_addr)));

		if (ia->ia_broadaddr.sin_addr.s_addr !=
		    ia->ia_netbroadcast.s_addr) { /* have two address'es */

			ia_bcast_net = in_bcast_alloc(ia);

			(void)hash_insert(&hashinfo_inaddr_bcast,
				&(ia_bcast_net->hashbucket),
				(caddr_t)(&(ia->ia_netbroadcast)));
		}
		status = 1;
	}
	return status;
}

/*
 * This is a hash lookup match procedure which compares the important
 * fields of the broadcast hash table entry for a complete match.
 * If a match is found we atomically increment the reference count in
 * the entry and return a match indication.
 * This allows us to avoid inserting duplicate entries for the same
 * interface into the broadcast hash table.
 *
 * in_bcast_find_match(struct hashbucket *h, struct in_addr *key,
 *  struct ifnet *arg1, struct in_ifaddr *arg2)
 *
 * NOTE: This procedure is called with only HTF_BROADCAST hash bucket entries,
 * since the're stored in a separate has table, so the no explict checking
 * is required.
 */
/* ARGSUSED */
int
in_bcast_find_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia;
	struct in_bcast *in_bcastp;
	struct ifnet *ifp;
	int match;

	if ((h->flags & HTF_BROADCAST) == 0) { /* not broadcast entry */
		return 0;
	}

	/*
	 * Check that supplied lookup key matches that the one used when
	 * the entry was inserted before we check the rest of the fields to be
	 * sure that the correct one is matched.
	 */
	if (bcmp(h->key, key, sizeof(struct in_addr) != 0)) { /* no match */
		return 0;
	}

	in_bcastp = (struct in_bcast *)h;
	ifp = (struct ifnet *)arg1;
	ia = (struct in_ifaddr *)arg2;
	match = 0;

	/*
	 * The address supplied must match all of the following fields in the
	 * in_bcast entry: ia_broadaddr, ia_netbroadcast, ia_subnet and ia_net
	 *
	 * NOTE: There are TWO entries inserted into the broadcast hash table
	 * that correspond to the full broadcast address AND the network
	 * broadcast address, so that either can be found via the hash lookup.
	 */
	if ((in_bcastp->ifp == ifp) &&
	    (in_bcastp->ia_broadaddr.sin_addr.s_addr ==
		ia->ia_broadaddr.sin_addr.s_addr) &&
	    (in_bcastp->ia_netbroadcast.s_addr ==
		ia->ia_netbroadcast.s_addr) &&
	    (in_bcastp->ia_net == ia->ia_net) &&
	    (in_bcastp->ia_subnet == ia->ia_subnet)) {

		(void)atomicAddUint(&in_bcastp->refcnt, 1);
		match = 1;
	}
	return match;
}

/*
 * Allocate and initialize the network address structure
 */
struct in_ifaddr *
in_inifaddr_alloc(struct ifnet *ifp)
{
	struct in_ifaddr *ia;

	/*
	 * allocate and initialize in_ifaddr structure
	 */
	ia = (struct in_ifaddr *)mcb_get(M_WAIT,
			sizeof(struct in_ifaddr), MT_IFADDR);

	ia->hashbucket.flags = HTF_INET;
	/*
	 * initialize protocol independent ifaddr info
	 */
	IA_SIN(ia)->sin_family = AF_INET;
	ia->ia_ifp = ifp;
	
	/* initialize the address reference count */
	ia->ia_ifa.ifa_refcnt = 1;

	/* save start address of struct in_ifaddr */
	ia->ia_ifa.ifa_start_inifaddr = (caddr_t)ia;

	/* set address of addr and dstaddr sockaddr in ifaddr struct */
	ia->ia_ifa.ifa_addr = (struct sockaddr *)&ia->ia_addr;
	ia->ia_ifa.ifa_dstaddr = (struct sockaddr *)&ia->ia_dstaddr;

	ia->ia_ifa.ifa_netmask =
#ifdef _HAVE_SIN_LEN
		(struct sockaddr *)&ia->ia_sockmask;
#else
		(struct sockaddr_new *)&ia->ia_sockmask;
#endif
	/*
	 * NOTE: The setting of the sockmask address family field is probaly
	 * not required since the netmask comparison code assumes that only
	 * the sa_data array of the sockaddr matches. That portion starts with
	 * the port number and the remaining bytes of the address. The XOR an
	 * AND for loops in ifa_ifwithnet() and ifaof_ifpforaddr() in if.c is
	 * where all the work gets done. What is CRITICAL is that the 'sa_len'
	 * field of the ifa_netmask sockaddr be set properly since that
	 * controls the byte comparison/matching loop.
	 *
	 * Somewhat orthogonal to all of this is the sin_len field which was
	 * added in 4.4 BSD to the 'struct sockaddr'. The first byte becomes
	 * the length or number of bytes used in the address structure.
	 */
	ia->ia_sockmask.sin_family = AF_UNSPEC;
	ia->ia_sockmask.sin_len = _SIN_ADDR_SIZE;

	if (ifp->if_flags & IFF_BROADCAST) {
#ifdef _HAVE_SIN_LEN
		ia->ia_broadaddr.sin_len = sizeof(ia->ia_addr);
#endif
		ia->ia_broadaddr.sin_family = AF_INET;
	}
	return ia;
}

/*
 * returns a printable string version of an internet address
 */
char *
inet_ntoa(struct in_addr in)
{
	u_char n, *p;
	char *b;
	static char buf[20];
	int i;

	p = (u_char *)&in;
	b = buf;
	for (i = 0; i < 4; i++) {
		if (i)
			*b++ = '.';
		n = *p;
		if (n > 99) {
			*b++ = '0' + (n / 100);
			n %= 100;
		}
		if (*p > 9) {
			*b++ = '0' + (n / 10);
			n %= 10;
		}
		*b++ = '0' + n;
		p++;
	}
	*b = 0;
	return (buf);
}

/*
 * Internet address interpretation routine.
 * All the network library routines call this
 * routine to interpret entries in the data bases
 * which are expected to be an address.
 * The value returned is in network order.
 */
__uint32_t
inet_addr(register char *cp)
{
	register __uint32_t val, base, n;
	register char c;
	__uint32_t parts[4], *pp = parts;

again:
	/*
	 * Collect number up to ``.''.
	 * Values are specified as for C:
	 * 0x=hex, 0=octal, other=decimal.
	 */
	val = 0; base = 10;
	if (*cp == '0')
		base = 8, cp++;
	if (*cp == 'x' || *cp == 'X')
		base = 16, cp++;
	while (c = *cp) {
		if ((c >= '0') && (c <= '9')) {
			val = (val * base) + (c - '0');
			cp++;
			continue;
		}
		if (base == 16) {
			if ((c >= 'a') && (c <= 'f')) {
				val = (val << 4) + (c + 10 - 'a');
				cp++;
				continue;
			}
			if ((c >= 'A') && (c <= 'F')) {
				val = (val << 4) + (c + 10 - 'A');
				cp++;
				continue;
			}
		}
		break;
	}
	if (*cp == '.') {
		/*
		 * Internet format:
		 *	a.b.c.d
		 *	a.b.c	(with c treated as 16-bits)
		 *	a.b	(with b treated as 24 bits)
		 */
		if (pp >= parts + 4) {
			return (-1);
		}
		*pp++ = val, cp++;
		goto again;
	}
	/*
	 * Check for trailing characters.
	 */
	if (*cp && *cp != ' ' && *cp != '\t' && *cp != '\n') {
		return (-1);
	}
	*pp++ = val;
	/*
	 * Concoct the address according to
	 * the number of parts specified.
	 */
	n = pp - parts;
	switch (n) {

	case 1:				/* a -- 32 bits */
		val = parts[0];
		break;

	case 2:				/* a.b -- 8.24 bits */
		val = (parts[0] << 24) | (parts[1] & 0xffffff);
		break;

	case 3:				/* a.b.c -- 8.8.16 bits */
		val = (parts[0] << 24) | ((parts[1] & 0xff) << 16) |
			(parts[2] & 0xffff);
		break;

	case 4:				/* a.b.c.d -- 8.8.8.8 bits */
		val = (parts[0] << 24) | ((parts[1] & 0xff) << 16) |
		      ((parts[2] & 0xff) << 8) | (parts[3] & 0xff);
		break;

	default:
		return (-1);
	}
	val = htonl(val);
	return (val);
}

/*
 * This is a hash enumeration match procedure which compares the network
 * number desired with the network number in the hash bucket entry.
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
in_net_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	int net = (__psint_t)key;

	return ((net == ia->ia_net) ? 1 : 0);
}

/*
 * Return the network number from an internet address.
 */
__uint32_t
in_netof(struct in_addr in)
{
	register __uint32_t i = ntohl(in.s_addr);
	register __uint32_t net;
	register struct in_ifaddr *ia;

	if (IN_CLASSA(i))
		net = i & IN_CLASSA_NET;
	else if (IN_CLASSB(i))
		net = i & IN_CLASSB_NET;
	else if (IN_CLASSC(i))
		net = i & IN_CLASSC_NET;
	else if (IN_CLASSD(i))
		net = i & IN_CLASSD_NET;
	else
		return (0);
	/*
	 * Check whether network is a subnet; if so, return subnet number.
	 */
	ia = (struct in_ifaddr *)hash_enum(&hashinfo_inaddr,
		in_net_enummatch, HTF_INET,
		(caddr_t)(__psint_t)net, (caddr_t)0, (caddr_t)0);
	if (ia) {
		net = (i & ia->ia_subnetmask);
	}
	return net;
}

/*
 * This is a hash enumeration match procedure which gets called when the
 * subnets are NOT local. It AND's the network number desired with the
 * subnet mask in this entry to see if it matches the subnet number
 * in the hash bucket entry. The desired network number is supplied in 'key'.
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
in_localaddr_net_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	uint i = (__psint_t)key;

	return (((i & ia->ia_subnetmask) == ia->ia_subnet) ? 1 : 0);
}

/*
 * This is a hash enumeration match procedure which is called as the match
 * procedure WHEN 'subnetsarelocal' is non-zero. We compare the network
 * number desired with (the ip address in 'key' AND'ed with the network mask)
 * in this hash bucket entry. If they are equal we have a match.
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
in_localaddr_subnet_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	uint i = (__psint_t)key;

	return (((i & ia->ia_netmask) == ia->ia_net) ? 1 : 0);
}

/*
 * Return 1 if an internet address is for a ``local'' host
 * (one to which we have a connection).  If subnetsarelocal
 * is true, this includes other subnets of the local net.
 * Otherwise, it includes only the directly-connected (sub)nets.
 */
int
in_localaddr(struct in_addr in)
{
	register struct in_ifaddr *ia;
	register __uint32_t i;

	i = ntohl(in.s_addr);

	if (subnetsarelocal) {
		ia = (struct in_ifaddr *)hash_enum(&hashinfo_inaddr,
			in_localaddr_subnet_enummatch, HTF_INET,
			(caddr_t)(__psint_t)i, (caddr_t)0, (caddr_t)0);
	} else {
		ia = (struct in_ifaddr *)hash_enum(&hashinfo_inaddr,
			in_localaddr_net_enummatch, HTF_INET,
			(caddr_t)(__psint_t)i, (caddr_t)0, (caddr_t)0);
	}
	return ((ia) ? 1 : 0);
}

/*
 * Determine whether an IP address is in a reserved set of addresses
 * that may not be forwarded, or whether datagrams to that destination
 * may be forwarded.
 */
int
in_canforward(struct in_addr in)
{
	register __uint32_t i = ntohl(in.s_addr);
	register __uint32_t net;

	if (IN_EXPERIMENTAL(i) || IN_MULTICAST(i))
		return (0);
	if (IN_CLASSA(i)) {
		net = i & IN_CLASSA_NET;
		if (net == 0 || net == (IN_LOOPBACKNET << IN_CLASSA_NSHIFT))
			return (0);
	}
	return (1);
}

/*
 * This enumaration match procedure is called after for each entry in the
 * hash table. We are checking for ANY Internet entry which matches the
 * interface address supplied in 'arg1'.
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
ifnet_any_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct ifnet *ifp = (struct ifnet *)arg1;

	return ((ia->ia_ifp == ifp) ? 1 : 0);
}

/*
 * This enumaration match procedure is called after for each entry in the
 * hash table. We are checking for the primary IP address associated
 * with this interface. We return a TRUE if a matching entry is founc.
 *
 * int ifnet_enummatch(struct hashbucket *h, caddr_t key,
 *   struct ifnet *arg1, caddr_t arg2)
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
ifnet_enummatch(struct hashbucket *h, caddr_t key, caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct ifnet *ifp = (struct ifnet *)arg1;
	int match = 0;

	if ((ia->ia_ifp == ifp) && (ia->ia_addrflags & IADDR_PRIMARY))
		match = 1;
	return match;
}

/*
 * This match procedure is called after the hashing lookup has indexed
 * into the hash table and is comparing hash bucket entries for a match
 * with the supplied key (IP address). We return the non-zero if the key
 * matches the ia_addr field of the 'struct in_ifaddr'.
 * NOTE: The key is the same as the ia_'addr' field.
 */
/* ARGSUSED */
int
inaddr_match(struct hashbucket *h, caddr_t key, caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia;
	struct in_addr *addr;
	int match = 0;

	if (h->flags & HTF_INET) { /* internet address */

		ia = (struct in_ifaddr *)h;
		addr = (struct in_addr *)key;

		if (IA_SIN(ia)->sin_addr.s_addr == addr->s_addr) {
			match = 1;
		}
	}
	return match;
}

/*
 * This match procedure is called after the hashing lookup has matched
 * the supplied IP address AND the supplied 'struct ifnet' address
 * matches that which was stored in the entry.
 *
 *inaddr_ifnetaddr_match(struct in_ifaddr *ia, struct in_addr *key,
 *	struct ifnet *arg1, caddr_t arg2)
 */
/* ARGSUSED */
int
inaddr_ifnetaddr_match(struct hashbucket *h, caddr_t key, caddr_t arg1,
	caddr_t arg2)
{
	struct in_ifaddr *ia;
	struct in_addr *addr;
	struct ifnet *ifp;
	int match = 0;

	if (h->flags & HTF_INET) { /* internet address */
		ia = (struct in_ifaddr *)h;
		addr = (struct in_addr *)key;
		ifp = (struct ifnet *)arg1;

		if ((ia->ia_ifp == ifp) &&
		    IA_SIN(ia)->sin_addr.s_addr == addr->s_addr) {
			match = 1;
		}
	}
	return match;
}

/*
 * Procedure that finds the interface (ifnet structure) corresponding
 * to one of our IP addresses. Used by the macro of the upper case
 * version of the same name. This is the optimized case where we
 * do the direct hash table lookup to see if we have a matching
 * source IP address.
 */
struct ifnet *
inaddr_to_ifp(struct in_addr addr)
{
	register struct in_ifaddr *ia;

	ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
			inaddr_match,
			(caddr_t)&(addr),
			(caddr_t)0,
			(caddr_t)0);

	return (((ia) ? ia->ia_ifp : (struct ifnet *)0));
}

/*
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
in_rtinit_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct ifnet *ifp = (struct ifnet *)key;

	return (((ia->ia_ifp == ifp) && (ia->ia_flags & IFA_ROUTE)) ? 1 : 0);
}

/*
 * Delete a host route for an in_ifaddr structure if one exists for
 * the supplied interface
 *
 * NOTE: This procedure is used by the slip and pp drivers to delete
 * there host route on closing. This procedure removes these drivers
 * from having to know about all the hash table ip address lookup stuff.
 */
int
in_rtinit(struct ifnet *ifp)
{
	struct in_ifaddr *ia;

	ia = (struct in_ifaddr *)hash_enum(&hashinfo_inaddr,
		in_rtinit_enummatch, HTF_INET,
		(caddr_t)ifp, (caddr_t)0, (caddr_t)0);
	if (ia) {
		(void)rtinit(&ia->ia_ifa, RTM_DELETE, RTF_HOST);
		ia->ia_flags &= ~IFA_ROUTE;

		ia->ia_addr.sin_family = AF_UNSPEC;
		ia->ia_dstaddr.sin_family = AF_UNSPEC;
	}
	return 1;
}

/*
 * Trim a mask in a sockaddr_in structure
 * Produce a length of 0 for an address of 0.
 * Otherwise produce the index of the first zero byte.
 */
void
in_socktrim(ap)
#ifdef _HAVE_SIN_LEN
	struct sockaddr_in *ap;
#else
	struct sockaddr_in_new *ap;
#endif
{
	register char *cp;

	if (ap->sin_addr.s_addr == 0) {
		ap->sin_len = 0;
		return;
	}
	cp = (char *)(&ap->sin_addr.s_addr+1);
	while (*--cp == 0)
		continue;
	ap->sin_len = cp - (char*)ap + 1;
	return;
}

/*
 * Generic internet control operations (ioctl's).
 * Ifp is 0 if not an interface-specific ioctl.
 */
/* ARGSUSED */
int
in_control(struct socket *so, __psint_t cmd, caddr_t data, struct ifnet *ifp)
{
	struct in_aliasreq *ifra;
	struct ifaliasreq *ifa_req;
	struct ifreq *ifr = (struct ifreq *)data;
	struct in_ifaddr *ia;
	__uint32_t i;
	int error = 0;
	NETSPL_DECL(s)
	NETSPL_DECL(s1)

	/*
	 * Find address for this interface, if it exists.
	 */
	if (ifp) {
		IFNET_LOCK(ifp, s);	
		/*
		 * The primary address associated with this ifnet record
		 * is stored in the ifnet record. In the old linked-list
		 * scheme it was the first 'struct ifaddr' hung off the
		 * ifnet structure which contained a matching ifnet address.
		 */
		ia = (struct in_ifaddr *)(ifp->in_ifaddr);
	} else {
		ia = (struct in_ifaddr *)0;
	}

	switch (cmd) {

	case SIOCGIFADDR:	/* Get interface address */
		if (ia) {
			ifr->ifr_addr = *(ia->ia_ifa.ifa_addr);
		} else {
			error = EADDRNOTAVAIL;
		}
		break;

	case SIOCGIFDSTADDR:	/* Get destination address */
		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}
		if ((ifp->if_flags & IFF_POINTOPOINT) == 0) {
			error = EINVAL;
		} else {
			if (ia) {
				ifr->ifr_dstaddr = *(ia->ia_ifa.ifa_dstaddr);
			} else {
				error = EADDRNOTAVAIL;
			}
		}
		break;

	case SIOCGIFBRDADDR:	/* Get broadcast address */
		if (ifp == 0) {
			error = EADDRNOTAVAIL;
			break;
		}
		if ((ifp->if_flags & IFF_BROADCAST) == 0) {
			error = EINVAL;
		} else {
			if (ia) {
				ifr->ifr_dstaddr = *(ia->ia_ifa.ifa_broadaddr);
			} else {
				error = EADDRNOTAVAIL;
			}
		}
		break;

	case SIOCGIFNETMASK:	/* Get network mask */

		if (ia) {
#ifdef _HAVE_SIN_LEN
			*satosin(&ifr->ifr_addr) = ia->ia_sockmask;
#else
			*satosin_new(&ifr->ifr_addr) = ia->ia_sockmask;
#endif
		} else {
			error = EADDRNOTAVAIL;
		}
		break;

	case SIOCAIFADDR:	/* Add interface alias address */
	{
		int iaIsNew = 0, hostIsNew = 1, maskIsNew = 0;
		int local_ia_addrflags = IADDR_ALIAS;
		int alias;

		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}
		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = curthreadp->k_error;
			break;
		}
		ifra = (struct in_aliasreq *)data;

		if (ia) {
			/*
			 * primary address set in ifnet record so check
			 * if we have match addresses. If so this is a
			 * duplicate of the primary address for this
			 * interface.
			 */
			if ((ia->ia_ifp == ifp) &&
				(IA_SIN(ia)->sin_addr.s_addr ==
				  ifra->ifra_addr.sin_addr.s_addr)) {
				hostIsNew = 0;
				goto foundit;
			}
		}
		/*
		 * Find address entry matching both ifnet and in_addr
		 */
		ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
			inaddr_ifnetaddr_match,
		(caddr_t)(&(satosin(&(ifra->ifra_addr))->sin_addr)),
			(caddr_t)ifp,
			(caddr_t)0);

		if (ia && (ia->ia_addr.sin_family == AF_INET)) {
			/*
			 * Found a matching entry so we do NOT
			 * have to add a new alias address
			 */
#ifdef _HAVE_SIN_LEN
			if (ifra->ifra_addr.sin_len == 0)
#else
			if (ifra->ifra_addr.sin_addr.s_addr == 0)
#endif
			{
				/*
				 * Typical case is where ifconfig
				 * set's the netmask first and only
				 * later sets the ifaddr. In that case
				 * the the primary address is zero.
				 * The other case when the supplied
				 * interfaces address is zero, we
				 * set the primary address in the
				 * request structure, but it's not
				 * returned to the user process.
				 */
				ifra->ifra_addr = ia->ia_addr;
				hostIsNew = 0;
			} else {
				/*
				 * Check if add alias is a no-op case
				 * if we already have that address
				 */
				if (ifra->ifra_addr.sin_addr.s_addr ==
				       ia->ia_addr.sin_addr.s_addr) {
					hostIsNew = 0;
				}
			}
		} else {
			/*
			 * Allocate and initialize in_ifaddr structure
			 * Also note that this address needs to be
			 * added to the ip address hash table.
			 */
			ia = in_inifaddr_alloc(ifp);
			iaIsNew = 1;

			if (ifp->in_ifaddr == (caddr_t)0) {
				/*
				 * Note the primary in_ifaddr for the
				 * interface has NOT been set AND this
				 * is being done via an alias "type"
				 * ioctl operation.
				 * We save the address flags to use
				 * later because it's the primary and
				 * save the address of the record for
				 * this interface so it's always valid.
				 */
				local_ia_addrflags = ia->ia_addrflags
					= IADDR_PRIMARY;

				ifp->in_ifaddr = (caddr_t)ia;
				if (ifp != &loif) {
				  (void)atomicAddInt(&in_interfaces,1);
				}
			 }
		}
foundit:
		if (ifra->ifra_mask.sin_len) {

			in_ifscrub(ifp, ia);

			ia->ia_sockmask = ifra->ifra_mask;
			/*
			 * NOTE: The af_family must NOT be set in the
			 * ia_sockmask field. The af_family is passed here
			 * via the alias ioctl but it's cleared here.
			 * The reason is the address comparison procedures
			 * require that only the length field be set. You
			 * can argue this either way but BSD 4.4lite2 doesn't
			 * set it in the routing socket code and so to avoid
			 * getting the wrong comparison results any code which
			 * calls the rn_refines procedures must adhere to the
			 * length only policy. The reason is that the radix
			 * tree code requires the first byte of the byte
			 * string to compare contains the length. Hence the
			 * ifa_ifwithnet* procedure(s) fall into this category
			 * since they call the radix tree comparison functions.
			 */
			ia->ia_sockmask.sin_family = AF_UNSPEC;

			ia->ia_subnetmask =
				ntohl(ia->ia_sockmask.sin_addr.s_addr);
			maskIsNew = 1;
		}

		if ((ifra->ifra_dstaddr.sin_family == AF_INET) &&
			(ifp->if_flags & IFF_POINTOPOINT)) {

			/* delete any existing route for this interface */
			in_ifscrub(ifp, ia);

			/* update destination address in 'struct in_ifaddr' */
			ia->ia_dstaddr =
				*((struct sockaddr_in *)(&ifra->ifra_dstaddr));
			maskIsNew  = 1;	/* We lie; but the effect's the same */
	        }

		if (ifra->ifra_addr.sin_family == AF_INET) {

			if (iaIsNew) {
				/*
				 * Add address record to IP address hash table
				 */
				IA_SIN(ia)->sin_family = AF_INET;
				IA_SIN(ia)->sin_addr= ifra->ifra_addr.sin_addr;
				ia->hashbucket.flags = HTF_INET;
				ia->ia_addrflags = local_ia_addrflags;

				(void)hash_insert(&hashinfo_inaddr,
					&(ia->hashbucket),
			(caddr_t)(&(satosin(&(ifra->ifra_addr))->sin_addr)));

				/* Bump number of configured address'es */
				(void)atomicAddInt(&in_ifaddr_count, 1);
			}

			if (hostIsNew || maskIsNew) {
				/*
				 * initialize address flags
				 */
				ia->ia_addrflags = local_ia_addrflags;

				/*
				 * initialize new route for this interface.
				 * The parameters mean don't scrub old
				 * interface and make host route since this
				 * is an alias address.
				 */
				alias = (local_ia_addrflags & IADDR_ALIAS)
						? 1 : 0;
				error = in_ifinit(ifp, ia,
					satosin(&ifra->ifra_addr), 0, alias);
		        }
	        }

		if ((ifra->ifra_broadaddr.sin_family == AF_INET) &&
			(ifp->if_flags & IFF_BROADCAST)) {

			/* update broadcast address in 'struct in_ifaddr' */
			ia->ia_broadaddr =
			     *((struct sockaddr_in *)(&ifra->ifra_broadaddr));

			if (iaIsNew &&
			    (ia->ia_broadaddr.sin_addr.s_addr != INADDR_ANY)) {
				/*
				 * Search for a matching broadcast address
				 * entry in the broadcast hash table.
				 * Create them if not found OR increment
				 * the reference counts if they already exist.
				 */
				(void)in_bcast_refcnt(ifp, ia);
			}
		} else {
			/*
			 * We arrive here when the broadcast address isn't
			 * explictly specified on the command line.
			 * In this case in_ifinit() would have set it to the
			 * default value depending on the class of the
			 * specified IP address. So we only have to handle
			 * creating the hash table entries for the broadcast
			 * addresses for this ip alias if the interface
			 * supports broadcast and we added the main address.
			 */
			if (hostIsNew) {

				if (iaIsNew &&
				    (ia->ia_broadaddr.sin_addr.s_addr
						!= INADDR_ANY) &&
				    (ifp->if_flags & IFF_BROADCAST)) {

					(void)in_bcast_refcnt(ifp, ia);
				}
			}
		}
	}
		break;

	case SIOCDIFADDR:	/* Remove interface alias address */

		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = curthreadp->k_error;
			break;
		}
		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}

		ifra = (struct in_aliasreq *)data;
		if (ifra->ifra_addr.sin_family == AF_INET) {
			/*
			 * Look for match of ifnet and s_addr fields.
			 */
			ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
				inaddr_ifnetaddr_match,
			(caddr_t)(&(satosin(&(ifra->ifra_addr))->sin_addr)),
				(caddr_t)ifp,
				(caddr_t)0);
		}
		if (ia == 0) {
			error = EADDRNOTAVAIL;
			break;
		}
		if (ifp->in_ifaddr == (caddr_t)ia) {
			/*
			 * Requested to delete primary address for interface!
			 * We make this an error since we assume the user
			 * doesn't know the consequences of this action.
			 * The problem is that the alias address'es have
			 * host routes associated with them and the primary
			 * address is used in the host route. Deleting the
			 * primary address when alias'es exist causes serious
			 * problems when trying to delete subsequent alias'es.
			 * This could return no error, aka a no-op instead.
			 */
			error = EOPNOTSUPP;
			break;
		}

		/* delete any existing route for this interface */
		in_ifscrub(ifp, ia);

		/* Decrement number of configured address'es */
		(void)atomicAddInt(&in_ifaddr_count, -1);

		/*
		 * Call generic ifa free procedure to delink the entry
		 * from hash table and free memory if reference count is zero.
		 */
		ROUTE_MPLOCK(s1);
		ifafree_delink(&(ia->ia_ifa));
		ROUTE_MPUNLOCK(s1);
		/* XXX can't ia have been freed by this point? */
		(void)in_bcast_free(ia, ifp);
		break;

	case SIOCLIFADDR:	/* Get list of interface network addresses */

		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}

		ifa_req = (struct ifaliasreq *)data;

		if (ifa_req->ifra_addr.sa_family == AF_INET) {

			/* Check if the requested 'nth' element is valid */
			if (ifa_req->cookie < 0) { /* bad nth element req */
				error = EINVAL;
				break;
			}

			if (ifa_req->cookie == 0) {
				/*
				 * return primary address for the interface
				 */
				ia = (struct in_ifaddr *)(ifp->in_ifaddr);
			} else {
				/*
				 * return all alias'es for the interface
				 */
				ia = (struct in_ifaddr *)hash_enum_getnext(
					&hashinfo_inaddr,
					in_getalias_enummatch,
					HTF_INET,
					1, /* skip primary addr which is zero*/
					ifa_req->cookie,
					(caddr_t)ifp);
			}

			if (ia == (struct in_ifaddr *)0) {
				ifa_req->cookie = -1;
				break;
			}
			ifa_req->cookie++;
			ifa_req->ifra_addr = *(ia->ia_ifa.ifa_addr);
			ifa_req->ifra_broadaddr = *(ia->ia_ifa.ifa_broadaddr);
#ifdef _HAVE_SIN_LEN
			*satosin(&ifa_req->ifra_mask) = ia->ia_sockmask;
#else
			*satosin_new(&ifa_req->ifra_mask) = ia->ia_sockmask;
#endif /* _HAVE_SIN_LEN */
		}
		break;

	case SIOCSIFBRDADDR:	/* Set interface broadcast address */
		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}
		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = curthreadp->k_error;
			break;
		}

		if ((ifp->if_flags & IFF_BROADCAST) == 0) {
			error = EINVAL;
		} else {
			if (ia) {
				ia->ia_broadaddr =
				*((struct sockaddr_in *)(&ifr->ifr_broadaddr));
			} else {
				error = EADDRNOTAVAIL;
			}
		}
		break;

	case SIOCSIFNETMASK:	/* Set interface network mask */

		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = curthreadp->k_error;
			break;
		}
		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}

		if (ia == (struct in_ifaddr *)0) {

			ia = in_inifaddr_alloc(ifp);
			/*
			 * We set the not-in-hash flag to indicate that
			 * the primary address is NOT available and so this
			 * entry hashn't been inserted into the address table
			 * This is because we allow ifconfig to set the
			 * network mask BEFORE setting the interface address
			 */
			ia->hashbucket.flags = HTF_INET | HTF_NOTINTABLE;

			if (ifp != &loif) {
				(void)atomicAddInt(&in_interfaces, 1);
			}

			if (ifp->in_ifaddr == (caddr_t)0) {
				/*
				 * Note the primary 'struct in_ifaddr' for this
				 * interface since it has NOT been set yet
				 */
				ia->ia_addrflags = IADDR_PRIMARY;
				ifp->in_ifaddr = (caddr_t)ia;
			}
		}
		i = satosin(&(ifr->ifr_addr))->sin_addr.s_addr;
		ia->ia_subnetmask = ntohl(ia->ia_sockmask.sin_addr.s_addr = i);

		in_socktrim(&ia->ia_sockmask);
		break;

	case SIOCSIFADDR:	/* Set interface address */
		{
		struct sockaddr_in ia_addr_old;
		struct sockaddr_in ia_broadaddr_old, *ia_broad;
		struct in_addr ia_netbroadcast_old, *ia_netbroad;
		int addBcast = 1;
		int iaAdjust = 0;


		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = curthreadp->k_error;
			break;
		}
		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}
		if (ia == (struct in_ifaddr *)0) {

			/* allocate and initialize in_ifaddr structure */
			ia = in_inifaddr_alloc(ifp);

			/* set the interface address */
			IA_SIN(ia)->sin_addr.s_addr =
				satosin(&ifr->ifr_addr)->sin_addr.s_addr;
			/*
			 * Add 'ia' to 'inaddr' hash table
			 * The majority of the fields in 'struct in_ifaddr'
			 * are set by the in_ifinit procedure.
			 */
			(void)hash_insert(&hashinfo_inaddr,
				&(ia->hashbucket),
			 (caddr_t)(&(satosin(&(ifr->ifr_addr))->sin_addr)));

			/* Bump number of configured address'es */
			(void)atomicAddInt(&in_ifaddr_count, 1);

			if (ifp->in_ifaddr == (caddr_t)0) {
				/*
				 * Note the primary address'es in_ifaddr' for
				 * the interface since it has NOT been set.
				 */
				ia->ia_addrflags = IADDR_PRIMARY;
				ifp->in_ifaddr = (caddr_t)ia;
			}

			if (ifp != &loif) {
				(void)atomicAddInt(&in_interfaces, 1);
			}
			/*
			 * NOTE: The default case is to add the broadcast
			 * address'es AFTER the main ia structure is fully
			 * initialized down below, since the netbroadcast
			 * and other fields are set by 'in_ifinit'.
			 */
		} else {
			/*
			 * If we arrive here the ifnet structure already has
			 * it's in_ifaddr address set. We check if the primary
			 * address needs to be inserted into the IP address
			 * hash table.
			 *
			 * NOTE: The not-in-table case arises when:
			 *
			 * 1. ifconfig does the ioctl to set the network mask
			 *    first then the ioctl to set the interface address
			 * 2. For the point-to-point interfaces we allow
			 *    the destination address to be set before the
			 *    interface address is set.
			 */
			if (ia->hashbucket.flags & HTF_NOTINTABLE) {

				ia->hashbucket.flags &= ~HTF_NOTINTABLE;
				ia->ia_addrflags = IADDR_PRIMARY;

				/* initialize the primary address info */
				IA_SIN(ia)->sin_family = AF_INET;
				IA_SIN(ia)->sin_addr.s_addr =
				  satosin(&ifr->ifr_addr)->sin_addr.s_addr;

				ia->ia_ifp = ifp;
				(void)hash_insert(&hashinfo_inaddr,
					&(ia->hashbucket),
			(caddr_t)(&(satosin(&(ifr->ifr_addr))->sin_addr)));

				/* Bump number of configured address'es */
				(void)atomicAddInt(&in_ifaddr_count, 1);
			} else {
				/*
				 * We arrive here when the 'ia_addr' field
				 * of an existing in_ifaddr structure changes.
				 * This may require the hash entry for the
				 * ia_addr be moved so that it will reside
				 * on the correct hash chain depending on
				 * the hash index computed from it's IP address
				 * NOTE; We must save ia_addr, ia_broadcast
				 * and the ia_netbroadcast values so that the
				 * adjustment of the hash table entries can be
				 * done after calling in_ifinit() since it
				 * initializes the new fields, updates the
				 * routes, etc.
				 * There are three entries to consider, the
				 * main ip address which resides in the inaddr
				 * hash table AND up to two broadcast addresses
				 * which reside in the bcast hash table.
				 * All of these entries may require adjustment
				 * in the hash table in the event any of there
				 * addresses change.
				 * We save the old ia_addr address so that
				 * we can do the adjustment after the
				 * in_ifinit call, since that's where the
				 * 'in_ifaddr' fields and the routes are set.
				 *
				 * We also reset the address flag since this
				 * address could have been an alias which is
				 * now going to be the primary address. This
				 * case is easily reproduced by the commands:
				 * ifconfig ec0 alias host => set primary
				 * ifconfig ec0 alias alt =>add alias 'alt'
				 * ifconfig ec0 -alias alt => clear alias
				 * ifconfig ec0 down => down net interface
				 * ifconfig ec0 alt => set 'alt' as primary
				 */
				ia->ia_addrflags = IADDR_PRIMARY;
				ia_addr_old = ia->ia_addr;

				/* save copies of both bcast addresses */
				ia_broadaddr_old = ia->ia_broadaddr;
				ia_netbroadcast_old = ia->ia_netbroadcast;

				/* Don't add new bcast hash table entries */
				addBcast = 0;

				/* Note adjust unicast and bcast entries */
				iaAdjust = 1;
			}
		}
		/*
		 * Scrub the old route before initializing the new
		 * one and then initialize the in_ifaddr structure.
		 */
		error = in_ifinit(ifp, ia, satosin(&ifr->ifr_addr), 1, 0);

		/*
		 * Now that the 'ia' structure is fully initialized we can
		 * create and insert into the hash table the broadcast entries
		 * for all interfaces which support broadcast. PPP ones don't
		 * an so a braodcast hash table entry won't be created.
		 */
		if (addBcast &&
		    (ia->ia_broadaddr.sin_addr.s_addr != INADDR_ANY) &&
		    (ifp->if_flags & IFF_BROADCAST)) {

			(void)in_bcast_refcnt(ifp, ia);
		}

		if (iaAdjust) {
			/*
			 * On IFF_POINTOPOINT interfaces only the ip address
			 * of the source is updated as the dst addr must be
			 * updated via a different ioctl operation.
			 */
			if (ifp->if_flags & IFF_POINTOPOINT) {
				ia_broad = (struct sockaddr_in *)0;
				ia_netbroad = (struct in_addr *)0;
			} else {
				ia_broad = &ia_broadaddr_old;
				ia_netbroad = &ia_netbroadcast_old;
			}
			in_adjust_hashentries(ifp, ia, &ia_addr_old,
					      ia_broad, ia_netbroad);
			break;
		}
	}
		break;

	case SIOCSIFDSTADDR:		/* Set interface destination address */
	    {
		struct sockaddr_in oldaddr;

		if (!_CAP_ABLE(CAP_NETWORK_MGT)) {
			error = curthreadp->k_error;
			break;
		}
		if (ifp == 0) {
			error = EOPNOTSUPP;
			break;
		}
		/*
		 * Operation is valid if only a point-to-pont interface
		 */
		if ((ifp->if_flags & IFF_POINTOPOINT) == 0) {
			error = EINVAL;
			break;
		}

		if (ia == (struct in_ifaddr *)0) {

			/* allocate and initialize in_ifaddr structure */
			ia = in_inifaddr_alloc(ifp);

			if (ifp != &loif) {
				(void)atomicAddInt(&in_interfaces, 1);
			}

			if (ifp->in_ifaddr == (caddr_t)0) {
				/*
				 * Note the primary 'struct in_ifaddr' for this
				 * interface since it has NOT been set yet
				 * SO we not that the in_ifaddr entry can't be
				 * in the hash table.
				 * NOTE: This is a little weird, allowing
				 * ifconfig or ifctl callers to set a number
				 * of other parameters before ever setting the
				 * primary IP address. Hence we handle this
				 * case by noting to insert the entry when we
				 * get the primary IP address.
				 */
				ia->hashbucket.flags |= HTF_NOTINTABLE;
				ifp->in_ifaddr = (caddr_t)ia;
			}
		}

		oldaddr = ia->ia_dstaddr;
		ia->ia_dstaddr = *(struct sockaddr_in *)&(ifr->ifr_dstaddr);

		if (ifp->if_ioctl &&
		    (error = (*ifp->if_ioctl)(ifp, SIOCSIFDSTADDR,
				&(ia->ia_ifa)))) {
			ia->ia_dstaddr = oldaddr;
			break;
		}
		if (ia->ia_flags & IFA_ROUTE) {
			/*
			 * entry has route associated with it so delete
			 * the old one and add one with the new dest addr
			 */
			ia->ia_ifa.ifa_dstaddr = (struct sockaddr *)&oldaddr;
			(void)rtinit(&(ia->ia_ifa), (int)RTM_DELETE, RTF_HOST);
			ia->ia_ifa.ifa_dstaddr =
				(struct sockaddr *)&ia->ia_dstaddr;
			(void)rtinit(&(ia->ia_ifa), (int)RTM_ADD,
				RTF_HOST|RTF_UP);
		}
	    }
		break;

	default:
		if (ifp == 0 || ifp->if_ioctl == 0) {
			error = EOPNOTSUPP;
			break;
		}
		error = ((*ifp->if_ioctl)(ifp, cmd, data));
	}

	if (ifp) {
		IFNET_UNLOCK(ifp, s);
	}
	return error;
}

/*
 * Delete any existing route for an interface.
 *
 * NOTE:  For alias address'es we delete the host route we created
 * for the alias address when it was added.
 */
void
in_ifscrub(struct ifnet *ifp, struct in_ifaddr *ia)
{
	struct sockaddr *save_addr, *save_dstaddr;
	int flags;

	if (ia->ia_flags & IFA_ROUTE) {

		if (ia->ia_addrflags & IADDR_ALIAS) {
			/*
			 * In the IP alias address case, since we have a host
			 * route we need to fixup the dstaddr field so that it
			 * points to the correct address. Host routes imply
			 * the rtinit proc uses 'dstaddr' as the IP address
			 * on which to search.
			 */
			flags = RTF_HOST;

			/* save original addr and dstaddr for later restore */
			save_dstaddr = ia->ia_ifa.ifa_dstaddr;

			/* now make dst address point to alias address */
			ia->ia_ifa.ifa_dstaddr = ia->ia_ifa.ifa_addr;

			/*
			 * now make addr address point to primary address
			 * so that the host route's gateway address will
			 * be the primary address of this interface.
			 */
			save_addr = ia->ia_ifa.ifa_addr;

			ia->ia_ifa.ifa_addr =
		       ((struct in_ifaddr *)(ifp->in_ifaddr))->ia_ifa.ifa_addr;

			(void)rtinit(&(ia->ia_ifa), (int)RTM_DELETE, flags);

			/* restore original dstaddr and addr address'es */
			ia->ia_ifa.ifa_dstaddr = save_dstaddr;
			ia->ia_ifa.ifa_addr = save_addr;

		} else {
			flags = (ifp->if_flags &(IFF_LOOPBACK|IFF_POINTOPOINT))
				? RTF_HOST : 0;
			(void)rtinit(&(ia->ia_ifa), (int)RTM_DELETE, flags);
		}
		ia->ia_flags &= ~IFA_ROUTE;
	}
	return;
}

/*
 * Initialize an interface's internet address and routing table entry.
 *
 * NOTE: The 'alias' parameter is non-zero this is requesting that
 * a host route be made for IP alias address on the interface 'ifp'.
 */
int
in_ifinit(struct ifnet *ifp, struct in_ifaddr *ia, struct sockaddr_in *sin,
	  int scrub, int alias)
{
	struct in_ifaddr *ia_xx;
	struct sockaddr_in oldaddr;
	struct sockaddr *save_addr, *save_dstaddr;
	__uint32_t i = ntohl(sin->sin_addr.s_addr);
	int flags = RTF_UP, error;
	int restore_addr = 0;
	int wasup = ifp->if_flags & IFF_UP;

	ASSERT(IFNET_ISLOCKED(ifp));

	oldaddr = ia->ia_addr;
	ia->ia_addr = *sin;

	if (alias == 0) {
		/*
		 * Not an alias address add so assume SIOCSIFADDR case.
		 * Give the interface a chance to initialize if this is its
		 * first address, and to validate the address if necessary.
		 */
		if (ifp->if_ioctl &&
		   (error = (*ifp->if_ioctl)(ifp, SIOCSIFADDR,
					(caddr_t)&ia->ia_ifa))) {
			ia->ia_addr = oldaddr;
			return error;
		}

		/*
		 * If the interface did not know its MAC address when it
		 * attached itself, then fix it up now. Use a pointer to
		 * NULL as the address which informs the procedure to look
		 * in the raw interface list for the address.
		 */
		rawif_fixlink(ifp);
	}

	if ((ia_xx = (struct in_ifaddr *)ifp->in_ifaddr)) {
		if (! ia_xx->ia_ifa.ifa_rtrequest) { /* not set yet */
			ia_xx->ia_ifa.ifa_rtrequest = link_rtrequest;
		}
	}
	if (!wasup && (ifp->if_flags & IFF_UP))
		ifp->if_lastchange = time;

	if (scrub) {
		ia->ia_ifa.ifa_addr = (struct sockaddr *)&oldaddr;
		in_ifscrub(ifp, ia);
		ia->ia_ifa.ifa_addr = (struct sockaddr *)&ia->ia_addr;
	}
	if (IN_CLASSA(i))
		ia->ia_netmask = IN_CLASSA_NET;
	else if (IN_CLASSB(i))
		ia->ia_netmask = IN_CLASSB_NET;
	else
		ia->ia_netmask = IN_CLASSC_NET;
	/*
	 * The subnet mask usually includes at least the standard network part,
	 * but may may be smaller in the supernetting case.
	 * If it is set, we believe it.
	 */
	if (ia->ia_subnetmask == 0) {
		ia->ia_subnetmask = ia->ia_netmask;
		ia->ia_sockmask.sin_addr.s_addr = htonl(ia->ia_subnetmask);
	} else
		ia->ia_netmask &= ia->ia_subnetmask;
	ia->ia_net = i & ia->ia_netmask;
	ia->ia_subnet = i & ia->ia_subnetmask;
	in_socktrim(&ia->ia_sockmask);

	/*
	 * Add route for the network.
	 *
	 * Check for hostroute override case where we add a host route
	 * for the IP alias address for this interface.
	 */
	ia->ia_ifa.ifa_metric = ifp->if_metric;

	if (ifp->if_flags & IFF_BROADCAST) {

		ia->ia_broadaddr.sin_addr.s_addr =
			htonl(ia->ia_subnet | ~ia->ia_subnetmask);
		ia->ia_netbroadcast.s_addr =
			htonl(ia->ia_net | ~ ia->ia_netmask);

		if (alias) { /* for alias address create a host route */

			/* save original addr and dstaddr for later restore */
			save_dstaddr = ia->ia_ifa.ifa_dstaddr;

			/* now make dst address point to alias address */
			ia->ia_ifa.ifa_dstaddr = ia->ia_ifa.ifa_addr;

			/*
			 * now make addr address point to primary address
			 * so that the host route's gateway address will
			 * be the primary address of this interface.
			 */
			save_addr = ia->ia_ifa.ifa_addr;

			ia->ia_ifa.ifa_addr =
		       ((struct in_ifaddr *)(ifp->in_ifaddr))->ia_ifa.ifa_addr;

			restore_addr = 1;
			flags |= (RTF_HOST | RTF_HOSTALIAS);
		}
	} else if (ifp->if_flags & IFF_LOOPBACK) {

		if (alias) {
			/* save original addr and dstaddr for later restore */
			save_dstaddr = ia->ia_ifa.ifa_dstaddr;
			save_addr = ia->ia_ifa.ifa_addr;

			/* now make dst address point to alias address */
			ia->ia_ifa.ifa_dstaddr = ia->ia_ifa.ifa_addr;

			/*
			 * Make address record's primary address point
			 * to the primary address of the loopback interface
			 */
			ia->ia_ifa.ifa_addr =
		       ((struct in_ifaddr *)(ifp->in_ifaddr))->ia_ifa.ifa_addr;

			restore_addr = 1;
			flags |= (RTF_HOST | RTF_HOSTALIAS);
		} else {
			/*
			 * Point destination address to local address storage
			 */
			ia->ia_ifa.ifa_dstaddr = ia->ia_ifa.ifa_addr;
			flags |= RTF_HOST;
		}
	} else {
		if (ifp->if_flags & IFF_POINTOPOINT) {
			if (ia->ia_dstaddr.sin_family != AF_INET)
				return (0);
			flags |= RTF_HOST;
		}
	}

	if ((error = rtinit(&(ia->ia_ifa), (int)RTM_ADD, flags)) == 0) {
		ia->ia_flags |= IFA_ROUTE;
	}
	if (alias && restore_addr) {
		/*
		 * restore original dstaddr and addr address'es
		 */
		ia->ia_ifa.ifa_dstaddr = save_dstaddr;
		ia->ia_ifa.ifa_addr = save_addr;
	}
	/*
	 * If the interface supports multicast, join the "all hosts"
	 * multicast group on that interface.
	 */
	if (ifp->if_flags & IFF_MULTICAST) {
		struct in_addr addr;

		addr.s_addr = htonl(INADDR_ALLHOSTS_GROUP);
		in_addmulti(addr, ifp);
	}
	return (0);
}

/*
 * This is a hash enumeration match procedure which is used to get the
 * 'nth' IP alias address associated with the interface specified by 'ifp'.
 * The starting index of the enumeration is 'start' and can be used to skip
 * the primary address (index zero) or start at the first alias (index 1).
 *
 * int
 * in_getalias_enummatch(struct hashbucket *h, caddr_t start,
 *	caddr_t nth, struct ifnet *arg2)
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
in_getalias_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct ifnet *ifp = (struct ifnet *)arg2;

	return (((ia->ia_ifp == ifp) && (ia->ia_addrflags & IADDR_ALIAS))
		? 1 : 0);
}

/*
 * This is a hash enumeration match procedure which compares the network
 * number desired with the subnet number in the hash bucket entry,
 * to see if we have a match.
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 */
/* ARGSUSED */
int
in_iaonnetof_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia;
	__uint32_t net;
	int match = 0;

	ia = (struct in_ifaddr *)h;
	net = (__psint_t)key;

	if (ia->ia_subnet == net)
		match = 1;
	return match;
}

/*
 * Return address info for specified internet network.
 */
struct in_ifaddr *
in_iaonnetof(__uint32_t net)
{
	return ((struct in_ifaddr *)hash_enum(&hashinfo_inaddr,
			in_iaonnetof_enummatch, HTF_INET,
			(caddr_t)(__psint_t)net, (caddr_t)0, (caddr_t)0));
}

/*
 * The hash enumerate match function to compare the address supplied in
 * 'in_addr' (aka key) to the address'es for the supplied network interface
 * in 'ifp'.
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the enumerate call. Here that is HTF_INET types.
 *
 * in_broadcast_enummatch(struct in_ifaddr *ia, struct in_addr *key,
 *	struct ifnet *arg1, caddr_t arg2)
 */
/* ARGSUSED */
int
in_broadcast_enummatch(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct ifnet *ifp = (struct ifnet *)arg1;
	struct in_addr *addr;
	__uint32_t t;
	int match;

	if (ia->ia_ifp != ifp)
		return 0;

	if ((ia->ia_ifp->if_flags & IFF_BROADCAST) == 0) {
		return 0;
	}

	addr = (struct in_addr *)key;
	t = ntohl(addr->s_addr);
	match = 0;

	if (ia->ia_ifa.ifa_addr->sa_family == AF_INET &&
	 (addr->s_addr == satosin(ia->ia_ifa.ifa_broadaddr)->sin_addr.s_addr ||
		     addr->s_addr == ia->ia_netbroadcast.s_addr ||
		     /*
		      * Check for old-style (host 0) broadcast.
		      */
		     t == ia->ia_subnet || t == ia->ia_net)) {
			    match = 1;
	}
	return match;
}

/*
 * The hash lookup match function to compare the broadcast IP address
 * in the 'struct in_addr' (aka key) to the broadcast address field
 * in the broadcast hash table entry supplied. We also check that this
 * entry corresponds to the network interface designated by ifp (aka arg1).
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the lookup call. Here that is HTF_BROADCAST hash
 * bucket entries.
 *
 * in_broadcast_full_match(struct in_ifaddr *ia, struct in_addr *key,
 *	struct ifnet *arg1, caddr_t arg2)
 */
/* ARGSUSED */
int
in_broadcast_full_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_bcast *in_bcastp;
	struct in_addr *addr;
	struct ifnet *ifp;
	int match;

	if ((h->flags & HTF_BROADCAST) == 0) { /* not broadcast entry */
		return 0;
	}

	/*
	 * Check that supplied lookup key matches the one used on entry
	 * insertion, before we check the rest of the fields to be sure
	 * this is the correct one since there can be a full broadcast
	 * entry and a netbroadcast entry depending on the netmask.
	 */
	if (bcmp(h->key, key, sizeof(struct in_addr) != 0)) { /* no match */
		return 0;
	}

	in_bcastp = (struct in_bcast *)h;
	addr = (struct in_addr *)key;
	ifp = (struct ifnet *)arg1;

	if ((in_bcastp->ifp == ifp) &&
	    (in_bcastp->ia_addr.sin_family == AF_INET) &&
	    (in_bcastp->ia_dstaddr.sin_addr.s_addr == addr->s_addr)) {
		/*
		 * match on the interface AND the full broadcast address
		 */
		match = 1;
	} else {
		match = 0;
	}
	return match;
}

/*
 * The hash lookup match function to compare the netbroadcast IP address
 * in the 'struct in_addr' (aka key) to the netbroadcast address field
 * in the broadcast hash table entry supplied. We also check that this
 * entry corresponds to the network interface designated by ifp (aka arg1).
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the lookup call. Here that is HTF_BROADCAST hash
 * bucket entries.
 *
 * in_broadcast_net_match(struct in_ifaddr *ia, struct in_addr *key,
 *	struct ifnet *arg1, caddr_t arg2)
 */
/* ARGSUSED */
int
in_broadcast_net_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_bcast *in_bcastp;
	struct in_addr *addr;
	struct ifnet *ifp;
	__uint32_t t;
	int match = 0;

	/*
	 * Check that supplied lookup key matches the one used on entry
	 * insertion, before we check the rest of the fields to be sure
	 * this is the correct one since there can be a full broadcast
	 * entry and a netbroadcast entry depending on the netmask.
	 */
	if (bcmp(h->key, key, sizeof(struct in_addr) != 0)) { /* no match */
		return 0;
	}

	in_bcastp = (struct in_bcast *)h;
	addr = (struct in_addr *)key;
	ifp = (struct ifnet *)arg1;
	t = ntohl(addr->s_addr);

	if ((in_bcastp->ifp == ifp) &&
	    (in_bcastp->ia_addr.sin_family == AF_INET) &&
	    ((in_bcastp->ia_netbroadcast.s_addr == addr->s_addr) ||
	     /* Check for old-style (host 0) broadcast */
	     (t == in_bcastp->ia_subnet) || (t == in_bcastp->ia_net))) {
		/*
		 * match on the interface AND the address family AND
		 * (either the net broadcast address or the subnet numbers)
		 */
		match = 1;
	}
	return match;
}

/*
 * The hash lookup match function to compare the broadcast IP address
 * in the 'struct in_addr' (aka key) to the any of the broadcast address
 * fields in the broadcast hash table entry supplied. We also check that this
 * entry corresponds to the network interface designated by ifp (aka arg1).
 *
 * NOTE: This procedure is called with only hash bucket entries of the
 * specified type on the lookup call. Here that is HTF_BROADCAST hash
 * bucket entries. 'arg2' is ignored here.
 *
 * in_broadcast_net_match(struct in_ifaddr *ia, struct in_addr *key,
 *	struct ifnet *arg1, caddr_t arg2)
 */
/* ARGSUSED */
int
in_broadcast_match(struct hashbucket *h, caddr_t key,
	caddr_t arg1, caddr_t arg2)
{
	struct in_bcast *in_bcastp;
	struct in_addr *addr;
	struct ifnet *ifp;
	__uint32_t t;
	int match;

	if ((h->flags & HTF_BROADCAST) == 0) { /* not broadcast entry */
		return 0;
	}

	in_bcastp = (struct in_bcast *)h;
	addr = (struct in_addr *)key;
	ifp = (struct ifnet *)arg1;

	if (in_bcastp->ifp != ifp) { /* not right interface */
		return 0;
	}

	if ((in_bcastp->ifp->if_flags & IFF_BROADCAST) == 0) {
		/* no bcast support on this network interface */
		return 0;
	}
	      
	t = ntohl(addr->s_addr);
	if ((in_bcastp->ia_addr.sin_family == AF_INET) &&

	     ((in_bcastp->ia_dstaddr.sin_addr.s_addr == addr->s_addr) ||
	      (in_bcastp->ia_netbroadcast.s_addr == addr->s_addr) ||
	       /* Check for old-style (host 0) broadcast */
	      (t == in_bcastp->ia_subnet) || (t == in_bcastp->ia_net))) {
		/*
		 * match on the address family AND
		 * (either full broadcast address or net broadcast address or
		 *  either the subnet or net numbers)
		 */
		 match = 1;
	} else {
		 match = 0;
	}
	return match;
}

/*
 * Return 1 if the address might be a local broadcast address.
 */
int
in_broadcast(struct in_addr in, struct ifnet *ifp)
{
	struct in_ifaddr *ia;
	struct in_bcast *in_bcastp;

	if (in.s_addr == INADDR_BROADCAST || in.s_addr == INADDR_ANY) {
		return 1;
	}

	/*
	 * Check if address matches broadcast address for the
	 * primary address supplied for the interface.
	 */
	if (ia = (struct in_ifaddr *)ifp->in_ifaddr) {
		if (ip_check_broadcast(ia, &in)) {
			return 1;
		}
	}

	/*
	 * First check if address matches broadcast address for the
	 * primary address for any of our network interfaces.
	 * This handles the case when we get the broadcast address
	 * for a particular network interface passed in with 'ifp'
	 * corresponding to the loopback interface.
	 */
	if (ip_accept_primary_broadcast(&in)) {
		return 1;
	}

	/* Check if this network interface supports broadcast */
	if ((ifp->if_flags & IFF_BROADCAST) == 0) { /* no cigar */
		return 0;
	}

	/*
	 * NOTE: The broadcast hash lookup will return either
	 * a HTF_BROADCAST entry or NULL.
	 */
	in_bcastp = (struct in_bcast *)hash_lookup(&hashinfo_inaddr_bcast,
			in_broadcast_match,
			(caddr_t)&in,
			(caddr_t)ifp,
			(caddr_t)0);
	return ((in_bcastp) ? 1 : 0);
}

/*
 * This procedure is used for the hash lookup match which compares the
 * multi-cast address in the entry with that supplied along with comparing
 * the given interface address pointers.
 * NOTE: The INADDR_ALLHOSTS_GROUP is added to the multicast address list
 * for each interface.
 *
 * in_multi_match(struct in_multi *inm, struct in_addr *key,
 *  struct ifnet *arg)
 */
/* ARGSUSED */
int
in_multi_match(struct hashbucket *h, caddr_t key, caddr_t arg1, caddr_t arg2)
{
	struct in_multi *inm;
	struct in_addr *addr;
	struct ifnet *ifp;
	int match = 0;

	if (h->flags & HTF_MULTICAST) { /* multicast address node */

		inm = (struct in_multi *)h;
		addr = (struct in_addr *)key;
		ifp = (struct ifnet *)arg1;

		if ((inm->inm_ifp == ifp) &&
		    (inm->inm_addr.s_addr == addr->s_addr)) {
			 match = 1;
		}
	}
	return match;
}

/*
 * Add an address to the list of IP multicast addresses for a given interface.
 */
struct in_multi *
in_addmulti(struct in_addr addr, struct ifnet *ifp)
{
	register struct in_multi *inm;
	struct sockaddr_in sin;

	ASSERT(IFNET_ISLOCKED(ifp));

	inm = (struct in_multi *)hash_lookup(&hashinfo_inaddr,
		in_multi_match, (caddr_t)&addr,
		(caddr_t)ifp, (caddr_t)0);
	
	if (inm) { /* found so simply increment refcount */

		(void)atomicAddUint(&(inm->inm_refcount), 1);
		return inm;
	}

	/*
	 * Ask the network driver to update its multicast reception
	 * filter appropriately for the new address.
	 */
	if (ifp->if_ioctl == NULL) {
		return NULL;
	}

	sin.sin_family = AF_INET;
	sin.sin_port = 0;
	sin.sin_addr = addr;
	if ((*ifp->if_ioctl)(ifp, SIOCADDMULTI, (caddr_t)&sin) !=  0) {
		return NULL;
	}
	/*
	 * New address; get a buffer for a new multicast record
	 * and link it into the interface's multicast list.
	 */
	inm = (struct in_multi *)mcb_get(M_WAIT,
			sizeof(struct in_multi), MT_IFADDR);

	inm->inm_refcount = 1;
	inm->hashbucket.flags = HTF_MULTICAST;
	inm->inm_addr = addr;
	inm->inm_ifp = ifp;

	/*
	 * Let IGMP know that we have joined a new IP multicast group.
	 * We must do this with the ifnet lock unlocked, because we
	 * grab the lock again to send the IGMP HOST MEMBERSHIP packet.
	 * Also do hash insert without interface lock to avoid deadlock.
	 */
	IFNET_UNLOCKNOSPL(ifp);

	/*
	 * NOTE:  With the hash based IP address scheme the ifnet lock is
	 * NOT required in order to insert/remove address records since these
	 * records are no longer linked to the ifnet structure.
	 */
	(void)hash_insert(&hashinfo_inaddr, &(inm->hashbucket),
			  (caddr_t)(&addr));

	igmp_joingroup(inm);
	IFNET_LOCKNOSPL(ifp);

	return inm;
}

/*
 * Delete a multicast address record.
 */
void
in_delmulti(struct in_multi *inm)
{
	struct sockaddr_in sin;
	struct ifnet *ifp = inm->inm_ifp;
	/*REFERENCED*/
	NETSPL_DECL(s)

	IFNET_LOCK(ifp, s);

	if (--inm->inm_refcount == 0) {
		/*
		 * Notify network driver to update it's
		 * multicast reception filter.
		 */
		sin.sin_family = AF_INET;
		sin.sin_port = 0;
		sin.sin_addr = inm->inm_addr;
		(*inm->inm_ifp->if_ioctl)(inm->inm_ifp, SIOCDELMULTI,
			(caddr_t)&sin);

		IFNET_UNLOCK(ifp, s);
		/*
		 * No remaining claims to this IP group; let IGMP know that
		 * we are leaving the multicast group. We must do this with
		 * the lock unlocked, because we grab the lock again for
		 * the sending of the IGMP LEAVE packet.
		 */
		igmp_leavegroup(inm);

		/* Remove from the IP hash table does NOT delete storage */
		hash_remove(&hashinfo_inaddr, (struct hashbucket *)inm);

		mcb_free(inm, sizeof(struct in_multi), MT_IFADDR);
	} else {
		IFNET_UNLOCK(ifp, s);
	}
	return;
}

