/*
 * ipfilter.c
 *
 * Screen inbound IP packets against filterdaemon's access control list.
 * Screening is based on source and destination IP address, IP protocol, and
 * source and destination ports.
 *
 * Copyright 1990, Silicon Graphics, Inc. 
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or 
 * duplicated in any form, in whole or in part, without the prior written 
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions 
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or 
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished - 
 * rights reserved under the Copyright Laws of the United States.
 */

#include <string.h>
#include "tcp-param.h"
#include "sys/param.h"
#include "sys/sema.h"
#include "sys/hashing.h"

#include "sys/debug.h"
#include "sys/systm.h"
#include "sys/mbuf.h"
#include "sys/errno.h"
#include "sys/par.h"
#include "sys/idbgentry.h"
#include "sys/conf.h"
#include "sys/socketvar.h"

#include "net/if.h"

#include "in.h"
#include "in_systm.h"
#include "in_var.h"
#include "ip.h"
#include "ip_var.h"
#include "ip_icmp.h"
#include "tcp.h"
#include "udp.h"
#include "igmp.h"

#include "sys/proc.h"
#include "sys/user.h"
#include "sys/ioctl.h"
#include "net/netisr.h"
#include "ipfilter.h"
#include "sys/capability.h"
#ifdef _MP_NETLOCKS
#include "sys/sockd.h"
#endif

sema_t ipfilterq_wait;
#ifdef _MP_NETLOCKS
mrlock_t ipfilterq_mrlock, ipkfilter_mrlock;
#define IPFILTER_INITLOCKS() { \
	mrinit(&ipkfilter_mrlock, "ipkf_lock"); \
	mrinit(&ipfilterq_mrlock, "ipfq_lock"); \
	initnsema(&ipfilterq_wait, 0, "ipfltwait"); \
}
#define IPFILTER_FREELOCKS() { \
	mrfree(&ipkfilter_mrlock); \
	mrfree(&ipfilterq_mrlock); \
	freesema(&ipfilterq_wait); \
}
#define IPKFILTER_MRRLOCK(s)	mrlock(&ipkfilter_mrlock, MR_ACCESS, PZERO-1)
#define IPKFILTER_MRWLOCK(s)	mrlock(&ipkfilter_mrlock, MR_UPDATE, PZERO-1)
#define IPKFILTER_MRUNLOCK(s)	mrunlock(&ipkfilter_mrlock)
#define IPFILTERQ_MRRLOCK(s)	mrlock(&ipfilterq_mrlock, MR_ACCESS, PZERO-1)
#define IPFILTERQ_MRWLOCK(s)	mrlock(&ipfilterq_mrlock, MR_ACCESS, PZERO-1)
#define IPFILTERQ_MRUNLOCK(s)	mrunlock(&ipfilterq_mrlock)
#else
/* protect against timeout call */
#define IPFILTER_INITLOCKS() 	initnsema(&ipfilterq_wait, 0, "ipfltwait")
#define IPFILTER_FREELOCKS()	freesema(&ipfilterq_wait)
#define IPKFILTER_MRRLOCK(s)	s = splnet()
#define IPKFILTER_MRWLOCK(s)	s = splnet()
#define IPKFILTER_MRUNLOCK(s)	splx(s)
#define IPFILTERQ_MRRLOCK(s)	s = splnet()
#define IPFILTERQ_MRWLOCK(s)	s = splnet()
#define IPFILTERQ_MRUNLOCK(s)	splx(s)
#endif

int ipfilterdevflag = D_OLD;

extern struct hashinfo hashinfo_inaddr;

/*
 * IP packet filtering global variables
 */
extern int 	ipfilterflag;	    /* activity state of daemon and filtering*/

/*
 * define IP packet behavior when daemon is killed or dies.
 * configurable in /var/sysgen/mtune/kernel
 */
extern int	ipfilterd_inactive_behavior;

extern int ipfilter_ttl;

/*
 * search depth in kernel cache before MRU reordering happens
 */
extern int 	filtercache_search_depth;

extern struct ipkernelfilter ipfiltercache[]; /* kernel IP filter cache */
extern int	numipkflt;		      /* size of kernel filter cache */

struct ipkernelfilter *ipkfq_head; 		/* current head of cache */
struct ifqueue	ipfiltqueue;   	    /* for packets awaiting daemon verdict */

/*
 * has daemon been active?
 */
int ipfilter_active_flag = 0;

extern struct ifqueue 	ipintrq;
extern void ipfiltertimer(void);

/* macro to jump to next protocol in packet */
#define nextproto(x, y, t)              ((t)((__psint_t)(x) + (__psint_t)(y)))

#ifdef DEBUG
#include "sys/idbgentry.h"

int ipfilter_idbgflag = 0;

#define idbg_ipfilt_vmask 0x7  /* Mask to avoid vnames index overflow */

char *idbg_ipfilt_vnames[] = {
	"FREE",          /* 0 => Free entry */
	"ACCEPTIT",  /* 1 Accept this packet */
	"DROPIT",    /* 2 Drop this packet with NO notification */
	"DROPNOTIFY" /* 3 Drop this packet BUT send ICMP error to sender */
 	"GRABIT"     /* 4 Grab this packet rather than forwarding it */
};

#define IDBG_IPPROT_NAMES_MAX 16

struct in_addr_bytes {
	unsigned char b[4];
};

struct idbg_ipproto {
	unsigned char proto_value;
	char *proto_name;
};

struct idbg_ipproto idbg_ipproto_names[] = {
	{IPPROTO_IP, "IP"},
	{IPPROTO_ICMP, "ICMP"},
	{IPPROTO_IGMP, "IGMP"},
	{IPPROTO_GGP, "GGP"},
	{IPPROTO_ENCAP, "ENCAP"},
	{IPPROTO_TCP, "TCP"},
	{IPPROTO_EGP, "EGP"},
	{IPPROTO_PUP, "PUP"},
	{IPPROTO_UDP, "UDP"},
	{IPPROTO_IDP, "IDP"},
	{IPPROTO_TP, "TP"},
	{IPPROTO_XTP, "XTP"},
	{IPPROTO_HELLO, "HELLO"},
	{IPPROTO_ND, "ND"},
	{IPPROTO_EON,"EON"},
	{IPPROTO_RAW, "RAW"},
	{IPPROTO_MAX, ""}
};

char *
idbg_ipfilt_pnames(unsigned char proto)
{
	unsigned short ix;

	for (ix=0; ix < IDBG_IPPROT_NAMES_MAX; ix++) {
		if (proto == idbg_ipproto_names[ix].proto_value) { /* match */

			return (idbg_ipproto_names[ix].proto_name);
		}
	}

	return (idbg_ipproto_names[IDBG_IPPROT_NAMES_MAX].proto_name);
}

void
idbg_ipfilt_print(struct ipkernelfilter *ipkf_p)
{
	struct in_addr_bytes src, dst;

	qprintf("  entry 0x%x, next 0x%x, ifnet 0x%x, idle %d\n",
		ipkf_p, ipkf_p->next, ipkf_p->interface, ipkf_p->idle);

	qprintf("   verdict %d [%s], proto %d [%s],",
		ipkf_p->verdict,
		idbg_ipfilt_vnames[(ipkf_p->verdict & idbg_ipfilt_vmask)],
		ipkf_p->proto,
		idbg_ipfilt_pnames(ipkf_p->proto)
		);

	switch(ipkf_p->proto) {

	  case IPPROTO_ICMP:
		qprintf(" icmp_type %d, icmp_code %d\n",
			ipkf_p->u1.s_port, ipkf_p->u2.d_port);
		break;

	  case IPPROTO_IGMP:
		qprintf(" igmp_type %d, igmp_code %d\n",
			ipkf_p->u1.s_port, ipkf_p->u2.d_port);
		break;

	  default: /* IPPROTO_TCP and IPPROTO_UDP also land here */
		qprintf(" s_port %d, d_port %d\n",
			ipkf_p->u1.s_port, ipkf_p->u2.d_port);
		break;
	}

	bcopy (&ipkf_p->src_addr, &src, sizeof(long));
	bcopy (&ipkf_p->dst_addr, &dst, sizeof(long));

       qprintf("   src_addr %d.%d.%d.%d [0x%x], dst_addr %d.%d.%d.%d [0x%x]\n",
		src.b[0], src.b[1], src.b[2], src.b[3],ipkf_p->src_addr.s_addr,
		dst.b[0], dst.b[1], dst.b[2], dst.b[3],ipkf_p->dst_addr.s_addr
		);

	return;
}

/*
 * idbg procedure, dynamically registered, which prints the IP filter cache
 *
 * kp ipfilt addr  -- print kernel IP filter cache from symmon
 */
void
idbg_ipfilt(__psint_t addr)
{
	register struct ipkernelfilter *ipkf;

	qprintf(" ipfilter_active_flag %d ipfilterd_inactive_behavior %d\n",
		ipfilter_active_flag, ipfilterd_inactive_behavior);

	qprintf(" numipkflt %d ipfilter_ttl %d filtercache_search_depth %d\n",
		numipkflt, ipfilter_ttl, filtercache_search_depth);

	qprintf(" ipfiltqueue 0x%x ipkfq_head 0x%x\n\n",
		&ipfiltqueue, ipkfq_head);

	if (!ipfilter_active_flag) {
		qprintf("kernel ipfilter not active\n");
		return;
	}

	if (addr == -1) { /* dump entire ip filter cache */

		ipkf = ipkfq_head;
		while (ipkf) {
			idbg_ipfilt_print(ipkf);
			ipkf = ipkf->next;
			if (ipkf == ipkfq_head) { /* back to head so quit */
				break;
			}
		}
	} else { /* use the supplied address */
		idbg_ipfilt_print((struct ipkernelfilter *)addr);
	}
	return;
}

/*
 * Print out an IP filter cache entry
 */
void
ipfilter_idbg(void)
{
	if (ipfilter_idbgflag == 0) { /* register the debug proc */
		ipfilter_idbgflag = 1;
		idbg_addfunc("ipfilt", idbg_ipfilt);
	}
	return;
}
#endif /* DEBUG */

/*
 * hash lookup procedure to check if this packet is for this system.
 * Returns 1 if for us and zero otherwise.
 *
 * ipfilter_kernel_match(struct hashbucket *h, struct in_addr *key,
 *  struct ifnet *arg1, caddr_t arg2)
 */
/* ARGSUSED */
static int
ipfilter_kernel_match(struct hashbucket *h, caddr_t key, caddr_t arg1,
	caddr_t arg2)
{
	struct in_ifaddr *ia = (struct in_ifaddr *)h;
	struct in_addr *addr = (struct in_addr *)key;
	int match = 0;

	if (h->flags & HTF_INET) {
		if (IA_SIN(ia)->sin_addr.s_addr == addr->s_addr)
			match = 1;
	}
	return match;
}

/* 
 * Screen inbound packets against access control database. 
 */

ipfilter_kernel(struct ifnet *ifp, struct mbuf *m)
{
	register struct ipkernelfilter 	*ipkf, *ipkfp;
	register struct mbuf 		*ipfq;
	struct ip    			*iphd, *iphdq;
	struct ifheader			*hdr;
	u_short				dport, qdport, sport, qsport;
	int				hlen;
	struct in_ifaddr 		*ia;
	int 				depth = 0;
	NETSPL_SPDECL(s);

	iphd = mtod(m, struct ip *);
	/*
 	 * Loopback packets are always accepted.
 	 */
	ia = (struct in_ifaddr *)hash_lookup(&hashinfo_inaddr,
		ipfilter_kernel_match, (caddr_t)(&(iphd->ip_src)),
		(caddr_t)ifp, (caddr_t)0);

	if (ia) { /* it's for us */
		return IPF_ACCEPTIT;
	}
	/*
	 * When daemon has died or been killed, obey configurable
	 * ipfilterd inactive behavior.
	 */
	if (!ipfilter_active_flag) {
		if (ipfilterd_inactive_behavior) {
			m_freem(m);
			return(IPF_DROPIT);
		} else return(IPF_ACCEPTIT);
	}
	/*
 	 * Accept all fragments without filtering, but not the head -
	 * the fragments are useless without the fragment head, 
 	 * and the head will contain port info needed for screening.
 	 */
	if(iphd->ip_off & IP_OFFMASK)
		return(IPF_ACCEPTIT);

	hlen = iphd->ip_hl << 2;
	switch(iphd->ip_p) {
	  case IPPROTO_TCP: {
		struct tcphdr *tcphd = nextproto(iphd, hlen, struct tcphdr *);
		dport = tcphd->th_dport;
		sport = tcphd->th_sport;
		}
		break;
	  case IPPROTO_UDP: {
		struct udphdr *udphd = nextproto(iphd, hlen, struct udphdr *);
		dport = udphd->uh_dport;
		sport = udphd->uh_sport;
		}
		break;
	  case IPPROTO_ICMP: {
		struct icmp *icmpp = nextproto(iphd, hlen, struct icmp *);
		sport = (u_short)(icmpp->icmp_type);
		dport = (u_short)(icmpp->icmp_code);
		}
		break;
	  case IPPROTO_IGMP: {
		struct igmp *igmpp = nextproto(iphd, hlen, struct igmp *);
		sport = (u_short)(igmpp->igmp_type);
		dport = (u_short)(igmpp->igmp_code);
		}
		break;
	  default:
		dport = 0;
		sport = 0;
		break;
	}
        /*
         * Return 0 if can keep it, 1 otherwise (dropped because it hit
         * a filter, or was placed on the ioctl pending queue). If it hit
         * a filter, and our search depth is greater than the
         * MRU-invoking depth (filtercache_search_depth),
         * put that entry at the head of the filter cache.
         * Refresh timer on hits.
         */
	IPKFILTER_MRRLOCK(s);

        for(ipkf = ipkfp = ipkfq_head; ipkf; ipkf = ipkf->next) {

                if (ipkf->verdict == 0) { /* end of filter queue */
                        break;
                }

                if ((sport == ipkf->srcport) &&
                   (dport == ipkf->dstport) &&
                   (ifp == ipkf->interface) &&
                   (iphd->ip_src.s_addr == ipkf->src_addr.s_addr) &&
                   (iphd->ip_dst.s_addr == ipkf->dst_addr.s_addr) &&
                   (iphd->ip_p == ipkf->proto)) {

			ipkf->idle = 0;		/* refresh idle timer */

                        if (depth > filtercache_search_depth) {

				if (ipkf != ipkfq_head) {
					IPKFILTER_MRUNLOCK(s);
					IPKFILTER_MRWLOCK(s);

                               		ipkfp->next = ipkf->next;
                               		ipkf->next = ipkfq_head;
                               		ipkfq_head = ipkf;
				}
                        }
                        switch (ipkf->verdict) {

                          case IPF_ACCEPTIT:
				IPKFILTER_MRUNLOCK(s);
                                return(IPF_ACCEPTIT);

                          case IPF_GRABIT:
				IPKFILTER_MRUNLOCK(s);
				return (IPF_GRABIT);

                          case IPF_DROPNOTIFY:
                          case IPF_DROPIT:
                          default:
				IPKFILTER_MRUNLOCK(s);
                                m_freem(m);
                                return(IPF_DROPIT);
                        }
                }
                ipkfp = ipkf;
                depth++;
        }
	IPKFILTER_MRUNLOCK(s);

	/*
 	 * Add this request to the pending ioctl queue, if there's room 
	 * and if we haven't queued it yet for a previous packet (walk 
	 * through queue checking for matches without actually touching queue).
 	 */
	IPFILTERQ_MRRLOCK(s);

	if (IF_QFULL(&ipfiltqueue)) {
		IPFILTERQ_MRUNLOCK(s);
		m_freem(m);
		return(IPF_DROPIT);
	}
	ipfq = ipfiltqueue.ifq_head;
	while(ipfq) {
		hdr = mtod(ipfq, struct ifheader *);
                iphdq = nextproto(hdr, sizeof(struct ifheader), struct ip *);
		hlen = iphdq->ip_hl << 2;
		if((iphd->ip_src.s_addr == iphdq->ip_src.s_addr) &&
		   (iphd->ip_dst.s_addr == iphdq->ip_dst.s_addr) &&
                   (ifp == hdr->ifh_ifp) &&
		   (iphd->ip_p == iphdq->ip_p)) { 

        		switch(iphdq->ip_p) {
        		  case IPPROTO_TCP: {
                		struct tcphdr *tcphd = nextproto(iphdq, hlen,
							struct tcphdr *);
                		qdport = tcphd->th_dport;
                		qsport = tcphd->th_sport;
				}
                		break;
        		  case IPPROTO_UDP: {
                		struct udphdr *udphd = nextproto(iphdq, hlen,
				    			struct udphdr *);
                		qdport = udphd->uh_dport;
                		qsport = udphd->uh_sport;
				}
                		break;
        		  case IPPROTO_ICMP: {
                		struct icmp *icmpp = nextproto(iphdq, hlen,
							struct icmp *);
                		qsport = (u_short)(icmpp->icmp_type);
                		qdport = (u_short)(icmpp->icmp_code);
				}
                		break;
        		  case IPPROTO_IGMP: {
                		struct igmp *igmpp = nextproto(iphdq, hlen,
							struct igmp *);
                		qsport = (u_short)(igmpp->igmp_type);
                		qdport = (u_short)(igmpp->igmp_code);
				}
                		break;
			  default:
				qdport = qsport = 0;
                		break;
        		}
			if((dport == qdport) && (sport == qsport)) {
				IPFILTERQ_MRUNLOCK(s);
				m_freem(m);
				return(IPF_DROPIT);
			}
		}
		ipfq = ipfq->m_act;
	}
	IPFILTERQ_MRUNLOCK(s);

	/*
 	 * Assume always have space to stick an ifheader, because one was just 
 	 * taken off it in ipintr()...
 	 */
	ASSERT(m->m_off >= (sizeof(struct ifheader) + MMINOFF));
	m->m_off -= sizeof(struct ifheader);
	m->m_len += sizeof(struct ifheader);
	hdr = mtod(m, struct ifheader *);
	hdr->ifh_ifp = ifp;
	hdr->ifh_hdrlen = sizeof(*hdr);
	IPFILTERQ_MRWLOCK(s);
	IF_ENQUEUE_NOLOCK(&ipfiltqueue, m);
	IPFILTERQ_MRUNLOCK(s);
	while (cvsema(&ipfilterq_wait));
	return (IPF_DROPIT);
}


/* 
 * IP Filtering ioctl handler 
 */
#define COPYSIZE sizeof(struct ipfilter_ioctl)

/*ARGSUSED*/
int
ipfilterioctl(int fd, int cmd, caddr_t addr)
{
	u_char	data[COPYSIZE + 2];

	if (!_CAP_ABLE(CAP_NETWORK_MGT))
		return EPERM;
	if (!ipfilter_active_flag)
		return EACCES;
        if (curthreadp->k_error = copyin(addr, (caddr_t)data, COPYSIZE))
		return(curthreadp->k_error);

	switch  (cmd) {

	  case ZIOCIPFILTER: {
		struct ipfilter_ioctl 	*ipf_i = (struct ipfilter_ioctl *)data;
		struct mbuf 		*m;
		struct ifheader 	*ifh;
		struct ip 		*iphd;
		int 			hlen;
		struct ipkernelfilter 	*ipkf, *ipkfp;
		int 			s;
                char                    *cp, *ep;

		ipf_i = (struct ipfilter_ioctl *)data;

		if (ipf_i->verdict == 0) 	/* during initialization */
			goto findanother;
		/*
		 * Find an empty ipkernelfilter entry, or else the tail of the
		 * queue, and copy the verdict and info there and put it at 
		 * the head. If verdict had been to drop packet, m_freem()
 		 * the packet. Then, go to ioctl pending queue. Look for next
		 * packet to screen. If none is available yet, sleep. If one
		 * is found, copy data into ipf_i and complete ioctl
 		 */
		IPKFILTER_MRRLOCK(s);

		/* 
 		 * Handle case when we run off the queue that we can
		 * then overwrite the last entry. Requires previous and
		 * current always set correctly.
		 */
		ipkf = ipkfp = ipkfq_head;
		while (ipkf) {
			if ((ipkf->verdict == 0) || (!ipkf->next)) {
				break;
			}
			ipkfp = ipkf;
			ipkf = ipkf->next;
		}

		IPFILTERQ_MRWLOCK(s);	
		IF_DEQUEUE_NOLOCK(&ipfiltqueue, m);
		IPFILTERQ_MRUNLOCK(s);	
		ASSERT(m != 0);
		ifh = mtod(m, struct ifheader *);
		iphd = nextproto(ifh, ifh->ifh_hdrlen, struct ip *);

		/*
		 * Use the TTL to avoid the possibility of looping forever 
		 * through the filter code and ip input queue.
		 */
		iphd->ip_ttl -= IPTTLDEC;
		if(iphd->ip_ttl <= IPTTLDEC) {
			IPKFILTER_MRUNLOCK(s);
			m_freem(m);
			goto findanother;
		}
		hlen = iphd->ip_hl<<2;
		ipkf->dst_addr.s_addr = iphd->ip_dst.s_addr;
		ipkf->src_addr.s_addr = iphd->ip_src.s_addr;
		ipkf->proto = iphd->ip_p;
		ipkf->verdict = ipf_i->verdict;
                ipkf->interface = ifh->ifh_ifp;
		ipkf->idle = 0;

                switch(iphd->ip_p) {
                  case IPPROTO_TCP: {
                        struct tcphdr *tcphd = nextproto(iphd, hlen,
							struct tcphdr *);
                        ipkf->dstport = tcphd->th_dport;
                        ipkf->srcport = tcphd->th_sport;
			}
                        break;
                  case IPPROTO_UDP: {
                        struct udphdr *udphd = nextproto(iphd, hlen,
							struct udphdr *);
                        ipkf->dstport = udphd->uh_dport;
                        ipkf->srcport = udphd->uh_sport;
			}
                        break;
                  case IPPROTO_ICMP: {
                        struct icmp *icmpp = nextproto(iphd, hlen,
							struct icmp *);
                        ipkf->srcport = (u_short)icmpp->icmp_type;
                        ipkf->dstport = (u_short)icmpp->icmp_code;
			}
                        break;
                  case IPPROTO_IGMP: {
                        struct igmp *igmpp = nextproto(iphd, hlen,
							struct igmp *);
                        ipkf->srcport = (u_short)igmpp->igmp_type;
                        ipkf->dstport = (u_short)igmpp->igmp_code;
			}
                        break;
		  default:
                        ipkf->srcport = 0;
                        ipkf->dstport = 0;
                        break;
                }
		/* 
 		 * ipkf points to EITHER the last entry in the queue (LRU)
		 * OR the first free one (if the queue hasn't filled up yet).
		 * In either event use that entry and make it the queue head
 		 */
		if (ipkf != ipkfq_head) {
			ipkfp->next = ipkf->next;
			ipkf->next = ipkfq_head;
			ipkfq_head = ipkf;
		}
		/* 
 		 * Now, act on verdict
 		 */
		switch(ipkf->verdict) {

		case IPF_ACCEPTIT:
		case IPF_GRABIT:
			IPKFILTER_MRUNLOCK(s);
			/*
			 * TTL was changed above, so recompute checksum
			 * before enqueueing to IP input queue.
			 * Set proper mbuf length and offset for checksum.
			 */
			m->m_off += sizeof(struct ifheader);
			m->m_len -= sizeof(struct ifheader);
			iphd->ip_sum = 0;
			iphd->ip_sum = in_cksum(m, hlen);
			m->m_off -= sizeof(struct ifheader);
			m->m_len += sizeof(struct ifheader);
			s = splimp();
			IFQ_LOCK(&ipintrq);
        		if (IF_QFULL(&ipintrq)) {
                		IF_DROP(&ipintrq);
				splx(s);
				IFQ_UNLOCK(&ipintrq);
                		m_freem(m);
                		break;
        		}
        		IF_ENQUEUE_NOLOCK(&ipintrq, m);
			IFQ_UNLOCK(&ipintrq);
			schednetisr(NETISR_IP);
			splx(s);
			break;

		case IPF_DROPIT:
		case IPF_DROPNOTIFY: 
			IPKFILTER_MRUNLOCK(s);
			m_freem(m);
			break;
		default:
			IPKFILTER_MRUNLOCK(s);
			m_freem(m);
			break;
		}
		/*
 		 * See if there are any other packets waiting for screening.
		 * If so, complete the ioctl by grabbing the info and returning
		 * it to the daemon. If not sleep until ipintr() calls wakeup()
		 * for another packet.
 		 */
findanother:
		IPFILTERQ_MRRLOCK(s);
		while(((m = ipfiltqueue.ifq_head) == 0) && 
						ipfilter_active_flag) {
			IPFILTERQ_MRUNLOCK(s);
			if(psema(&ipfilterq_wait, (PZERO+1)|PCATCH)) {
				curthreadp->k_error = EINTR;
				return EINTR;
			}
			IPFILTERQ_MRRLOCK(s);
		}
		if (!ipfilter_active_flag) {
			IPFILTERQ_MRUNLOCK(s);
			return EACCES;
		}
                ifh = mtod(m, struct ifheader *);
		iphd = nextproto(ifh, sizeof(struct ifheader), struct ip *);
		bzero((caddr_t)ipf_i, sizeof(struct ipfilter_ioctl));
		bcopy((caddr_t)iphd, (caddr_t)ipf_i->packet, IPFILTERSIZE);
                ep = (caddr_t)((__psunsigned_t)ipf_i->ifname + IFNAMSIZE - 1);
		(void) strncpy(ipf_i->ifname, ifh->ifh_ifp->if_name,IFNAMSIZE);
                for (cp = ipf_i->ifname; cp < ep && *cp; cp++)
                        ;
		sprintf(cp, "%d", ifh->ifh_ifp->if_unit);
		IPFILTERQ_MRUNLOCK(s);
		break;
	  }
	  default:
		curthreadp->k_error = EOPNOTSUPP;
		break;
	}
        /*
         * Copy any data to user, size was
         * already set and checked above.
         */
        curthreadp->k_error = copyout(data, addr, COPYSIZE);
        return curthreadp->k_error;
}

/* 
 * Initialize kernel filter cache when filtering is enabled 
 */
#ifdef _MP_NETLOCKS
#define ipfilter_timeout	sockd_timeout	/* changes to sockd_timeout */
#else
#define ipfilter_timeout	timeout
#endif

void
ipkernelfilter_init(void)
{
	register struct ipkernelfilter *ipk, *ipkp;
	int i;

	ipkfq_head = (struct ipkernelfilter *)ipfiltercache;
        bzero ((caddr_t)ipfiltercache,
               (sizeof(struct ipkernelfilter) * numipkflt));
	ipk = ipkp = ipkfq_head;

	for (i = 0; i < (numipkflt - 1); i++, ipkp++) {
		ipkp->next = ++ipk;
	}

	ipfiltqueue.ifq_maxlen = IFQ_MAXLEN;
	IPFILTER_INITLOCKS();

	/* XXX back to timeout after converting to mr spin lock  huy */
        ipfilter_timeout(ipfiltertimer, (caddr_t) 0, HZ);

#ifdef DEBUG
	ipfilter_idbg();
#endif
	return;
}

ipfilteropen(void)
{
	if(ipfilter_active_flag) {
		curthreadp->k_error = EBUSY;
		return EBUSY;
	}
	ipkernelfilter_init();
	ipfilterflag = 1;		/* set on first open, stays set */
	ipfilter_active_flag = 1;	/* current state of daemon activity */
	return(0);
}

ipfilterclose(void)
{
	struct mbuf	*m;
	NETSPL_SPDECL(s);

	IPKFILTER_MRWLOCK(s);
	ipfilter_active_flag = 0;
        bzero ((caddr_t)ipfiltercache,
               (sizeof(struct ipkernelfilter) * numipkflt));
	IPKFILTER_MRUNLOCK(s);

	IPFILTERQ_MRWLOCK(s);
next:
	IF_DEQUEUE_NOLOCK(&ipfiltqueue, m);
	if (m) {
		m_freem(m);
		goto next;
	}
	IPFILTERQ_MRUNLOCK(s);

	while (cvsema(&ipfilterq_wait));
	IPFILTER_FREELOCKS();
	return(0);
}

/*
 * age and recover idle cache entries
 */
void
ipfiltertimer(void)
{
        struct ipkernelfilter *ipkf, *ipkfp, *ipkf_savenext;
        struct ipkernelfilter *free_head, *free_tail, *last_alive;
        int    s1;
	NETSPL_SPDECL(s2); 

        /* Lower to net, but no lower to protect the values in the loop */
        s1 = spl1();

	/* Only age entries and schedule new timeout if daemon is active */
	IPKFILTER_MRWLOCK(s2);

	if (ipfilter_active_flag) { /* user daemon is active */

        	ipfilter_timeout(ipfiltertimer, (caddr_t) 0, HZ);

        	last_alive = free_head = free_tail =(struct ipkernelfilter *)0;
		ipkf = ipkfp = ipkfq_head;

		while (ipkf) {

                	if (ipkf->verdict) {
				ipkf->idle++;
				if (ipkf->idle <= ipfilter_ttl) {
					last_alive = ipkfp = ipkf;
					ipkf = ipkf->next;
					continue;
                		}
			}
			ipkf_savenext = ipkf->next;

			if (ipkf == ipkfq_head) { /* reclaim head */
				ipkfq_head = ipkfp = ipkf->next;
			} else {
				ipkfp->next = ipkf->next;
			}
			bzero ((caddr_t)ipkf, sizeof(*ipkf));

			if (free_tail) {
				free_tail->next = ipkf;
                               	free_tail = ipkf;
                       	} else {
				free_head = free_tail = ipkf;
			}
			ipkf = ipkf_savenext;
        	}
		/*
		 * Now append the list of recovered nodes, if we got any,
		 * onto the tail of the last active node found or the head
		 * in the case we recovered the whole list.
		 */
		if (free_head) {
			if (last_alive) {
				ASSERT(last_alive->next == NULL);
        			last_alive->next = free_head;
			} else {
				ASSERT(ipkfq_head == NULL);
				ipkfq_head = free_head;
			}
		}
	}
	IPKFILTER_MRUNLOCK(s2);
        splx(s1);
}

