/*
 * Copyright (c) 1982, 1986, 1988, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution is only permitted until one year after the first shipment
 * of 4.4BSD by the Regents.  Otherwise, redistribution and use in source and
 * binary forms are permitted provided that: (1) source distributions retain
 * this entire copyright notice and comment, and (2) distributions including
 * binaries display the following acknowledgement:  This product includes
 * software developed by the University of California, Berkeley and its
 * contributors'' in the documentation or other materials provided with the
 * distribution and in all advertising materials mentioning features or use
 * of this software.  Neither the name of the University nor the names of
 * its contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)tcp_subr.c	7.19 (Berkeley) 7/25/90
 */

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/systm.h"
#include "sys/mbuf.h"
#include "sys/socket.h"
#include "sys/socketvar.h"
#include "sys/tcpipstats.h"
#include "sys/protosw.h"

#include "sys/sysmacros.h"
#include "sys/errno.h"
#include "net/route.h"
#include "net/if.h"
#include "sys/types.h"

#include "in.h"
#include "in_systm.h"
#include "ip.h"
#include "in_pcb.h"
#include "ip_var.h"
#include "ip_icmp.h"
#include "tcp.h"
#include "tcp_fsm.h"
#include "tcp_seq.h"
#include "tcp_timer.h"
#include "tcp_var.h"
#include "tcpip.h"
#include "tcpip_ovly.h"

#ifdef TCPPRINTFS
extern int tcpprintfs;
#endif

/* patchable/settable parameters for tcp */
extern int tcp_ttl;
int 	tcp_mssdflt = TCP_MSS;
int 	tcp_rttdflt = TCPTV_SRTTDFLT / PR_SLOWHZ;

extern int tcp_winscale;  	/* disable/enable RFC 1323 window scaling */
extern int tcp_tsecho;		/* disable/enable RFC 1323 timestamps */
extern int tcp_mtudisc;		/* MTUDisc */

/*
 * Each of the following TWO constants MUST be power of 2
 *
 * The dynamic sizing algorithm is computed in the procedure
 * in_pcb_hashtablesize() which is sized depending on the physical size of
 * main memory. The maximum for TCP is TCP_MAXHASHTABLESZ but NOT less
 * than TCP_MINHASHTABLESZ.
 */
#define TCP_MINHASHTABLESZ		64
#define TCP_MAXHASHTABLESZ		8192

extern int in_pcb_hashtablesize(void);
extern int tcp_hashtablesz;

/*
 * Tcp initialization
 */
void
tcp_init(void)
{

	tcp_iss = 1;		/* wrong */

	if (tcp_hashtablesz == 0) { /* dynamic sizing */
		tcp_hashtablesz = in_pcb_hashtablesize();
	}
	if (tcp_hashtablesz < TCP_MINHASHTABLESZ)
		tcp_hashtablesz = TCP_MINHASHTABLESZ;

	if (tcp_hashtablesz > TCP_MAXHASHTABLESZ)
		tcp_hashtablesz = TCP_MAXHASHTABLESZ;
	
	tcp_hashtablesz *= 2; 	/* room for TIME-WAITers */
	(void)in_pcbinitcb(&tcb, tcp_hashtablesz, 0, TCP_PCBSTAT);
	return;
}

/*
 * Create template to be used to send tcp packets on a connection.
 * Call after host entry created, allocates an mbuf and fills
 * in a skeletal tcp/ip header, minimizing the amount of work
 * necessary when the connection is used.
 */
struct tcpiphdr *
tcp_template(struct tcpcb *tp)
{
	register struct inpcb *inp = tp->t_inpcb;
	register struct mbuf *m;
	register struct tcpiphdr *n;

	if ((n = tp->t_template) == 0) {
		m = m_get(M_DONTWAIT, MT_HEADER);
		if (m == NULL)
			return (0);
		m->m_off = MMAXOFF - sizeof (struct tcpiphdr);
		m->m_len = sizeof (struct tcpiphdr);
		n = mtod(m, struct tcpiphdr *);
	}
	n->ti_next = n->ti_prev = 0;
	n->ti_x1 = 0;
	n->ti_pr = IPPROTO_TCP;
	n->ti_len = htons(sizeof (struct tcpiphdr) - sizeof (struct ip));
	n->ti_src = inp->inp_laddr;
	n->ti_dst = inp->inp_faddr;
	n->ti_sport = inp->inp_lport;
	n->ti_dport = inp->inp_fport;
	n->ti_seq = 0;
	n->ti_ack = 0;
	n->ti_x2 = 0;
	n->ti_off = 5;
	n->ti_flags = 0;
	n->ti_win = 0;
	n->ti_sum = 0;
	n->ti_urp = 0;
	return (n);
}

/*
 * Send a single message to the TCP at address specified by
 * the given TCP/IP header.  If m == 0, then we make a copy
 * of the tcpiphdr at ti and send directly to the addressed host.
 * This is used to force keep alive messages out using the TCP
 * template for a connection tp->t_template.  If flags are given
 * then we send a message back to the TCP which originated the
 * segment ti, and discard the mbuf containing it and any other
 * attached mbufs.
 *
 * In any case the ack and sequence number of the transmitted
 * segment are as specified by the parameters.
 */
void
tcp_respond(
	struct tcpcb *tp,
	register struct tcpiphdr *ti,
	register struct mbuf *m,
	tcp_seq ack,
	tcp_seq seq,
	int flags)
{
	register int tlen;
	int win = 0;
	struct route *ro = 0;

	if (tp) {
		win = sbspace(&tp->t_inpcb->inp_socket->so_rcv);
		ro = &tp->t_inpcb->inp_route;
	}
	if (m == 0) {
		m = m_get(M_DONTWAIT, MT_HEADER);
		if (m == NULL)
			return;
		/* TCP_COMPAT_42 */
		tlen = 1;
		*mtod(m, struct tcpiphdr *) = *ti;
		ti = mtod(m, struct tcpiphdr *);
		flags = TH_ACK;
	} else {
		m_freem(m->m_next);
		m->m_next = 0;
		m->m_off = (__psint_t)ti - (__psint_t)m;
		ASSERT((char *)ti == mtod(m, char *));
		tlen = 0;
#define xchg(a,b,type) { type t; t=a; a=b; b=t; }
		xchg(ti->ti_dst.s_addr, ti->ti_src.s_addr, u_long);
		xchg(ti->ti_dport, ti->ti_sport, u_short);
#undef xchg
	}
	ti->ti_len = htons((u_short)(sizeof (struct tcphdr) + tlen));
	tlen += sizeof (struct tcpiphdr);
	m->m_len = tlen;
	ti->ti_next = ti->ti_prev = 0;
	ti->ti_x1 = 0;
	ti->ti_seq = htonl(seq);
	ti->ti_ack = htonl(ack);
	ti->ti_x2 = 0;
	ti->ti_off = sizeof (struct tcphdr) >> 2;
	ti->ti_flags = flags;

	if ( tp )
		ti->ti_win = htons( (u_short) (win >> tp->rcv_scale) );
	else
		ti->ti_win = htons((u_short)win);
	ti->ti_urp = 0;
	/* Don't assume this field has been zeroed */
	ti->ti_sum = 0;
	ti->ti_sum = in_cksum(m, tlen);
	/* Make sure the board doesn't do a checksum */
	m->m_flags &= ~M_CKSUMMED;
	((struct ip *)ti)->ip_len = tlen;
	((struct ip *)ti)->ip_ttl = tcp_ttl;
	(void) ip_output(m, (struct mbuf *)0, ro, 0, (struct mbuf *)NULL);
}

/*
 * Create a new TCP control block, making an
 * empty reassembly queue and hooking it to the argument
 * protocol control block.
 */
struct tcpcb *
tcp_newtcpcb(struct inpcb *inp)
{
	register struct tcpcb *tp = (struct tcpcb *)
		mcb_get( M_DONTWAIT, sizeof(struct tcpcb), MT_PCB );
	if (tp == NULL)
		return((struct tcpcb *)0);

	TCP_NEXT_PUT((struct tcpiphdr *)tp, (struct tcpiphdr *)tp);
	TCP_PREV_PUT((struct tcpiphdr *)tp, (struct tcpiphdr *)tp);

	tp->t_maxseg = tcp_mssdflt;
	tp->t_maxseg0 = tcp_mssdflt;

	tp->t_flags = 0;		/* sends options! */
	tp->t_inpcb = inp;
	/*
	 * Init srtt to TCPTV_SRTTBASE (0), so we can tell that we have no
	 * rtt estimate.  Set rttvar so that srtt + 2 * rttvar gives
	 * reasonable initial retransmit time.
	 */
	tp->t_srtt = TCPTV_SRTTBASE;
	tp->t_rttvar = tcp_rttdflt * PR_SLOWHZ << 2;
	tp->t_rttmin = TCPTV_MIN;
	TCPT_RANGESET(tp->t_rxtcur, 
	    ((TCPTV_SRTTBASE >> 2) + (TCPTV_SRTTDFLT << 2)) >> 1,
	    TCPTV_MIN, TCPTV_REXMTMAX);

	tp->snd_cwnd = TCP_MAXWIN<<TCP_MAX_WINSHIFT;
	tp->snd_ssthresh = TCP_MAXWIN<<TCP_MAX_WINSHIFT;

	inp->inp_ip.ip_ttl = tcp_ttl;
	inp->inp_ppcb = (caddr_t)tp;
	return (tp);
}

/*
 * Drop a TCP connection, reporting
 * the specified error.  If connection is synchronized,
 * then send a RST to peer.
 */
struct tcpcb *
tcp_drop(register struct tcpcb *tp, int errno)
{
	struct socket *so = tp->t_inpcb->inp_socket;

	if (TCPS_HAVERCVDSYN(tp->t_state)) {
		tp->t_state = TCPS_CLOSED;
		(void) tcp_output(tp);
		TCPSTAT(tcps_drops);
	} else
		TCPSTAT(tcps_conndrops);
	if (errno == ETIMEDOUT && tp->t_softerror)
		errno = tp->t_softerror;
	so->so_error = errno;
	return (tcp_close(tp));
}

/*
 * Close a TCP control block:
 *	discard all space held by the tcp
 *	discard internet protocol block
 *	wake up any sleepers
 */
struct tcpcb *
tcp_close(register struct tcpcb *tp)
{
	register struct tcpiphdr *t;
	struct inpcb *inp = tp->t_inpcb;
	struct socket *so = inp->inp_socket;
	register struct mbuf *m;
	register struct rtentry *rt;
	NETSPL_MPDECL(s);

	ASSERT(SOCKET_ISLOCKED(so));

	/* clobber input pcb cache if we're closing the cached connection */
	INHEAD_MRWLOCK(inp->inp_head);
	if (inp == inp->inp_head->inp_last) {
		inp->inp_head->inp_last = 0;
	}
	INHEAD_MRUNLOCK(inp->inp_head);
	inp->inp_ppcb = 0;

	/*
	 * If we sent enough data to get some meaningful characteristics,
	 * save them in the routing entry.  'Enough' is arbitrarily 
	 * defined as the sendpipesize (default 4K) * 16.  This would
	 * give us 16 rtt samples assuming we only get one sample per
	 * window (the usual case on a long haul net).  16 samples is
	 * enough for the srtt filter to converge to within 5% of the correct
	 * value; fewer samples and we could save a very bogus rtt.
	 *
	 * Don't update the default route's characteristics and don't
	 * update anything that the user "locked".
	 */
	ROUTE_MPLOCK(s);
	if (SEQ_LT(tp->iss + so->so_snd.sb_hiwat * 16, tp->snd_max) &&
	    (rt = inp->inp_route.ro_rt) &&
	    ((struct sockaddr_in *)rt_key(rt))->sin_addr.s_addr != INADDR_ANY) {
		register __uint32_t i;

		if ((rt->rt_rmx.rmx_locks & RTV_RTT) == 0) {
			i = tp->t_srtt *
			    (RTM_RTTUNIT / (PR_SLOWHZ * TCP_RTT_SCALE));
			if (rt->rt_rmx.rmx_rtt && i)
				/*
				 * filter this update to half the old & half
				 * the new values, converting scale.
				 * See route.h and tcp_var.h for a
				 * description of the scaling constants.
				 */
				rt->rt_rmx.rmx_rtt =
				    (rt->rt_rmx.rmx_rtt + i) / 2;
			else
				rt->rt_rmx.rmx_rtt = i;
		}
		if ((rt->rt_rmx.rmx_locks & RTV_RTTVAR) == 0) {
			i = tp->t_rttvar *
			    (RTM_RTTUNIT / (PR_SLOWHZ * TCP_RTTVAR_SCALE));
			if (rt->rt_rmx.rmx_rttvar && i)
				rt->rt_rmx.rmx_rttvar =
				    (rt->rt_rmx.rmx_rttvar + i) / 2;
			else
				rt->rt_rmx.rmx_rttvar = i;
		}
		/*
		 * update the pipelimit (ssthresh) if it has been updated
		 * already or if a pipesize was specified & the threshhold
		 * got below half the pipesize.  I.e., wait for bad news
		 * before we start updating, then update on both good
		 * and bad news.
		 */
		if ((rt->rt_rmx.rmx_locks & RTV_SSTHRESH) == 0 &&
		    (i = tp->snd_ssthresh) && rt->rt_rmx.rmx_ssthresh ||
		    i < (rt->rt_rmx.rmx_sendpipe / 2)) {
			/*
			 * convert the limit from user data bytes to
			 * packets then to packet data bytes.
			 */
			i = (i + tp->t_maxseg / 2) / tp->t_maxseg;
			if (i < 2)
				i = 2;
			i *= (__uint32_t)(tp->t_maxseg + sizeof (struct tcpiphdr));
			if (rt->rt_rmx.rmx_ssthresh)
				rt->rt_rmx.rmx_ssthresh =
				    (rt->rt_rmx.rmx_ssthresh + i) / 2;
			else
				rt->rt_rmx.rmx_ssthresh = i;
		}
	}
	ROUTE_MPUNLOCK(s);

	/* free the reassembly queue, if any */
	t = TCP_NEXT_GET((struct tcpiphdr *)tp);
	while (t != (struct tcpiphdr *)tp) {
		t = TCP_NEXT_GET(t);
		m = REASS_MBUF_GET(TCP_PREV_GET(t));
		remque_tcb(TCP_PREV_GET(t));
		m_freem(m);
	}
	if (tp->t_template)
		(void) m_free(tcptmplttom(tp->t_template));
	mcb_free(tp, sizeof(struct tcpcb), MT_PCB);
	soisdisconnected(so);

	/*
	 * Wait for sosend()/soreceive() to exit uiomove() critical section.
	 * This used to sleep in a loop, but on systems with only a single
	 * rtnetd, that could deadlock with a process hung waiting on NFS.
	 *
	 * So, now if anybody is in sosend() or soreceive(), they simply
	 * exit, ensuring that so_holds is decremented.  When soclose() is
	 * called, in_pcbdetach() will be called by tcp_usrreq() in response
	 * to the PRU_DETACH operation.
	 */
	so->so_state |= SS_CLOSING;
	if (so->so_holds == 0) {
		/*
		 * In the case of being called by soabort(), if the connection
		 * is not freed, unlock the socket; that is what callers of
		 * soabort() expect.
		 */
		if (!in_pcbdetach(inp) && so->so_error == ECONNABORTED)
			SOCKET_MPUNLOCK(so);
	} else {
		/*
		 * In order to ensure that no incoming segments find this
		 * PCB in its semi-dead state, remove it from the list of
		 * valid PCBs
		 */
		in_pcbunlink(inp);
		/*
		 * Leave socket locked for caller to unlock, unless it is
		 * soabort().
		 */
		if (so->so_error == ECONNABORTED) {
			SOCKET_MPUNLOCK(so);
		}
		/*
		 * Note that we exit this section of code with the PCB
		 * reference count inflated.  The call to in_pcbdetach() when
		 * the deferred close happens will account for this.
		 */
	}
	TCPSTAT(tcps_closed);
	return ((struct tcpcb *)0);
}

void
tcp_drain(void)
{

}

/*
 * Notify a tcp user of an asynchronous error;
 * store error as soft error, but wake up user
 * (for now, won't do anything until can select for soft error).
 */
/*ARGSUSED*/
void
tcp_notify(register struct inpcb *inp, int error, void *data)
{
	struct tcpcb *tp;
	/* if the connection is already closed, don't bother */
	if ((tp = intotcpcb(inp)) == 0)
		return;

	/* ignore some errors if we are hooked up */
	if (intotcpcb(inp)->t_state == TCPS_ESTABLISHED &&
	    (error == EHOSTUNREACH || error == ENETUNREACH ||
	     error == EHOSTDOWN)) {
		return;
	}

	if (tp->t_state == TCPS_SYN_RECEIVED) {
		/* help stop SYN-bombing */
		tp->t_timer[TCPT_KEEP] = 1;
		return;
	}

	tp->t_softerror = error;
	wakeup((caddr_t) &inp->inp_socket->so_timeo);
	sorwakeup(inp->inp_socket, NETEVENT_TCPUP);
	sowwakeup(inp->inp_socket, NETEVENT_TCPUP);
}

void
tcp_ctlinput(cmd, sa, icp)
	int cmd;
	struct sockaddr *sa;
	struct icmp *icp;
{
	register struct tcphdr *th;
	extern struct in_addr zeroin_addr;
	extern u_char inetctlerrmap[];
	void tcp_quench(struct inpcb *, int, void *);
	void (*notify)(struct inpcb *, int, void *) = tcp_notify;
	void tcp_msgsize(struct inpcb *inp, int errno, struct icmp *icp);
	register struct ip *ip = icp ? &icp->icmp_ip : NULL;
	struct in_addr src;
	register u_short dport = 0, sport = 0;

	if (cmd == PRC_QUENCH)
		notify = tcp_quench;
	else if ( cmd == PRC_MSGSIZE )
		notify = (void(*)(struct inpcb *, int, void *))tcp_msgsize;
	else if (!PRC_IS_REDIRECT(cmd) &&
		    ((unsigned)cmd >= PRC_NCMDS || inetctlerrmap[cmd] == 0))
		return;
	if (ip) {
		/* Make sure the packet contains at least the ports. */
		th = (struct tcphdr *)((caddr_t)ip + (ip->ip_hl << 2));
		if (((caddr_t)th+sizeof(th->th_sport)+sizeof(th->th_dport)) <=
		    (caddr_t)ip + ip->ip_len) {
			sport = th->th_sport;
			dport = th->th_dport;
		}
		src = ip->ip_src;
	} else
		src = zeroin_addr;
	in_pcbnotify(&tcb, sa, dport, src, sport, cmd, notify, (void *) icp );
}

/*
 * When a source quench is received, close congestion window
 * to one segment.  We will gradually open it again as we proceed.
 */
/*ARGSUSED*/
void
tcp_quench(struct inpcb *inp, int flag, void *data)
{
	struct tcpcb *tp = intotcpcb(inp);

	if (tp)
		tp->snd_cwnd = tp->t_maxseg;
}


/*
 * Receive a message size error from a router.
 * This means that we're doing (RFC 1191) MTU discovery and our
 * mss is too big.  If we're lucky, the router will tell us
 * what the far side mtu size is.
 */
/* ARGSUSED */
void
tcp_msgsize(
	struct inpcb *inp,
	int     errno,
	register struct icmp *icp)
{
	int nextmtu, mss, tcp_optionslen;		/* MUST BE signed ! */
	struct tcpcb *tp = intotcpcb(inp);
	extern int tcp_mtutable[];

	if ( ! tcp_mtudisc || ! tp || ! icp )
		return;		/* Must be a stale message */

	/* If we're using RFC 1323 time-stamps, adjust for them. */
	tcp_optionslen = (tp->t_flags&(TF_REQ_TSTMP|TF_RCVD_TSTMP)) ==
		(TF_REQ_TSTMP|TF_RCVD_TSTMP) ? TCPOLEN_TSTAMP_HDR : 0;

	/* Get next MTU from icmp message (RFC1191 extension to routers) */
	nextmtu = ntohs( icp->icmp_nextmtu );

	/* If router doesn't give us an MTU value, then try the next
	 * smaller value in our table of typical MTU values.  This table
	 * is in master.d/bsd.
	 */
	if (nextmtu == 0 ) {
		register int i = 0;
		int previous_len = ntohs( icp->icmp_ip.ip_len );

		/* Correct for broken BSD based routers.  They return
		 * a total length field that is the total length plus
		 * the header length.  See Note in section 5 (page 8)
		 * of RFC 1191.
		 */
		if ( previous_len > tp->t_maxseg+sizeof(struct tcpiphdr)+
		   tcp_optionslen )
			previous_len -= icp->icmp_ip.ip_hl<<2;

		while ( previous_len <= tcp_mtutable[i] && tcp_mtutable[i] )
			i++;

		nextmtu = tcp_mtutable[i];	/* Could be zero. */
	}

	if (nextmtu==0 || nextmtu-sizeof(struct tcpiphdr) <= tcp_mssdflt)
		mss = tcp_mssdflt-tcp_optionslen;
	else
		mss = nextmtu-sizeof(struct tcpiphdr)-tcp_optionslen;
	
	/* straight out of tcp_mss()... */
#if (MCLBYTES & (MCLBYTES - 1)) == 0
	if (mss > MCLBYTES)
		mss &= ~(MCLBYTES-1);
#else
	if (mss > MCLBYTES)
		mss = mss / MCLBYTES * MCLBYTES;
#endif

	/* Only start retransmission if this lowers the mss.  Otherwise,
	 * it might be a duplicate message in which case we don't want
	 * to retransmit smaller segments.  If something is really wrong,
	 * the retransmission timer should clean things up.
	 */
	if ( mss < tp->t_maxseg ) {

#ifdef DEBUG
		static int tcp_mtudebug = 1;
		if ( tcp_mtudebug )
			printf( "tcp_msgsize: tp=%x,nextmtu=%d, IP.len=%d, oldmss=%d, newmss=%d\n",
			(long) tp, nextmtu, icp->icmp_ip.ip_len, tp->t_maxseg,
			mss );
#endif /* DEBUG */


		tp->t_maxseg = mss;

		/* start slow-start phase again but don't
		 * change snd_ssthresh.
		 */
		tp->snd_cwnd = mss;

		/* Set up TCP timer to expire this smaller mtu in
		 * 5 minutes.
		 */
		tp->t_timer[TCPT_MTUEXP] = TCPTV_MTUEXP;

		/* Begin retransmission.  Reset retransmit timer and 
		 * turn off rtt measurement.
		 */
		tp->snd_nxt = tp->snd_una;
		tp->t_rtt = 0;
		tp->t_timer[TCPT_REXMT] = 0;
		tcp_output(tp);
	}
}

void
insque_tcb(register struct tcpiphdr *ep, register struct tcpiphdr *pp)
{
#ifdef TCPPRINTFS
	if (tcpprintfs & 0x02)
		printf("tcp_reass, insert, ep=0x%x, pp=0x%x\n", ep, pp);
#endif
	TCP_NEXT_PUT(ep, TCP_NEXT_GET(pp));
	TCP_PREV_PUT(ep, pp);
	TCP_PREV_PUT(TCP_NEXT_GET(pp), ep);
	TCP_NEXT_PUT(pp, ep);
}

void
remque_tcb(register struct tcpiphdr *ep)
{
#ifdef TCPPRINTFS
	if (tcpprintfs & 0x02)
		printf("tcp_reass, remove, ep=0x%x\n", ep);
#endif
	TCP_NEXT_PUT(TCP_PREV_GET(ep), TCP_NEXT_GET(ep));
	TCP_PREV_PUT(TCP_NEXT_GET(ep), TCP_PREV_GET(ep));
}

