/*
 * Copyright (c) 1982, 1986, 1988, 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution is only permitted until one year after the first shipment
 * of 4.4BSD by the Regents.  Otherwise, redistribution and use in source and
 * binary forms are permitted provided that: (1) source distributions retain
 * this entire copyright notice and comment, and (2) distributions including
 * binaries display the following acknowledgement:  This product includes
 * software developed by the University of California, Berkeley and its
 * contributors'' in the documentation or other materials provided with the
 * distribution and in all advertising materials mentioning features or use
 * of this software.  Neither the name of the University nor the names of
 * its contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)udp_usrreq.c	7.7 (Berkeley) 6/29/88 plus MULTICAST 1.2
 *	plus parts of 7.18 (Berkeley) 7/25/90
 */

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/sema.h"
#include "sys/hashing.h"

#include "sys/errno.h"
#include "sys/mbuf.h"
#include "sys/protosw.h"
#include "sys/socket.h"
#include "sys/socketvar.h"
#include "sys/tcpipstats.h"
#include "sys/sysmacros.h"

#include "sys/user.h"
#include "sys/systm.h"
#include "net/if.h"
#include "net/route.h"

#include "in.h"
#include "in_systm.h"
#include "in_var.h"
#include "ip.h"
#include "in_pcb.h"
#include "ip_var.h"
#include "ip_icmp.h"
#include "icmp_var.h"
#include "udp.h"
#include "udp_var.h"

#define UDP_MINHASHTABLESZ	64
#define UDP_MAXHASHTABLESZ	2048

extern	int	udp_hashtablesz;

/* Forward references. */
void udp_detach(struct inpcb *);
static  struct mbuf *udp_saveopt(caddr_t, int, int);

/* External references. */
extern int in_pcb_hashtablesize(void);

/*
 * UDP protocol implementation.
 * Per RFC 768, August, 1980.
 */
void
udp_init(void)
{
	if (udp_hashtablesz == 0) {
	      udp_hashtablesz = in_pcb_hashtablesize();
	}
	if (udp_hashtablesz < UDP_MINHASHTABLESZ)
	      udp_hashtablesz = UDP_MINHASHTABLESZ;
	if (udp_hashtablesz > UDP_MAXHASHTABLESZ)
	      udp_hashtablesz = UDP_MAXHASHTABLESZ;

	(void)in_pcbinitcb(&udb, udp_hashtablesz, INPFLAGS_CLTS, UDP_PCBSTAT);

	return;
}

extern int	udp_ttl;
struct	inpcb udb;	/* head of UDP protocol control block list */
int nfs_port = -1;	/* "special" port number for NFS packets */

extern void nfs_input(struct mbuf *m);	/* fast path hook to NFS */

#ifndef _MP_NETLOCKS
struct	sockaddr_in udp_in = { AF_INET };
#endif

void
udp_input(m, ifp)
	struct mbuf *m;
	struct ifnet *ifp;
{
	register struct udpiphdr *ui;
	register struct inpcb *inp, *lastinp;
	int len;
	struct ip ip;
	struct in_pcbhead *hinp;
	u_short hashval;
	struct mbuf *opts = 0;

#ifdef _MP_NETLOCKS
	struct sockaddr_in udp_in;
	extern struct sockaddr_in in_zeroaddr;
#endif
	struct ifnet *ifns;

	NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
		 NETEVENT_UDPUP, NETCNT_NULL, NETRES_PROTCALL);
	UDPSTAT(udps_ipackets);

	/*
	 * ip_intr has pulled up entire IP header with options.
	 * Strip IP options, if any; should skip this,
	 * make available to user, and use on returned packets,
	 * but we don't yet have a way to check the checksum
	 * with options still present.
	 *
	 */
	ui = mtod(m, struct udpiphdr *);
	if (((struct ip *)ui)->ip_hl > (sizeof (struct ip) >> 2))
		ip_stripoptions(m, (struct mbuf *)0);
	if (m->m_len < sizeof (struct udpiphdr)) {
		if ((m = m_pullup(m, sizeof (struct udpiphdr))) == 0) {
			UDPSTAT(udps_hdrops);
			NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
				 NETEVENT_UDPUP, m->m_len, NETRES_HEADER);
			return;
		}
		ui = mtod(m, struct udpiphdr *);
	}

	/*
	 * Make mbuf data length reflect UDP length.
	 * If not enough data to reflect UDP length, drop.
	 */
	len = ntohs((u_short)ui->ui_ulen);
	if (((struct ip *)ui)->ip_len != len) {
		if (len > ((struct ip *)ui)->ip_len) {
			UDPSTAT(udps_badlen);
			NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
				 NETEVENT_UDPUP, len, NETRES_SIZE);
			goto bad;
		}
		m_adj(m, len - ((struct ip *)ui)->ip_len);
	}
	/*
	 * Save a copy of the IP header in case we want restore it for ICMP.
	 */
	ip = *(struct ip*)ui;

	/*
	 * Checksum extended UDP header and data.
	 * Believe the link layer if possible.
	 */
	if (ui->ui_sum) {
		ui->ui_next = ui->ui_prev = 0;
		ui->ui_x1 = 0;
		ui->ui_len = ui->ui_ulen;
		if (!(m->m_flags & M_CKSUMMED) && 
		    in_cksum(m, len + sizeof (struct ip))) {
			UDPSTAT(udps_badsum);
			m_freem(m);
			NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
			    NETEVENT_UDPUP, len, NETRES_CKSUM);
			return;
		}
	}

	if (ui->ui_dport == nfs_port) {
		/*
		 * Special-case fast path for NFS
		 */
		nfs_input(m);
		return;
	}
		

#if _MP_NETLOCKS
	udp_in = in_zeroaddr;
#endif

	if (IN_MULTICAST(ntohl(ui->ui_dst.s_addr)) ||
	    in_broadcast(ui->ui_dst, ifp)) {
		struct socket *last;
		struct inpcb *inpnxt;
		struct inpcb *head = &udb;
		int done = 0;

		/*
		 * Deliver a multicast or broadcast datagram to *all* sockets
		 * for which the local and remote addresses and ports match
		 * those of the incoming datagram.  This allows more than
		 * one process to receive multi/broadcasts on the same port.
		 * (This really ought to be done for unicast datagrams as
		 * well, but that would cause problems with existing
		 * applications that open both address-specific sockets and
		 * a wildcard socket listening to the same port -- they would
		 * end up receiving duplicates of every unicast datagram.
		 * Those applications open the multiple sockets to overcome an
		 * inadequacy of the UDP socket interface, but for backwards
		 * compatibility we avoid the problem here rather than
		 * fixing the interface.  Maybe 4.4BSD will remedy this?)
		 */

		/*
		 * Construct sockaddr format source address.
		 */
		udp_in.sin_port = ui->ui_sport;
		udp_in.sin_addr = ui->ui_src;
		m->m_len -= sizeof (struct udpiphdr);
		m->m_off += sizeof (struct udpiphdr);
		/*
		 * Locate pcb(s) for datagram.
		 * (Algorithm copied from raw_intr().)
		 */
		INHEAD_MRRLOCK(head);
		hashval = (*head->inp_hashfun)(head, ui->ui_src,
			ui->ui_sport, ui->ui_dst, ui->ui_dport);

resync:
                /*
                 * This section of code takes advantage of the fact that
                 * we know that UDP inpcb's hash into exactly one
                 * list per port number.
                 */
		last = NULL;
		hinp = &head->inp_table[hashval];
		for (inp = hinp->hinp_next; !done && 
			inp != (struct inpcb *)hinp; inp = inpnxt) {
			ASSERT(INHEAD_ISMRLOCKED(head));
			inpnxt = inp->inp_next;
			/*
			 * Make sure sockets with no multicast options
			 * do not get multicast packets.
			 */
			if (IN_MULTICAST(ntohl(ui->ui_dst.s_addr)) &&
				(inp->inp_moptions == NULL)) {
					continue;
			}
			if (inp->inp_lport != ui->ui_dport) {
				continue;
			}
			if (inp->inp_laddr.s_addr != INADDR_ANY) {
				if (inp->inp_laddr.s_addr !=
					ui->ui_dst.s_addr)
					continue;
			}
			if (inp->inp_faddr.s_addr != INADDR_ANY) {
				if (inp->inp_faddr.s_addr !=
					ui->ui_src.s_addr ||
				    inp->inp_fport != ui->ui_sport)
					continue;
			}

			INPCB_HOLD(inp);
			if (last != NULL) {
				struct mbuf *n;

				if ((n = m_copy(m, 0, M_COPYALL)) != NULL) {
					lastinp = sotoinpcb(last);
					INHEAD_MRUNLOCK(head);
					SOCKET_MPLOCK(last);
					if (last->so_options & SO_PASSIFNAME)
					    ifns = ifp;
					else
					    ifns = 0;
					if (sbappendaddr(&last->so_rcv,
					  (struct sockaddr *)&udp_in, n, 
					  (struct mbuf *)0, ifns) == 0) {
						m_freem(n);
						UDPSTAT(udps_fullsock);
						NETPAR(NETFLOW, NETDROPTKN,
						  NETPID_NULL, NETEVENT_UDPUP,
						  len, NETRES_SBFULL);
					} else {
						sorwakeup(last, NETEVENT_UDPUP);
					}
					if (!INPCB_RELE(lastinp))
						SOCKET_MPUNLOCK(last);
					INHEAD_MRRLOCK(head);
					if (inp->inp_next != inpnxt) {
						struct socket *so = 
							inp->inp_socket;
			
						INHEAD_MRUNLOCK(head);	
						SOCKET_MPLOCK(so);
						if (!INPCB_RELE(inp))
							SOCKET_MPUNLOCK(so);
						INHEAD_MRRLOCK(head);
						goto resync;
					}
				}
			}
			last = inp->inp_socket;
			/*
			 * Don't look for additional matches if this one does
			 * not have one of the SO_REUSE{ADDR,PORT} options set.
			 * This heuristic avoids searching through all pcbs
			 * in the common case of a non-shared port.  It
			 * assumes that an application will never clear
			 * the SO_REUSE{ADDR,PORT} option after setting it.
			 */
			if ((last->so_options & 
				(SO_REUSEADDR|SO_REUSEPORT))==0) {
				done = 1;
				break;
			}
		} /* per-hash for loop */
		INHEAD_MRUNLOCK(head);

		if (last == NULL) {
			/*
			 * No matching pcb found; discard datagram.
			 * (No need to send an ICMP Port Unreachable
			 * for a broadcast or multicast datgram.)
			 */

			NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
			       NETEVENT_UDPUP, len, NETRES_UNREACH);
			UDPSTAT(udps_noportbcast);
			goto bad;
		}
		lastinp = sotoinpcb(last);
		SOCKET_MPLOCK(last);
		if (last->so_options & SO_PASSIFNAME)
		    ifns = ifp;
		else
		    ifns = 0;
		if (sbappendaddr(&last->so_rcv, (struct sockaddr *)&udp_in,
		     m, (struct mbuf *)0, ifns) == 0)
		{
			    UDPSTAT(udps_fullsock);
			    NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
				 NETEVENT_UDPUP, len, NETRES_SBFULL);
			    if (!INPCB_RELE(lastinp))
			    	SOCKET_MPUNLOCK(last);
			    goto bad;
		}

		NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
			 NETEVENT_UDPUP, len, NETRES_NULL);
		sorwakeup(last, NETEVENT_UDPUP);
		if (!INPCB_RELE(lastinp))
			SOCKET_MPUNLOCK(last);
		return;
	}
	/*
	 * Locate pcb for datagram.
	 */
	INHEAD_MRRLOCK(&udb);
	inp = udb.inp_last;
	if (!inp || !IP_ADDRS_EQUAL(inp, ui->ui_dst, ui->ui_dport, 
			ui->ui_src, ui->ui_sport)) {
		INHEAD_MRUNLOCK(&udb);
		UDPSTAT(udpps_pcbcachemiss);
		inp = in_pcblookup(&udb,
			ui->ui_src, ui->ui_sport, ui->ui_dst, ui->ui_dport,
			INPLOOKUP_WILDCARD);
		if (inp == 0) {		
			struct in_addr null_inaddr = {0};
			NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
					NETEVENT_UDPUP, len, NETRES_UNREACH);
			UDPSTAT(udps_noport);
			*(struct ip *)ui = ip;
			ASSERT(ui == mtod(m, struct udpiphdr *));

			icmp_error(m, ICMP_UNREACH, ICMP_UNREACH_PORT, ifp,
				   null_inaddr, NULL);
			return;
		}
	} else {
		INPCB_HOLD(inp);
		INHEAD_MRUNLOCK(&udb);
	}

	/*
	 * Construct sockaddr format source address.
	 * Stuff source address and datagram in user buffer.
	 */
	udp_in.sin_port = ui->ui_sport;
	udp_in.sin_addr = ui->ui_src;
	m->m_len -= sizeof (struct udpiphdr);
	m->m_off += sizeof (struct udpiphdr);
	if (inp->inp_socket->so_options & SO_PASSIFNAME)
	    ifns = ifp;
	else
	    ifns = 0;

	if (inp->inp_flags & INP_CONTROLOPTS) {
		struct mbuf **mp = &opts;

		if (inp->inp_flags & INP_RECVDSTADDR) {
			*mp = udp_saveopt((caddr_t) &ui->ui_dst,
			    sizeof(struct in_addr), IP_RECVDSTADDR);
			if (*mp)
				mp = &(*mp)->m_next;
		}
	}
	SOCKET_MPLOCK(inp->inp_socket);
	if (sbappendaddr(&inp->inp_socket->so_rcv, 
		(struct sockaddr *)&udp_in, m, opts, ifns) == 0) {
		UDPSTAT(udps_fullsock);
		NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
			 NETEVENT_UDPUP, len, NETRES_SBFULL);
		if (!INPCB_RELE(inp))
			SOCKET_MPUNLOCK(inp->inp_socket);
		goto bad;
	}
	NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
		 NETEVENT_UDPUP, len, NETRES_NULL);
	/*
	 * Our cheezeball hack of using MT_RIGHTS for this requires us to
	 * free it here, since sbappendaddr() makes a copy.
	 */
	if (opts) {
		m_freem(opts);
	}
	sorwakeup(inp->inp_socket, NETEVENT_UDPUP);
	if (!INPCB_RELE(inp))
		SOCKET_MPUNLOCK(inp->inp_socket);
	return;
bad:
	m_freem(m);
	if (opts) {
		m_freem(opts);
	}
}

/*
 * Notify a udp user of an asynchronous error;
 * just wake up so that he can collect error status.
 */
#ifdef sgi
/*ARGSUSED*/
void
udp_notify(
	register struct inpcb *inp,
	int errno,
	void *data)
#else
udp_notify(inp, errno)
	register struct inpcb *inp;
	int errno;
#endif
{
	inp->inp_socket->so_error = errno;
	sorwakeup(inp->inp_socket, NETEVENT_UDPUP);
	sowwakeup(inp->inp_socket, NETEVENT_UDPUP);
}

#ifdef sgi /* MTUDisc */
void
udp_ctlinput(cmd, sa, icp)
	int cmd;
	struct sockaddr *sa;
	struct icmp *icp;
#else
udp_ctlinput(cmd, sa, ip)
	int cmd;
	struct sockaddr *sa;
	register struct ip *ip;
#endif /* MTUDisc */
{
	register struct udphdr *uh;
	extern struct in_addr zeroin_addr;
	extern u_char inetctlerrmap[];
	struct in_addr src;
	register u_short dport = 0, sport = 0;
#ifdef sgi /* MTUDisc */
	register struct ip *ip = icp ? &icp->icmp_ip : NULL;
#endif /* MTUDisc */

	if (!PRC_IS_REDIRECT(cmd) && 
	    ((unsigned)cmd >= PRC_NCMDS || inetctlerrmap[cmd] == 0))
		return;
	if (ip) {
		uh = (struct udphdr *)((caddr_t)ip + (ip->ip_hl << 2));
		if (((caddr_t)uh+sizeof(uh->uh_sport)+sizeof(uh->uh_dport)) <=
		    (caddr_t)ip + ip->ip_len) {
			sport = uh->uh_sport;
			dport = uh->uh_dport;
		}
		src = ip->ip_src;
	} else
		src = zeroin_addr;
#ifdef sgi
	in_pcbnotify(&udb, sa, dport, src, sport, cmd, udp_notify, 0);
#else
	in_pcbnotify(&udb, sa, dport, src, sport, cmd, udp_notify);
#endif
}

udp_output(
	register struct inpcb *inp,
	struct mbuf *m0,
	struct inaddrpair *iap)
{
	register struct mbuf *m;
	register struct udpiphdr *ui;
	register int len = 0;

	ASSERT(SOCKET_ISLOCKED(inp->inp_socket));
	/*
	 * Calculate data length and get a mbuf
	 * for UDP and IP headers.
	 */
	NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
		 NETEVENT_UDPDOWN, NETCNT_NULL, NETRES_PROTCALL);
	for (m = m0; m; m = m->m_next)
		len += m->m_len;
	if (len > (IP_MAXPACKET - sizeof(struct udpiphdr))) {
		NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
			 NETEVENT_UDPDOWN, len, NETRES_SIZE);
		m_freem(m0);
		return (EMSGSIZE);
	}
	MGET(m, M_DONTWAIT, MT_HEADER);
	if (m == 0) {
		NETPAR(NETFLOW, NETDROPTKN, NETPID_NULL,
			 NETEVENT_UDPDOWN, len, NETRES_MBUF);
		m_freem(m0);
		return (ENOBUFS);
	}

	/*
	 * Fill in mbuf with extended UDP header
	 * and addresses and length put into network format.
	 */
	m->m_off = MMAXOFF - sizeof (struct udpiphdr);
	m->m_len = sizeof (struct udpiphdr);
	m->m_next = m0;
	ui = mtod(m, struct udpiphdr *);
	ui->ui_next = ui->ui_prev = 0;
	ui->ui_x1 = 0;
	ui->ui_pr = IPPROTO_UDP;
	ui->ui_len = htons((u_short)len + sizeof (struct udphdr));
	ui->ui_src = iap->iap_laddr;
	ui->ui_dst = iap->iap_faddr;
	ui->ui_sport = iap->iap_lport;
	ui->ui_dport = iap->iap_fport;
	ui->ui_ulen = ui->ui_len;

	/*
	 * Stuff checksum (required now) and output datagram.
	 */
	ui->ui_sum = 0;
	{
		/*
		 * Let the link layer compute the checksum if it wants
		 */
		struct rtentry *rt;
		struct ifnet *ifp;
		extern struct socket *ip_mrouter;

		if ((rt = inp->inp_route.ro_rt)
		    && 0 != (ifp = rt->rt_ifp)
		    && 0 != (ifp->if_flags & IFF_CKSUM)
		    && sizeof(struct udpiphdr) + len < ifp->if_mtu
		    && 0 != (rt->rt_flags & RTF_UP)
		    && (((struct sockaddr_in *)&inp->inp_route.ro_dst
			 )->sin_addr.s_addr == ui->ui_dst.s_addr)
		    && (0 == ip_mrouter ||
			!IN_MULTICAST(ntohl(ui->ui_dst.s_addr)))
		    ) {
			ui->ui_sum = 0xffff;
			m->m_flags |= M_CKSUMMED;
		} else {
			if ((ui->ui_sum = in_cksum(m, sizeof (struct udpiphdr)
						   + len)) == 0)
			    ui->ui_sum = 0xffff;
		}
	}
	((struct ip *)ui)->ip_len = sizeof (struct udpiphdr) + len;
	((struct ip *)ui)->ip_ttl = inp->inp_ip.ip_ttl;	/* XXX */
	((struct ip *)ui)->ip_tos = inp->inp_ip.ip_tos;	/* XXX */
	NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
		 NETEVENT_UDPDOWN, len, NETRES_NULL);
	UDPSTAT(udps_opackets);
	return ip_output(m, inp->inp_options, &inp->inp_route,
	    (inp->inp_socket->so_options & (SO_DONTROUTE |
					    SO_BROADCAST))
	    | IP_MULTICASTOPTS, inp->inp_moptions);

}

extern u_long udp_sendspace, udp_recvgrams;

/*ARGSUSED*/
udp_usrreq(so, req, m, nam, rights)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *rights;
{
	struct inpcb *inp = sotoinpcb(so);
	int error = 0;
	u_long udp_recvspace;
	struct in_addr oladdr;
	struct in_addr *from = 0;

	ASSERT(SOCKET_ISLOCKED(so));
	if (req == PRU_CONTROL) 
		return (in_control(so, (__psint_t)m, (caddr_t)nam,
			(struct ifnet *)rights));
	if ((req != PRU_SEND) && rights && rights->m_len) {
		error = EINVAL;
		goto release;
	}
	if (inp == NULL && req != PRU_ATTACH) {
		error = EINVAL;
		goto release;
	}
	/*
	 * Note: need to block udp_input while changing
	 * the udp pcb queue and/or pcb addresses.
	 */
	switch (req) {

	case PRU_ATTACH:
		if (inp != NULL) {
			error = EINVAL;
			break;
		}
		error = in_pcballoc(so, &udb);
		if (error)
			break;
		udp_recvspace = udp_recvgrams
			* (udp_sendspace+sizeof(struct sockaddr_in));
		error = soreserve(so, udp_sendspace, udp_recvspace);
		if (error)
			break;
		((struct inpcb *) so->so_pcb)->inp_ip.ip_ttl = udp_ttl;
		break;

	case PRU_DETACH:
		udp_detach(inp);
		inp = 0;
		break;

	case PRU_BIND:
		error = in_pcbbind(inp, nam);
		break;

	case PRU_LISTEN:
		error = EOPNOTSUPP;
		break;

	case PRU_CONNECT:
		if (inp->inp_faddr.s_addr != INADDR_ANY) {
			error = EISCONN;
			break;
		}
		NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
			 NETEVENT_UDPDOWN, NETCNT_NULL, NETRES_CONN);
		error = in_pcbconnect(inp, nam);
		NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
			 NETEVENT_UDPDOWN, NETCNT_NULL, NETRES_CONNDONE);
		if (error == 0)
			soisconnected(so);
		break;

	case PRU_CONNECT2:
		error = EOPNOTSUPP;
		break;

	case PRU_ACCEPT:
		error = EOPNOTSUPP;
		break;

	case PRU_DISCONNECT:
		if (inp->inp_faddr.s_addr == INADDR_ANY) {
			error = ENOTCONN;
			break;
		}
		in_pcbdisconnect(inp);
		inp->inp_laddr.s_addr = INADDR_ANY;
		so->so_state &= ~SS_ISCONNECTED;		/* XXX */
		break;

	case PRU_SHUTDOWN:
		socantsendmore(so);
		break;

	case PRU_SEND: {
		struct inaddrpair ipr;
		struct sockaddr_in *sin;

		if (nam) {
			/*
			 * Hack to allow sendmsg() call to specify source
			 * address to use.  This avoids having to open
			 * thousands of sockets when using large numbers of
			 * aliases.
			 */
			oladdr = inp->inp_laddr;
			if (rights) {
				struct cmsghdr *cmsg;
				struct sockaddr_in sin;
#if _K64U64
				extern int irix5_to_cmsghdr(struct mbuf *);

				irix5_to_cmsghdr(rights);
#endif
				cmsg = mtod(rights, struct cmsghdr *);
				bzero((caddr_t)&sin, sizeof(sin));
#define _LEN	(sizeof(*cmsg) + sizeof(*from))
				if (rights->m_len < _LEN) {
					error = EINVAL;
					break;
				}
				if (cmsg->cmsg_len != _LEN ||
				    cmsg->cmsg_level != IPPROTO_IP ||
				    cmsg->cmsg_type != IP_SENDSRCADDR) {
					error = EINVAL;
					break;
				}
				from = (struct in_addr *)CMSG_DATA(cmsg);
				sin.sin_family = AF_INET;
				sin.sin_addr = *from;

				/* must be one of our addresses */
				if (!ifa_ifwithaddr((struct sockaddr *)&sin)) {
					error = EADDRNOTAVAIL;
					break;
				}
			}
#undef _LEN
			if (inp->inp_faddr.s_addr != INADDR_ANY) {
				error = EISCONN;
				break;
			}
			sin = mtod(nam, struct sockaddr_in *);
			if (nam->m_len != sizeof(*sin)) {
				error = EINVAL;
				break;
			}
			if (from) {
				if (inp->inp_lport == 0) {
					error = in_pcbbind(inp, 
						(struct mbuf *)0);
					if (error) {
						break;
					}
				}
				inp->inp_laddr = *from;
			}
			if (error = in_pcbsetaddr(inp, sin, &ipr)) {
				inp->inp_laddr = oladdr;
				break;
			}
			error = udp_output(inp, m, &ipr);
			inp->inp_laddr = oladdr;
		} else {
			if (inp->inp_faddr.s_addr == INADDR_ANY) {
				error = ENOTCONN;
				break;
			}
			error = udp_output(inp, m, &inp->inp_iap);
		}
		m = NULL;
		}
		break;

	case PRU_ABORT:
		soisdisconnected(so);
		udp_detach(inp);
		break;

	case PRU_SOCKADDR:
		in_setsockaddr(inp, nam);
		break;

	case PRU_PEERADDR:
		in_setpeeraddr(inp, nam);
		break;

	case PRU_SENSE:
		/*
		 * stat: don't bother with a blocksize.
		 */
		return (0);

	case PRU_SENDOOB:
	case PRU_FASTTIMO:
	case PRU_SLOWTIMO:
	case PRU_PROTORCV:
	case PRU_PROTOSEND:
		error =  EOPNOTSUPP;
		break;

	case PRU_RCVD:
	case PRU_RCVOOB:
		return (EOPNOTSUPP);	/* do not free mbuf's */

	case PRU_SOCKLABEL:
		error = EOPNOTSUPP;
		break;

	default:
		panic("udp_usrreq");
	}
release:
	if (m != NULL)
		m_freem(m);
	return (error);
}

#ifdef sgi
void
#endif
udp_detach(struct inpcb *inp)
{
	in_pcbdetach(inp);
}

/*
 * Create a "control" mbuf containing the specified data
 * with the specified type for presentation with a datagram.
 */
struct mbuf *
udp_saveopt(caddr_t p, int size, int type)
{
	register struct cmsghdr *cp;
	struct mbuf *m;

	if ((m = m_get(M_DONTWAIT, MT_RIGHTS)) == NULL)	/* XXX control */
		return ((struct mbuf *) NULL);
	cp = (struct cmsghdr *) mtod(m, struct cmsghdr *);
	bcopy(p, CMSG_DATA(cp), size);
	size += sizeof(*cp);
	m->m_len = size;
	cp->cmsg_len = size;
	cp->cmsg_level = IPPROTO_IP;
	cp->cmsg_type = type;
	return (m);
}
