/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)uipc_socket.c	7.10 (Berkeley) 6/29/88
 */

#ident	"$Revision: 4.70 $"

#include "tcp-param.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/cmn_err.h"
#include "sys/domain.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/mbuf.h"
#include "sys/proc.h"
#include "sys/protosw.h"
#include "sys/signal.h"
#include "sys/socket.h"
#include "sys/socketvar.h"
#include "sys/strmp.h"
#include "sys/uio.h"
#include "sys/user.h"
#include "sys/systm.h"
#include "sys/kmem.h"
#include "bstring.h"
#include "sys/sat.h" 
#include "sys/capability.h"
#include "sys/atomic_ops.h"

static void sorflush(struct socket *so);
extern struct pollhead *phalloc(int);
extern void phfree(struct pollhead *);
extern void sbdroprecord(struct sockbuf *sb);
#ifdef _MP_NETLOCKS
static void mp_sorflush(struct socket *so);
#endif

extern int os_has_new_pflip;

/*
 * Lock a socket buffer sb for socket so.
 * returns 0 if successfully locked
 * if the intr flag is set, wait interruptable by signals
 *	returns EINTR if interrupted, with socket NOT locked
 * else
 *	wait ignoring signals, return 0 with the socket locked
 */
/* ARGSUSED */
static int
sblock(struct sockbuf	*sb,
	int		event,
	struct socket	*so,
	int		*ospl,
	int 		intr)
{
	NETSPL_SPDECL(ts)

	SOCKET_LOCK(so, ts);
	while ((sb)->sb_flags & SB_LOCK) {
                (sb)->sb_flags |= SB_WANT;
		NETPAR(NETSCHED, NETSLEEPTKN,  NETPID_NULL, event,
                       NETCNT_NULL, NETRES_SBLOCK);
		if (intr) {
		    if (SOCKET_MPUNLOCK_SBWANT(so, sb, PCATCH|PZERO+1) != 0)
			return EINTR;
		} else  {
			SOCKET_MPUNLOCK_SBWANT_NO_INTR(so, sb, PZERO+1);
		}	
		NETPAR(NETSCHED, NETWAKEUPTKN, NETPID_NULL, event,
                       NETCNT_NULL, NETRES_SBLOCK);
                SOCKET_MPLOCK(so);
        }
        (sb)->sb_flags |= SB_LOCK; 
#ifndef _MP_NETLOCKS
	*ospl = ts;
#endif
	return 0;
}

/*
 * Unlock a socket buffer sb for socket so.
 */
/* ARGSUSED */
static void
sbunlock(struct sockbuf	*sb,
	int		event,
	struct socket	*so,
	int		ospl)
{
	ASSERT(SOCKET_ISLOCKED(so));
	(sb)->sb_flags &= ~SB_LOCK;
	if ((sb)->sb_flags & SB_WANT) {
		(sb)->sb_flags &= ~SB_WANT;
		NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL, (event),
			NETCNT_NULL, NETRES_SBUNLOCK);
		SOCKET_SBWANT_WAKEALL(sb);
	}
	SOCKET_UNLOCK(so, ospl);
}


/*
 * Socket operation routines.
 * These routines are called by the routines in
 * sys_socket.c or from a system process, and
 * implement the semantics of socket operations by
 * switching out to the protocol specific routines.
 *
 * TODO:
 *	test socketpair
 *	clean up async
 *	out-of-band is a kludge
 */
/*ARGSUSED*/
socreate(int dom,
	 struct socket **aso,
	 register int type,
	 int proto)
{
	register struct protosw *prp;
	register struct socket *so;
	register int error;
	NETSPL_SPDECL(s)

	if (proto)
		prp = pffindproto(dom, proto, type);
	else
		prp = pffindtype(dom, type);
	if (prp == 0)
		return (EPROTONOSUPPORT);
	if (prp->pr_type != type)
		return (EPROTOTYPE);
	so = (struct socket *)mcb_get(M_WAIT, sizeof *so, MT_SOCKET);
	so->so_type = type;
	initpollhead(&so->so_ph);
	SOCKET_INITLOCKS(so);
	/*
	 * Audit: No capability check needed here.
	 * Privileged ports are a uid based (must be user root) mechanism.
	 * see bsd/net/raw_usrreq.c and below with SS_PRIVCHECKED.
	 */
	if (curprocp->p_cred->cr_uid == 0)
		so->so_state |= SS_PRIV;
	so->so_proto = prp;
	so->so_refcnt = 1;
	SOCKET_LOCK(so, s);
	error =
	    (*prp->pr_usrreq)(so, PRU_ATTACH,
		(struct mbuf *)0, (struct mbuf *)(NULL+proto),
			      (struct mbuf *)0);
	/* 
	 * SS_PRIVCHECKED used for auditing: 
	 */
	if (so->so_state & SS_PRIVCHECKED)
		/* acknowledge privilege check */
		(void) _CAP_CRABLE(curprocp->p_cred, CAP_NETWORK_MGT);
	if (error) {
		so->so_state |= SS_NOFDREF;
		if (!sofree(so))
			SOCKET_UNLOCK(so, s);
		return (error);
	}
	SOCKET_UNLOCK(so, s);
	*aso = so;
	return (0);
}

sobind( struct socket *so,
	struct mbuf *nam)
{
	int error;
	NETSPL_SPDECL(s)

	SOCKET_LOCK(so, s);
	error =
	    (*so->so_proto->pr_usrreq)(so, PRU_BIND,
		(struct mbuf *)0, nam, (struct mbuf *)0);
	SOCKET_UNLOCK(so, s);
	return (error);
}

solisten(register struct socket *so,
	 int backlog)
{
	int error;
	NETSPL_SPDECL(s)

	SOCKET_LOCK(so, s);
	error =
	    (*so->so_proto->pr_usrreq)(so, PRU_LISTEN,
		(struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0);
	if (error) {
		SOCKET_UNLOCK(so, s);
		return (error);
	}
	if (so->so_q == 0) {
		so->so_q = so;
		so->so_q0 = so;
		so->so_options |= SO_ACCEPTCONN;
	}
	if (backlog < 0)
		backlog = 0;
	so->so_qlimit = MIN(backlog, SOMAXCONN);
	SOCKET_UNLOCK(so, s);
	return (0);
}

sofree(register struct socket *so)
{
	struct socket *head;
#ifdef DEBUGDISCARD
	int save_state;
#endif
#ifndef _MP_NETLOCKS
#define mp_sorflush(x)	sorflush(x)
#endif

	/* 
	 * Decision to dequeue and the dequeue operation must be atomic.
	 * MP protection for soqremque() 
	 */
	ASSERT(SOCKET_ISLOCKED(so));
	if (so->so_pcb || (so->so_state & SS_NOFDREF) == 0) {
		/* soclose can't go here */
		return(0);
	}
#ifdef DEBUGDISCARD
	save_state = so->so_state;
#endif

	if (head = so->so_head) {
		register s1;
		SOCKET_MPLOCK(head);
		if (!soqremque(so, 0)) {
			/* synchronize with accept() */
			SOCKET_MPQLOCK(head->so_qlock, s1);
			/*
			 * We used to panic if we could not remove the
			 * connection from the list.
			 * In MP systems, it's possible that we could be
			 * removed between the check of head and the
			 * call to soqremque().  So, just proceed.
			 */
			(void) soqremque(so, 1);
			SOCKET_MPQUNLOCK(head->so_qlock, s1);
		}
		SOCKET_MPUNLOCK(head);
		so->so_head = 0;
	}
	return sorele(so);
}

int
sorele(struct socket *so)
{

	ASSERT(so);
	ASSERT(SOCKET_ISLOCKED(so));
	if (atomicAddInt(&so->so_refcnt, -1)) {
		return 0;
	}

	ASSERT(!(so->so_state & SS_SOFREE));
#if DEBUG
	so->so_state |= SS_SOFREE;
#endif
#ifdef XTP
	if (!(so->so_state & SS_XTP)) {
		sbrelease(&so->so_snd);
		mp_sorflush(so);
		SOCKET_MPUNLOCK(so);
	}
#else
	sbrelease(&so->so_snd);
	mp_sorflush(so);
	SOCKET_MPUNLOCK(so);
#endif

	ASSERT(so->so_holds == 0);
	spinlock_destroy(&so->so_ph.ph_lock);
	SOCKET_FREELOCKS(so);
	mcb_free(so, sizeof *so, MT_SOCKET);
#ifdef DEBUGDISCARD
	if (save_state & SS_CONN_IND_SENT) {
		printf("sofree: socket had outstanding T_CONN_IND, so_state = 0x%x\n",
			save_state);
	}
#endif
	return(1);
}

/*
 * Close a socket on last file table reference removal.
 * Initiate disconnect if connected.
 * Free socket when disconnect complete.
 */
soclose(register struct socket *so)
{
	NETSPL_SPDECL(s)
	register s1;
	int error = 0;

#ifndef _MP_NETLOCKS
	s = splnet();
#endif

restart:
	SOCKET_MPLOCK(so);
	if (so->so_options & SO_ACCEPTCONN) {
		register struct socket *soq;

		so->so_state |= SS_CLOSING;
		while (so->so_q0len + so->so_qlen > 0) {
			if ((((soq = so->so_q0) != so) && soq != NULL) ||
			    (((soq = so->so_q ) != so) && soq != NULL)) {
#ifdef _MP_NETLOCKS
				if (!SOCKET_CMPLOCK(soq)) {
					SOCKET_MPUNLOCK(so);
					qswtch(RESCHED_Y);
					goto restart;
				}
#endif
				if (soq == so->so_q0) {
					(void)soqremque(soq, 0);
				} else {
					SOCKET_MPQLOCK(so->so_qlock, s1);
					(void)soqremque(soq, 1);
					SOCKET_MPQUNLOCK(so->so_qlock, s1);
				}
				SOCKET_MPUNLOCK(so);
				error = soabort(soq);
				if (error) {
					ASSERT(error == EINVAL);
					SOCKET_MPUNLOCK(soq);
				}
				SOCKET_MPLOCK(so);
			} else {
				panic("soclose: q sum len > 0");
			}
		}
        }

	ASSERT(SOCKET_ISLOCKED(so));
	if (so->so_pcb == 0)
		goto discard;
	if (so->so_state & SS_ISCONNECTED) {
		if ((so->so_state & SS_ISDISCONNECTING) == 0) {
			error = sodisconnect(so);
			ASSERT(SOCKET_ISLOCKED(so));
			if (error)
				goto drop;
		}
		if (so->so_options & SO_LINGER) {
			if ((so->so_state & SS_ISDISCONNECTING) &&
			    (so->so_state & SS_NBIO))
				goto drop;
			while (so->so_state & SS_ISCONNECTED) {
				int rv;

				SOCKET_MPUNLOCK_SOTIMEO(so, PZERO+1, rv); 
				SOCKET_MPLOCK(so);
				if (rv != 0) {
					error = EINTR;
					break;
				}
			}
		}
	}
drop:
	ASSERT(SOCKET_ISLOCKED(so));
	if (so->so_pcb) {
		int error2 =
		    (*so->so_proto->pr_usrreq)(so, PRU_DETACH,
			(struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0);
		if (error == 0)
			error = error2;
	}
discard:
	if (so->so_state & SS_NOFDREF)
		panic("soclose: NOFDREF");
	so->so_state |= SS_NOFDREF;
	if (!sofree(so))
		SOCKET_MPUNLOCK(so);
#ifndef _MP_NETLOCKS
	splx(s);
#endif
	return (error);
}

/*
 * Must be called at splnet...
 */
soabort(struct socket *so)
{
	int error;

	ASSERT(SOCKET_ISLOCKED(so));
	error =
	    (*so->so_proto->pr_usrreq)(so, PRU_ABORT,
		(struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0);
	return (error);
}

soaccept(register struct socket *so,
	 struct mbuf *nam)
{
	int error;

	ASSERT(SOCKET_ISLOCKED(so));
	error = (*so->so_proto->pr_usrreq)(so, PRU_ACCEPT,
	    (struct mbuf *)0, nam, (struct mbuf *)0);
	return (error);
}

soconnect(register struct socket *so,
	  struct mbuf *nam)
{
	int error;

	ASSERT(SOCKET_ISLOCKED(so));
	if (so->so_options & SO_ACCEPTCONN)
		return (EOPNOTSUPP);
	/*
	 * If protocol is connection-based, can only connect once.
	 * Otherwise, if connected, try to disconnect first.
	 * This allows user to disconnect by connecting to, e.g.,
	 * a null address.
	 */
	if (so->so_state & (SS_ISCONNECTED|SS_ISCONNECTING) &&
	    ((so->so_proto->pr_flags & PR_CONNREQUIRED) ||
	    (error = sodisconnect(so))))
		error = EISCONN;
	else
		error = (*so->so_proto->pr_usrreq)(so, PRU_CONNECT,
		    (struct mbuf *)0, nam, (struct mbuf *)0);
	return (error);
}

soconnect2(register struct socket *so1,
	   register struct socket *so2)
{
	NETSPL_SPDECL(s)
	int error;

	SOCKET_PAIR_LOCK(so1, so2, s);
	error = (*so1->so_proto->pr_usrreq)(so1, PRU_CONNECT2,
	    (struct mbuf *)0, (struct mbuf *)so2, (struct mbuf *)0);
	SOCKET_PAIR_UNLOCK(so1, so2, s);
	return (error);
}

sodisconnect(register struct socket *so)
{
	int error;

	ASSERT(SOCKET_ISLOCKED(so));
	if ((so->so_state & SS_ISCONNECTED) == 0) {
		error = ENOTCONN;
		goto bad;
	}
	if (so->so_state & SS_ISDISCONNECTING) {
		error = EALREADY;
		goto bad;
	}
	error = (*so->so_proto->pr_usrreq)(so, PRU_DISCONNECT,
	    (struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0);
bad:
	return (error);
}

static int xmitbreak = 1024;

/*
 * Send on a socket.
 * If send must go all at once and message is larger than
 * send buffering, then hard error.
 * Lock against other senders.
 * If must go all at once and not enough room now, then
 * inform user that this would block and do nothing.
 * Otherwise, if nonblocking, send as much as possible.
 *
 * When attaching the user's page, we may do lazy syncing of the tlbs
 * to avoid inter-CPU overhead. Since other CPUs may have pre-attach
 * information in their TLBs, we cannot allow the socket buffer to
 * drain until the TLBs are synced. This we do before we release
 * the lock on the socket buffer.
 */
sosend( register struct socket *so,
	struct mbuf *nam,
	register struct uio *uio,
	int flags,
	struct mbuf *rights)
{
	struct mbuf *top = 0;
	register struct mbuf *m, **mp;
	register int space;
	int len, rlen = 0, error = 0, dontroute, first = 1;
	NETSPL_DECL(s)
	int mwait, mflipped = 0; /* if mflipped, tlbs need to be synced */
	static int sendnombufs = 0;

	NETPAR(NETSCHED,NETEVENTKN, NETPID_NULL,
	       NETEVENT_SODOWN, NETCNT_NULL, NETRES_SYSCALL);
	if (sosendallatonce(so) && uio->uio_resid > so->so_snd.sb_hiwat) {
		NETPAR(NETFLOW, NETDROPTKN,  NETPID_NULL,
			 NETEVENT_SODOWN, uio->uio_resid, NETRES_SBFULL);
		return (EMSGSIZE);
	}

#ifdef XTP
	if (so->so_state & SS_XTP) {
		SOCKET_LOCK(so, s);	/* XXX xtp_sosend needs lock? */
		error = xtp_sosend(so, nam, uio, flags, rights);
		SOCKET_UNLOCK(so, s);
		return (error);
	}
#endif

	mwait = (so->so_state & SS_NBIO) ? M_DONTWAIT : M_WAIT;
	dontroute =
	    (flags & MSG_DONTROUTE) && (so->so_options & SO_DONTROUTE) == 0 &&
	    (so->so_proto->pr_flags & PR_ATOMIC);
	if (!(so->so_state & SS_WANT_CALL))
	        u.u_ru.ru_msgsnd++;	/* if WANT_CALL,may be no u. */
	if (rights)
		rlen = rights->m_len;
#define	snderr(errno)	{ error = errno; goto release; }


	if ((error = sblock(&so->so_snd, NETEVENT_SODOWN, so, &s, 1)) != 0)
		return (error);
	do {
		ASSERT(SOCKET_ISLOCKED(so));
		if (so->so_state & SS_CANTSENDMORE)
			snderr(EPIPE);
		if (so->so_error) {
			error = so->so_error;
			so->so_error = 0;			/* ??? */
			goto release;
		}
		if ((so->so_state & SS_ISCONNECTED) == 0) {
			if (so->so_proto->pr_flags & PR_CONNREQUIRED)
				snderr(ENOTCONN);
			if (nam == 0 &&
			    (so->so_proto->pr_flags & PR_DESTADDROPT) == 0)
				snderr(EDESTADDRREQ);
		}
		if (flags & MSG_OOB)
			space = 1024;
		else {
			space = sbspace(&so->so_snd);
			if (space <= rlen ||
			   (sosendallatonce(so) &&
				space < uio->uio_resid + rlen) ||
			   (uio->uio_resid >= xmitbreak && space < xmitbreak &&
			   so->so_snd.sb_cc >= xmitbreak &&
			   (so->so_state & SS_NBIO) == 0)) {
				if (so->so_state & SS_NBIO) {
					if (first)
						error = EWOULDBLOCK;
					goto release;
				}
				/*
				 * if we've pattach'd, we can't release the
				 * lock and allow things to drain without first
				 * making sure that all the tlb's have the pages
				 * marked for copy on write. do the tlb syncing.
				 */
				if (mflipped) {
					mflipped = 0;
					m_shget_finished();
				}
				NETPAR(NETSCHED, NETSLEEPTKN, (char)&so->so_snd,
					NETEVENT_SODOWN,
                                       	NETCNT_NULL, NETRES_SBFULL);
				/*
				 * This is a real mess.  In 6.2-MR we used
				 * to pass 'first' in here, so that n-1
				 * times through the loop we would sleep
				 * uninterruptably.  I think this was so
				 * we would not return EINTR if we had already
				 * written some data; an EINTR return after
				 * a partial write would confuse the
				 * application.  However, this would leave
				 * processes hanging, unable to be killed.
				 * A check of 4.4BSD-Lite and the X/Open 4.2
				 * spec leads me to believe that returning a
				 * short count on a signal is acceptable.
				 * So, that's what we now do.  I could fix
				 * sendit() to handle EINTR as a special
				 * error, which is what 4.4 does, but other 
				 * callers of sosend() (e.g. BDS) might get
				 * confused, so we fix up the error here.
				 * uiomove() has already adjusted uio_resid,
				 * so we don't have to do it.
				 */
				error = sbunlock_wait(&so->so_snd, so, s, 1);
				NETPAR(NETSCHED, NETWAKEUPTKN,
				    (char)&so->so_snd, NETEVENT_SODOWN,
				    NETCNT_NULL, NETRES_SBFULL);
				if (error) {
					/* Don't have lock here; safe to exit.*/
					if (error != EINTR) {
						return error;
					} else {
						if (first) {
							return EINTR;
						}
						/* sendit() will adjust count.*/
						return 0;
					}
				}
				error = sblock(&so->so_snd,
				       NETEVENT_SODOWN, so, &s, 1);
				if (error) {
					/* Didn't get lock here; safe to exit.*/
					if (error != EINTR) {
						return error;
					} else {
						if (first) {
							return EINTR;
						}
						/* sendit() will adjust count.*/
						return 0;
					}
				}
				continue;
			}
		}

		/*
		 * The hold count keeps a socket from being closed by a
		 * protocol (on MPs).  Below the socket lock is dropped
		 * while doing uiomove and associated bookkeepping.  This
		 * is done because uiomove could fault and could wait a
		 * very long time (nfs); the protocol timeout routines
		 * become hung as they need to grab socket locks to do
		 * timeout function processing.
		 */
		so->so_holds++;
		SOCKET_MPUNLOCK(so);

		mp = &top;
		space -= rlen;
		while (space > 0) {
			int m_shget_dosync;
			int max_write;
			
			struct iovec *iov;
			iov = uio->uio_iov;

			ASSERT(uio->uio_iovcnt >= 1);
			
			len = iov->iov_len;
			if (len <= 0) {
				uio->uio_iov++;
				if (--uio->uio_iovcnt == 0) {
					ASSERT(uio->uio_resid == 0);
					break;
				}
				continue;
			}
			len = MIN(len,
				  NBPP - ((__psint_t)iov->iov_base & (NBPP-1)));
			if (flags & MSG_OOB)
				len = MIN(len, space);
			/*
			 * If we are called from a service routine of 
			 * a streams driver, then we are not in KUSEG.
			 * But in this case, do not try the page flipping.
			 * m_shget: only sync tlbs if we can flip at least
			 * three pages on this call.
			 */
			if (os_has_new_pflip == 0)
				m_shget_dosync = 1;
			else {
				max_write = (space < uio->uio_resid) ?
					space : uio->uio_resid;
				m_shget_dosync =
					(max_write < NBPP*3 && mflipped == 0);
			}
			if ((len >= SOPFLIPMIN) && IS_KUSEG(iov->iov_base) &&
			    0 != (m = m_shget(mwait, MT_DATA, iov->iov_base,
						len, m_shget_dosync))) {
				/* we can page-flip */
				if (!m_shget_dosync)
					mflipped = 1;	/* tlbs need syncing */
				uio->uio_resid -= len;
				uio->uio_offset += len;
				iov->iov_base = (char *)iov->iov_base + len;
				if (0 == (iov->iov_len -= len)) {
					uio->uio_iov = iov+1;
					if (--uio->uio_iovcnt == 0) {
						ASSERT(uio->uio_resid == 0);
					}
				}
			} else {
				len = MIN(VCL_MAX,
					  MIN(uio->uio_resid, space));
				if (0 == (m = m_vget(mwait,len,MT_DATA))) {
					sendnombufs++;
					if (first)
					    error = EWOULDBLOCK;
					goto holdoff;
				}
				error = uiomove(mtod(m, caddr_t),
						len, UIO_WRITE, uio);
			}
			space -= len;
			*mp = m;
			if (error)
				goto holdoff;
			mp = &m->m_next;
			if (uio->uio_resid <= 0)
				break;
		}

		SOCKET_MPLOCK(so);
		so->so_holds--;

		if (so->so_state & SS_CANTSENDMORE)
			snderr(EPIPE);
		if (so->so_error) {
			error = so->so_error;
			so->so_error = 0;			/* ??? */
			goto release;
		}
		if (dontroute)
			so->so_options |= SO_DONTROUTE;
		NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
			 NETEVENT_SODOWN, len, NETRES_NULL);
		error = (*so->so_proto->pr_usrreq)(so,
		    (flags & MSG_OOB) ? PRU_SENDOOB : PRU_SEND,
		    top, nam, rights);
		if (dontroute)
			so->so_options &= ~SO_DONTROUTE;

		rights = 0;
		rlen = 0;
		top = 0;
		first = 0;
		if (error)
			break;
	} while (uio->uio_resid);

release:
	ASSERT(SOCKET_ISLOCKED(so));

	if (mflipped)
		m_shget_finished(); /* sync tlbs */
	sbunlock(&so->so_snd, NETEVENT_SODOWN, so, s);
	if (top)
		m_freem(top);
	if (error == EPIPE)
		sigtoproc(curprocp, SIGPIPE, SYNC_SIGNAL);
	NETPAR(error ? NETFLOW : 0,
	       NETDROPTKN, NETPID_NULL, NETEVENT_SODOWN,
	       uio->uio_resid, NETRES_ERROR);
	return (error);
holdoff:
	SOCKET_MPLOCK(so);
	so->so_holds--;
	goto release;
}

/*
 * Implement receive operations on a socket.
 * We depend on the way that records are added to the sockbuf
 * by sbappend*.  In particular, each record (mbufs linked through m_next)
 * must begin with an address if the protocol so specifies,
 * followed by an optional mbuf containing access rights if supported
 * by the protocol, and then zero or more mbufs of data.
 * In order to avoid blocking network interrupts for the entire time here,
 * we splx() while doing the actual copy to user space.
 * Although the sockbuf is locked, new data may still be appended,
 * and thus we must maintain consistency of the sockbuf during that time.
 *
 * When pageflipping, we may do lazy syncing of the tlbs to avoid inter-CPU
 * overhead. Since other CPUs may have the old pre-flip page in their TLBs,
 * we cannot free the old pages until the TLBs are synced.
 * M_FLIP_MAX_UNFREED_CNT is the maximum number of pages we allow to be
 * accumulated before syncing the TLBs and freeing them.
 */
#define M_FLIP_MAX_UNFREED_CNT	(MBUFCL_SIZE / sizeof(caddr_t))
#define unfreed_pages (mtod(unfreed_mbuf, caddr_t *))

soreceive(register struct socket *so,
	  struct mbuf **aname,
	  register struct uio *uio,
	  int flags,
	  struct mbuf **rightsp)
{
	register struct mbuf *m;
	register int len, error = 0, offset;
	struct protosw *pr = so->so_proto;
	struct mbuf *nextrecord;
	int moff;
	int orig_resid = uio->uio_resid;
	caddr_t mflip_page;
	int	unfreed_cnt	= 0;
	struct  mbuf *unfreed_mbuf = NULL;
	NETSPL_DECL(s)

	NETPAR(NETSCHED, NETEVENTKN, NETPID_NULL,
		 NETEVENT_SOUP, NETCNT_NULL, NETRES_SYSCALL);
	if (rightsp)
		*rightsp = 0;
	if (aname)
		*aname = 0;
	if (flags & MSG_DONTWAIT)
		flags &= ~MSG_WAITALL;	/* conflicting options */
	if (flags & MSG_OOB) {
		m = m_get(M_WAIT, MT_DATA);
		SOCKET_LOCK(so, s);
		error = (*pr->pr_usrreq)(so, PRU_RCVOOB,
		    m, (struct mbuf *)(NULL+(flags & MSG_PEEK)),
					 (struct mbuf *)0);
		SOCKET_UNLOCK(so, s);
		if (error)
			goto bad;
		do {
			len = uio->uio_resid;
			if (len > m->m_len)
				len = m->m_len;
			error =
			    uiomove(mtod(m, caddr_t), (int)len, UIO_READ, uio);
			m = m_free(m);
		} while (uio->uio_resid && error == 0 && m);
bad:
		if (m)
			m_freem(m);
		NETPAR(error ? NETFLOW : 0,
		       NETDROPTKN, NETPID_NULL, NETEVENT_SOUP,
		       NETCNT_NULL, NETRES_ERROR);
		return (error);
	}
#ifdef XTP
	if (so->so_state & SS_XTP) {
		SOCKET_LOCK(so, s);	/* XXX is this needed? */
		error = xtp_soreceive(so, aname, uio, flags, rightsp);
		SOCKET_UNLOCK(so, s);
		return (error);
	}
#endif

	if ((error = sblock(&so->so_rcv, NETEVENT_SOUP, so, &s, 1)) != 0)
		return (error);
restart:
	ASSERT((signed)so->so_rcv.sb_cc >= 0);
	if (so->so_rcv.sb_cc == 0) {
		if (so->so_error) {
			if (orig_resid == uio->uio_resid) {
				error = so->so_error;
				so->so_error = 0;
			}
			goto release;
		}
		if (so->so_state & SS_CANTRCVMORE)
			goto release;
		if ((so->so_state & SS_ISCONNECTED) == 0 &&
		    (so->so_proto->pr_flags & PR_CONNREQUIRED)) {
			error = ENOTCONN;
			goto release;
		}
		if (uio->uio_resid == 0)
			goto release;
		if ((so->so_state & SS_NBIO) || (flags & MSG_DONTWAIT)) {
			error = EWOULDBLOCK;
			goto release;
		}
		if (unfreed_cnt >= 16) { /* ethan: hack arbitrary value */
			m_flip_finished(unfreed_pages, unfreed_cnt);
			unfreed_cnt = 0;
		}
		NETPAR(NETSCHED, NETSLEEPTKN, (char)&so->so_rcv,
			 NETEVENT_SOUP, NETCNT_NULL, NETRES_SBEMPTY);
		error = sbunlock_wait(&so->so_rcv, so, s, 1);
		if (error)
			goto out_error;
		NETPAR(NETSCHED, NETWAKEUPTKN, (char)&so->so_rcv,
			 NETEVENT_SOUP, NETCNT_NULL, NETRES_SBEMPTY);
		error = sblock(&so->so_rcv, NETEVENT_SOUP, so, &s, 1);
		if (error)
			goto out_error;
		goto restart;
	}
	/*
	 * For streams interface, this routine MAY be called from a 	
	 * service routine, which means NO u area, OR the wrong one.
	 */
	if (!(so->so_state & SS_WANT_CALL)) 
	        u.u_ru.ru_msgrcv++;	/* may be no u. */
	m = so->so_rcv.sb_mb;
	if (m == 0)
		panic("receive 1");
	GOODMT(m->m_type);
	nextrecord = m->m_act;
	if (pr->pr_flags & PR_ADDR) {
		if (m->m_type != MT_SONAME)
			panic("receive 1a");
		if (flags & MSG_PEEK) {
			if (aname)
				*aname = m_copy(m, 0, m->m_len);
			m = m->m_next;
		} else {
			sbfree(&so->so_rcv, m);
			ASSERT((signed)so->so_rcv.sb_cc >= 0);
			if (aname) {
				*aname = m;
				m = m->m_next;
				(*aname)->m_next = 0;
				so->so_rcv.sb_mb = m;
			} else {
				MFREE(m, so->so_rcv.sb_mb);
				m = so->so_rcv.sb_mb;
			}
			if (m)
				m->m_act = nextrecord;
		}
	}
	if (m && m->m_type == MT_RIGHTS) {
		if ((pr->pr_flags & PR_RIGHTS) == 0)
			panic("receive 2");
		if (flags & MSG_PEEK) {
			if (rightsp)
				*rightsp = m_copy(m, 0, m->m_len);
			m = m->m_next;
		} else {
			sbfree(&so->so_rcv, m);
			ASSERT((signed)so->so_rcv.sb_cc >= 0);
			if (rightsp) {
				*rightsp = m;
				so->so_rcv.sb_mb = m->m_next;
				m->m_next = 0;
				m = so->so_rcv.sb_mb;
			} else {
				MFREE(m, so->so_rcv.sb_mb);
				m = so->so_rcv.sb_mb;
			}
			if (m)
				m->m_act = nextrecord;
		}
	}
	moff = 0;
	offset = 0;
	while (m && uio->uio_resid > 0 && error == 0) {
		int m_flip_dosync;
		int maxread;
		struct iovec *iov;

		len = uio->uio_resid;
		iov = uio->uio_iov;

		ASSERT(uio->uio_iovcnt >= 1);
#ifndef XTP				/* allow XTP strange types */
		ASSERT(m->m_type == MT_DATA || m->m_type == MT_HEADER);
#endif /* XTP */
		so->so_state &= ~SS_RCVATMARK;
		if (so->so_oobmark && len > so->so_oobmark - offset)
			len = so->so_oobmark - offset;
		/*
		 * the holdcount keeps a socket from being closed by a protocol
		 * while executing the loop below
		 */
		so->so_holds++;
		SOCKET_MPUNLOCK(so);

		if (len > m->m_len - moff)
			len = m->m_len - moff;
		NETPAR(NETFLOW, NETFLOWTKN, NETPID_NULL,
		       NETEVENT_SOUP, (int)len, NETRES_NULL);

		/* m_flip does tlb sync only if we can flip at least
		 * three pages on this call */

		if (os_has_new_pflip == 0)
			m_flip_dosync = 1;
		else {
			maxread = uio->uio_resid;
			if ((flags & MSG_WAITALL) == 0 && so->so_rcv.sb_cc < maxread)
				maxread = so->so_rcv.sb_cc;
			m_flip_dosync = (maxread < NBPP*3 && unfreed_cnt == 0);
		}
		/* page-flip if we can - but onlu if data in KUSEG. */
		if (IS_KUSEG(iov->iov_base)
		    && len == NBPP
		    && iov->iov_len >= NBPP
		    && 0 == ((__psint_t)iov->iov_base & (NBPP-1))
		    && moff == 0
		    && !(flags & MSG_PEEK)
		    && (unfreed_mbuf || (unfreed_mbuf = m_vget(M_WAIT, MBUFCL_SIZE, MT_DATA)))
		    && (mflip_page = m_flip(m, iov->iov_base, m_flip_dosync)))
		{
			if (!m_flip_dosync) {
				unfreed_pages[unfreed_cnt++] = mflip_page;
				if (unfreed_cnt >= M_FLIP_MAX_UNFREED_CNT) {
					m_flip_finished(unfreed_pages, unfreed_cnt);
					unfreed_cnt = 0;
				}
			}
			uio->uio_resid -= len;
			uio->uio_offset += len;
			iov->iov_base = (char *)iov->iov_base + len;
			if (0 == (iov->iov_len -= len)) {
				uio->uio_iov = iov+1;
				if (--uio->uio_iovcnt == 0) {
					ASSERT(uio->uio_resid == 0);
				}
			}
		} else {
#ifdef DEBUG
		    if (len == 0 ) {
			printf("Zero length mbuf in soreceive()\n");
		    } else
#endif
			error = uiomove(mtod(m, caddr_t) + moff,
					len, UIO_READ, uio);
		}
		ASSERT(len <= m->m_len - moff); /* m_len must not shrink */

		SOCKET_MPLOCK(so);
		so->so_holds--;

		if (len == m->m_len - moff) {
			if (flags & MSG_PEEK) {
				m = m->m_next;
				moff = 0;
			} else {
				nextrecord = m->m_act;
				sbfree(&so->so_rcv, m);
				ASSERT((signed)so->so_rcv.sb_cc >= 0);
				if ((m->m_len == NBPP) && (m->m_off == MMINOFF)) /* m_flipped */
					m->m_len = 0;
				MFREE(m, so->so_rcv.sb_mb);
				m = so->so_rcv.sb_mb;
				if (m)
					m->m_act = nextrecord;
			}
		} else {
			if (flags & MSG_PEEK)
				moff += len;
			else {
				m->m_off += len;
				m->m_len -= len;
				so->so_rcv.sb_cc -= len;
				ASSERT((signed)so->so_rcv.sb_cc >= 0);
			}
		}
		if (so->so_oobmark) {
			if ((flags & MSG_PEEK) == 0) {
				so->so_oobmark -= len;
				if (so->so_oobmark == 0) {
					so->so_state |= SS_RCVATMARK;
					break;
				}
			} else
				offset += len;
		}
	}
	if ((flags & MSG_PEEK) == 0) {
		if (m == 0)
			so->so_rcv.sb_mb = nextrecord;
		else if (pr->pr_flags & PR_ATOMIC)
			(void) sbdroprecord(&so->so_rcv);
		if (pr->pr_flags & PR_WANTRCVD && so->so_pcb)
			(*pr->pr_usrreq)(so, PRU_RCVD, (struct mbuf *)0,
			    (struct mbuf *)0, (struct mbuf *)0);
		if (error == 0 && rightsp && *rightsp &&
		    pr->pr_domain->dom_externalize) {
			error = (*pr->pr_domain->dom_externalize)(*rightsp);
			orig_resid = uio->uio_resid; /* keep this error */
		}
	}

	if ((flags & MSG_WAITALL) && uio->uio_resid > 0 && error == 0)
		goto restart;
release:
	ASSERT(SOCKET_ISLOCKED(so));
	sbunlock(&so->so_rcv, NETEVENT_SOUP, so, s);
	NETPAR(error ? NETFLOW : 0,
	       NETDROPTKN, NETPID_NULL,
	       NETEVENT_SOUP, NETCNT_NULL, NETRES_ERROR);
out_error:
	if (unfreed_mbuf) {
		if (unfreed_cnt)
			m_flip_finished(unfreed_pages, unfreed_cnt);
		m_free(unfreed_mbuf);
	}

	/* if data has been moved, return it, give the error next time */

	if (orig_resid > uio->uio_resid)
		return 0;
	else
		return (error);
}
#undef unfreed_pages

soshutdown(register struct socket *so,
	   register int how)
{
	register struct protosw *pr = so->so_proto;
	NETSPL_SPDECL(s)

	how++;
	if (how & FREAD)
#ifdef XTP
		if (!(so->so_state & SS_XTP))
			sorflush(so);
#else
		sorflush(so);
#endif
	if (how & FWRITE) {
		int rv;

		SOCKET_LOCK(so, s);
		rv =  ((*pr->pr_usrreq)(so, PRU_SHUTDOWN,
		    (struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0));
		SOCKET_UNLOCK(so, s);
		return(rv);
	}
	return (0);
}

static void
sorflush(struct socket *so)
{
	register struct sockbuf *sb = &so->so_rcv;
	register struct protosw *pr = so->so_proto;
	int s;
	struct sockbuf asb;

	sblock(sb, NETEVENT_SOUP, so, &s, 0);
	socantrcvmore(so);
	sbunlock(sb, NETEVENT_SOUP, so, s);
	asb = *sb;
	SOCKBUF_CLEAR(sb);
	if (pr->pr_flags & PR_RIGHTS && pr->pr_domain->dom_dispose)
		(*pr->pr_domain->dom_dispose)(asb.sb_mb);
	sbrelease(&asb);
}

#ifdef _MP_NETLOCKS
static void
mp_sorflush(struct socket *so)
{
	register struct sockbuf *sb = &so->so_rcv;
        register struct protosw *pr = so->so_proto;
        struct sockbuf asb;

	ASSERT(SOCKET_ISLOCKED(so));
        socantrcvmore(so);
        asb = *sb;
        SOCKBUF_CLEAR(sb);
        if (pr->pr_flags & PR_RIGHTS && pr->pr_domain->dom_dispose)
                (*pr->pr_domain->dom_dispose)(asb.sb_mb);
        sbrelease(&asb);
}
#endif /* _MP_NETLOCKS */


sosetopt(register struct socket *so,
	 int level,
	 int optname,
	 struct mbuf *m0)
{
	int error = 0;
	register struct mbuf *m = m0;
	NETSPL_SPDECL(s)
	

	SOCKET_LOCK(so, s);
	if (level != SOL_SOCKET) {
		if (so->so_proto && so->so_proto->pr_ctloutput) {
			int rv;

			rv = ((*so->so_proto->pr_ctloutput)
				  (PRCO_SETOPT, so, level, optname, &m0));
			SOCKET_UNLOCK(so, s);
			return(rv);
		}
		error = ENOPROTOOPT;
	} else {
		switch (optname) {

		case SO_LINGER:
			if (m == NULL || m->m_len != sizeof (struct linger)) {
				error = EINVAL;
				goto bad;
			}
			so->so_linger = mtod(m, struct linger *)->l_linger;
			/* fall thru... */

		case SO_DEBUG:
		case SO_KEEPALIVE:
		case SO_DONTROUTE:
		case SO_USELOOPBACK:
		case SO_BROADCAST:
		case SO_REUSEADDR:
		case SO_OOBINLINE:
		case SO_REUSEPORT:
		case SO_PASSIFNAME:
			if (m == NULL || m->m_len < sizeof (int)) {
				error = EINVAL;
				goto bad;
			}
			if (*mtod(m, int *))
				so->so_options |= optname;
			else
				so->so_options &= ~optname;
			break;

		case SO_SNDBUF:
		case SO_RCVBUF:
		case SO_SNDLOWAT:
		case SO_RCVLOWAT:
		case SO_SNDTIMEO:
		case SO_RCVTIMEO:
			if (m == NULL || m->m_len < sizeof (int)) {
				error = EINVAL;
				goto bad;
			}
			switch (optname) {

			case SO_SNDBUF:
			case SO_RCVBUF:
				if (sbreserve(optname == SO_SNDBUF ?
				    &so->so_snd : &so->so_rcv,
				    (__psunsigned_t) *mtod(m, int *)) == 0) {
					error = ENOBUFS;
					goto bad;
				}
				break;

			case SO_SNDLOWAT:
				so->so_snd.sb_lowat = *mtod(m, int *);
				break;
			case SO_RCVLOWAT:
				so->so_rcv.sb_lowat = *mtod(m, int *);
				break;
			case SO_SNDTIMEO:
				so->so_snd.sb_timeo = *mtod(m, int *);
				break;
			case SO_RCVTIMEO:
				so->so_rcv.sb_timeo = *mtod(m, int *);
				break;
			}
			break;

		default:
			error = ENOPROTOOPT;
			break;
		}
	}
bad:
	SOCKET_UNLOCK(so, s);
	if (m)
		(void) m_free(m);
	return (error);
}

sogetopt(register struct socket *so,
	 int level,
	 int optname,
	 struct mbuf **mp)
{
	register struct mbuf *m;
	NETSPL_SPDECL(s)
	int error = 0;

	SOCKET_LOCK(so, s);
	if (level != SOL_SOCKET) {
		if (so->so_proto && so->so_proto->pr_ctloutput) {
			error = ((*so->so_proto->pr_ctloutput)
				  (PRCO_GETOPT, so, level, optname, mp));
		} else 
			error = ENOPROTOOPT;
	} else {
	    if (!(so->so_state & SS_TPISOCKET) || (m = *mp) == NULL)
		m = m_get(M_WAIT, MT_SOOPTS);
		m->m_len = sizeof (int);

		switch (optname) {

		case SO_LINGER:
			m->m_len = sizeof (struct linger);
			mtod(m, struct linger *)->l_onoff =
				so->so_options & SO_LINGER;
			mtod(m, struct linger *)->l_linger = so->so_linger;
			break;

		case SO_USELOOPBACK:
		case SO_DONTROUTE:
		case SO_DEBUG:
		case SO_KEEPALIVE:
		case SO_REUSEADDR:
		case SO_BROADCAST:
		case SO_OOBINLINE:
		case SO_REUSEPORT:
		case SO_PASSIFNAME:
			*mtod(m, int *) = so->so_options & optname;
			break;

		case SO_TYPE:
			*mtod(m, int *) = so->so_type;
			break;

		case SO_ERROR:
			*mtod(m, int *) = so->so_error;
			so->so_error = 0;
			break;

		case SO_SNDBUF:
			*mtod(m, int *) = so->so_snd.sb_hiwat;
			break;

		case SO_RCVBUF:
			*mtod(m, int *) = so->so_rcv.sb_hiwat;
			break;

		case SO_SNDLOWAT:
			*mtod(m, int *) = so->so_snd.sb_lowat;
			break;

		case SO_RCVLOWAT:
			*mtod(m, int *) = so->so_rcv.sb_lowat;
			break;

		case SO_SNDTIMEO:
			*mtod(m, int *) = so->so_snd.sb_timeo;
			break;

		case SO_RCVTIMEO:
			*mtod(m, int *) = so->so_rcv.sb_timeo;
			break;

		default:
		    if (!(so->so_state & SS_TPISOCKET) || *mp == NULL)
			(void)m_free(m);
			error = ENOPROTOOPT;
			goto out;
		}
	    if (!(so->so_state & SS_TPISOCKET) || *mp == NULL)
		*mp = m;
	}
out:
	SOCKET_UNLOCK(so, s);
	return (error);
}

void
sohasoutofband(register struct socket *so)
{
	struct proc *p;

	ASSERT(SOCKET_ISLOCKED(so));
	if (so->so_state & SS_WANT_CALL) 
		streams_interrupt((void(*)())so->so_callout, so,
				  (void *)(NULL+SSCO_OOB_INPUT_ARRIVED),
				  so->so_callout_arg);
	else if (so->so_pgrp < 0)
		signal(-so->so_pgrp, SIGURG);
	else if (so->so_pgrp > 0 && (p = procfind(so->so_pgrp, ZNO)) != 0) {
		procunlock(p);
		psignal(p, SIGURG);
	} 
	pollwakeup(&so->so_ph, POLLIN|POLLPRI|POLLRDNORM|POLLRDBAND);
}
