/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)uipc_socket2.c	7.5 (Berkeley) 6/29/88
 *	plus portions of 7.15 (Berkeley) 6/28/90
 */

#ident	"$Revision: 4.57 $"

#include "tcp-param.h"
#include "sys/types.h"
#include "sys/param.h"
#include "sys/systm.h"
#include "bstring.h"
#include "sys/debug.h"
#include "sys/errno.h"
#include "sys/mbuf.h"
#include "sys/pda.h"
#include "sys/proc.h"
#include "sys/protosw.h"
#include "sys/signal.h"
#include "sys/socket.h"
#include "sys/socketvar.h"
#include "sys/strmp.h"
#include "sys/kmem.h"
#include "sys/user.h"
#include "sys/atomic_ops.h"


extern struct pollhead *phalloc(int);
extern void phfree(struct pollhead *);

#define SIGIO_PSIGNAL(p)	psignal(p, SIGIO)
#define SIGIO_SIGNAL(pgrp)	signal(pgrp, SIGIO)

#ifdef DEBUG
extern	void cksplnet(void);
#endif

void soqinsque(struct socket *, struct socket *, int);
void sbcompress(struct sockbuf *, struct mbuf *, struct mbuf *);
void sbdrop(struct sockbuf *, int);

/*
 * Primitive routines for operating on sockets and socket buffers
 */

/*
 * Procedures to manipulate state flags of socket
 * and do appropriate wakeups.  Normal sequence from the
 * active (originating) side is that soisconnecting() is
 * called during processing of connect() call,
 * resulting in an eventual call to soisconnected() if/when the
 * connection is established.  When the connection is torn down
 * soisdisconnecting() is called during processing of disconnect() call,
 * and soisdisconnected() is called when the connection to the peer
 * is totally severed.  The semantics of these routines are such that
 * connectionless protocols can call soisconnected() and soisdisconnected()
 * only, bypassing the in-progress calls when setting up a ``connection''
 * takes no time.
 *
 * From the passive side, a socket is created with
 * two queues of sockets: so_q0 for connections in progress
 * and so_q for connections already made and awaiting user acceptance.
 * As a protocol is preparing incoming connections, it creates a socket
 * structure queued on so_q0 by calling sonewconn().  When the connection
 * is established, soisconnected() is called, and transfers the
 * socket structure to so_q, making it available to accept().
 * 
 * If a socket is closed with sockets on either
 * so_q0 or so_q, these sockets are dropped.
 *
 * If higher level protocols are implemented in
 * the kernel, the wakeups done here will sometimes
 * cause software-interrupt process scheduling.
 */

void
soisconnecting(register struct socket *so)
{
	ASSERT(SOCKET_ISLOCKED(so));
	so->so_state &= ~(SS_ISCONNECTED|SS_ISDISCONNECTING);
	so->so_state |= SS_ISCONNECTING;
	SOCKET_SOTIMEOWAKE(so);
}

int so_wakeone = 1;

void
soisconnected(register struct socket *so)
{
	register struct socket *head = so->so_head;

	ASSERT(SOCKET_ISLOCKED(so));
	if (head) {
		/*
		 * Incoming connection: wake up an accepter
		 * Use a special lock here because this is the one contended
		 * for by user programs doing "accepts".
		 */
		SOCKET_MPLOCK(head);
		if (soqremque(so, 0) == 0)
			panic("soisconnected");
		mp_spinlock(&head->so_qlock);
		soqinsque(head, so, 1);
		sorwakeup(head, NETEVENT_SOUP);

		if (so_wakeone) {
			SOCKET_SOTIMEOWAKEONE(head);
		} else {
			SOCKET_SOTIMEOWAKE(head);
		}
		mp_spinunlock(&head->so_qlock);
		SOCKET_MPUNLOCK(head);
	}
		
	so->so_state &= ~(SS_ISCONNECTING|SS_ISDISCONNECTING);
	so->so_state |= SS_ISCONNECTED;
	SOCKET_SOTIMEOWAKE(so);
	sorwakeup(so, NETEVENT_SOUP);
	sowwakeup(so, NETEVENT_SODOWN);
}

void
soisdisconnecting(register struct socket *so)
{
	ASSERT(SOCKET_ISLOCKED(so));
	so->so_state &= ~SS_ISCONNECTING;
	so->so_state |= (SS_ISDISCONNECTING|SS_CANTRCVMORE|SS_CANTSENDMORE);
	SOCKET_SOTIMEOWAKE(so);
	sowwakeup(so, NETEVENT_SODOWN);
	sorwakeup(so, NETEVENT_SOUP);
}

void
soisdisconnected(register struct socket *so)
{
	ASSERT(SOCKET_ISLOCKED(so));
	so->so_state &= ~(SS_ISCONNECTING|SS_ISCONNECTED|SS_ISDISCONNECTING);
	so->so_state |= (SS_CANTRCVMORE|SS_CANTSENDMORE);
	SOCKET_SOTIMEOWAKE(so);
	sowwakeup(so, NETEVENT_SODOWN);
	sorwakeup(so, NETEVENT_SOUP);
}

/*
 * When an attempt at a new connection is noted on a socket
 * which accepts connections, sonewconn is called.  If the
 * connection is possible (subject to space constraints, etc.)
 * then we allocate a new structure, propoerly linked into the
 * data structure of the original socket, and return this.
 */
struct socket *
sonewconn(register struct socket *head)
{
	register struct socket *so;

	ASSERT(SOCKET_ISLOCKED(head));
	if (head->so_qlen + head->so_q0len > 3 * head->so_qlimit / 2 ||
		head->so_state & SS_CLOSING)
		goto bad;
	so = (struct socket *)mcb_get(M_DONTWAIT, sizeof *so, MT_SOCKET);
	if (so == NULL)
		goto bad;
	so->so_type = head->so_type;
	so->so_options = head->so_options &~ SO_ACCEPTCONN;
	so->so_linger = head->so_linger;
	so->so_state = (head->so_state | SS_NOFDREF) &
			~(SS_WANT_CALL | SS_TPISOCKET | SS_CONN_IND_SENT);
	so->so_proto = head->so_proto;
	so->so_timeo = head->so_timeo;
	so->so_pgrp = head->so_pgrp;
	initpollhead(&so->so_ph);
	so->so_refcnt = 1;
	SOCKET_INITLOCKS(so);
	SOCKET_MPLOCK(so);	
	(void) soreserve(so, head->so_snd.sb_hiwat, head->so_rcv.sb_hiwat);
	soqinsque(head, so, 0);
	if (0 == (*so->so_proto->pr_usrreq)(so, PRU_ATTACH,
	    (struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0)) {
		return (so);	/* success */
	}

	/* failure, unwind all the allocations */
	(void) soqremque(so, 0);
	SOCKET_MPUNLOCK(so);
	SOCKET_FREELOCKS(so);
	spinlock_destroy(&so->so_ph.ph_lock);
	mcb_free(so, sizeof *so, MT_SOCKET);
bad:
	return ((struct socket *)0);
}

void
soqinsque(register struct socket *head,
	  register struct socket *so,
	  int q)
{
	register struct socket *prev, *next;

	ASSERT(SOCKET_ISLOCKED(so));
	so->so_head = prev = head;

	if (q == NULL) {
		ASSERT(SOCKET_ISLOCKED(head));
		while ((next = prev->so_q0) != head ) {
			prev = next;
		}
		so->so_q0 = next;
		prev->so_q0 = so;
		head->so_q0len++;
	} else {
#ifdef _MP_NETLOCKS
		ASSERT(spinlock_islocked(&head->so_qlock));
#endif
		while ((next = prev->so_q) != head ) {
			prev = next;
		}
		so->so_q = next;
		prev->so_q = so;
		head->so_qlen++;
	}
}

soqremque(register struct socket *so, int q)
{
	register struct socket *head, *prev, *next;

	head = so->so_head;
	ASSERT(SOCKET_ISLOCKED(so));
	/*
	 * XXX Hack to fix a bug reported in comp.bugs.4sd by John LoVerso 
	 * (10/9/87).  If mget() fails in tcp_newtcpcb(), this routine is 
	 * eventually called twice.  so->so_head is null on the 2nd call, 
	 * hence the following to prevent a null ptr. dereference.
	 */
    	if (head == 0) {
		return (0);
	}
#ifdef DEBUG
	if (q == 0) {
		ASSERT(SOCKET_ISLOCKED(head));
	} else if (q == 1) {
#ifdef _MP_NETLOCKS
		ASSERT(spinlock_islocked(&head->so_qlock));
#endif
	}
#endif
	prev = head;
	for (;;) {
		next = q ? prev->so_q : prev->so_q0;
		if (next == so)
			break;
		if (next == head)
			return (0);
		prev = next;
	}
	if (q == 0) {
		prev->so_q0 = next->so_q0;
		head->so_q0len--;
	} else {
		prev->so_q = next->so_q;
		head->so_qlen--;
	}
	next->so_q0 = next->so_q = 0;
	next->so_head = 0;
	return (1);
}

/*
 * Socantsendmore indicates that no more data will be sent on the
 * socket; it would normally be applied to a socket when the user
 * informs the system that no more data is to be sent, by the protocol
 * code (in case PRU_SHUTDOWN).  Socantrcvmore indicates that no more data
 * will be received, and will normally be applied to the socket by a
 * protocol when it detects that the peer will send no more data.
 * Data queued for reading in the socket may yet be read.
 */

void
socantsendmore(struct socket *so)
{
	ASSERT(SOCKET_ISLOCKED(so));
	so->so_state |= SS_CANTSENDMORE;
	sowwakeup(so, NETEVENT_SODOWN);
}

void
socantrcvmore(struct socket *so)
{

	ASSERT(SOCKET_ISLOCKED(so));
	so->so_state |= SS_CANTRCVMORE;
	sorwakeup(so, NETEVENT_SOUP);
}


/*
 * Wait for data to arrive at/drain from a socket buffer.
 * XXX implement same interface? 
 */
/* ARGSUSED */
sbwait(struct sockbuf *sb, struct socket *so)
{
	ASSERT(SOCKET_ISLOCKED(so));
        sb->sb_flags |= SB_WAIT;
	if (SOCKET_MPUNLOCK_SBWAIT(so, sb, PZERO+1))
		return EINTR;
        return 0;
}

/* ARGSUSED */
sbunlock_wait(struct sockbuf *sb, struct socket *so, int ospl, int intr)
{
	ASSERT(SOCKET_ISLOCKED(so));
	sb->sb_flags |= SB_WAIT;

	/* do sbunlock in line */
	(sb)->sb_flags &= ~SB_LOCK;  
        if ((sb)->sb_flags & SB_WANT) { 
                (sb)->sb_flags &= ~SB_WANT; 
                NETPAR(NETSCHED, NETWAKINGTKN, NETPID_NULL, (event), \
                       NETCNT_NULL, NETRES_SBUNLOCK);   
		SOCKET_SBWANT_WAKE(sb);
        }  
	if (intr) {
		if (SOCKET_MPUNLOCK_SBWAIT(so, sb, PZERO+1) != 0)
			return (EINTR);
	} else
		SOCKET_MPUNLOCK_SBWAIT_NO_INTR(so, sb, PZERO+1);
	return 0;
}

/*
 * Wakeup socket readers and writers.
 * Do asynchronous notification via SIGIO
 * if the socket has the SS_ASYNC flag set.
 */
void
sowakeup(register struct socket *so,
	 struct sockbuf *sb,
	 int event)
{
	register struct proc *p;

	ASSERT(SOCKET_ISLOCKED(so));
	if (so->so_state & SS_WANT_CALL) {
		streams_interrupt((void(*)())so->so_callout, so, 
		                  (void *)((sb == &so->so_rcv)
					? (NULL+SSCO_INPUT_ARRIVED)
		                        : (NULL+SSCO_OUTPUT_AVAILABLE)),
		                  so->so_callout_arg);
		return;
	}
	pollwakeup(&so->so_ph, event);
	if (sb->sb_flags & SB_WAIT) {
		sb->sb_flags &= ~SB_WAIT;
		SOCKET_SBWAIT_WAKEALL(sb);
	}
	if (so->so_state & SS_ASYNC) {
		if (so->so_pgrp < 0) {
			SIGIO_SIGNAL(-so->so_pgrp);
		} else if (so->so_pgrp > 0 && (p = procfind(so->so_pgrp, ZNO)) != 0){
			procunlock(p);
			SIGIO_PSIGNAL(p);
		}
	}
}

/*
 * Socket buffer (struct sockbuf) utility routines.
 *
 * Each socket contains two socket buffers: one for sending data and
 * one for receiving data.  Each buffer contains a queue of mbufs,
 * information about the number of mbufs and amount of data in the
 * queue, and other fields allowing select() statements and notification
 * on data availability to be implemented.
 *
 * Data stored in a socket buffer is maintained as a list of records.
 * Each record is a list of mbufs chained together with the m_next
 * field.  Records are chained together with the m_act field. The upper
 * level routine soreceive() expects the following conventions to be
 * observed when placing information in the receive buffer:
 *
 * 1. If the protocol requires each message be preceded by the sender's
 *    name, then a record containing that name must be present before
 *    any associated data (mbuf's must be of type MT_SONAME).
 * 2. If the protocol supports the exchange of ``access rights'' (really
 *    just additional data associated with the message), and there are
 *    ``rights'' to be received, then a record containing this data
 *    should be present (mbuf's must be of type MT_RIGHTS).
 * 3. If a name or rights record exists, then it must be followed by
 *    a data record, perhaps of zero length.
 *
 * Before using a new socket structure it is first necessary to reserve
 * buffer space to the socket, by calling sbreserve().  This should commit
 * some of the available buffer space in the system buffer pool for the
 * socket (currently, it does nothing but enforce limits).  The space
 * should be released by calling sbrelease() when the socket is destroyed.
 */

soreserve(register struct socket *so,
	  u_int sndcc,
	  u_int rcvcc)
{
	int error = 0;

	ASSERT(SOCKET_ISLOCKED(so));
	if (sbreserve(&so->so_snd, sndcc) == 0) 
		error = ENOBUFS;
	else if (sbreserve(&so->so_rcv, rcvcc) == 0) {
		error = ENOBUFS;
		sbrelease(&so->so_snd);
	}
	return (error);
}

/*
 * Allot mbufs to a sockbuf.
 * Attempt to scale cc so that mbcnt doesn't become limiting
 * if buffering efficiency is near the normal case.
 */
sbreserve(struct sockbuf *sb, u_int cc)
{
	/*
	 * There used to be a check for SB_MAX here, but since
	 * this is not a real "reserve" anyway, there is no need
	 * for any specific limit.
	 */
	sb->sb_hiwat = cc;
	return (1);
}

/*
 * Free mbufs held by a socket, and reserved mbuf space.
 */
void
sbrelease(struct sockbuf *sb)
{
	sbflush(sb);
	sb->sb_hiwat = 0;
}

/*
 * Routines to add and remove
 * data from an mbuf queue.
 *
 * The routines sbappend() or sbappendrecord() are normally called to
 * append new mbufs to a socket buffer, after checking that adequate
 * space is available, comparing the function sbspace() with the amount
 * of data to be added.  sbappendrecord() differs from sbappend() in
 * that data supplied is treated as the beginning of a new record.
 * To place a sender's address, optional access rights, and data in a
 * socket receive buffer, sbappendaddr() should be used.  To place
 * access rights and data in a socket receive buffer, sbappendrights()
 * should be used.  In either case, the new data begins a new record.
 * Note that unlike sbappend() and sbappendrecord(), these routines check
 * for the caller that there will be enough space to store the data.
 * Each fails if there is not enough space, or if it cannot find mbufs
 * to store additional information in.
 *
 * Reliable protocols may use the socket send buffer to hold data
 * awaiting acknowledgement.  Data is normally copied from a socket
 * send buffer in a protocol with m_copy for output to a peer,
 * and then removing the data from the socket buffer with sbdrop()
 * or sbdroprecord() when the data is acknowledged by the peer.
 */

#if defined(sgi) && defined(DEBUG)
void
sbccverify(struct sockbuf *sb)
{
	register struct mbuf *m;
	register int cc;

	m = sb->sb_mb;
	if (m == NULL) {
		ASSERT(sb->sb_cc == 0);
		return;
	}
	cc = m_length(m);
	while (m = m->m_act) {
		cc += m_length(m);
	}
	ASSERT(sb->sb_cc == cc);
}
#endif

/*
 * Append mbuf chain m to the last record in the
 * socket buffer sb.  The additional space associated
 * the mbuf chain is recorded in sb.  Empty mbufs are
 * discarded and mbufs are compacted where possible.
 */
void
sbappend(struct sockbuf *sb,
	 struct mbuf *m)
{
	register struct mbuf *n;

	if (m == 0)
		return;
	GOODMT(m->m_type);
	if (n = sb->sb_mb) {
		while (n->m_act) {
			n = n->m_act;
			GOODMT(n->m_type);
		}
		while (n->m_next) {
			n = n->m_next;
			GOODMT(n->m_type);
		}
	} 
	sbcompress(sb, m, n);
}

/*
 * As above, except the mbuf chain
 * begins a new record.
 */
void
sbappendrecord(register struct sockbuf *sb,
	       register struct mbuf *m0)
{
	register struct mbuf *m;

	if (m0 == 0)
		return;
	GOODMT(m0->m_type);
	if (m = sb->sb_mb) {
		while (m->m_act) {
			m = m->m_act;
			GOODMT(m->m_type);
		}
	}
	/*
	 * Put the first mbuf on the queue.
	 * Note this permits zero length records.
	 */
	sballoc(sb, m0);
	if (m)
		m->m_act = m0;
	else
		sb->sb_mb = m0;
	m = m0->m_next;
	m0->m_next = 0;
	sbcompress(sb, m, m0);
}

/*
 * Append address and data, and optionally, rights
 * to the receive queue of a socket.  Return 0 if
 * no space in sockbuf or insufficient mbufs.
 */
sbappendaddr(register struct sockbuf *sb,
	     struct sockaddr *asa,
	     struct mbuf *m0,
	     struct mbuf *rights0,
	     struct ifnet *ifp
	     )
{
	register struct mbuf *m, *n, *mif;
	int space = sizeof (*asa);

	mif = 0;

	for (m = m0; m; m = m->m_next) {
		GOODMT(m->m_type);
		space += m->m_len;
	}
	if (rights0)
		space += rights0->m_len;
	if (ifp)
		space += IFNAMSIZ;
	if (space > sbspace(sb))
		return (0);
	MGET(m, M_DONTWAIT, MT_SONAME);
	if (m == 0)
		return (0);
	*mtod(m, struct sockaddr *) = *asa;
	m->m_len = sizeof (*asa);

	if (ifp) {
		MGET(mif, M_DONTWAIT, MT_DATA);
		if (mif == 0) {
			m_freem(m);
			return (0);
		}
		mif->m_len = IFNAMSIZ;
		bzero(mtod(mif, char *), IFNAMSIZ);
		sprintf(mtod(mif, char *), "%s%d",
			ifp->if_name, (int)ifp->if_unit);
		mif->m_next = m0;
		m0 = mif;
	}

	if (rights0 && rights0->m_len) {
		m->m_next = m_copy(rights0, 0, rights0->m_len);
		if (m->m_next == 0) {
			m_freem(m);
			if (mif)
				m_free(mif);
			return (0);
		}
		sballoc(sb, m->m_next);
	}
	sballoc(sb, m);

	if (n = sb->sb_mb) {
		GOODMT(n->m_type);
		while (n->m_act) {
			n = n->m_act;
			GOODMT(n->m_type);
		}
		n->m_act = m;
	} else
		sb->sb_mb = m;
	if (m->m_next)
		m = m->m_next;
	if (m0)
		sbcompress(sb, m0, m);
	return (1);
}

sbappendrights(struct sockbuf *sb,
	       struct mbuf *m0,
	       struct mbuf *rights)
{
	register struct mbuf *m, *n;
	int space = 0;

	if (rights == 0)
		panic("sbappendrights");
	for (m = m0; m; m = m->m_next) {
		GOODMT(m->m_type);
		space += m->m_len;
	}
	space += rights->m_len;
	if (space > sbspace(sb))
		return (0);
	m = m_copy(rights, 0, rights->m_len);
	if (m == 0)
		return (0);
	sballoc(sb, m);
	if (n = sb->sb_mb) {
		GOODMT(n->m_type);
		while (n->m_act) {
			n = n->m_act;
			GOODMT(n->m_type);
		}
		n->m_act = m;
	} else
		sb->sb_mb = m;
	if (m0)
		sbcompress(sb, m0, m);
	return (1);
}

/*
 * Compress mbuf chain m into the socket
 * buffer sb following mbuf n.  If n
 * is null, the buffer is presumed empty.
 */
void
sbcompress(register struct sockbuf *sb,
	   register struct mbuf *m,
	   register struct mbuf *n)
{

	while (m) {
		GOODMT(m->m_type);
		if (m->m_len == 0) {
			if (n == 0 && m->m_next != 0)
				m->m_next->m_act = m->m_act;
			m = m_free(m);
			continue;
		}
		if (m->m_len <= MLEN*4 && M_HASCL(m) && n) {
			/*
			 * There is a very sparse cluster about to be put on a
			 * non-empty socket input queue. Copy it to a bunch of 
			 * short mbufs to conserve memory. We do this here
			 * to avoid the time/space trade-off in the case of NFS
			 * and TCP acks, since those packets are never queued
			 * for a significant period of time.
			 */
			struct mbuf *new;
			struct mbuf *first;
			struct mbuf **prev;

			first = (struct mbuf *)0;
			while (m->m_len > 0) {
				new = m_get(M_DONTWAIT, m->m_type);
				if (first == (struct mbuf *)0) {
					first = new;
				} else {
					*prev = new;
				}
				if (new == (struct mbuf *)0)
				    break;
				prev = &new->m_next;
				if (m->m_len > MLEN)
				    new->m_len = MLEN;
				else
				    new->m_len = m->m_len;
				bcopy( mtod(m, caddr_t), mtod(new, caddr_t),
				      new->m_len);
				m->m_len -= new->m_len;
				m->m_off += new->m_len;
			}
			if (new) {
				/*
				 * We got new small mbufs, free the OLD one,
				 * and use the new chain below.
				 */
				new->m_next = m_free(m);
				m = first;
			} else {
				/*
				 * One of the m_gets failed, so in this case
				 * we need to free the NEW ones (if any),
				 * and re-adjust the counts in the old one.
				 */
				while (first) {
					new = first;
					first = new->m_next;
					m->m_len += new->m_len;
					m->m_off -= new->m_len;
					m_free(new);
				}
			}
		}
		if (n && !M_HASCL(n) &&
		    (n->m_off + n->m_len + m->m_len) <= MMAXOFF &&
		    n->m_type == m->m_type) {
			bcopy(mtod(m, caddr_t), mtod(n, caddr_t) + n->m_len,
			    (unsigned)m->m_len);
			n->m_len += m->m_len;
			sb->sb_cc += m->m_len;
			m = m_free(m);
			continue;
		}
		sballoc(sb, m);
		if (n)
			n->m_next = m;
		else
			sb->sb_mb = m;
		n = m;
		m = m->m_next;
		n->m_next = 0;
	}
#ifdef DEBUG
	sbccverify(sb);
#endif
}

/*
 * Free all mbufs in a sockbuf.
 * Check that all resources are reclaimed.
 */
void
sbflush(register struct sockbuf *sb)
{
	ASSERT(!(sb->sb_flags & SB_LOCK));
	while (sb->sb_cc)
		sbdrop(sb, (int)sb->sb_cc);
	ASSERT(0 == sb->sb_cc && 0 == sb->sb_mb);
}

/*
 * Drop data from (the front of) a sockbuf.
 */
void
sbdrop(register struct sockbuf *sb,
       register int len)
{
	register struct mbuf *m, *mn;
	struct mbuf *next;

	ASSERT((signed)sb->sb_cc >= 0);
	ASSERT(len >= 0);
#if defined(sgi) && defined(DEBUG)
	sbccverify(sb);
#endif
	next = (m = sb->sb_mb) ? m->m_act : 0;
	while (len > 0) {
		if (m == 0) {
			if (next == 0)
				panic("sbdrop");
			m = next;
			next = m->m_act;
			continue;
		}
		GOODMT(m->m_type);
		if (m->m_len > len) {
			m->m_len -= len;
			m->m_off += len;
			sb->sb_cc -= len;
			ASSERT((signed)sb->sb_cc >= 0);
			break;
		}
		len -= m->m_len;
		sbfree(sb, m);
		ASSERT((signed)sb->sb_cc >= 0);
		MFREE(m, mn);
		m = mn;
	}
	while (m && m->m_len == 0) {
		sbfree(sb, m);		/* why bother to subtract 0 ? */
		MFREE(m, mn);
		m = mn;
	}
	if (m) {
		GOODMT(m->m_type);
		sb->sb_mb = m;
		m->m_act = next;
	} else
		sb->sb_mb = next;
}

/*
 * Drop a record off the front of a sockbuf
 * and move the next record to the front.
 */
void
sbdroprecord(register struct sockbuf *sb)
{
	register struct mbuf *m, *mn;

	m = sb->sb_mb;
	if (m) {
		sb->sb_mb = m->m_act;
		do {
			sbfree(sb, m);
			MFREE(m, mn);
		} while (m = mn);
	}
}

#ifdef _MP_NETLOCKS
int
socket_pair_cmplock(struct socket *so1, struct socket *so2)
{
	if (so1 == so2)
		return(1);
	ASSERT(SOCKET_ISLOCKED(so1));
	if ((so1) < (so2))  {
		SOCKET_MPLOCK(so2);
		return(1);
	}
	else if (SOCKET_MPTRYLOCK(so2) == 0) {
		SOCKET_MPUNLOCK(so1);
		qswtch(RESCHED_Y);
		SOCKET_MPLOCK(so1);
		return(0);
	}
	return(1);
}
		
void
socket_pair_mplock(struct socket *so1, struct socket *so2)
{
	ASSERT(!SOCKET_ISLOCKED(so1));
	ASSERT(!SOCKET_ISLOCKED(so2));
	if ((so1) == (so2)) {
		SOCKET_MPLOCK(so1);
		return;
	}
	if ((so1) < (so2)) { 
                SOCKET_MPLOCK(so1); 
		ASSERT(&(so1)->so_sem != &(so2)->so_sem);
                SOCKET_MPLOCK(so2);  
        } else { 
                SOCKET_MPLOCK(so2); 
		ASSERT(&(so1)->so_sem != &(so2)->so_sem);
                SOCKET_MPLOCK(so1);  
        } 
}

void
socket_pair_mpunlock(struct socket *so1, struct socket *so2)
{
	if ((so1) == (so2)) {
		SOCKET_MPUNLOCK(so1);
		return;
	}
	if ((so1) < (so2)) { 
		ASSERT(&(so1)->so_sem != &(so2)->so_sem);
                SOCKET_MPUNLOCK(so2); 
                SOCKET_MPUNLOCK(so1); 
        } else { 
		ASSERT(&(so1)->so_sem != &(so2)->so_sem);
		SOCKET_MPUNLOCK(so1); 
                SOCKET_MPUNLOCK(so2); 
        }
}
#endif
