/*
 * System call interface to the socket abstraction.
 *
 * Copyright 1992, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	%W% (Berkeley) %G%
 */
#ident "$Revision: 4.51 $"

#include <limits.h>
#include <tcp-param.h>
#include <sys/types.h>
#include <sys/buf.h>
#include <sys/debug.h>
#include <sys/domain.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/kmem.h>
#include <sys/mbuf.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/protosw.h>
#include <sys/sat.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/user.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/xlate.h>
#include <sys/kuio.h>

extern struct vfs uipc_vfs;
extern struct vnodeops uipc_vnodeops;

extern int soaccept(struct socket *, struct mbuf *);

static int sockargs(struct mbuf **, void *, int, int);
static int sendit(int, struct msghdr *, int, rval_t *);

static int recvit(int, struct msghdr *, int, void *, void *, 
	   rval_t *);

#if _K64U64
int cmsghdr_to_irix5(struct mbuf *);
#endif

struct socketa {
	sysarg_t	domain;
	sysarg_t	type;
	sysarg_t	protocol;
};

int
socket(struct socketa *uap, rval_t *rvp)
{
	int error, fd;
	struct file *fp;
	struct socket *so;

	error = socreate(uap->domain, &so, uap->type, uap->protocol);
	if (error) {
		return error;
	}

	error = makesockvp(&fp, &fd, (void *)so);
	if (error) {
		(void)soclose(so);
		return error;
	}

	fready(fp);
	rvp->r_val1 = fd;

#ifdef sgi /* Audit */
	_SAT_BSDIPC_CREATE(fd, so, uap->domain, uap->type, 0);
#endif
	return 0;
}

struct binda {
	sysarg_t	s;
	caddr_t		name;
	sysarg_t	namelen;
};

/* ARGSUSED */
int
bind(struct binda *uap, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct mbuf *nam;

	if (error = getsock(uap->s, &so))
		return error;
	error = sockargs(&nam, uap->name, uap->namelen, MT_SONAME);
	if (error)
		return error;
	error = sobind(so, nam);

#ifdef sgi /* Audit */
	if (!error)
	    _SAT_BSDIPC_ADDR(uap->s, so, nam, curprocp, 0);
#endif
	m_freem(nam);
	return error;
}

struct listena {
	sysarg_t	s;
	sysarg_t	backlog;
};

/* ARGSUSED */
int
listen(struct listena *uap, rval_t *rvp)
{
	int error;
	struct socket *so;

	if (error = getsock(uap->s, &so))
		return error;
	return solisten(so, uap->backlog);
}

struct accepta {
	sysarg_t	s;
	caddr_t		name;
	sysarg_t	*anamelen;
};

int
accept(struct accepta *uap, rval_t *rvp)
{
	int error, fd;
	struct file *fp;
	struct mbuf *nam;
	int namelen;
	NETSPL_DECL(s)
	struct socket *so, *aso;

	if (uap->name != NULL) {
		error = copyin(uap->anamelen, &namelen, sizeof (namelen));
		if (error)
			return error;
		if (useracc(uap->name, namelen, B_READ) == 0)
			return EFAULT;
		 /*
		  * useracc acquired shdlock (SH_ACC), but we release it now,
		  * therefore the copyout below could fail if a shared process
		  * makes the address unavailable.
		  */
		shdunlock(curprocp);
	}
	if (error = getsock(uap->s, &so))
		return error;

restart:
	if ((so->so_options & SO_ACCEPTCONN) == 0) {
		return EINVAL;
	}
	SOCKET_QLOCK(so->so_qlock, s);	
	if ((so->so_state & SS_NBIO) && so->so_qlen == 0) {
		SOCKET_QUNLOCK(so->so_qlock, s);
		return EWOULDBLOCK;
	}
	while (so->so_qlen == 0 && so->so_error == 0) {
		int rv;

		if (so->so_state & SS_CANTRCVMORE) {
			so->so_error = ECONNABORTED;
			break;
		}
		rv = SOCKET_WAIT_SIG(&so->so_timeosem, PZERO+1, 
			&so->so_qlock, s);
		if (rv) {
			return EINTR;
		}
		SOCKET_MPQLOCK(so->so_qlock, s);
	}
	if (so->so_error) {
		error = so->so_error;
		so->so_error = 0;
		SOCKET_QUNLOCK(so->so_qlock, s);
		return error;
	}

	aso = so->so_q;
#ifdef _MP_NETLOCKS
	if (!SOCKET_CMPLOCK(aso)) {
		SOCKET_QUNLOCK(so->so_qlock, s);
		qswtch(RESCHED_Y);
		goto restart;
	}
#endif /* _MP_NETLOCKS */

	if (soqremque(aso, 1) == 0)
		panic("accept");

	SOCKET_MPQUNLOCK(so->so_qlock, s);

	/* 
	 * prevent protocol from dropping this connection while we get a 
	 * vnode. Notice that the following code is not necessary on MP
	 * since we have aso locked already.
	 */
	aso->so_state &= ~SS_NOFDREF;

	if (error = makesockvp(&fp, &fd, (void *)aso)) {
		aso->so_state |= SS_NOFDREF;
		if (aso->so_pcb == 0)
			sofree(aso);
		else 
			soabort(aso);
		return error;
	}
	fready(fp);

	nam = m_get(M_WAIT, MT_SONAME);
	if (aso->so_pcb == 0 || (aso->so_state & SS_CLOSING)) {
		/* lost the race: undo allocations and try again */
		SOCKET_UNLOCK(aso, s);
		setf(fd, NULLFP);
		closef(fp);
		m_freem(nam);
		goto restart;
	}
#ifdef DEBUG
	error = soaccept(aso, nam);
	ASSERT(!error);
#else
	(void)soaccept(aso, nam);
#endif /* DEBUG */
	SOCKET_UNLOCK(aso, s);
#ifdef sgi /* Audit */
	_SAT_BSDIPC_CREATE(u.u_rval1, aso,
			       aso->so_proto->pr_domain->dom_family,
			       aso->so_proto->pr_type, 0);
	_SAT_BSDIPC_ADDR(uap->s, so, nam, curprocp, 0);
#endif
	if (uap->name) {
		if (namelen > nam->m_len)
			namelen = nam->m_len;
		/* SHOULD COPY OUT A CHAIN HERE */
		(void) copyout(mtod(nam, caddr_t), uap->name, namelen);
		(void) copyout(&namelen, uap->anamelen, sizeof (int));
	}
	rvp->r_val1 = fd;
	m_freem(nam);
	return error;
}

struct connecta {
	sysarg_t	s;
	caddr_t		name;
	sysarg_t	namelen;
};

/* ARGSUSED */
int
connect(struct connecta *uap, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct mbuf *nam;
	NETSPL_SPDECL(s)

	if (error = getsock(uap->s, &so))
		return error;
	SOCKET_LOCK(so, s);
	if ((so->so_state & SS_NBIO) &&
	    (so->so_state & SS_ISCONNECTING)) {
		SOCKET_UNLOCK(so, s);
		return EALREADY;
	}
	error = sockargs(&nam, uap->name, uap->namelen, MT_SONAME);
	if (error) {
		SOCKET_UNLOCK(so, s);
		return error;
	}
	error = soconnect(so, nam);
	if (error)
		goto bad;
	if ((so->so_state & SS_NBIO) &&
	    (so->so_state & SS_ISCONNECTING)) {
		SOCKET_UNLOCK(so, s);
		m_freem(nam);
		return EINPROGRESS;
	}
	while ((so->so_state & SS_ISCONNECTING) && so->so_error == 0) {
		int rv;

		SOCKET_MPUNLOCK_SOTIMEO(so, (PZERO+1)|PCATCH, rv);
		SOCKET_LOCK(so, s);
		if (rv) {
			error = EINTR;
			goto bad;
		}
	}
	error = so->so_error;
	so->so_error = 0;

bad:
	so->so_state &= ~SS_ISCONNECTING;
	SOCKET_UNLOCK(so, s);
#ifdef sgi /* Audit */
	if (! so->so_error)
		_SAT_BSDIPC_ADDR(uap->s, so, nam, curprocp, 0);
#endif
	m_freem(nam);
	return error;
}

struct sendtoa {
	sysarg_t	s;
	caddr_t		buf;
	sysarg_t	len;
	sysarg_t	flags;
	caddr_t		to;
	sysarg_t	tolen;
};

int
sendto(struct sendtoa *uap, rval_t *rvp)
{
	struct msghdr msg;
	struct iovec aiov;

	msg.msg_name = uap->to;
	msg.msg_namelen = uap->tolen;
	msg.msg_iov = &aiov;
	msg.msg_iovlen = 1;
	aiov.iov_base = uap->buf;
	aiov.iov_len = uap->len;
	msg.msg_accrights = 0;
	msg.msg_accrightslen = 0;
	return sendit(uap->s, &msg, uap->flags, rvp);
}

struct senda {
	sysarg_t	s;
	caddr_t		buf;
	sysarg_t	len;
	sysarg_t	flags;
};

int
send(struct senda *uap, rval_t *rvp)
{
	struct msghdr msg;
	struct iovec aiov;

	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = &aiov;
	msg.msg_iovlen = 1;
	aiov.iov_base = uap->buf;
	aiov.iov_len = uap->len;
	msg.msg_accrights = 0;
	msg.msg_accrightslen = 0;
	return sendit(uap->s, &msg, uap->flags, rvp);
}

struct sendmsga {
	sysarg_t	s;
	caddr_t		msg;
	sysarg_t	flags;
};

#if _K64U64
struct irix5_msghdr {
	app32_ptr_t	msg_name;
	app32_int_t	msg_namelen;
	app32_ptr_t	msg_iov;
	app32_int_t	msg_iovlen;
	app32_ptr_t	msg_accrights;
	app32_int_t	msg_accrightslen;
};


/*ARGSUSED*/
static int
irix5_to_msghdr(
	enum xlate_mode mode,
	void *to,
	int count,
	register xlate_info_t *info)
{
	register struct msghdr *mh;
	register struct irix5_msghdr *i5_mh;

	ASSERT(info->smallbuf != NULL);
	ASSERT(mode == SETUP_BUFFER || mode == DO_XLATE);
	ASSERT(count == 1);

	if (mode == SETUP_BUFFER) {
		ASSERT(info->copybuf == NULL);
		ASSERT(info->copysize == 0);
		if (sizeof(struct irix5_msghdr) <= info->inbufsize)
			info->copybuf = info->smallbuf;
		else
			info->copybuf = kern_malloc(
					 sizeof(struct irix5_msghdr));
		info->copysize = sizeof(struct irix5_msghdr);
		return 0;
	}

	ASSERT(info->copysize == sizeof(struct irix5_msghdr));
	ASSERT(info->copybuf != NULL);

	mh = to;

	i5_mh = info->copybuf;

	mh->msg_name = (caddr_t)(__psint_t)i5_mh->msg_name;
	mh->msg_namelen = i5_mh->msg_namelen;
	mh->msg_iov = (struct iovec *)(__psint_t)i5_mh->msg_iov;
	mh->msg_iovlen = i5_mh->msg_iovlen;
	mh->msg_accrights = (caddr_t)(__psint_t)i5_mh->msg_accrights;
	mh->msg_accrightslen = i5_mh->msg_accrightslen;

	return 0;
}

#endif /* _K64U64 */


int
sendmsg(struct sendmsga *uap, rval_t *rvp)
{
	int error;
	struct msghdr msg;
	struct iovec aiov[MSG_MAXIOVLEN];

#ifdef __sgi
	error = COPYIN_XLATE(uap->msg, &msg, sizeof (msg), irix5_to_msghdr,
			     curprocp->p_abi, 1);
#else
	error = copyin(uap->msg, &msg, sizeof (msg));
#endif
	if (error)
		return error;
	if ((u_int)msg.msg_iovlen >= sizeof (aiov) / sizeof (aiov[0]))
		return EMSGSIZE;
#ifdef __sgi
	error =
	    COPYIN_XLATE(msg.msg_iov, aiov, msg.msg_iovlen * sizeof (aiov[0]),
				irix5_to_iovec, curprocp->p_abi,
				msg.msg_iovlen);
#else
	error =
	    copyin(msg.msg_iov, aiov, msg.msg_iovlen * sizeof (aiov[0]));
#endif
	if (error)
		return error;
	msg.msg_iov = aiov;
	return sendit(uap->s, &msg, uap->flags, rvp);
}

static int
sendit(int s, register struct msghdr *mp, int flags, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct uio auio;
	register struct iovec *iov;
	register int i;
	struct mbuf *to, *rights;
	int len;
	
	if (error = getsock(s, &so))
		return error;
	auio.uio_iov = mp->msg_iov;
	auio.uio_iovcnt = mp->msg_iovlen;
	auio.uio_segflg = UIO_USERSPACE;
	auio.uio_offset = 0;			/* XXX */
	auio.uio_resid = 0;
	auio.uio_pio = 0;
	auio.uio_pbuf = 0;
	for (i = 0, iov = mp->msg_iov; i < mp->msg_iovlen; i++, iov++) {
		if (iov->iov_len > SSIZE_MAX)
			return EINVAL;
		auio.uio_resid += iov->iov_len;
	}
	/*
	 * Note: only do useracc() check if we could page-flip, otherwise
	 * avoid it because of the high cost of useracc() for small writes.
	 */
	if (auio.uio_resid >= SOPFLIPMIN) {
	    for (i = 0, iov = mp->msg_iov; i < mp->msg_iovlen; i++, iov++) {
		if (iov->iov_len == 0)
			continue;
		if (useracc(iov->iov_base, iov->iov_len, B_WRITE) == 0)
			return EFAULT;
		/*
		 * useracc acquired shdlock (SH_ACC), but we release it now,
		 * therefore the copyin below could fail if a shared process
		 * makes the address unavailable.
		 */
		shdunlock(curprocp);
	    }
	}
	if (mp->msg_name) {
		error = sockargs(&to, mp->msg_name, mp->msg_namelen, MT_SONAME);
		if (error)
			return error;
	} else
		to = 0;
	if (mp->msg_accrights) {
		error =
		    sockargs(&rights, mp->msg_accrights, mp->msg_accrightslen,
		    MT_RIGHTS);
		if (error)
			goto bad;
	} else
		rights = 0;
	len = auio.uio_resid;
	error = sosend(so, to, &auio, flags, rights);
	if (auio.uio_resid != len && error == EINTR)
		error = 0;
	rvp->r_val1 = len - auio.uio_resid;
	if (rights)
		m_freem(rights);
bad:
	if (to)
		m_freem(to);
	return error;
}

struct recvfroma {
	sysarg_t	s;
	caddr_t		buf;
	sysarg_t	len;
	sysarg_t	flags;
	caddr_t		from;
	sysarg_t	*fromlenaddr;
};

int
recvfrom(struct recvfroma *uap, rval_t *rvp)
{
	int error;
	struct msghdr msg;
	struct iovec aiov;
	int len;

	if (uap->fromlenaddr != NULL) {
	    error = copyin(uap->fromlenaddr, &len, sizeof (len));
	    if (error)
		    return error;
	} else
	    len = 0;
	msg.msg_name = uap->from;
	msg.msg_namelen = len;
	msg.msg_iov = &aiov;
	msg.msg_iovlen = 1;
	aiov.iov_base = uap->buf;
	aiov.iov_len = uap->len;
	msg.msg_accrights = 0;
	msg.msg_accrightslen = 0;
	return recvit((int)uap->s, &msg, (int)uap->flags,
		(caddr_t)uap->fromlenaddr,
	       (void *)0, rvp);
}

struct recva {
	sysarg_t	s;
	caddr_t		buf;
	sysarg_t	len;
	sysarg_t	flags;
};

int
recv(struct recva *uap, rval_t *rvp)
{
	struct msghdr msg;
	struct iovec aiov;

	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = &aiov;
	msg.msg_iovlen = 1;
	aiov.iov_base = uap->buf;
	aiov.iov_len = uap->len;
	msg.msg_accrights = 0;
	msg.msg_accrightslen = 0;
	return recvit(uap->s, &msg, uap->flags, 0, 0, rvp);
}

struct recvmsga {
	sysarg_t	s;
	struct msghdr	*msg;
	sysarg_t	flags;
};

struct irix5_recvmsga {
	sysarg_t		s;
	struct irix5_msghdr	*msg;
	sysarg_t		flags;
};

int
recvmsg(struct recvmsga *uap, rval_t *rvp)
{
	int error;
	struct msghdr msg;
	struct irix5_recvmsga *i5_uap;
	struct iovec aiov[MSG_MAXIOVLEN];

#ifdef __sgi
	error = COPYIN_XLATE(uap->msg, &msg, sizeof (msg), irix5_to_msghdr,
		curprocp->p_abi, 1);
#else
	error = copyin(uap->msg, &msg, sizeof (msg));
#endif
	if (error)
		return error;
	if ((u_int)msg.msg_iovlen >= sizeof (aiov) / sizeof (aiov[0]))
		return EMSGSIZE;
#ifdef __sgi
	error =
	    COPYIN_XLATE(msg.msg_iov, aiov, msg.msg_iovlen * sizeof (aiov[0]),
			irix5_to_iovec, curprocp->p_abi, msg.msg_iovlen);
#else
	error =
	    copyin(msg.msg_iov, aiov, msg.msg_iovlen * sizeof (aiov[0]));
#endif
	if (error)
		return error;
	msg.msg_iov = aiov;
	if (msg.msg_accrights) {
		if (!useracc(msg.msg_accrights, msg.msg_accrightslen, B_READ))
			return EFAULT;
		/*
		 * useracc acquired shdlock (SH_ACC), but we release it now,
		 * therefore the copyout below could fail if a shared process
		 * makes the address unavailable.
		 */
		shdunlock(curprocp);
	}
#if _K64U64
	if (!ABI_IS_IRIX5_64(curprocp->p_abi)) {
		i5_uap = (struct irix5_recvmsga *)(__psint_t)uap;
		return recvit(uap->s, &msg, uap->flags, &i5_uap->msg->msg_namelen,
		      &i5_uap->msg->msg_accrightslen, rvp);
	}
	else
#endif	/* _K64U64 */
	return recvit(uap->s, &msg, uap->flags, &uap->msg->msg_namelen,
		      &uap->msg->msg_accrightslen, rvp);
}


static int
recvit( int s,
	register struct msghdr *mp,
	int flags,
	void *namelenp, 
       	void *rightslenp,
	rval_t *rvp)
{
	int error;
	struct socket *so;
	struct uio auio;
	register struct iovec *iov;
	register int i;
	struct mbuf *from, *rights;
	int len;
	
	if (error = getsock(s, &so)) {
		return error;
	}
	auio.uio_iov = mp->msg_iov;
	auio.uio_iovcnt = mp->msg_iovlen;
	auio.uio_segflg = UIO_USERSPACE;
	auio.uio_offset = 0;			/* XXX */
	auio.uio_resid = 0;
	auio.uio_pio = 0;
	auio.uio_pbuf = 0;
	iov = mp->msg_iov;
	for (i = 0; i < mp->msg_iovlen; i++, iov++) {
		if (iov->iov_len > SSIZE_MAX)
			return EINVAL;
		if (iov->iov_len == 0)
			continue;
		if (useracc(iov->iov_base, iov->iov_len, B_READ) == 0) {
			return EFAULT;
		}
		/*
		 * useracc acquired shdlock (SH_ACC), but we release it now,
		 * therefore the copyout below could fail if a shared process
		 * makes the address unavailable.
		 */
		shdunlock(curprocp);
		auio.uio_resid += iov->iov_len;
	}
	len = auio.uio_resid;
	error = soreceive(so, &from, &auio, flags, &rights);
	if (auio.uio_resid != len && error == EINTR)
		error = 0;
	rvp->r_val1 = len - auio.uio_resid;
	if (mp->msg_name) {
		len = mp->msg_namelen;
		if (len <= 0 || from == 0)
			len = 0;
		else {
			if (len > from->m_len)
				len = from->m_len;
			(void) copyout(mtod(from, caddr_t),
			    mp->msg_name, len);
		}
		(void) copyout(&len, namelenp, sizeof (int));
	}
	if (mp->msg_accrights) {
		len = mp->msg_accrightslen;
		if (len <= 0 || rights == 0)
			len = 0;
		else {
			if (len > rights->m_len)
				len = rights->m_len;

#if _K64U64
			/*
			 * hack for binary compat w/ 32-bit systems
			 */
			if (so->so_proto->pr_domain->dom_family == AF_INET) {
				cmsghdr_to_irix5(rights);
				len = (len < rights->m_len) ? len :
					rights->m_len;
			}
#endif
			(void) copyout(mtod(rights, caddr_t),
			    mp->msg_accrights, len);
		}
		(void) copyout(&len, rightslenp, sizeof (int));
	}
	if (rights)
		m_freem(rights);
	if (from)
		m_freem(from);
	return error;
}

struct shutdowna {
	sysarg_t	s;
	sysarg_t	how;
};

/* ARGSUSED */
int
shutdown(struct shutdowna *uap, rval_t *rvp)
{
	int error;
	struct socket *so;

	if (error = getsock(uap->s, &so))
		return error;
	error = soshutdown(so, uap->how);
	if (!error)
	    	_SAT_BSDIPC_SHUTDOWN(uap->s, so, uap->how, 0);
	return error;
}

struct setsockopta {
	sysarg_t	s;
	sysarg_t	level;
	sysarg_t	name;
	caddr_t		val;
	sysarg_t	valsize;
};

/* ARGSUSED */
int
setsockopt(struct setsockopta *uap, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct mbuf *m = NULL;

	if (error = getsock(uap->s, &so))
		return error;
	if (uap->valsize > MLEN)
		return EINVAL;
	if (uap->val) {
		m = m_get(M_WAIT, MT_SOOPTS);
		if (m == NULL)
			return ENOBUFS;
		error = copyin(uap->val, mtod(m, caddr_t), uap->valsize);
		if (error) {
			(void) m_free(m);
			return error;
		}
		m->m_len = uap->valsize;
	}
	return sosetopt(so, uap->level, uap->name, m);
}

struct getsockopta {
	sysarg_t	s;
	sysarg_t	level;
	sysarg_t	name;
	caddr_t		val;
	sysarg_t	*avalsize;
};

/* ARGSUSED */
int
getsockopt(struct getsockopta *uap, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct mbuf *m = NULL;
	int valsize;

	if (error = getsock(uap->s, &so))
		return error;
	if (uap->val) {
		error = copyin(uap->avalsize, &valsize, sizeof (valsize));
		if (error)
			return error;
	} else
		valsize = 0;
	error = sogetopt(so, uap->level, uap->name, &m);
	if (error)
		goto bad;
	if (uap->val && valsize && m != NULL) {
		if (valsize > m->m_len)
			valsize = m->m_len;
		error = copyout(mtod(m, caddr_t), uap->val, valsize);
		if (error)
			goto bad;
		error = copyout(&valsize, uap->avalsize, sizeof (valsize));
	}
bad:
	if (m != NULL)
		(void) m_free(m);
	return error;
}

/*
 * Get socket name.
 */
struct getsocknamea {
	sysarg_t	fdes;
	caddr_t		asa;
	sysarg_t	*alen;
};

/* ARGSUSED */
int
getsockname(struct getsocknamea *uap, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct mbuf *m;
	int len;

	if (error = getsock(uap->fdes, &so))
		return error;
	error = copyin(uap->alen, &len, sizeof (len));
	if (error)
		return error;
	m = m_getclr(M_WAIT, MT_SONAME);
	if (m == NULL)
		return ENOBUFS;
	SOCKET_MPLOCK(so);
	error = (*so->so_proto->pr_usrreq)(so, PRU_SOCKADDR, 0, m, 0);
	SOCKET_MPUNLOCK(so);
	if (error)
		goto bad;
	if (len > m->m_len)
		len = m->m_len;
	error = copyout(mtod(m, caddr_t), uap->asa, len);
	if (error)
		goto bad;
	error = copyout(&len, uap->alen, sizeof (len));
bad:
	m_freem(m);
	return error;
}

/*
 * Get name of peer for connected socket.
 */
struct getpeernamea {
	sysarg_t	fdes;
	caddr_t		asa;
	sysarg_t	*alen;
};

/* ARGSUSED */
int
getpeername(struct getpeernamea *uap, rval_t *rvp)
{
	int error;
	struct socket *so;
	struct mbuf *m;
	int len;

	if (error = getsock(uap->fdes, &so))
		return error;
	if ((so->so_state & SS_ISCONNECTED) == 0)
		return ENOTCONN;
	m = m_getclr(M_WAIT, MT_SONAME);
	if (m == NULL)
		return ENOBUFS;
	error = copyin(uap->alen, &len, sizeof (len));
	if (error)
		goto bad;
	SOCKET_MPLOCK(so);
	error = (*so->so_proto->pr_usrreq)(so, PRU_PEERADDR, 0, m, 0);
	SOCKET_MPUNLOCK(so);
	if (error)
		goto bad;
	if (len > m->m_len)
		len = m->m_len;
	error = copyout(mtod(m, caddr_t), uap->asa, len);
	if (error)
		goto bad;
	error = copyout(&len, uap->alen, sizeof (len));
bad:
	m_freem(m);
	return error;
}

static int
sockargs(struct mbuf **aname, void *name, int namelen, int type)
{
	register struct mbuf *m;
	int error;

	if (((u_int)namelen > MLEN) || (namelen == 0)) {
		return EINVAL;
	}
	/* Ensure that internalized access rights can still fit in 1 mbuf.
	   In 64-bit mode, a pointer and an int have different sizes. */
	if (type == MT_RIGHTS && 
		(u_int)namelen / sizeof(int) * sizeof(struct file *) > MLEN) {
		return EINVAL;
	}
	m = m_get(M_WAIT, type);
	if (m == NULL)
		return (ENOBUFS);
	m->m_len = namelen;
	error = copyin(name, mtod(m, caddr_t), namelen);
	if (error)
		(void) m_free(m);
	else
		*aname = m;
	return (error);
}

/*
 * This function is exported to nfs_server.c
 */
int
getsock(int fd, struct socket **sop)
{
	int error;
	struct file *fp;
	struct vnode *vp;

	if (error = getf(fd, &fp))
		return error;
	vp = fp->f_vnode;
	if (vp->v_op != &uipc_vnodeops)
		return ENOTSOCK;
	*sop = vp->v_data;
	return 0;
}

/*
 * hack to make S5 copyin/out look like bsd
 * XXX remove these and their evil macros in socketvar.h when the SVR4.1
 *     DDI/DKI copyin and copyout are done
 */
#undef copyin
int					/* return a good errno */
bsd_copyin(void *from, void *to, int len)
{
	return (copyin(from,to,len) ? EFAULT : 0);
}

#undef copyout
int					/* return a good errno */
bsd_copyout(void *from, void *to, int len)
{
	return (copyout(from,to,len) ? EFAULT : 0);
}

#if _K64U64
struct irix5_cmsghdr {
	app32_int_t	cmsg_len;
	app32_int_t	cmsg_level;
	app32_int_t	cmsg_type;
};

int
irix5_to_cmsghdr(struct mbuf *m)
{
	struct irix5_cmsghdr *i5_mh;
	struct cmsghdr *mh;
	char buf[64];	/* XXX */
	int len;
	caddr_t d, s;

	if (ABI_IS_IRIX5_64(u.u_procp->p_abi)) {
		return 0;
	}

	i5_mh = mtod(m, struct irix5_cmsghdr *);

	len = i5_mh->cmsg_len;
	len += (sizeof(*mh) - sizeof(*i5_mh));

	if (len > sizeof(buf)) {
		return 0;
	}

	mh = (struct cmsghdr *)buf;

	mh->cmsg_len = len;
	mh->cmsg_level = i5_mh->cmsg_level;
	mh->cmsg_type = i5_mh->cmsg_type;

	d = (caddr_t)CMSG_DATA(mh);
	s = (caddr_t)(i5_mh + 1);

	/* copy remaining data */
	bcopy(s, d, len - sizeof(*mh));

	m->m_len = len;

	bcopy(buf, mtod(m, caddr_t), len);

	return 1;
}

int
cmsghdr_to_irix5(struct mbuf *m)
{
	struct irix5_cmsghdr *i5_mh;
	struct cmsghdr *mh;
	int len;
	caddr_t d, s;
	char buf[64];	/* XXX */

	if (ABI_IS_IRIX5_64(u.u_procp->p_abi)) {
		return 0;
	}

	i5_mh = (struct irix5_cmsghdr *)buf;
	mh = mtod(m, struct cmsghdr *);

	len = mh->cmsg_len;
	len -= (sizeof(*mh) - sizeof(*i5_mh));

	if (len > sizeof(buf)) {
		return 0;
	}

	i5_mh->cmsg_len = len;
	i5_mh->cmsg_level = mh->cmsg_level;
	i5_mh->cmsg_type = mh->cmsg_type;

	d = (caddr_t)(i5_mh + 1);
	s = (caddr_t)(mh + 1);

	/* copy remaining data */
	bcopy(s, d, len - sizeof(*i5_mh));

	m->m_len = len;

	bcopy(buf, mtod(m, caddr_t), len);

	return 1;
}
#endif	/* _K64U64 */
