/*
 * Socket vfs operations.
 *
 * Only statvfs is supported.  Sockets all share a single, static vfs that
 * has no private data.  A vfs is needed because generic vnode code chases
 * vp->v_vfsp to inspect vfs_dev and update vfs_bcount.
 */
#include <sys/types.h>
#include <sys/debug.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/fs_subr.h>
#include <sys/mbuf.h>		/* just for MLEN assertion */
#include <sys/param.h>
#include <sys/protosw.h>
#include <sys/sema.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/statvfs.h>
#include <sys/systm.h>
#include <sys/un.h>		/* just for sun_path size, see uipc_statvfs */
#include <sys/vnode.h>
#include <sys/vfs.h>
#include <sys/atomic_ops.h>
#include <sys/cmn_err.h>
#include <sys/kmem.h>

extern struct vnodeops uipc_vnodeops;
extern void bsd_init(void);
extern char *strcpy(char *, const char *);

struct vfs 	uipc_vfs;
vnlist_t	*uipc_vnlist;	/* for idbg */
static u_int	uipc_vncount;
static mutex_t	*uipc_vnlock;

#define LOCK_VNLIST(i)	mp_mutex_lock(&uipc_vnlock[i], PZERO)
#define UNLOCK_VNLIST(i)	mp_mutex_unlock(&uipc_vnlock[i])

/*
 * Make a new open file and vnode for a socket.
 */
int
makesockvp(struct file **fpp, int *fdp, void *data)
{
	struct vnode *vp;
	int error;
	int zcpu = cpuid();

	vp = vn_alloc(&uipc_vnodeops, &uipc_vfs, VSOCK, NODEV, data);
	if (error = falloc(vp, FREAD|FWRITE, fpp, fdp)) {
		VN_RELE(vp);
		return error;
	}
	LOCK_VNLIST(zcpu);
	vn_append(vp, &uipc_vnlist[zcpu]);
	atomicAddUint(&uipc_vncount, 1);
	((struct socket *)vp->v_data)->so_cpu = zcpu;
	UNLOCK_VNLIST(zcpu);
	return 0;
}

/*
 * Drop vp from the active socket vnode list.
 */
void
dropsockvp(struct vnode *vp)
{
	int zcpu;
	if (vp->v_data == 0) {
		cmn_err(CE_WARN, "dropsockvp: null socket: vp=0x%x\n", vp);
		/* shouldn't happen, but OK */
		return;
	}
	zcpu = ((struct socket *)vp->v_data)->so_cpu;
	LOCK_VNLIST(zcpu);
	vn_unlink(vp);
	atomicAddUint(&uipc_vncount, -1);
	UNLOCK_VNLIST(zcpu);
	vp->v_data = 0;
}

/*
 * Given a sockaddr file identifier, find the vnode of a socket bound or
 * connected to the sockaddr (used by fuser).
 */
/* ARGSUSED */
static int
uipc_vget(struct vfs *vfsp,
	struct vnode **vpp,
	struct fid *fidp)
{
	struct sockaddr *sa;
	struct mbuf *m;
	struct vnode *vp;
	vmap_t vmap;
	struct socket *so;
	int i;
	int gotit = 0;
	NETSPL_SPDECL(s);

	if (fidp->fid_len != sizeof *sa)
		return EINVAL;

	sa = (struct sockaddr *)fidp->fid_data;
	m = m_getclr(M_WAIT, MT_SONAME);
	if (m == NULL)
		return ENOBUFS;

	for (i = 0; i < maxcpus; i++) {
	LOCK_VNLIST(i);

	for (vp = uipc_vnlist[i].vl_next; vp != (struct vnode *) &uipc_vnlist[i];
	     vp = vp->v_next) {

		so = vp->v_data;
		SOCKET_LOCK(so, s);
		if ((*so->so_proto->pr_usrreq)(so, PRU_SOCKADDR, 0, m, 0)) {
			SOCKET_UNLOCK(so, s);
			continue;
		}
		if (bcmp(sa, mtod(m, struct sockaddr *), sizeof *sa) == 0) {
			SOCKET_UNLOCK(so, s);
			break;
		}
		if ((*so->so_proto->pr_usrreq)(so, PRU_PEERADDR, 0, m, 0)) {
			SOCKET_UNLOCK(so, s);
			continue;
		}
		if (bcmp(sa, mtod(m, struct sockaddr *), sizeof *sa) == 0) {
			SOCKET_UNLOCK(so, s);
			break;
		}
		SOCKET_UNLOCK(so, s);
	}
	if (vp != (struct vnode *) &uipc_vnlist[i]) {
		gotit = 1;
		VMAP(vp, vmap);
		UNLOCK_VNLIST(i);
		break;
	}
	UNLOCK_VNLIST(i);
	}
	m_free(m);
	if (!gotit || !(vp = vn_get(vp, &vmap)))
		return ENOENT;
	*vpp = vp;
	return 0;
}

/*
 * Get socket filesystem status.
 */
/* ARGSUSED */
static int
uipc_statvfs(
	struct vfs *vfsp,
	struct statvfs *sp,
	struct vnode *vp)
{
	if (vfsp == NULL)
		return EINVAL;

	bzero(sp, sizeof *sp);
	sp->f_bsize = NBPC;
	sp->f_frsize = NBPC;
	sp->f_files = uipc_vncount;
	sp->f_ffree = sp->f_favail = -1;
	sp->f_fsid = vfsp->vfs_dev;
	strcpy(sp->f_basetype, vfssw[vfsp->vfs_fstype].vsw_name);
	sp->f_flag = vf_to_stf(vfsp->vfs_flag);	
	sp->f_namemax = sizeof(((struct sockaddr_un *) 0)->sun_path) - 1;
	return 0;
}

struct vfsops uipc_vfsops = {
	(int(*)()) fs_nosys,	/* mount */
        (int(*)()) fs_nosys,	/* unmount */
	(int(*)()) fs_nosys,	/* root */
	(int(*)()) uipc_statvfs,
	(int(*)(struct vfs *, short, struct cred *)) fs_sync,
	(int(*)()) uipc_vget,
	(int(*)()) fs_nosys,	/* mountroot */
	(int(*)()) fs_nosys,	/* swapvp */
};

/* ARGSUSED */
int
uipc_init(struct vfssw *vswp, int fstype)
{
	int i;

	/*
	 * Make sure a sockaddr fits in a file identifier.
	 */
	ASSERT(sizeof(struct sockaddr) <= MAXFIDSZ);

	/*
	 * Initialize a single, extern vfs for all sockets to use.
	 */
	VFS_INIT(&uipc_vfs, &uipc_vfsops, NULL);
	uipc_vfs.vfs_bsize = 1024;
	uipc_vfs.vfs_fstype = fstype;
	uipc_vfs.vfs_fsid.val[0] = NODEV;
	uipc_vfs.vfs_fsid.val[1] = fstype;
	uipc_vfs.vfs_dev = NODEV;

	/*
	 * Initialize the list of all active socket vnodes.
	 */
	uipc_vnlist = (vnlist_t *)kmem_zalloc(maxcpus * sizeof(*uipc_vnlist), KM_NOSLEEP);
	if (uipc_vnlist == 0) {
		cmn_err(CE_PANIC, "uipc_vnlist allocation failed");
	}
	uipc_vnlock = (mutex_t *)kmem_zalloc(maxcpus * sizeof(*uipc_vnlock), KM_NOSLEEP);
	if (uipc_vnlock == 0) {
		cmn_err(CE_PANIC, "uipc_vnlock allocation failed");
	}
	for (i = 0; i < maxcpus; i++) {
		vn_initlist(&uipc_vnlist[i]);
		mutex_init(&uipc_vnlock[i], MUTEX_DEFAULT, "uipcvnl");
	}

	return 0;
}
