/*
 * Socket vnode operations.
 *
 * BSD sockets are connected to the rest of the system by making them
 * live in a special filesystem.  This allows us to avoid having to put
 * 'file ops' in each file table entry, or having to modify all of the
 * places where a user might hand a descriptor opened on a socket to a
 * system call expecting a vnode file descriptor.
 */
#ident "$Revision: 1.37 $"

#include <sys/tcp-param.h>
#include <sys/types.h>
#include <sys/debug.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/fs_subr.h>
#include <sys/ioctl.h>
#include <sys/kmem.h>
#include <sys/mbuf.h>
#include <sys/param.h>
#include <sys/pda.h>
#include <sys/poll.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/stat.h>
#include <sys/systm.h>
#include <sys/termio.h>
#include <sys/uio.h>
#include <sys/user.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/proc.h>

/* Trusted IRIX */
#include <sys/mac_label.h>
#include <sys/so_dac.h>

#include <net/soioctl.h>	/* XXX fold into ioctl.h */

extern int ifioctl(struct socket *, int, caddr_t);
extern int rtioctl(int, caddr_t);

/*
 * Close a socket
 */
/* ARGSUSED */
int
uipc_close(struct vnode *vp,
	   int flag,
	   lastclose_t lastclose,
	   off_t offset,
	   struct cred *cr)
{
        register struct socket *so;

        if (lastclose == L_FALSE)
                return 0;

        so = (struct socket *)vp->v_data;
        dropsockvp(vp);

        if (so == (struct socket *)0)
	    return(0);
	return(soclose(so));
}

/*
 * read from a socket
 */
/* ARGSUSED */
static int
uipc_read(struct vnode *vp,
	  uio_t *uio,
	  int ioflag,
	  struct cred *cr)
{
	return soreceive(vp->v_data, 0, uio, 0, 0);
}

/*
 * write to a socket
 */
/* ARGSUSED */
static int
uipc_write(struct vnode *vp,
	   uio_t *uio,
	   int ioflag,
	   struct cred *cr)
{
	return sosend(vp->v_data, 0, uio, 0, 0);
}

/*
 * do IOCTLs for sockets
 */
/* ARGSUSED */
static int
uipc_ioctl(struct vnode *vp,
	   int cmd,
	   void *arg,
	   int flag,
	   struct cred *cr,
	   rval_t *rvalp)
{
	int error = 0;
	register struct socket *so = vp->v_data;
	register u_int size;
	char data[IOCPARM_MASK+1];
	NETSPL_SPDECL(s)

	/*
	 * Interpret high order word to find
	 * amount of data to be copied to/from the user address space.
	 */
	size = (cmd &~ (IOC_INOUT|IOC_VOID)) >> 16;
	if (size > sizeof (data))
		return EFAULT;
	if (cmd & IOC_IN) {
		if (size != 0) {
			if (error = copyin((caddr_t)arg, data, size)) 
				return error;
		} else
			*(caddr_t *)data = (caddr_t)arg;
	} else if ((cmd & IOC_OUT) && size != 0) {
		/*
		 * Zero the buffer on the stack so the user
		 * always gets back something deterministic.
		 */
		bzero(data, size);
	} else if (cmd & IOC_VOID) {
		*(caddr_t *)data = (caddr_t)arg;
	}

	SOCKET_LOCK(so, s);
	switch (cmd) {

	case FIONBIO:
		if (*(int *)data)
			so->so_state |= SS_NBIO;
		else
			so->so_state &= ~SS_NBIO;
		break;

	case FIOASYNC:
		if (*(int *)data)
			so->so_state |= SS_ASYNC;
		else
			so->so_state &= ~SS_ASYNC;
		break;

	case FIONREAD:
		*(int *)data = so->so_rcv.sb_cc;
		break;

	case SIOCNREAD:
		if ((*(int *)data = so->so_rcv.sb_cc) == 0) {
			if (so->so_error) {
				error = so->so_error;
				so->so_error = 0;
			} else if (so->so_state & SS_CANTRCVMORE) {
				/*
				 * For UDP sockets, this isn't really true...
				 */
				error = ECONNRESET;
			}
		}
		break;

	case FIOSETOWN:
	case SIOCSPGRP:
		so->so_pgrp = *(int *)data;
		break;

	case FIOGETOWN:
	case SIOCGPGRP:
		*(int *)data = so->so_pgrp;
		break;

	case SIOCATMARK:
		*(int *)data = (so->so_state&SS_RCVATMARK) != 0;
		break;

/* XXX - old Trusted IRIX cases */
	case SIOCGETLABEL:	
	case SIOCSETLABEL:		
        case SIOCGETACL:
	case SIOCSETACL:
	case SIOCGETUID:
	case SIOCSETUID:
	case SIOCGETRCVUID:
		error = ENOPKG;
		break;
/* End of old Trusted IRIX cases */

	default:
		/*
		 * Interface/routing/protocol specific ioctls:
		 * interface and routing ioctls should have a
		 * different entry since a socket is unnecessary
		 */
#define	cmdbyte(x)	(((x) >> 8) & 0xff)
		if (cmdbyte(cmd) == 'i')
			error = (ifioctl(so, cmd, data));
		else if (cmdbyte(cmd) == 'r')
			error = (rtioctl(cmd, data));
		else
			error = (*so->so_proto->pr_usrreq)(so,
				PRU_CONTROL, (struct mbuf *)(NULL+cmd),
				(struct mbuf *)data, (struct mbuf *)0);
	}
	SOCKET_UNLOCK(so, s);

	/*
	 * Copy any data to user, size was
	 * already set and checked above.
	 */
	if (!error && (cmd & IOC_OUT) && size)
		error = copyout(data, (caddr_t)arg, size);
	return error;
}

/* ARGSUSED */
static int
uipc_setfl(struct vnode *vp,
	   int oflags,
	   int nflags,
	   struct cred *cr)
{
	register struct socket *so = vp->v_data;
	NETSPL_SPDECL(s)

	SOCKET_LOCK(so, s);
	if (nflags & (FNONBLK|FNDELAY))
		so->so_state |= SS_NBIO;
	else
		so->so_state &= ~SS_NBIO;
	if (nflags & FASYNC)
		so->so_state |= SS_ASYNC;
	else
		so->so_state &= ~SS_ASYNC;
	SOCKET_UNLOCK(so, s);
	return 0;
}

/*
 * Get socket attributes.  From the BUGS section of 4.3BSD stat(2):
 *
 *	"Applying fstat to a socket ... returns a zero'd [sic] buffer,
 *	 except for the blocksize field, and a unique device and vnode
 *	 number."
 */
/*ARGSUSED*/
static int
uipc_getattr(struct vnode *vp,
	     struct vattr *vap,
	     int flags,
	     struct cred *cr)
{
	register struct socket *so = vp->v_data;
	NETSPL_SPDECL(s)
	int error;

	bzero(vap, sizeof *vap);
	vap->va_type = VSOCK;
	SOCKET_LOCK(so, s);
	error = (*so->so_proto->pr_usrreq)(so, PRU_SENSE, (struct mbuf *)vap,
					   (struct mbuf *)0, (struct mbuf *)0);
	SOCKET_UNLOCK(so, s);
	return(error);
}

/*
 * Socket file control
 */
/* ARGSUSED */
int
uipc_fcntl(
	struct vnode *vp,
	int cmd,
	void *arg,
	int flag,
	off_t offset,
	struct cred *cr,
	rval_t *rvp)
{
	register struct socket *so = vp->v_data;

	switch (cmd) {
	case F_GETOWN:
		rvp->r_val1 = so->so_pgrp;
		break;

	case F_SETOWN:
		so->so_pgrp = (__psint_t)arg;
		break;

	default:
		return EINVAL;
	}
	return 0;
}

static int
uipc_fid(struct vnode *vp,
	 struct fid **fidpp)
{
	struct mbuf *m;
	struct socket *so;
	int error;
	struct fid *fidp;
	NETSPL_SPDECL(s)

	m = m_getclr(M_WAIT, MT_SONAME);
	if (m == NULL)
		return ENOBUFS;
	so = vp->v_data;
	SOCKET_LOCK(so, s);
	if (error = (*so->so_proto->pr_usrreq)(so, PRU_SOCKADDR, 0, m, 0)) {

		SOCKET_UNLOCK(so, s);
		m_free(m);
		return error;
	}
	SOCKET_UNLOCK(so, s);
	fidp = kmem_alloc(sizeof *fidp - MAXFIDSZ + m->m_len, KM_SLEEP);
	fidp->fid_len = m->m_len;
	(void) m_datacopy(m, 0, fidp->fid_len, fidp->fid_data);
	m_free(m);
	*fidpp = fidp;
	return 0;
}

/* ARGSUSED */
static int
uipc_seek(struct vnode *vp,
	  off_t ooff,
	  off_t *noffp)
{
	/* BSD4.3 doesn't permit seek on sockets */
	return ESPIPE;
}

/* ARGSUSED */
static int
uipc_poll(struct vnode *vp,
	  short events,
	  int anyyet,
	  short *reventsp,
	  struct pollhead **phpp)
{
	register unsigned revents;
	struct socket *so;
	NETSPL_SPDECL(s)

#define sbselqueue(sb) (sb)->sb_flags |= SB_SEL

	/* we can set phpp outside lock, saves two arg restores */
	so = vp->v_data;
	if (!anyyet)
		*phpp = &so->so_ph;

	SOCKET_LOCK(so, s);
	mp_spinlock(&so->so_qlock);
	revents = events;
	if ((revents & (POLLIN|POLLRDNORM)) && !soreadable(so)) {
		revents &= ~(POLLIN|POLLRDNORM);
		sbselqueue(&so->so_rcv);
	}
	if ((revents & (POLLPRI|POLLRDBAND))
	    && !(so->so_oobmark | (so->so_state & SS_RCVATMARK))) {
		revents &= ~(POLLPRI|POLLRDBAND);
		sbselqueue(&so->so_rcv);
	}
	if ((revents & POLLOUT) && !sowriteable(so)) {
		revents &= ~POLLOUT;
		sbselqueue(&so->so_snd);
	}
	*reventsp = revents;
	mp_spinunlock(&so->so_qlock);
	SOCKET_UNLOCK(so, s);
	return 0;
}

struct vnodeops uipc_vnodeops = {
	fs_nosys,
	uipc_close,
	uipc_read,
	uipc_write,
	(int(*)())uipc_ioctl,
	uipc_setfl,
	uipc_getattr,
	fs_nosys,	/* setattr */
	fs_nosys,	/* access */
	fs_nosys,	/* lookup */
	fs_nosys,	/* create */
	fs_nosys,	/* remove */
	fs_nosys,	/* link */
	fs_nosys,	/* rename */
	fs_nosys,	/* mkdir */
	fs_nosys,	/* rmdir */
	fs_nosys,	/* readdir */
	fs_nosys,	/* symlink */
	fs_nosys,	/* readlink */
	fs_nosys,	/* fsync */
	fs_noval,	/* inactive */
	uipc_fid,
	fs_nosys,	/* fid2 */
	fs_rwlock,
	fs_rwunlock,
	uipc_seek,
	fs_cmp,
	fs_frlock,
	fs_nosys,	/* realvp */
	fs_nosys,	/* bmap */
	fs_noval,	/* strategy */
	fs_nosys,	/* map */
	fs_nosys,	/* addmap */
	fs_nosys,	/* delmap */
	uipc_poll,
	fs_nosys,	/* dump */
	fs_pathconf,
	fs_nosys,	/* allocstore */
	uipc_fcntl,	/* fcntl */
	fs_noerr,	/* reclaim */
	fs_nosys,	/* attr_get */
	fs_nosys,	/* attr_set */
	fs_nosys,	/* attr_remove */
	fs_nosys,	/* attr_list */
};
