/*
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *	@(#)uipc_usrreq.c	7.7 (Berkeley) 6/29/88
 */

#include <tcp-param.h>
#include <sys/param.h>
#include <sys/debug.h>
#include <sys/domain.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <sys/unpcb.h>
#include <sys/user.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/systm.h>
#include <sys/atomic_ops.h>

#include <sys/mac_label.h>


extern	struct vfs uipc_vfs;
extern	struct vnodeops uipc_vnodeops;
extern	mac_label *mac_equal_equal_lp;

/*
 * Forward References.
 */
int unp_attach(struct socket *);
int unp_bind(struct unpcb *, struct mbuf *);
int unp_connect(struct socket *, struct mbuf *);
int unp_connect2(struct socket *, struct socket *);
void unp_detach(struct unpcb *);
void unp_discard(struct file *);
void unp_disconnect(struct unpcb *);
static int unp_drop(struct unpcb *, int);
void unp_gc(void);
int unp_internalize(struct mbuf *);
void unp_usrclosed(struct unpcb *);
void unp_scan(struct mbuf *, void (*)(struct file *));
int uipc_three_way_lock(struct socket *, struct socket *, int);

/*
 * Unix communications domain.
 *
 * TODO:
 *	SEQPACKET, RDM
 *	rethink name space problems
 *	need a proper out-of-band
 */
struct	sockaddr sun_noname = { AF_UNIX };
ino_t	unp_ino;			/* prototype for fake inode numbers */

#ifdef _MP_NETLOCKS

#ifdef _SCA_DEBUG
mutex_t	unp_lock;
struct proc *unp_locker;
#define UNP_MPLOCK(s)		{ unp_locker = curprocp; \
	mutex_lock(&unp_lock, PZERO); }
#define UNP_MPUNLOCK(s)		{ unp_locker = 0; \
	mutex_unlock(&unp_lock); }
#else

lock_t	unp_lock;
#define UNP_MPLOCK(s)		s = mp_mutex_spinlock(&unp_lock)
#define UNP_MPUNLOCK(s)		mp_mutex_spinunlock(&unp_lock, s)
#endif
#define UNP_PROTECT(code)	{  \
	int _s;  UNP_MPLOCK(_s); code; UNP_MPUNLOCK(_s); }
#else
#define UNP_MPLOCK(s)		
#define UNP_MPUNLOCK(s)	
#define UNP_PROTECT(code)	{ code; }
#endif


int
uipc_usrreq(struct socket *so,
	    int req,
	    struct mbuf *m,
	    struct mbuf *nam,
	    struct mbuf *rights)
{
	struct unpcb *unp = sotounpcb(so);
	register struct socket *so2;
	int error = 0;

	ASSERT(SOCKET_ISLOCKED(so));
	if (req == PRU_CONTROL)
		return (EOPNOTSUPP);
	if (req != PRU_SEND && rights && rights->m_len) {
		error = EOPNOTSUPP;
		goto release;
	}

	if (unp == 0 && req != PRU_ATTACH) {
		error = EINVAL;
		goto release;
	}
	switch (req) {

	case PRU_ATTACH:
		if (unp) {
			error = EISCONN;
			break;
		}
		error = unp_attach(so);
		break;

	case PRU_DETACH:
		unp_detach(unp);
		break;

	case PRU_BIND:
		error = unp_bind(unp, nam);
		break;

	case PRU_LISTEN:
		if (unp->unp_vnode == 0)
			error = EINVAL;
		break;

	case PRU_CONNECT:
		error = unp_connect(so, nam);
		break;

	case PRU_CONNECT2:
		ASSERT(SOCKET_ISLOCKED((struct socket *)nam));
		error = unp_connect2(so, (struct socket *)nam);
		break;

	case PRU_DISCONNECT:
		unp_disconnect(unp);
		break;

	case PRU_ACCEPT:
		/*
		 * Pass back name of connected socket,
		 * if it was bound and we are still connected
		 * (our peer may have closed already!).
		 */
		if (unp->unp_conn && unp->unp_conn->unp_addr) {
			nam->m_len = unp->unp_conn->unp_addr->m_len;
			bcopy(mtod(unp->unp_conn->unp_addr, caddr_t),
			    mtod(nam, caddr_t), (unsigned)nam->m_len);
		} else {
			nam->m_len = sizeof(sun_noname);
			*(mtod(nam, struct sockaddr *)) = sun_noname;
		}
		break;

	case PRU_SHUTDOWN:
		socantsendmore(so);
		unp_usrclosed(unp);
		break;

	case PRU_RCVD:
		switch (so->so_type) {

		case SOCK_DGRAM:
			panic("uipc 1");
			/*NOTREACHED*/

		case SOCK_STREAM:
#define	rcv (&so->so_rcv)
#define snd (&so2->so_snd)
rcv_strloop:
			if (unp->unp_conn == 0) {
				break;
			}
			so2 = unp->unp_conn->unp_socket;
			if (so2->so_state & SS_CLOSING) {
				break;
			}
#ifdef _MP_NETLOCKS
			if (SOCKET_PAIR_CMPLOCK(so, so2) == 0) 
				goto rcv_strloop;
#endif /* _MP_NETLOCKS */
			ASSERT(SOCKET_ISLOCKED(so2));		
			/*
			 * Adjust backpressure on sender
			 * and wakeup any waiting to write.
			 */
			snd->sb_hiwat += unp->unp_cc - rcv->sb_cc;
			unp->unp_cc = rcv->sb_cc;
			sowwakeup(so2, NETEVENT_SODOWN);
#ifdef _MP_NETLOCKS
			if (so2 != so)
				SOCKET_MPUNLOCK(so2);	
#endif /* _MP_NETLOCKS */
#undef snd
#undef rcv
			break;

		default:
			panic("uipc 2");
		}
		break;

	case PRU_SEND:
		if (rights) {
			error = unp_internalize(rights);
			if (error)
				break;
		}
		switch (so->so_type) {

		case SOCK_DGRAM: {
			struct sockaddr *from;

			/*
			 * No other process clears a SOCK_DGRAM socket's
			 * unp_conn, so we don't need UNP_LOCK.
			 */
			if (nam) {
				if (unp->unp_conn) {
					error = EISCONN;
					break;
				}
				error = unp_connect(so, nam);
				if (error)
					break;
			} else {
				if (unp->unp_conn == 0) {
					error = ENOTCONN;
					break;
				}
			}
#ifndef _MP_NETLOCKS
			so2 = unp->unp_conn->unp_socket;
			if (so2->so_state & SS_CLOSING) {
				error = ENOTCONN;
				break;
			}
#else /* _MP_NETLOCKS */
snd_dloop: 
			so2 = unp->unp_conn->unp_socket;
			if (so2->so_state & SS_CLOSING) {
				error = ENOTCONN;
				break;
			}
			if (SOCKET_PAIR_CMPLOCK(so, so2) == 0) {
				if ((unp->unp_conn == 0) ||
				    (so2->so_state & SS_CLOSING)) {
					error = ENOTCONN;
                                        break;
                                }
				goto snd_dloop;
			}
#endif /* _MP_NETLOCKS */
			ASSERT(SOCKET_ISLOCKED(so2));

			if (unp->unp_addr)
				from = mtod(unp->unp_addr, struct sockaddr *);
			else
				from = &sun_noname;
			if (sbspace(&so2->so_rcv) > 0 &&
			    sbappendaddr(&so2->so_rcv, from, 
				    m, rights, 0)) {
				sorwakeup(so2, NETEVENT_SOUP);
				m = 0;
			} else
				error = ENOBUFS;
#ifdef _MP_NETLOCKS
			if (so2 != so)
				SOCKET_MPUNLOCK(so2);
#endif /* _MP_NETLOCKS */
			if (nam)
				unp_disconnect(unp);
			break;
		}

		case SOCK_STREAM:
#define	rcv (&so2->so_rcv)
#define	snd (&so->so_snd)
			if (so->so_state & SS_CANTSENDMORE) {
				error = EPIPE;
				break;
			}
			if (unp->unp_conn == 0) {
				/*
				 * Previously we called
				 * panic("uipc 3"); why panic?
				 * In this case it's unnecessary
				 */
				error = ENOTCONN;
				break;
			}
#ifndef _MP_NETLOCKS
			so2 = unp->unp_conn->unp_socket;
			if (so2->so_state & SS_CLOSING) {
				error = ENOTCONN;
				break;
			}
#else /* _MP_NETLOCKS */
snd_strloop:
			so2 = unp->unp_conn->unp_socket;
			if (so2->so_state & SS_CLOSING) {
				error = ENOTCONN;
				break;
			}
			if (SOCKET_PAIR_CMPLOCK(so, so2) == 0) {
				if (unp->unp_conn == 0) {
					error = ENOTCONN;
					break;
				}
				goto snd_strloop;
			}
#endif /* _MP_NETLOCKS */
			/*
			 * Send to paired receive port, and then reduce
			 * send buffer hiwater marks to maintain backpressure.
			 * Wake up readers.
			 */
			if (rights)
				(void)sbappendrights(rcv, m, rights);
			else
				sbappend(rcv, m);
			snd->sb_hiwat -= rcv->sb_cc - unp->unp_conn->unp_cc;
			unp->unp_conn->unp_cc = rcv->sb_cc;
			sorwakeup(so2, NETEVENT_SOUP);
#ifdef _MP_NETLOCKS
			if (so2 != so)
				SOCKET_MPUNLOCK(so2);
#endif /* _MP_NETLOCKS */
			m = 0;
#undef snd
#undef rcv
			break;

		default:
			panic("uipc 4");
		}
		break;

	case PRU_ABORT:
		(void)unp_drop(unp, ECONNABORTED);
		break;

	case PRU_SENSE:
#define vap ((struct vattr *) m)
		vap->va_blksize = so->so_snd.sb_hiwat;
		if (so->so_type == SOCK_STREAM && unp->unp_conn != 0) {
			so2 = unp->unp_conn->unp_socket;
			vap->va_blksize += so2->so_rcv.sb_cc;
		}
		vap->va_fsid = NODEV;
		if (unp->unp_ino == 0)
			unp->unp_ino = ++unp_ino ? unp_ino : ++unp_ino;
		vap->va_nodeid = unp->unp_ino;
		vap->va_mac = mac_equal_equal_lp;
#undef vap
		return (0);

	case PRU_RCVOOB:
		return (EOPNOTSUPP);

	case PRU_SENDOOB:
		error = EOPNOTSUPP;
		break;

	case PRU_SOCKADDR:
		if (unp->unp_addr) {
			nam->m_len = unp->unp_addr->m_len;
			bcopy(mtod(unp->unp_addr, caddr_t),
			    mtod(nam, caddr_t), (unsigned)nam->m_len);
		} else
			nam->m_len = 0;
		break;

	case PRU_PEERADDR:
		if (unp->unp_conn && unp->unp_conn->unp_addr) {
			nam->m_len = unp->unp_conn->unp_addr->m_len;
			bcopy(mtod(unp->unp_conn->unp_addr, caddr_t),
			    mtod(nam, caddr_t), (unsigned)nam->m_len);
		} else
			nam->m_len = 0;
		break;

	case PRU_SLOWTIMO:
		break;

	case PRU_SOCKLABEL:
		error = EOPNOTSUPP;
		break;

	default:
		panic("piusrreq");
	}
release:
	if (m)
		m_freem(m);
	return (error);
}

/*
 * Both send and receive buffers are allocated PIPSIZ bytes of buffering
 * for stream sockets, although the total for sender and receiver is
 * actually only PIPSIZ.
 * Datagram sockets really use the sendspace as the maximum datagram size,
 * and don't really want to reserve the sendspace.  Their recvspace should
 * be large enough for at least one max-size datagram plus address.
 */
extern u_int	unpst_sendspace;
extern u_int	unpst_recvspace;
extern u_int	unpdg_sendspace;	/* really max datagram size */
extern u_int	unpdg_recvspace;

struct	unpcb *unp_bound;		/* list of bound unpcbs */
int	unp_rights;			/* file descriptors in flight */

int
unp_attach(struct socket *so)
{
	register struct mbuf *m;
	register struct unpcb *unp;
	int error;

	ASSERT(SOCKET_ISLOCKED(so));	
	if (so->so_snd.sb_hiwat == 0 || so->so_rcv.sb_hiwat == 0) {
		switch (so->so_type) {

		case SOCK_STREAM:
			error = soreserve(so, unpst_sendspace, unpst_recvspace);
			break;

		case SOCK_DGRAM:
			error = soreserve(so, unpdg_sendspace, unpdg_recvspace);
			break;
		}
		if (error)
			return (error);
	}
	m = m_getclr(M_DONTWAIT, MT_PCB);
	if (m == NULL)
		return (ENOBUFS);
	unp = mtod(m, struct unpcb *);
	so->so_pcb = (caddr_t)unp;
	unp->unp_socket = so;
	return (0);
}

void
unp_detach(register struct unpcb *unp)
{
	NETSPL_MPDECL(s)
	struct vnode *vp;
	struct unpcb *unq;
	struct socket *so = unp->unp_socket;

	ASSERT(SOCKET_ISLOCKED(so));

	UNP_MPLOCK(s);
	so->so_state |= SS_CLOSING;
	ASSERT((unp->unp_vnode != NULL) ^ (unp->unp_prevp == NULL));
	if (vp = unp->unp_vnode) {
		unp->unp_vnode = 0;
		if (unq = unp->unp_next)
			unq->unp_prevp = unp->unp_prevp;
		*unp->unp_prevp = unq;
	}
	UNP_MPUNLOCK(s);
	if (unp->unp_conn || (so->so_type == SOCK_DGRAM)) {
		/* ensure everybody disconnected */
		unp_disconnect(unp);
	}
restart:
	UNP_MPLOCK(s);
	while (unp->unp_refs) {
		struct socket *so = (unp->unp_refs)->unp_socket;
#ifdef _MP_NETLOCKS
		if (uipc_three_way_lock(unp->unp_socket, so, s) == 0) {
			goto restart;
		}
#endif
		UNP_MPUNLOCK(s);
		/*
		 * Make sure we use the same socket we just locked above,
		 * and that it was not detached while we blocked.
		 */
		if (so->so_pcb == 0 ||
		    unp_drop((struct unpcb *)so->so_pcb, ECONNRESET) == 0)
			SOCKET_MPUNLOCK(so);
		UNP_MPLOCK(s);
	}
	UNP_MPUNLOCK(s);
	soisdisconnected(unp->unp_socket);
	unp->unp_socket->so_pcb = 0;
	if (vp)
		VN_RELE(vp);
	m_freem(unp->unp_addr);
	(void) m_free(unptom(unp));
#ifdef NEVER
	if (unp_rights)
		unp_gc();
#endif
}

unp_bind(struct unpcb *unp,
	 struct mbuf *nam)
{
	struct sockaddr_un *soun = mtod(nam, struct sockaddr_un *);
	struct vnode *vp;
	int error, s;
	struct vattr vattr;
	struct unpcb *unq;

	ASSERT(SOCKET_ISLOCKED(unp->unp_socket));
#ifdef sgi
	/* 
	 * Audit:
	 * Since bind() NEVER checks the sun_family, programs often
	 * fail to initialize it properly.  Make it right here for audit.
	 */
	soun->sun_family = AF_UNIX;
#endif

	if (unp->unp_vnode != NULL || nam->m_len == MLEN)
		return (EINVAL);
	*(mtod(nam, caddr_t) + nam->m_len) = 0;
/* SHOULD BE ABLE TO ADOPT EXISTING AND wakeup() ALA FIFO's */
	error = lookupname(soun->sun_path, UIO_SYSSPACE, FOLLOW,
			   NULLVPP, &vp);
	if (!error) {
		VN_RELE(vp);
		return (EADDRINUSE);
	}
	vattr.va_mask = AT_TYPE|AT_MODE|AT_RDEV;
	vattr.va_type = VSOCK;
	vattr.va_mode = 0777;
	vattr.va_rdev = NODEV;
	error = vn_create(soun->sun_path, UIO_SYSSPACE, &vattr, EXCL,
			  VREAD|VWRITE, &vp, CRMKNOD);
	if (error) {
		if (error == EEXIST)
			error = EADDRINUSE;
		return (error);
	}
	UNP_MPLOCK(s);
	unp->unp_vnode = vp;
	if (unq = unp_bound)
		unq->unp_prevp = &unp->unp_next;
	unp->unp_next = unq;
	unp->unp_prevp = &unp_bound;
	unp_bound = unp;
	UNP_MPUNLOCK(s);
	unp->unp_addr = m_copy(nam, 0, (int)M_COPYALL);
	return (0);
}

unp_connect(struct socket *so,
	    struct mbuf *nam)
{
	register struct sockaddr_un *soun = mtod(nam, struct sockaddr_un *);
	struct vnode *vp;
	register struct socket *so2 = 0, *so3;
	struct unpcb *unp2, *unp3;
	int error, s;

	ASSERT(SOCKET_ISLOCKED(so));
#ifdef sgi
	/* 
	 * Audit:
	 * Since unp_connect() NEVER checks the sun_family, programs often
	 * fail to initialize it properly.  Make it right here for audit.
	 */
	soun->sun_family = AF_UNIX;
#endif
	if (nam->m_len + (nam->m_off - MMINOFF) == MLEN)
		return (EMSGSIZE);
	*(mtod(nam, caddr_t) + nam->m_len) = 0;
	error = lookupname(soun->sun_path, UIO_SYSSPACE, FOLLOW,
			   NULLVPP, &vp);
	if (error)
		return (error);
	if (error = VOP_ACCESS(vp, VWRITE, 0, curprocp->p_cred))
		goto bad;
	if (vp->v_type != VSOCK) {
		error = ENOTSOCK;
		goto bad;
	}
retry:
	so2 = 0;
	UNP_MPLOCK(s);
	for (unp2 = unp_bound; unp2 != 0; unp2 = unp2->unp_next) {
		if ((unp2->unp_vnode == vp) && unp2->unp_socket) {
			so2 = unp2->unp_socket;
			SO_HOLD(so2);
			break;
		}
	}
	UNP_MPUNLOCK(s);
	if (so2 == 0) {
		error = ECONNREFUSED;
		goto bad;
	}
#ifdef _MP_NETLOCKS
	if (!SOCKET_PAIR_CMPLOCK(so, so2)) {
		/*
		 * Couldn't get second lock:
		 *	- drop first lock
		 *	- acquire second lock
		 *	- release second PCB
		 *	- reacquire first lock
		 * We don't need to hold so because it is not marked
		 * SS_NOFDREF; there should be no way that can happen while
		 * we are in unp_connect().
		 * (relies on the fact that SOCKET_PAIR_CMPLOCK() will return
		 * 1 if both sockets are the same)
		 */
		SOCKET_MPUNLOCK(so);
		SOCKET_MPLOCK(so2);
		if (SO_RELE(so2) == 0) {
			SOCKET_MPUNLOCK(so2);
		}
		SOCKET_MPLOCK(so);
		goto retry;
	}
#endif
	if (so->so_type != so2->so_type) {
		if ((SO_RELE(so2) == 0) && (so2 != so)) {
			SOCKET_MPUNLOCK(so2);
		}
		error = EPROTOTYPE;
		goto bad;
	}
	if ((so2->so_pcb == 0) || (so2->so_state & SS_CLOSING)) {
		if ((SO_RELE(so2) == 0) && (so2 != so)) {
			SOCKET_MPUNLOCK(so2);
		}
		error = ECONNREFUSED;
		goto bad;
	}
	if (so->so_proto->pr_flags & PR_CONNREQUIRED) {
		struct mbuf *nam2 = 0;

		if ((so2->so_options & SO_ACCEPTCONN) == 0) {
			error = ECONNREFUSED;
			if ((SO_RELE(so2) == 0) && (so2 != so)) {
				SOCKET_MPUNLOCK(so2);
			}
			goto bad;
		}
		unp2 = sotounpcb(so2);
		if (unp2->unp_addr)
			nam2 = m_copy(unp2->unp_addr, 0, M_COPYALL); 
		if ((so3 = sonewconn(so2)) == 0) {
			if (nam2) m_freem(nam2);
			/* sonewconn() no longer unlocks so2 */
			if ((SO_RELE(so2) == 0) && (so2 != so)) {
				SOCKET_MPUNLOCK(so2);
			}
			error = ECONNREFUSED;
			goto bad;
		}
		unp3 = sotounpcb(so3);
		unp3->unp_addr = nam2;
		/* sonewconn() no longer unlocks so2 */
		if ((SO_RELE(so2) == 0) && (so2 != so)) {
			SOCKET_MPUNLOCK(so2);
		}
		SO_HOLD(so3);	/* simplifies unlock below */
		so2 = so3;
	}
	error = unp_connect2(so, so2);
	if ((SO_RELE(so2) == 0) && (so2 != so)) {
		SOCKET_MPUNLOCK(so2);
	}
bad:
	ASSERT(SOCKET_ISLOCKED(so));
	VN_RELE(vp);
	return (error);
}

unp_connect2(register struct socket *so,
	     register struct socket *so2)
{
	register struct unpcb *unp = sotounpcb(so);
	register struct unpcb *unp2;
	int s;

	ASSERT(SOCKET_ISLOCKED(so));
	ASSERT(SOCKET_ISLOCKED(so2));
	if (so2->so_type != so->so_type)
		return (EPROTOTYPE);
	if (so2->so_state & SS_CLOSING) {
		return ECONNREFUSED;
	}
	unp2 = sotounpcb(so2);
	unp->unp_conn = unp2;
	ASSERT(so2->so_head != so);
	ASSERT(so->so_head != so2);
	switch (so->so_type) {

	case SOCK_DGRAM:
		UNP_MPLOCK(s);
		unp->unp_nextref = unp2->unp_refs;
		unp2->unp_refs = unp;
		UNP_MPUNLOCK(s);
		soisconnected(so);
		break;

	case SOCK_STREAM:
		unp2->unp_conn = unp;
		soisconnected(so2);
		soisconnected(so);
		break;

	default:
		panic("unp_connect2");
	}
	return (0);
}

void
unp_disconnect(struct unpcb *unp)
{
	register struct unpcb *unp2 = unp->unp_conn;
	int s;

	ASSERT(SOCKET_ISLOCKED(unp->unp_socket));
	if (unp2 == 0)
		return;
	switch (unp->unp_socket->so_type) {

	case SOCK_DGRAM:
		unp->unp_conn = 0;
		UNP_MPLOCK(s);
		if (unp2->unp_refs == unp)
			unp2->unp_refs = unp->unp_nextref;
		else {
			unp2 = unp2->unp_refs;
			for (;;) {
				if (unp2 == 0)
					panic("unp_disconnect");
				/* avoid panic in PRU_SEND */
				if (unp2->unp_conn == unp) {
					unp2->unp_conn = (struct unpcb *)0;
				}
				if (unp2->unp_nextref == unp)
					break;
				unp2 = unp2->unp_nextref;
			}
			unp2->unp_nextref = unp->unp_nextref;
		}
		unp->unp_nextref = 0;
		UNP_MPUNLOCK(s);
		unp->unp_socket->so_state &= ~SS_ISCONNECTED;
		break;

	case SOCK_STREAM:
#ifndef _MP_NETLOCKS
		if (unp2->unp_socket == 0) {
			return;
		}
#else /* _MP_NETLOCKS */
loop:
		if (unp2->unp_socket == 0) {
			return;
		}
		if (SOCKET_PAIR_CMPLOCK(unp->unp_socket, 
		    unp2->unp_socket) == 0) {
			if (unp->unp_conn == 0) {
				return;
			} else {
				unp2 = unp->unp_conn;
				goto loop;
			}
		}
#endif /* _MP_NETLOCKS */
		ASSERT(SOCKET_ISLOCKED(unp2->unp_socket));
		unp->unp_conn = 0;
		soisdisconnected(unp->unp_socket);
		unp2->unp_conn = 0;
		soisdisconnected(unp2->unp_socket);
#ifdef _MP_NETLOCKS
		if (unp->unp_socket != unp2->unp_socket)
			SOCKET_MPUNLOCK(unp2->unp_socket);
#endif /* _MP_NETLOCKS */
		ASSERT(SOCKET_ISLOCKED(unp->unp_socket));
		break;
	}
}

void
unp_usrclosed(struct unpcb *unp)
{
	if (unp->unp_socket->so_type == SOCK_STREAM) {
		struct socket *so;
sync:
		if (unp->unp_conn && (so = unp->unp_conn->unp_socket)) {
#ifdef _MP_NETLOCKS 
			if (SOCKET_PAIR_CMPLOCK(unp->unp_socket, so) == 0)
				goto sync;
#endif /* _MP_NETLOCKS */
			socantrcvmore(so);
#ifdef _MP_NETLOCKS
			if (unp->unp_socket != so)
				SOCKET_MPUNLOCK(so);
#endif /* _MP_NETLOCKS */
		}
	}
}

int
unp_drop(struct unpcb *unp,
	 int errno)
{
	struct socket *so = unp->unp_socket;
	int rv = 0;

	ASSERT(SOCKET_ISLOCKED(so));
	so->so_error = errno;
	unp_disconnect(unp);
	if ((so->so_state & SS_NOFDREF) && (errno == ECONNABORTED)) {
		so->so_pcb = (caddr_t) 0;
		m_freem(unp->unp_addr);
		(void) m_free(unptom(unp));
		rv = sofree(so);
	}
	return(rv);
}

#ifdef notdef
unp_drain()
{

}
#endif

int
unp_externalize(struct mbuf *rights)
{
	int newfds = rights->m_len / sizeof (struct file *);
	register int i, *ip;
	register struct file **rp = mtod(rights, struct file **);
	struct file *fp;
	int s;
	int *fdsbase, *fds ;

	fdsbase = kern_malloc(newfds * sizeof(int *));
	ip = (int *)rp;

	/* allocate new fds */
	for (i = 0, fds = fdsbase; i < newfds; i++, fds++) {
		if (ufalloc(0, fds) != 0) {
			/* failed to get enough - undo ufallocs */
			while (--i >= 0)
				setf(fdsbase[i], NULLFP);
			for (i = 0; i < newfds; i++) {
				fp = *rp;
				unp_discard(fp);
				*rp++ = 0;
			}
			kern_free(fdsbase);
			return (EMSGSIZE);
		}
	}

	/* have new fd's, assign the fp's to them */
	for (i = 0, fds = fdsbase; i < newfds; i++, fds++) {
		fp = *rp++;
		s = FLOCK(fp);
		fp->f_msgcount--;
		FUNLOCK(fp, s);
		setf(*fds, fp);
		UNP_PROTECT(unp_rights++);
		*ip++ = *fds;
	}
	kern_free(fdsbase);

	/* adjust mbuf length to reflect type change */
	rights->m_len = newfds * sizeof(int);

	return 0;
}

/*
 * the first getf will mark each fd as inuse so that any share group
 * members that share fd's can't close any of them. The second getf
 * then doesn't have to mark anything...
 */
unp_internalize(struct mbuf *rights)
{
	register struct file **rp;
	int oldfds = rights->m_len / sizeof (int);
	register int i, *ip;
	struct file *fp;
	int error, s;

	/* Make sure fds are OK */
	rp = mtod(rights, struct file **);
	ip = (int *)rp;
	for (i = 0; i < oldfds; i++)
		if (error = getf(*ip++, &fp))
			return (error);

	/* Convert fds to file pointers.  Walk backward because a pointer
	   is bigger than a fd in 64-bit mode. */
	rp = mtod(rights, struct file **) + oldfds - 1;
	ip = mtod(rights, int *) + oldfds - 1;
	for (i = 0; i < oldfds; i++, ip--, rp--) {
		error = getff(*ip, &fp, FDNOMARK, NULL);
		ASSERT(!error);
		*rp = fp;;
		s = FLOCK(fp);
		fp->f_count++;
		fp->f_msgcount++;
		FUNLOCK(fp, s);
		UNP_PROTECT(unp_rights++);
	}

	/* adjust mbuf length to reflect type change */
	rights->m_len = oldfds * sizeof(struct file *);

	return (0);
}

/*
 * Garbage collecting - its possible that there are fd's caught in transit
 * they have been passed into messages but not ever received.
 * XXXjwag - seems like all messages are cleaned vi soclose->sorflush->
 * the unp_dispose function which frees all descriptors.
 *
 * This code is very bogus:
 * 1) filescan holds the spinlock while calling its functions - one of these
 *	functions calls closef! this immediately double trips on the file lock
 * 2) The entire model - marking ALL file descriptors, then scanning
 *	them for unmarked descriptors doesn't work in an MP environment -
 *	between the marking and the final scan, other file descriptors could
 *	have been added - these will be closed! by mistake.
 * 3) holding the file lock for a scan is way too long to hold off any
 *	new file opens
 */
#ifdef NEVER
int	unp_defer, unp_gcing;
int	unp_mark();
extern	struct domain unixdomain;

static int
unp_clearscan(struct file *fp, void *arg)
{
	int s;

	s = FLOCK(fp);
	fp->f_flag &= ~(FMARK|FDEFER);
	FUNLOCK(fp, s);
	return 0;
}

static int
unp_markscan(struct file *fp, void *arg)
{
	int s;
	register struct socket *so;
	struct vnode *vp;

	s = FLOCK(fp);
	if (fp->f_flag & FDEFER) {
		fp->f_flag &= ~FDEFER;
		UNP_PROTECT(unp_defer--);
	} else {
		if ((fp->f_flag & FMARK) ||
		    fp->f_count == fp->f_msgcount) {
			FUNLOCK(fp, s);
			return 0;
		}
		fp->f_flag |= FMARK;
	}
	FUNLOCK(fp, s);

	/*
	 * There appears to be a small window where a newly created
	 * file struct may not have its f_vnode pointer set. This
	 * sanity check should handle that.
	 */
	if ((vp = fp->f_vnode) == NULL)
		return 0;

	if (vp->v_op != &uipc_vnodeops ||
	    (so = vp->v_data) == 0)
		return 0;
	if (so->so_proto->pr_domain != &unixdomain ||
	    (so->so_proto->pr_flags&PR_RIGHTS) == 0)
		return 0;

#ifndef sgi /* reno */
	/*
	 * This is problematical; it's not clear
	 * we need to wait for the sockbuf to be
	 * unlocked (on a uniprocessor, at least),
	 * and it's also not clear what to do
	 * if sbwait returns an error due to receipt
	 * of a signal.  If sbwait does return
	 * an error, we'll go into an infinite
	 * loop.  Delete all of this for now.
	 */
	if (so->so_rcv.sb_flags & SB_LOCK) {
		(void) sbwait(&so->so_rcv);
		return 1;
	}
#endif
	unp_scan(so->so_rcv.sb_mb, unp_mark);
	return 0;
}

static int
unp_sweepscan(struct file *fp, void *arg)
{
	int s;

	s = FLOCK(fp);
	if (fp->f_count == fp->f_msgcount && (fp->f_flag & FMARK) == 0)
		while (fp->f_msgcount) {
			/*
			 * Must expand unp_discard in-line here, in order
			 * to test and decrement f_msgcount atomically.
			 */
			fp->f_msgcount--;
			FUNLOCK(fp, s);
			UNP_PROTECT(unp_rights--);
			closef(fp);
			s = FLOCK(fp);
		}
	FUNLOCK(fp, s);
	return 0;
}

void
unp_gc(void)
{
	int s;

	return;
	s = UNP_LOCK();
	if (unp_gcing) {
		UNP_UNLOCK(s);
		return;
	}
	unp_gcing = 1;
	UNP_UNLOCK(s);
restart:
	unp_defer = 0;
	(void) filescan(unp_clearscan, 0);
	do {
		if (filescan(unp_markscan, 0))
			goto restart;
	} while (unp_defer);
	(void) filescan(unp_sweepscan, 0);
	unp_gcing = 0;
}
#endif

int
unp_dispose(struct mbuf *m)
{
	if (m)
		unp_scan(m, unp_discard);
	return (0);
}

void
unp_scan(register struct mbuf *m0,
	void (*op)(struct file *))
{
	register struct mbuf *m;
	register struct file **rp;
	register int i;
	int qfds;

	while (m0) {
		for (m = m0; m; m = m->m_next)
			if (m->m_type == MT_RIGHTS && m->m_len) {
				qfds = m->m_len / sizeof (struct file *);
				rp = mtod(m, struct file **);
				for (i = 0; i < qfds; i++, rp++)
					(*op)(*rp);
				break;		/* XXX, but saves time */
			}
		m0 = m0->m_act;
	}
}

void
unp_discard(struct file *fp)
{
	int s;

	s = FLOCK(fp);
	fp->f_msgcount--;
	FUNLOCK(fp, s);
	UNP_PROTECT(unp_rights--);
	closef(fp);
}

/*
 * This routine is a hacked up variant of socket_pair_cmp_lock(), which
 * knows about the unp_list lock.  It exists only to avoid livelock between
 * unp_detach() and the PRU_SEND code.
 */
/* ARGSUSED */
int
uipc_three_way_lock(struct socket *so1, struct socket *so2, int ospl)
{
	if (so1 == so2)
		return(1);
	ASSERT(SOCKET_ISLOCKED(so1));
	if (SOCKET_MPTRYLOCK(so2) == 0) {
		/*
		 * Couldn't get lock.  Hold so2 so that it won't go away
		 * while we drop the lock on so1.
		 */
		SO_HOLD(so2);
		UNP_MPUNLOCK(ospl);
		SOCKET_MPUNLOCK(so1);
		/*
		 * This really sticks in my craw, but we need to avoid
		 * livelock.
		 */
		qswtch(RESCHED_Y);
		SOCKET_MPLOCK(so2);
		if (SO_RELE(so2) == 0) {
			SOCKET_MPUNLOCK(so2);
		}
		SOCKET_MPLOCK(so1);
		return(0);
	}
	return(1);
}
