/*
 * |-----------------------------------------------------------|
 * | Copyright (c) 1991, 1990 MIPS Computer Systems, Inc.      |
 * | All Rights Reserved                                       |
 * |-----------------------------------------------------------|
 * |          Restricted Rights Legend                         |
 * | Use, duplication, or disclosure by the Government is      |
 * | subject to restrictions as set forth in                   |
 * | subparagraph (c)(1)(ii) of the Rights in Technical        |
 * | Data and Computer Software Clause of DFARS 252.227-7013.  |
 * |         MIPS Computer Systems, Inc.                       |
 * |         950 DeGuigne Avenue                               |
 * |         Sunnyvale, California 94088-3650, USA             |
 * |-----------------------------------------------------------|
 */
#ident "$Revision: 1.14 $"

/*
 * mbuf/mblk conversion routines
 */

#include "sys/types.h"
#include "sys/param.h"
#include "sys/stream.h"
#include "sys/stropts.h"
#include "sys/strmp.h"
#include "sys/systm.h"
#include "bsd/sys/mbuf.h"
#include "sys/errno.h"
#include "sys/debug.h"

#undef STATIC
#ifdef DEBUG
#define STATIC
#else
#define STATIC static
#endif

#define CONST /* esballoc has bad prototype, needs const */

/* forward function declarations */
struct mbuf *mblk_to_mbuf( mblk_t *, int, int, int);
#ifdef DEBUG
STATIC void mblk_to_mbuf_free( struct mbuf *m, inst_t *);
STATIC void mblk_to_mbuf_dup( struct mbuf *m, inst_t *);
#else
STATIC void mblk_to_mbuf_free( struct mbuf *m);
STATIC void mblk_to_mbuf_dup( struct mbuf *m);
#endif
mblk_t *mbuf_to_mblk( struct mbuf *, int);
STATIC void mbuf_to_mblk_free( struct mbuf *m);

/* some statistics stuff */
int	mblk_to_mbuf_attempts;
int	mblk_to_mbuf_failures;
int	mblk_to_mbuf_mbufs;
int	mbuf_to_mblk_attempts;
int	mbuf_to_mblk_failures;
int	mbuf_to_mblk_mblks;

/*
 * Convert mblk chain to mbuf chain
 *
 *	If mbuf_flags includes M_PKTHDR, sets up a packet header with the
 *	length set.   All mbuf's in the chain are marked iwth mbuf_type.
 *
 *	mblk chain is freed or reallocated if successful, and unchanged
 *	if not.
 *	streams monitor is held by the caller.
 */
struct mbuf *
mblk_to_mbuf(
	mblk_t	*mp,		/* pointer to message being passed */
	int	alloc_flags,
	int	mbuf_type,
	int	mbuf_flags)
{
	struct mbuf *m;		/* last (newest) mbuf on chain */
	struct mbuf **mn;
	struct mbuf *m0;	/* head of new mbuf chain */
	mblk_t      *bp;
	int         len = 0;
	int  	    s;

	mblk_to_mbuf_attempts++;
	/*
	 * First pass:  allocate mbuf chain and count bytes.
	 */
	m0 = NULL;
	mn = &m0;
	for (bp = mp; bp; bp = bp->b_cont) {
		MGET(m, alloc_flags, mbuf_type);
		if (m == NULL) {
			m_freem(m0);
			mblk_to_mbuf_failures++;
			return (NULL);
		}
		*mn = m;
		mn = &m->m_next;
		len += bp->b_wptr - bp->b_rptr;
	}

	/*
	 * Fill in header and trailer info.
	m->m_flags  |= mbuf_flags & M_EOR;
	 */
	m0->m_flags |= mbuf_flags & M_COPYFLAGS;

	/*
	 * Second pass:  make each mbuf into an IRIX cluster.
	 * Bump the datab ref count since we now reference it.
	 */
	for (m = m0, bp = mp; m; m = m->m_next, bp = bp->b_cont) {
		m->m_len = bp->b_wptr - bp->b_rptr;
		if (m->m_len == 0)
			continue;
		m->m_off      = (__psint_t)bp->b_rptr - (__psint_t)m;
		m->m_flags    = MCL_CLUSTER;
		m->m_freefunc = mblk_to_mbuf_free;
		m->m_farg     = (__psunsigned_t)bp->b_datap;
		m->m_dupfunc  = mblk_to_mbuf_dup;
		m->m_darg     = (__psunsigned_t)bp->b_datap;

		s = splimp();
		mblk_to_mbuf_mbufs++;
		if (bp->b_datap->db_mbufref == 0)
			bp->b_datap->db_ref++;
		bp->b_datap->db_mbufref++;
		splx(s);
	}
	freemsg(mp);
	return (m0);
}

/*
 * caller holds splimp.
 */

#ifdef DEBUG
/* ARGSUSED */
STATIC void
mblk_to_mbuf_free( struct mbuf *m, inst_t *notused)
#else
STATIC void
mblk_to_mbuf_free( struct mbuf *m)
#endif
{
	register struct datab *databp;
	extern void freedatab(struct datab *);


	databp = (struct datab *)m->m_darg;

	ASSERT(databp->db_mbufref > 0);
	if (--databp->db_mbufref) {
		mblk_to_mbuf_mbufs--;
		return;
	}
	/*
	 * We're protected by the global resource lock so it's O.K. to
	 * call freeb (inside freedatab).
	 */
	freedatab(databp);

	mblk_to_mbuf_mbufs--;
}

/*
 * caller holds splimp.
 */
#ifdef DEBUG
/* ARGSUSED */
STATIC void
mblk_to_mbuf_dup( struct mbuf *m, inst_t *notused)
#else
STATIC void
mblk_to_mbuf_dup( struct mbuf *m)
#endif
{
	register struct datab *databp;
	int s;

	databp = (struct datab *)m->m_darg;
	ASSERT(databp->db_mbufref > 0);
	s = splstr();
	databp->db_mbufref++;
	splx(s);
	mblk_to_mbuf_mbufs++;
}

/*
 * Convert mbuf chain to data buffer.
 *
 *	Frees or reallocates the mbuf chain if successful.
 */
mblk_t *
mbuf_to_mblk( struct mbuf *m0, int pri)
{
	mblk_t *mp;
	mblk_t *bp;
	mblk_t **nbp;
	struct mbuf *m;
	struct mbuf *n;
	static CONST struct free_rtn free_dummy = { NULL, NULL };

	mbuf_to_mblk_attempts++;
	/*
	 * First pass:  allocate external storage mblks.  Free_dummy is
	 * a placeholder to allow us to back out with freemsg().
	 */
	mp = NULL;
	nbp = &mp;
	for (m = m0; m; m = m->m_next) {
		if (m->m_len == 0)
			continue;
		bp = esballoc(mtod(m, u_char *), m->m_len, pri, &free_dummy);
		if (bp == NULL) {
			if (mp != NULL)
				freemsg(mp);
			mbuf_to_mblk_failures++;
			return (NULL);
		}
		*nbp = bp;
		nbp = &bp->b_cont;
		bp->b_wptr = bp->b_datap->db_lim;
	}

	/* if zero-length message, just get single mblk. */
	if (mp == NULL) {
		mp = allocb(-1, pri);
		if (mp)
			m_freem(m0);
		else
			mbuf_to_mblk_failures++;
		return (mp);
	}

	/*
	 * Second pass:  fill in free_rtn struct.
	 */
	m = m0;
	for (bp = mp; bp; bp = bp->b_cont) {
		struct free_rtn *frp;

		while (m->m_len == 0) {
			MFREE(m, n);
			m = n;
		}
		n = m->m_next;

		frp = &bp->b_datap->db_frtn;
		frp->free_func = (void (*)())mbuf_to_mblk_free;
		frp->free_arg  = (char *)m;

		bp->b_datap->db_frtnp = frp;
		mbuf_to_mblk_mblks++;
		m = n;
	}
	return (mp);
}

STATIC void
mbuf_to_mblk_free( struct mbuf *m)
{
	(void) m_free(m);
	mbuf_to_mblk_mblks--;
}
